
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

enum { ZONE = 1, REG, HOST, HUB, NODE, PVT, HOLD, DOWN };

static char NodelistRecord[256];
static void replace_underscore(char *p);

/* GetNodelistRecord(max number of fields, address, [ndl_block_name] */
static XS(perl_GetNodelistRecord)
{
    dXSARGS;
    STRLEN n_a;
    struct ndlTags *ndlTag = NULL;
    char *s, *caddr;
    char *err = "perl_GetNodelistRecord()";
    int max;
    ADDR addr;
	
    if (items < 2 || items > 3) {
	fprintf(stderr, "%s argument count\n", err);
	abort();
    }

    NodelistRecord[0] = 0;
    
    sp_s = FromSP(0);
    
    sscanf(sp_s, "%d", &max);
    
    sp_s = FromSP(1);
    caddr = XCAL;
    strncpy(caddr, sp_s, n_a);
    strip(caddr);
    expand_str(caddr);
	
    if ( !str_to_addr(caddr , &addr) ) {
	fprintf(stderr, "%s incorrect address : `%s'\n", err, caddr);
	abort();
    }
	
    if (items == 3) {
	sp_s = FromSP(2);
	s = XCAL;
    	strncpy(s, sp_s, n_a);
    	strip(s);
	expand_str(s);
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next ) {
	    word j1, j2;
	    if (!stricmp(ndlTag->tag, s)) {
		is_listed(ndlTag, addr, &j1, &j2);
		break;
	    }
	}
    } else {
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next ) {
	    word j1, j2;
	    if (is_listed(ndlTag, addr, &j1, &j2)) break;
	}
    }

    if (strlen(NodelistRecord)) {
	if (max) {
	    char *p;
	    int i, pos;
	    char save[256];
	    char **prelines = NULL;

	    strcpy (save, NodelistRecord);

	
	    for(pos = i = 0, p = strtok(NodelistRecord, ","); p && (i < max-1);
				p = strtok(NULL, ",")) {
		if (i == 0) {
		    if (stricmp(p, "Zone") == 0 || stricmp(p, "Region") == 0 ||
			stricmp(p, "Host") == 0 || stricmp(p, "Hub") == 0 ||
			stricmp(p, "Pvt") == 0 || stricmp(p, "Hold") == 0 ||
			stricmp(p, "Down") == 0) pos = 1;
		    prelines = (char**)xmalloc((i+1)*sizeof(char*));
		} else {
		    prelines = (char**)xrealloc(prelines, (i+1)*sizeof(char*));
		}
		if (i == pos+1 || i == pos+3) replace_underscore(p);
		prelines[i] = (char*)xcalloc(strlen(p)+1, sizeof(char));
		strcpy(prelines[i], p);
		i++;
	    }

	    if (p) {
		if (i == 0) {
		    prelines = (char**)xmalloc((i+1)*sizeof(char*));
		} else {
		    prelines = (char**)xrealloc(prelines, (i+1)*sizeof(char*));
		}
		prelines[i] = (char*)xcalloc(strlen(&save[p-NodelistRecord])+1, sizeof(char));
		strcpy(prelines[i++], &save[p-NodelistRecord]);
	    }
	    
	    EXTEND(SP, i);
    
	    for (max = 0; max < i; max++) {
		ST(max) = sv_newmortal();
		sv_setpv((SV*)ST(max), prelines[max]);
		free(prelines[max]);

	    }
	
	    free(prelines);
	} else {
	
	    max = 1;
	    EXTEND(SP, max);
    
    	    ST(max) = sv_newmortal();
	    sv_setpv((SV*)ST(0), NodelistRecord);
	}
    }
    XSRETURN(max);

}

/************************************************/
BOOL
is_listed(struct ndlTags *ndlTag, ADDR addr, word* uplink, word* region)
{
	FILE *h_ndl;
	int h_idx, h_inc;
	BOOL zone_ok = No, abort_ = No, rec1 = Yes;
	word defzone = 0;
	dword len, pos;
	char tmp[128], ext[4], path[128], s[256], buf[256];
	struct nodeidx nodeidx;
	struct nodeinc nodeinc;
	char n_inc_buf[NODEINC_SIZE], *pbuf;
	char n_idx_buf[NODEIDX_SIZE];
	long size;
	
	sprintf(buf, "is_listed: nodelist look-up for address: %s\n",
				addr_to_str( addr, s) );
	writeToLog(4, buf);
	sprintf(buf, "is_listed: current nodelist block name: %s\n", ndlTag->tag);
	writeToLog(4, buf);

	if (!ndlTag->active)
		return Fail; /*  ndl disabled  */

	*uplink = minus_one;

	ndlTag->rec = ndlTag->recfirst;
	
	/* scanning the list of records in the selected tag */
	while(ndlTag->rec)  
	{

		strcpy(ext, ndlTag->rec->ext);
		strcpy(path, ndlTag->rec->path);
		defzone = ndlTag->rec->zone;
		*region = 0;
		
		sprintf(s, "%snodeinc.%s", path, ext);
		sprintf(buf, "trying to open index file `%s'... ", s);
		if ( -1 == (h_inc = open(s, O_RDONLY|O_BINARY))) {
			fprintf(stderr, "failed.\n");
			abort();
		}
		else {
			strcat(buf, "ok.\n");
			writeToLog(4, buf);
		}

		sprintf(s,"%snodeidx.%s", path, ext);
		sprintf(buf, "trying to open index file `%s'... ", s);
		if ( -1 == (h_idx = open(s, O_RDONLY|O_BINARY))) {
			fprintf(stderr, "failed.\n");
			abort();
		}
		else {
			strcat(buf, "ok.\n");
			writeToLog(4, buf);
		}
		

		rec1 = Yes;
		zone_ok = abort_ = No;
		
		/*   scanning current nodeidx   */
		
		size = filelength(h_idx);
		
		while(size > tell(h_idx))  
		{
			len = NODEIDX_SIZE;
			if (len != read(h_idx, &n_idx_buf, len)) {
				close(h_idx);
				close(h_inc);
				return Fail; /*  failed  */
			}
			
			pbuf = n_idx_buf;
			nodeidx.stat = *pbuf;
			pbuf++;
			nodeidx.number = get_word(pbuf);
			pbuf += 2;
			nodeidx.listno = *pbuf;
			pbuf++;
			nodeidx.pos = get_dword(pbuf);
			
			switch (nodeidx.stat+1) {
				
			case ZONE: /* got zone */
				
				sprintf(buf, "is_listed: zone found: %d -- ", nodeidx.number);
				if ( (word)nodeidx.number == addr.zone) {
					strcat(buf, "zone ok.\n");
					zone_ok = Yes;
					
					if (addr.zone == addr.net) {
					    *region = addr.net;
					    if (addr.node == 0) {
						*uplink = (word)0;
						abort_ = Yes;
					    } else {
/*					if (addr.zone == addr.net) { */
						/*   detecting the nodelist name   */
						
						pos = (nodeidx.listno-1)*NODEINC_SIZE;
						if (pos != lseek(h_inc, pos, SEEK_SET)) {
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: seek error in *.inc file\n");
							return Fail;                                  /* failed */
						}
						len = NODEINC_SIZE;
						if (len != read(h_inc, &n_inc_buf, len)) {
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: read error in *.inc file\n");
							return Fail;                                  /* failed */
						}
						
						pbuf = n_inc_buf;
						nodeinc.namelen = *pbuf;
						pbuf++;
						memmove(&(nodeinc.name), pbuf, 12);
						pbuf += 12;
						nodeinc.crc32 = get_dword(pbuf);

						memset(tmp, 0, 13);
						strncpy(tmp, nodeinc.name, 12); tmp[12] = 0;
						sprintf(s, "%s%s", path, tmp);
						sprintf(buf, "is_listed: got nodelist name: %s\n", s);
						writeToLog(4, buf);

						/* opening actual nodelist */
						if ( (h_ndl = fopen(s, "rb")) == NULL ) {
							fprintf(stderr, "is_listed: error opening nodelist file `%s'\n", s);
							abort();
						}

						if (0 != fseek(h_ndl, nodeidx.pos, SEEK_SET)) {
							fclose(h_ndl);
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: seek error in nodelist file\n");
							return minus_one;                           /* failed */
						}
						
						/* start look-up from the current position set by fseek */
						if (Fail == nodeLookUp(h_ndl, addr.net, addr.node, uplink)) {
							*uplink = -1;
							sprintf(buf, "is_listed: node %d not found in the current region/net (%d)\n",
										addr.node, addr.net);
							writeToLog(4, buf);
						}
						else {
							sprintf(buf, "is_listed: node found OK (uplink: /%hu)\n", *uplink);
							writeToLog(4, buf);
/*              sprintf(buf, "trying to find next match\n");
                writeToLog(4, buf);
*/							abort_ = Yes;
						}
						fclose(h_ndl); /* close current nodelist */
						/* stop scanning current set of nodelists */
					    }
					}
				}
				else {
					strcat(buf, "wrong zone, go on scanning current idx.\n");
				}
				
				writeToLog(4, buf);
				
				break;
				
			case REG:  /* got region */
				*region = (word) nodeidx.number;
			case HOST: /* got net    */
				
				if (!rec1 && !zone_ok)
					break;

				if (rec1) {
					defzone = defzone ? defzone : MainAddress.zone; /* ??? */
					sprintf(buf, "is_listed: no zone at start, assuming the default (%d) -- ",
									defzone);
					if (defzone != addr.zone) {
						strcat(buf, "wrong zone, go to next idx\n");
						writeToLog(4, buf);
						abort_ = Yes;
						break;
					}
					else {
						zone_ok = Yes;
						strcat(buf, "zones match ok.\n");
						writeToLog(4, buf);
					}
				}   /* 1st record in file */

				sprintf(buf, "is_listed: net/region found: %d -- ", nodeidx.number);

				if ((word)nodeidx.number == addr.net) /* net ok */
				{
					strcat(buf, "net or region found ok.\n");
					writeToLog(5, buf);
					
					if (addr.node == 0)    /*  NC, RC, or ZC, no need to do look-up  */
						*uplink = (word) 0;   /*  he's the uplink to himself            */
					
					else {                 /*  do nodelist look-up                   */
						
						/*   detecting the nodelist name   */
						
						pos = (nodeidx.listno-1)*NODEINC_SIZE;
						if (pos != lseek(h_inc, pos, SEEK_SET)) {
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: seek error in *.inc file\n");
							return Fail;                                  /* failed */
						}
						len = NODEINC_SIZE;
						if (len != read(h_inc, &n_inc_buf, len)) {
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: read error in *.inc file\n");
							return Fail;                                  /* failed */
						}
						
						pbuf = n_inc_buf;
						nodeinc.namelen = *pbuf;
						pbuf++;
						memmove(&(nodeinc.name), pbuf, 12);
						pbuf += 12;
						nodeinc.crc32 = get_dword(pbuf);

						memset(tmp, 0, 13);
						strncpy(tmp, nodeinc.name, 12); tmp[12] = 0;
						sprintf(s, "%s%s", path, tmp);
						sprintf(buf, "is_listed: got nodelist name: %s\n", s);
						writeToLog(4, buf);

						/* opening actual nodelist */
						if ( (h_ndl = fopen(s, "rb")) == NULL ) {
							fprintf(stderr, "is_listed: error opening nodelist file `%s'\n", s);
							abort();
						}

						if (0 != fseek(h_ndl, nodeidx.pos, SEEK_SET)) {
							fclose(h_ndl);
							close(h_inc);
							close(h_idx);
							writeToLog(4, "is_listed: seek error in nodelist file\n");
							return minus_one;                           /* failed */
						}
						
						/* start look-up from the current position set by fseek */
						if (Fail == nodeLookUp(h_ndl, addr.net, addr.node, uplink)) {
							*uplink = -1;
							sprintf(buf, "is_listed: node %d not found in the current region/net (%d)\n",
										addr.node, addr.net);
							writeToLog(4, buf);
						}
						else {
							sprintf(buf, "is_listed: node found OK (uplink: /%hu)\n", *uplink);
							writeToLog(4, buf);
/*              sprintf(buf, "trying to find next match\n");
                writeToLog(4, buf);
*/						}
						fclose(h_ndl); /* close current nodelist */
					} /* nodelist look-up block */
					
					/* stop scanning current set of nodelists */
					abort_ = Yes;
					
				}   /* net ok, do look-up */
				
				else {
					strcat(buf, "wrong net, go on scanning current idx.\n");
					writeToLog(5, buf);
				}
				
				break;
				
			default:
				break;
				
			} /* end of switch(stat) */
	
			/* stop scanning current nodeidx, go to next record of nodeinc */
			if (abort_ == Yes)
				break;  
				
			rec1 = No; /* record no. != 1 anymore */
				
		}  /* scanning current nodeidx */
		
		close(h_idx);
		close(h_inc);
		
		writeToLog(4, "is_listed: end of scanning current idx\n");
		
		if (*uplink != minus_one) {
			writeToLog(4, "is_listed: node found, no need to go to next idx\n");
			break;
		}
		
		/* go to the next set of nodelists */
		
		ndlTag->rec = ndlTag->rec->next; 
		
	} /* outmost loop - scanning records in tag */
	
	writeToLog(4, "is_listed: done scanning all nodelists\n");
	
	return (*uplink) == minus_one ? Fail : Ok;
	
}
/****************************************************/
static BOOL nodeLookUp(FILE *h_ndl, word host, word node, word* uplink)

/* return: -1 on fail, uplink node no. on success */

{
	char temp[256], buf[256], junk[256], tag_name[4], cnode[6];
	register char *p = NULL, *q = NULL, c;

	*uplink = 0;
	
	while (NULL != fgets(temp, 255, h_ndl))
	{
		/* skip comments */
		
		if (temp[0] == ';')
			continue;
			

		if (!strlen(temp))
			continue;

		strcpy(junk, temp);
		
		/* get first nodelist field (tag name) */
		
		p = strchr(junk, ',');
		if (p == NULL) {
			sprintf(buf, "ndl_lookup: error parsing string `%s'\n", junk);
			writeToLog(3, buf);
			continue;
		}
			
		junk[p-junk] = 0;
		strncpy(tag_name, junk, 3);
		
		tag_name[3] = 0;         /* only the first 3 chars matter */

		/* get second nodelist field (value of tag - zone, net, etc.) */
			
		++p;                     /* skip over 1st comma               */
		
		q = strchr(p, ',');      /* search for next comma             */
		
		if (q == NULL) {
			sprintf(buf, "ndl_lookup: error parsing string `%s'\n", temp);
			writeToLog(3, buf);
			continue;              /* no comma found, discard the whole string */
		}
			
		p[q-p] = '\0';
		strncpy(cnode, p, 5);    /* cnode holds a char value of current number */
		cnode[5] = '\0';         /* 5 chars max (enough to stuff ushort)       */
			
		/*  if cnode holds junk, discard the whole string  */
		if (No == __isnumber(cnode))
			continue;
		
		c = tag_name[0];
		
		/* resolve ambiguity */
		if (c == 'H')       
			c = tag_name[2];	/* `s' for hoSt */
						/* `b' for huB  */
						/* `l' for hoLd */

		
		/*  next Zone/Region/hoSt found */
		
		if (c=='Z' || c=='R' || c=='s') {
			if ( (word)atoi(cnode) != host )                /* see if it's ours */
			{
				writeToLog(4, "ndl_lookup: next zone/region/net found, look-up failed\n");
				
				return Fail; /* failed */
			}
			
				/*  node,   huB,  Pvt  or hoLd     */
		} else if (c=='\0' || c=='b' || c=='P' || c=='l') {

			/* store the head of the current hub segment for future reference */
			if (c == 'b')
				*uplink = (word)atoi(cnode);
			
			/* node match found */
			if ((word)atoi(cnode) == node) {
				sprintf(buf, "ndl_lookup: uplink node # = %d\n", *uplink);
				writeToLog(4, buf);
				strcpy(NodelistRecord, temp);
				if (NodelistRecord[strlen(NodelistRecord)-1] == '\n')
					NodelistRecord[strlen(NodelistRecord)-1] = '\0';
				if (NodelistRecord[strlen(NodelistRecord)-1] == '\r')
					NodelistRecord[strlen(NodelistRecord)-1] = '\0';
				return Ok;
			}
			
		}               /* node  hub  pvt */
		
	}                 /* main ndl loop */
	
	return Fail; /* failed */
}


static void replace_underscore(char *p)
{
	while (*p)
	{
		if (*p == '_')
			*p = ' ';
		++p;
	}
	return;
}
