
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

extern int errno;


/*
 *  put_dword
 *
 *  Writes a 4 byte word in little endian notation, independent of the local
 *  system architecture.
 */

void put_dword(byte *ptr, dword value)
{
    ptr[0] = (value & 0xFF);
    ptr[1] = (value >> 8) & 0xFF;
    ptr[2] = (value >> 16) & 0xFF;
    ptr[3] = (value >> 24) & 0xFF;
}

/*
 *  put_word
 *
 *  Writes a 4 byte word in little endian notation, independent of the local
 *  system architecture.
 */

void put_word(byte *ptr, word value)
{
    ptr[0] = (value & 0xFF);
    ptr[1] = (value >> 8) & 0xFF;
}



static XS(perl_ReadPkt)
{
    dXSARGS;
    char *err = "perl_ReadPkt()";
    int rc = 0;
    BOOL need_inc;
	
    if (items) {
	fprintf(stderr, "%s argument count\n", err);
	abort();
    }
	
    if (CurMsg.pkt.icur == -1) {
	CurMsg.pkt.icur = 0;
	need_inc = 0;
    } else {
	if (CurMsg.pkt.rinc) {
	    CurMsg.pkt.rinc = 0;
	    need_inc = 0;
	} else
	    need_inc = 1;
    }
	 


    if (CurMsg.type == PKT) {
	int h;

	if (need_inc)
	    ++CurMsg.pkt.icur;
		
	if (CurMsg.pkt.icur < CurMsg.pkt.total) {
		
	    if (CurMsg.n)
		update_msg (PKT);

	    if (CurMsg.pkt.off)
		update_old_pkt(CurMsg.pkt.fname, CurMsg.pkt);
		
	    free_text(TextFirst);
	    TextFirst = NULL;
	    free_text(ViaTextFirst);
	    ViaTextFirst = NULL;
	    memset(&ftn, 0, sizeof(FTN));
	    CurMsg.pkt.off = 0L;
	    CurMsg.n = 0;

	    h = open_read_pkt (CurMsg.pkt.fnames[CurMsg.pkt.icur].fname, &CurMsg.pkt);
	    if (h != -1) {
		close (h);
		rc = 1;
	    } else
		rc = 0;
	}
    }

    XSRETURN_IV(rc);
}

static XS(perl_KillPkt)
{
    dXSARGS;
    char *err="perl_KillPkt()";
    int rc = 0;
	
    if (items) {
	fprintf(stderr, "%s argument count\n", err);
	abort();
    }

    if (CurMsg.type == PKT && CurMsg.n) {
	killpkt ();
	rc = 1;
    } 
    XSRETURN_IV(rc);
}


static XS(perl_CopyPkt)
{
    dXSARGS;
    STRLEN n_a;
    char *areaname, *err="perl_CopyPkt()";
    int rc = 0;

    if (items != 1) {
	fprintf(stderr, "%s argument count (need 1, exist %d)\n", err, (int)items);
	abort();
    }

    sp_s = FromSP(0);
    areaname = XCAL;	
    strncpy(areaname, sp_s, n_a);
    strip(areaname);
    expand_str(areaname);

    if (CurMsg.type == PKT && CurMsg.n) rc = copypkt (areaname);
	
    XSRETURN_IV(rc);
}

static XS(perl_MovePkt)
{
    dXSARGS;
    STRLEN n_a;
    char *areaname, *err="perl_MovePkt()";
    int rc = 0;

    if (items != 1) {
	fprintf(stderr, "%s argument count (need 1, exist %d)\n", err, (int)items);
	abort();
    }

    sp_s = FromSP(0);
    areaname = XCAL;	
    strncpy(areaname, sp_s, n_a);
    strip(areaname);
    expand_str(areaname);

    if (CurMsg.type == PKT && CurMsg.n) {
	if (copypkt (areaname)) {
	    killpkt ();
	    rc = 1;
	}
    }

    XSRETURN_IV(rc);	
}

static int copypkt (char* areaname)
{
	int rc = 0;
	if (areaname_to_type (areaname) == PKT) {
		char areapath[128], fname[256], fullname[256];
		if (!areaname_to_path (areaname, areapath)) {
			writeToLog(4, "copypkt: couldn't get areapath");
			return 0;
		}
		if (Fail == make_new_pkt_name(areapath, fname)) {
			writeToLog(4, "get_pkt_name: fail to get a new name\n");
			return 0;
		}
		sprintf(fullname, "%s%s", areapath, fname);
    update_old_pkt (fullname, CurMsg.pkt);
		rc = 1;
	}
	return rc;
}

static void killpkt (void)
{
	free_text(TextFirst);
	TextFirst = NULL;
	free_text(ViaTextFirst);
	ViaTextFirst = NULL;
	memset(&ftn, 0, sizeof(FTN));

	update_msg (PKT);
	CurMsg.pkt.off = CurMsg.pkt.len;
	update_old_pkt(CurMsg.pkt.fname, CurMsg.pkt);
	CurMsg.pkt.off = 0;
	CurMsg.pkt.len = 0;
	CurMsg.pkt.new_off = 0;
	CurMsg.n = 0;
}

BOOL is_pkt_name(char *s)
{
	char *q = s;
	while(*q && isxdigit(*q))
		++q;

	return ((strlen(s) == 12) && (stricmp(q, ".pkt") == 0));
}

static char *is_fbox_dir (int fbox, char *dirname)
{
	char *q;
	word z, n, f, p;
	
	if (fbox == BBOX) { /* brake has domain first */
		if (NULL == (q = strchr (dirname, '.')))
			return No;
		dirname = q+1;
	}
	
	return (char*)(4 == sscanf (dirname, "%hu.%hu.%hu.%hu", &z,&n,&f,&p));

}
/*************************************************************************/
static void _select_pkt_area(char *areapath, struct Pktnames **f, int* total)
{
	DIR *dir;
	struct dirent *d;

	char buf[256];
	struct PktAreas  *PktArea;
	int fbox = 0;
	
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea->next )
	{
		if ( !stricmp( PktArea->path, areapath ) ) {
			fbox = PktArea->fbox;
		}
	}

#ifdef __MINGW32__
	dirNameLen = strlen(areapath)-1;
	areapath[dirNameLen] = 0;
#endif

	dir = opendir(areapath);

#ifdef __MINGW32__
	areapath[dirNameLen] = '\\';
#endif

	sprintf (buf, "select_pkt: scanning %s\n", areapath);
	writeToLog(5, buf);
	
	if (dir) {
		while (NULL != (d = readdir(dir))) {
			
			if (!strcmp(d->d_name, ".") || !strcmp(d->d_name, ".."))
				continue;

			if (fbox == 0) {
				if (is_pkt_name(d->d_name)) {
					if (!*f)
						*f = (struct Pktnames *)malloc(sizeof(struct Pktnames));
					else
						*f = (struct Pktnames *)realloc(*f, ((*total)+1)*sizeof(struct Pktnames));
					sprintf (buf, "select_pkt: found pktfile `%s'\n", d->d_name);
					writeToLog(4, buf);
					sprintf(((*f)[(*total)++]).fname, "%s%s", areapath, d->d_name);
				}
			}
			else {
				char buf[256];
				struct stat st;
				sprintf (buf, "%s%s", areapath, d->d_name);

				if ((stat(buf, &st) == 0) && (st.st_mode & S_IFDIR)) {
					if (is_fbox_dir (fbox, d->d_name)) {
						strcat(buf, SEP);
						_select_pkt_area (buf, f, total);
					}
				}

			} /* fbox */
		}
		closedir(dir);
	}

	return;
}

void select_pkt_area(char *areapath, PKT_AREA *pkt)
{
	struct Pktnames *f = NULL;
	int total=0;

	free_pkt_cache ();
	
	_select_pkt_area (areapath, &f, &total);
	pkt->fnames = f;
	pkt->total = total;
	pkt->icur = -1;
	pkt->rinc = 0;
}

/******************************/
BOOL readpkt_wrapper(char *err)
{
	char buf[256];
	BOOL rc = Fail;

	if (CurMsg.pkt.icur == -1) {
		CurMsg.pkt.icur = 0;
	}

	while (CurMsg.pkt.icur < CurMsg.pkt.total)
	{
		
		/* read next file */
		if (0L == CurMsg.pkt.off) {

			strcpy(CurMsg.pkt.fname, CurMsg.pkt.fnames[CurMsg.pkt.icur].fname);
			
			if (Fail == read_pkt(CurMsg.pkt.fname, &CurMsg.pkt)) {
				sprintf(buf, "%s error reading file `%s'\n", err, CurMsg.pkt.fname);
				writeToLog(3, buf);
				CurMsg.pkt.off = 0L;
				CurMsg.n = 0;
				if (errno != EACCES)
					(void)rename_bad_pkt(CurMsg.pkt.fname);
				CurMsg.pkt.rinc = 1;
				++CurMsg.pkt.icur;
				continue;
			}
			else
				writeToLog(4, "read_msg (pkt): pkt read in memory OK\n");

		} /* read next pkt */

		if ( Ok == read_pkt_msg(&(CurMsg.pkt), &TextFirst) )
		{
			/* end of current file */
			if (CurMsg.pkt.off == CurMsg.pkt.len) {
				CurMsg.pkt.rinc = 1;
				++CurMsg.pkt.icur;
			}
	
			pktarea_to_ftn(CurMsg.pkt, &ftn);
			writeToLog(4, "read_msg (pkt): next pkt msg read OK\n");
			CurMsg.n = ++CurMsg.pkt.n;
			rc = Ok;
			break;  /* break from main loop */

		} /* msg read OK */

		else {

			sprintf(buf, "%s error reading pkt msg, read next *.pkt\n", err);
			writeToLog(3, buf);
			(void)rename_bad_pkt(CurMsg.pkt.fname);
			CurMsg.pkt.off = 0L;
			CurMsg.n = 0;
			
		}  /* bad pkt msg */

		CurMsg.pkt.rinc = 1;
		++CurMsg.pkt.icur;

	} /* main loop */

	if (rc == Fail) {
		sprintf(buf, "%s end of pkt area found\n", err);
		writeToLog(3, buf);
	}

	return rc;

}

static int open_read_pkt (char* PKT_name, PKT_AREA* pkt)
{
	int h;
	char buf[256], caddrTo[128], caddrFrom[128];
	
	if( (h = sopen(PKT_name, O_RDONLY|O_BINARY, SH_DENYNO, ACCFILE)) == -1 ) {

		sprintf(buf, "read_pkt: error opening  file: `%s'\n", PKT_name);
		writeToLog(0, buf);
		return  -1;
	}
	
	if (PKT_HDR_SIZE != ReadPktHdr(h, &(pkt->pkt_hdr))){
		sprintf(buf, "read_pkt: error reading pkt/?ut file: `%s'\n", PKT_name);
		writeToLog(0, buf);
		close(h);
		return -1;
	}
	
		
	if (pkt->pkt_hdr.pktType != 0x0002) {
		sprintf(buf, "read_pkt error: `%s' is no type 2 pkt file\n", PKT_name);
		writeToLog(0, buf);
		close(h);
		return Fail;
	}
	
	pkt->to.zone    = pkt->pkt_hdr.toZone;
	pkt->to.net     = pkt->pkt_hdr.toNet;
	pkt->to.node    = pkt->pkt_hdr.toNode;
	pkt->to.point   = pkt->pkt_hdr.toPoint;
	pkt->from.zone  = pkt->pkt_hdr.fromZone;
	pkt->from.net   = pkt->pkt_hdr.fromNet;
	pkt->from.node  = pkt->pkt_hdr.fromNode;
	pkt->from.point = pkt->pkt_hdr.fromPoint;

	if (pkt->from.point != 0  &&  pkt->from.net == Wild_1)
		pkt->from.net = pkt->pkt_hdr.AuxNet;
	
	addr_to_str(pkt->from, caddrFrom);
	addr_to_str(pkt->to, caddrTo);
	
	sprintf(buf, "Opening 2+ Pkt `%s'  from %s to %s", PKT_name, caddrFrom, caddrTo);
	writeToLog(1, buf);
	
	return h;
}

/****************************************************************************/
BOOL read_pkt(char *PKT_name, PKT_AREA *pkt)
{
	int h; 
	dword len;           /* sizeof of pkt file without header */
	dword flen;          /* total filelength */
	int zlen;            /* terminating zeroes count */

	if( (h = open_read_pkt (PKT_name, pkt)) == -1 )
		return  Fail;

	
	flen = filelength(h);
	len = flen - PKT_HDR_SIZE;
	
	if (len < PKT_HDR_SIZE + PKT_MSG_HDR_SIZE + 4 + 2) {
		writeToLog(0, "read_pkt error: incorrect length of file: `%s'\n", PKT_name);
		close (h);
		return  Fail;
	}
	
	if (pkt->text == NULL)
		pkt->text = (char *)malloc(len+1);
	else
		pkt->text = (char *)realloc(pkt->text, len+1);

	if (pkt->text == NULL) {
		writeToLog(0, "read_pkt error: low memo (text)\n");
		close(h);
		return Fail;
	}
	
	if ( len != read( h, pkt->text, len ) ){
		writeToLog(0, "read_pkt: error reading pkt/?ut file: `%s'\n", PKT_name);
		close(h);
		return Fail;
	}
	
	close(h);
	
	pkt->off0 = 0L;  /* previous offset */
	pkt->off = 0L;   /* current offset */
	pkt->len = len;
	pkt->new_off = 0L;

	zlen = 0;

	if (pkt->text[len-1] == '\0') {
	  zlen++;
	  if (pkt->text[len-2] == '\0') {
	    zlen++;
	    if (pkt->text[len-3] == '\0')
	      zlen++;
	  }
	} else {
	  pkt->text[len] = '\0';
	}

	if (zlen != 3)
	  writeToLog(0, "read_pkt warning: grunged pkt/?ut file: `%s' (%u:%u/%u.%u)\n", PKT_name, pkt->from.zone, pkt->from.net, pkt->from.node, pkt->from.point);

	if (zlen == 0)
	  pkt->len++;
	else
	  if (zlen>1)
	    pkt->len -= (dword)(zlen-1); /* exclude terminating zeroes */
	
	return Ok;
}
/****************************************************************************/
BOOL read_pkt_msg(PKT_AREA *pkt, TEXT **txt)
{
	dword off = pkt->off;
	dword len = PKT_MSG_HDR_SIZE;
	char buf[256], *ptr;
	time_t ut;
	
	ptr = &pkt->text[off];
	
	pkt->pkt_msg_hdr.type = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.fromNode = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.toNode = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.fromNet = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.toNet = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.attr = get_word(ptr);
	ptr += 2;
	pkt->pkt_msg_hdr.cost = get_word(ptr);
	ptr += 2;
	memmove(&(pkt->pkt_msg_hdr.DateTime), ptr, 20);
	
	if (pkt->pkt_msg_hdr.type != 0x0002) {
		sprintf(buf, "read_pkt_msg error:  message grunged in file `%s'\n",
						pkt->fname);
		writeToLog(0, buf);
		return Fail;
	}
	
	/* CHECK DATE */
	
	if (Fail == (ut = ftscdate_to_unix(pkt->pkt_msg_hdr.DateTime)) &&
			Fail == (ut = seadog_to_unix(pkt->pkt_msg_hdr.DateTime)) ) {
		sprintf(buf, "read_pkt_msg error: invalid date `%s', set current\n",
						pkt->pkt_msg_hdr.DateTime);
		writeToLog(0, buf);
		ut = time(NULL);
/*    return Fail; */
	}
	else
		unix_to_ftscdate(ut, pkt->pkt_msg_hdr.DateTime);
	
	off += len;

	/* TO NAME */
	if (off < pkt->len) {
		strncpy(pkt->toName, &pkt->text[off], 36);
		pkt->toName[35] = '\0';
		off += strlen(pkt->toName)+1;
		if (off >= pkt->len)
			return Fail;
	}
	
	/* FROM NAME */
	if (off < pkt->len) {
		strncpy(pkt->fromName, &pkt->text[off], 36);
		pkt->fromName[35] = '\0';
		off += strlen(pkt->fromName)+1;
		if (off >= pkt->len)
			return Fail;
	}
	
	/* SUBJECT */
	if (off < pkt->len) {
		strncpy(pkt->subject, &pkt->text[off], 72);
		pkt->subject[71] = '\0';
		off += strlen(pkt->subject)+1;
		if (off >= pkt->len)
			return Fail;
	}

	if (off < pkt->len) {
		*txt = str_to_text(&pkt->text[off]);
		off += textlen(*txt)+1;
	}

	pkt->off0 = pkt->off;
	pkt->off = off;
	
	
	return Ok;
}

/**********************************************/
void pktarea_to_ftn(struct pktarea pkt, FTN *ftn)
{
	ftn->to.net    = pkt.pkt_msg_hdr.toNet;
	ftn->to.node   = pkt.pkt_msg_hdr.toNode;
	ftn->from.net  = pkt.pkt_msg_hdr.fromNet;
	ftn->from.node = pkt.pkt_msg_hdr.fromNode;
	ftn->attr      = pkt.pkt_msg_hdr.attr;
	
	ftn->to.zone = ftn->from.zone = 0;
	
	strcpy(ftn->toName, pkt.toName);
	strcpy(ftn->fromName, pkt.fromName);
	strcpy(ftn->subject, pkt.subject);
	strcpy(ftn->DateTime, pkt.pkt_msg_hdr.DateTime);
	
	ftn->utime = ftscdate_to_unix(ftn->DateTime);
	
	
	return;
}

/*********************************************/
void ftn_to_pktarea(FTN ftn, struct pktarea *pkt)
{
	ADDR from;

	if (ftn.echo_from.zone)
		from = ftn.echo_from;
	else
		from = ftn.from;
	
	pkt->pkt_msg_hdr.fromNet = from.net;
	pkt->pkt_msg_hdr.fromNode = from.node;
		
	pkt->pkt_msg_hdr.toNet = ftn.to.net;
	pkt->pkt_msg_hdr.toNode = ftn.to.node;
	strcpy(pkt->pkt_msg_hdr.DateTime, ftn.DateTime);
	pkt->pkt_msg_hdr.attr = ftn.attr;
	pkt->pkt_msg_hdr.type = 0x0002;
	
	memset (&pkt->toName, 0, sizeof(pkt->toName));
	strcpy(pkt->toName, ftn.toName);
	
	memset (&pkt->fromName, 0, sizeof(pkt->fromName));
	strcpy(pkt->fromName, ftn.fromName);
	
	memset (&pkt->subject, 0, sizeof(pkt->subject));
	strcpy(pkt->subject, ftn.subject);
	
	return;
}
/*************************************************/
static BOOL make_new_pkt_name(char *areapath, char *fname)
{
/*	struct find_t ffblk; */
	char fullname[256];
	int i;
	
	for(i = 0 ; i < 100; ++i)
	{
		sprintf(fname, "%08x.pkt", (unsigned)get_MsgId() );
		sprintf(fullname, "%s%s", areapath, fname);
#if 0
		if (_dos_findfirst(fullname, 0, &ffblk))
			return Ok;
#endif
		if (access(fullname, 0))
			return Ok;
	}
	
	return Fail;
}


/***************************/
void update_old_pkt(char *fname, PKT_AREA pkt)
{
	BOOL   isFLO(char *fname);
	int h;
	char buf[256];
	dword len = PKT_HDR_SIZE;
	
	sprintf(buf, "update_pkt: end of pkt/?ut, updating the original file `%s'\n",
					fname);
	writeToLog(4, buf);
	
	/* nothing to update */
	if (!pkt.new_off && pkt.off == pkt.len) {
		if (CurMsg.type == PKT || (CurMsg.type == BINK && !isFLO(fname))) {
			sprintf(buf, "update_pkt: kill empty file `%s'\n", fname);
			writeToLog(4, buf);
			remove(fname);
			return;
		}
	}
		
	if( -1 == (h = sopen(fname, O_RDWR|O_BINARY|O_CREAT, SH_DENYNO, ACCFILE))) {
		perror(fname);
		sprintf(buf, "update_pkt: error opening pkt/?ut file (%s)\n", fname);
		writeToLog(0, buf);
		return;
	}

	lseek(h, 0L, SEEK_SET);
	setfsize(h, 0L);
	lseek(h, 0L, SEEK_SET);
	

	/* PKT_HEADER */
	if (CurMsg.type == PKT || (CurMsg.type == BINK && !isFLO(fname))) {
		len = PKT_HDR_SIZE;
		if (len != WritePktHdr(h, &pkt.pkt_hdr)) {
			sprintf(buf, "update_pkt: error writing pkt/?ut file `%s'\n",
							fname);
				writeToLog(0, buf);
				close(h);
				return;
		}
	}

	/* read but not modified part (stored in memory as pkt.new_buf) */
	if (pkt.new_off) {
		
		len = pkt.new_off;
	
		if (len != write(h, pkt.new_buf, len)) {
			sprintf(buf, "update_pkt: error writing pkt/?ut file `%s'\n",	fname);
			writeToLog(0, buf);
			close(h);
			return;
		}

	}

	/* unread original text (stored in memory as pkt.text) */
	len = pkt.len - pkt.off;
	if (len) {
		if (len != write(h, &pkt.text[pkt.off], len)) {
			sprintf(buf, "update_pkt: error writing pkt/?ut file `%s'\n",
							fname);
			writeToLog(0, buf);
		}
	}

	/* footer */
	if (CurMsg.type == PKT || (CurMsg.type == BINK && !isFLO(fname))) {
		char z[2] = {0,0};
		write(h, z, sizeof z);
	}
	

	close(h);
	
	return;
}


static int open_pkt (BOOL add_new, char *areatag,
										 struct pktarea* pkt, TEXT *t,	word flv, char* Fname)
{
	int h;
	char buf[256], areapath[256], fname[256], pktname[256], *p;
	
	areaname_to_path(areatag, areapath);
	
	if (!add_new) {
		if (!get_pkt_name(pkt, areapath, t, pktname))
			return -1;
	}
	else {
		if (Fail == make_new_pkt_name(areapath, pktname)) {
			writeToLog(4, "get_pkt_name: fail to get a new name\n");
			return -1;
		}
	}
	
	if (NULL != (p = is_fbox(areatag, pkt->to, flv))) {
		strcat(areapath, p);
		fix_path(areapath);
		strcat(areapath, SEP);
	}
	
	sprintf(fname, "%s%s", areapath, pktname);
	
	h = sopen(fname, O_CREAT|O_RDWR|O_BINARY, SH_DENYNO, ACCFILE);

	if (h == -1)
	{
		if (errno == EACCES) {
			sprintf(buf, "pkt file locked by another process: `%s'\n", fname);
			writeToLog(0, buf);
			return -2;
		}
		sprintf(buf, "write_pkt_msg: error opening pkt file : `%s'\n", fname);
		perror(fname);
/*		fprintf(stderr, */
		writeToLog(0, buf);
	}
	else
		strcpy (Fname, fname);
	
  return h;
}

/**********************************************/
BOOL write_pkt_msg(char *areatag, struct pktarea* pkt, TEXT *t,	word flv)
{
	int h;
	char buf[256], fname[256];
	dword len;
	char z[3], subj[72];

	if (-2 == (h = open_pkt (0, areatag, pkt, t, flv, fname)))
		h = open_pkt (1, areatag, pkt, t, flv, fname);

	if (h == -1)
		return Fail;
	
	if (0L == filelength(h)) {
		if (PKT_HDR_SIZE != WritePktHdr(h, &(pkt->pkt_hdr))) {
			sprintf(buf, "write_pkt_msg: error writing pktfile (hdr) `%s'\n", fname);
			writeToLog(0, buf);
			close(h);
			return Fail;
		}
	}
	else {
		if (-1L == lseek(h, -2L, SEEK_END)) {
			sprintf(buf, "write_pkt_msg: seek error in `%s'\n", fname);
			writeToLog(0, buf);
			close(h);
			return Fail;
		}
	}

	{
	    if (PKT_MSG_HDR_SIZE != WritePktMsgHdr(h, &(pkt->pkt_msg_hdr))) {
		sprintf(buf, "write_pkt_msg: error writing pktfile (msg) `%s'\n", fname);
		writeToLog(0, buf);
		close(h);
		return Fail;
	    }
	}
	
	pkt->toName[35] = '\0';
	pkt->fromName[35] = '\0';
	pkt->subject[71] = '\0';

	strcpy(subj, pkt->subject);
	if (pkt->pkt_msg_hdr.attr & a_File) {
		char *p = strrchr(pkt->subject, DELIM);
		if (p)
			strcpy(subj, &pkt->subject[p-pkt->subject+1]);
	}
	
	write(h, pkt->toName, strlen(pkt->toName)+1);
	write(h, pkt->fromName, strlen(pkt->fromName)+1);
	write(h, subj, strlen(subj)+1);
	

	for( ; t ; t = t->next )
	{
		len = strlen(t->str);
		if (len != write(h, t->str, len)) {
			sprintf(buf, "write_pkt_msg: error writing pktfile (msg) `%s'\n", fname);
			writeToLog(0, buf);
			close(h);
			return Fail;
		}
	}

	memset(z, 0, 3);
	write(h, z, 3);
	
	close(h);
	strcpy (OutFileName, fname);
	
	return Ok;
	
}

/********************************************************************/
static BOOL get_pkt_name(PKT_AREA* pkt, char *areapath, TEXT *t, char *fname)
{
	struct pktcache *cache = PktCache;
	char buf[256], caddr[128];
	dword len = 0L, maxlen = get_maxpktsz(pkt->to, areapath);

	if (maxlen)
		len = textlen(t) + PKT_MSG_HDR_SIZE + strlen(pkt->toName)
			+ strlen(pkt->fromName) + strlen(pkt->subject) + 4;

	addr_to_str(pkt->to, caddr);
	sprintf(buf, "get_pkt_name: max length for %s = %lu\n", caddr, maxlen);
	writeToLog(4, buf);
	
	for( ; cache; cache = cache->next )
	{
		if (addr_cmp(cache->to, pkt->to) &&
				addr_cmp(cache->from, pkt->from) &&
				!stricmp(cache->areapath, areapath) &&
				0 != stricmp(cache->fname, pkt->fname))
		{
			sprintf(buf, "get_pkt_name: curmsg len = %lu, curpkt len = %lu, maxlen = %lu\n",
							len, cache->len, maxlen);
			writeToLog(4, buf);
			
			if (!maxlen || (cache->len + len <= maxlen)) {
				strcpy(fname, cache->fname);
				sprintf(buf, "get_pkt_name: get name from cache, `%s'\n",
								fname);
				writeToLog(4, buf);
				
				if (maxlen)
					cache->len += len;
			}
			else {

				if (Fail == make_new_pkt_name(areapath, fname)) {
					writeToLog(4, "get_pkt_name: fail to get a new name\n");
					return Fail;
				}
				
				sprintf(buf, "get_pkt_name: file too long, assign new name `%s'\n",
								fname);
				writeToLog(4, buf);
				
				strcpy(cache->fname, fname);
				
				if (maxlen)
					cache->len = len + PKT_HDR_SIZE;
				
			} /* pkt too long, assign new one */

			
			return Ok;
		} /* cach record found */
	}   /* cache walk */

	
	/* no cache entry found, add new file */
	
	if (Fail == make_new_pkt_name(areapath, fname)) {
		writeToLog(4, "get_pkt_name: fail to get a new name\n");
		return Fail;
	}

	if (maxlen)
		len += PKT_HDR_SIZE;
	
	sprintf(buf, "get_pkt_name: add new name to cache, `%s', len=%lu\n",
					fname, len);
	writeToLog(4, buf);

	add_to_pktcache(pkt->to, pkt->from, areapath, fname, len);
	return Ok;
}

/*************************************/
static dword get_maxpktsz(ADDR to, char *areapath)
{
	struct LengthAddrTags *PktLenAddrTag = NULL;
	struct PktAreas *PktArea = NULL;
	char buf[256], pktlentag[20];
	
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea ->next )
		if (!stricmp(PktArea->path, areapath))
			break;

	if (!PktArea)
		return 0L;

	
	strcpy(pktlentag, PktArea->pktlentag);
	sprintf(buf, "get_maxpktsz: got pktlentag: `%s'\n", pktlentag);
	writeToLog(5, buf);

	if (!strlen(pktlentag))
		return 0L;
	
	PktLenAddrTag = PktLenAddrTagFirst;
	for( ; PktLenAddrTag; PktLenAddrTag = PktLenAddrTag->next )
		if (!stricmp(PktLenAddrTag->tag, pktlentag))
			break;

	if (!PktLenAddrTag)
		return 0L;

	
	PktLenAddrTag->rec = PktLenAddrTag->recfirst;
	for(; PktLenAddrTag->rec; PktLenAddrTag->rec = PktLenAddrTag->rec->next )
		if (comp_addrs(to, PktLenAddrTag->rec->addr))
			return PktLenAddrTag->rec->len;
	
	return 0L;
}
/****************************************************************/
void add_to_pktcache(ADDR to, ADDR from, char *areapath, char *fname, dword len)
{
	struct pktcache *cache = PktCache;
	
	cache = (struct pktcache *)calloc(1, sizeof(struct pktcache));
	if (cache) {
		cache->next = NULL;
		/*/////////////////*/
		strcpy(cache->fname, fname);
		strcpy(cache->areapath, areapath);
		cache->to = to;
		cache->from = from;
		cache->len = len;
		/*/////////////////*/
		if (PktCacheLast) {
			PktCacheLast->next = cache;
		}
		else {
			PktCache = cache;
		}
		PktCacheLast = cache;
	}
	
	return;
}

/***********************************************/
void rename_bad_pkt(char *fname)
{
	char buf[256], bad_name[256], *p;

	strcpy(buf, fname);
	
	if (NULL != (p = strrchr(buf, '.')))
		buf[p-buf] = '\0';
	
	sprintf(bad_name, "%s.bad", buf);

	if (0 != rename(fname, bad_name)) {
		
		if (NULL != (p = strrchr(buf, DELIM)))
			buf[p-buf] = '\0';
		else
			buf[0] = '\0';
		
		sprintf(bad_name, "%s%c%08x.bad", buf, DELIM, (unsigned)get_MsgId() );
		rename(fname, bad_name);
	}

	
	return;
}

char* get_fbox_path (int type, char* areatag, ADDR to, word flv)
{
	static char p[256];
	
	if (FBox)
		return FBox;
	
	if (type == BBOX) {
		char system[20];
		ADDR aka;
		areaname_to_system(areatag, system);
		get_def_addr(system, to, &aka);
		sprintf(p, "%s.%hu.%hu.%hu.%hu.%s",
						aka.domain, to.zone, to.net, to.node, to.point, get_flv(flv));
		return p;
	}
	else if (type == TBOX) {
		sprintf(p, "%hu.%hu.%hu.%hu", to.zone, to.net, to.node, to.point);
		if (flv & flv_Hold)
			strcat(p, ".H");
		return p;
	}
	return NULL;
}

static char * is_fbox(char *areatag, ADDR to, word flv)
{
	struct PktAreas  *PktArea;

	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea->next )
	{
		if ( !stricmp( PktArea->name, areatag ) ) {
			return get_fbox_path (PktArea->fbox, areatag, to, flv);
		}
	}
	
	return NULL;

}

static char *get_flv(word flv)
{
	static char *p;
	
	if (flv & flv_Imm)
		p = "Immediate";
	else if (flv &flv_Crash)
		p = "Crash";
	else if (flv &flv_Dir)
		p = "Direct";
	else if (flv &flv_Hold)
		p = "Hold";
	else
		p = "Normal";
	
	return p;
}


dword ReadPktHdr(int handle, PKT_HDR *pkt_hdr)
{
    char buf[PKT_HDR_SIZE], *pbuf = buf;
    dword len;
    
    len = read(handle, &buf, PKT_HDR_SIZE);
    
    if (PKT_HDR_SIZE != len) {
	return len;
    }
	
    pkt_hdr->fromNode = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->toNode = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->year = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->month = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->day = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->hour = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->minute = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->second = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->baud = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->pktType = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->fromNet = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->toNet = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->prodCodeL = *(pbuf++);
    pkt_hdr->verMajor =  *(pbuf++);
    memmove(&(pkt_hdr->password), pbuf, 8);
    pbuf += 8;
    pkt_hdr->QfromZone = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->QtoZone = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->AuxNet = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->CapWordS = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->prodCodeH = *(pbuf++);
    pkt_hdr->verMinor = *(pbuf++);
    pkt_hdr->CapWord = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->fromZone = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->toZone = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->fromPoint = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->toPoint = get_word(pbuf);
    pbuf += 2;
    pkt_hdr->prodData = get_dword(pbuf);
    pbuf += 4;

    assert (pbuf - buf == PKT_HDR_SIZE);
    
    return len;
}

dword WritePktHdr(int handle, PKT_HDR *pkt_hdr)
{
    char buf[PKT_HDR_SIZE], *pbuf = buf;
	    
    put_word(pbuf, pkt_hdr->fromNode);
    pbuf += 2;
    
    put_word(pbuf, pkt_hdr->toNode);
    pbuf += 2;
    
    put_word(pbuf, pkt_hdr->year);
    pbuf += 2;
    
    put_word(pbuf, pkt_hdr->month);
    pbuf += 2;
    
    put_word(pbuf, pkt_hdr->day);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->hour);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->minute);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->second);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->baud);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->pktType);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->fromNet);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->toNet);
    pbuf += 2;

    *(pbuf++) = pkt_hdr->prodCodeL;
    *(pbuf++) = pkt_hdr->verMajor;

    memmove(pbuf, pkt_hdr->password, 8);
    pbuf += 8;

    put_word(pbuf, pkt_hdr->QfromZone);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->QtoZone);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->AuxNet);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->CapWordS);
    pbuf += 2;

    *(pbuf++) = pkt_hdr->prodCodeH;
    *(pbuf++) = pkt_hdr->verMinor;

    put_word(pbuf, pkt_hdr->CapWord);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->fromZone);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->toZone);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->fromPoint);
    pbuf += 2;

    put_word(pbuf, pkt_hdr->toPoint);
    pbuf += 2;

    put_dword(pbuf, pkt_hdr->prodData);
    pbuf += 4;

    assert (pbuf - buf == PKT_HDR_SIZE);
	    
    return write(handle, &buf, PKT_HDR_SIZE);
}


dword WritePktMsgHdr(int handle, struct pkt_msg_hdr *pkt_msg_hdr)
{

    char buf[PKT_MSG_HDR_SIZE], *pbuf=buf;
	    
    put_word(pbuf, pkt_msg_hdr->type);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->fromNode);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->toNode);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->fromNet);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->toNet);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->attr);
    pbuf += 2;
    put_word(pbuf, pkt_msg_hdr->cost);
    pbuf += 2;
    memmove(pbuf, &(pkt_msg_hdr->DateTime), 20);
    pbuf += 20;
    
    assert(pbuf - buf == PKT_MSG_HDR_SIZE);
    
    return write(handle, &buf, PKT_MSG_HDR_SIZE);
}
