
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "regex.h"
#define FROM 0
#define TO 1
#define Direct 0
#define Block 1


static XS(perl_SelectFrom)
{
    dXSARGS;
    STRLEN n_a;
    char *s, *err="perl_SelectTo()";
    BOOL rc = 0;
	
    if (items != 1) {
	fprintf(stderr, "%s argument count (need 1, exist %d)\n", err, (int)items);
	abort();
    }

    sp_s = FromSP(0);	
    s = XCAL;
    strcpy(s, sp_s);
    strip(s);
    expand_str(s);
    rc = do_select( s , err , FROM );
    free(s);
    /*///////////////////////////////////////////*/

    XSRETURN_IV(rc);	
}
/*****************************************************************************/
static XS(perl_SelectTo)
{
    dXSARGS;
    STRLEN n_a;
    char *s, *err="perl_SelectTo()";
    BOOL rc = 0;
	
    if (items != 1) {
	fprintf(stderr, "%s argument count (need 1, exist %d)", err, (int)items);
	abort();
    }

    sp_s = FromSP(0);
    s = XCAL;
    strcpy(s, sp_s);
    strip(s);
    expand_str(s);
    rc = do_select( s , err , TO );
    free(s);
/*///////////////////////////////////////////*/
    XSRETURN_IV(rc);
}
/**************************************************************************/

static XS(perl_Select)
{
    dXSARGS;
    STRLEN n_a;
    char *s, *err="perl_Select()";
    BOOL rc = 0;
	
    if (items != 1) {
	fprintf(stderr, "%s argument count (need 1, exist %d)", err, (int)items);
	abort();
    }
	
    sp_s = FromSP(0);
    s = XCAL;
    strcpy(s, sp_s);
    strip(s);
    expand_str(s);
	
    rc = do_select( s , err , FROM );
	
    if (!rc) {
	strcpy(s, sp_s);
	s[n_a] = '\0';
	strip(s);
	expand_str(s);
	rc = do_select( s , err , TO   );
    }
    free(s);
/*///////////////////////////////////////////*/
    XSRETURN_IV(rc);
}

/****************************************************************************/
static BOOL do_select(char *tag, char *err_diag, BOOL fromTo)
{
	struct Tags *Tag = NULL;
	int type = -1;
	char buf[256], token[32]={0}, *p, savestr[256];
	BOOL mode = Block, relaxed = No, direct = No, negate = No, rc = 0;

	strcpy(savestr, tag);
	sprintf(buf, "select: initial string to select:`%s'\n", savestr);
	writeToLog(5, buf);
	
	expand_str( tag );
	sprintf(buf, "select: string to select after expanding:`%s'\n", tag);
	writeToLog(5, buf);

	strip( tag );
	sprintf(buf, "select: string to select after stripping blanks:`%s'\n", tag);
	writeToLog(4, buf);
	
	while( *tag && (*tag=='!' || *tag=='~' || *tag=='@') )
	{
		if (*tag == '!') { /* negate data */
			++tag;
			strip(tag);
			if (negate) {
				fprintf(stderr, "%s syntax: `%s'\n",  err_diag, savestr);
				abort();
			}
			negate = Yes;
			sprintf(buf, "select: do negative comparing for string:`%s'\n", tag);
			writeToLog(4, buf);
		}
		if (*tag == '~') {
			++tag;
			strip(tag);
			if (relaxed) {
				fprintf(stderr, "%s syntax: `%s'\n",  err_diag, savestr);
				abort();
			}
			relaxed = Yes;
			sprintf(buf, "select: do relaxed comparing for string:`%s'\n", tag);
			writeToLog(4, buf);
		}
		if (*tag == '@') { /* direct data */
			++tag;
			strip(tag);
			if (direct) {
				fprintf(stderr, "%s syntax: `%s'\n",  err_diag, savestr);
				abort();
			}
			direct = Yes;
			sprintf(buf, "select: do direct comparing for string:`%s'\n", tag);
			writeToLog(4, buf);
		}
	} /* while !~@ */

	if (!strlen(tag)) {
		fprintf(stderr, "%s argument missing: `%s'\n", err_diag, savestr);
		abort();
	}
		
	if (direct) {
		
		p = token;
		while( *tag && *tag!=' ' && *tag!='\t' ) /* skip to blank */
		{
			*p++ = *tag++;
		}
		*p = '\0';
		
		strip(tag);
		
		if (!strlen(tag)) {
			fprintf(stderr, "%s argument missing: `%s'\n", err_diag, savestr);
			abort();
		}
		
		sprintf(buf, "select: direct comparing with token:`%s' and data:`%s'\n",
						token, tag);
		writeToLog(5, buf);
		
		remove_quotes( tag );
		sprintf(buf, "select: string to select after removing quotes:`%s'\n", tag);
		writeToLog(4, buf);

		type = getblocktype(token);
		
		if (type == -1) {
			fprintf(stderr, "%s incorrect data type `%s' in string: `%s'\n", 
							err_diag, token, savestr);
			abort();
		}
		
		switch (type) {
			
		case ADDRESS:          break;
		case NAME:             break;
		case ADDRESSNAME:      break;
		case ATTRIBUTE:        break;
		case SUBJECT:          break;
		case TEXTSTRING:       break;
		case FILES:            break;
		case MESSAGEAGE:       break;
		case PROCESSAGE:       break;
		case FILEAGE:          break;
		case PASSWORDNAME:     break;
		case PASSWORDADDRESS:  break;
		case MESSAGELENGTH:    break;
		case DATE_ :            break;
		case DAYOFWEEK :       break;
		case KLUDGE :          break;
		case ECHONAME :        break;
		
			
		default:
			fprintf(stderr, "%s no direct data select possible with this token:`%s'\n",
							err_diag, token);
			abort();
			
			/*NOT REACHED */
			break;
		}
		mode = Direct;
	}
	else { /* block data */
		
		sprintf(buf, "select: do block comparing for string:`%s'\n", tag);
		writeToLog(4, buf);
	
		for(Tag = TagFirst; Tag; Tag = Tag->next)
		{
			if (!stricmp(Tag->tag, tag)) {
				type = Tag->i;
				break;
			}
		}
		
		mode = Block;
	}
	
	
	if ( !CurMsg.n         &&
			type != FILES     &&
			type != FILEAGE   &&
			type != DATE_      &&
			type != DAYOFWEEK    )
	{
		sprintf(buf, "select: no active message in memory\n");
		writeToLog(3, buf);
		return Fail;
	}
	
	switch (type) {
		
	/* (mode = block|direct) */

	case    NODELIST :  rc = select_nodelist( tag , fromTo );
		break;
	case    VERSION9 :  rc = select_v9_nodelist( tag , fromTo );
		break;
	case    ADDRESS :   rc = select_address( tag , fromTo , mode , relaxed ); 
		break;
	case    NAME :      rc = select_name( tag , fromTo , mode , relaxed); 
		break;
	case    ADDRESSNAME:rc = select_addrname( tag , fromTo , mode , relaxed); 
		break;
	case    SUBJECT :   rc = select_subject( tag , mode , relaxed); 
		break;
	case    TEXTSTRING :rc = select_textstr( tag , mode , relaxed); 
		break;
	case    ATTRIBUTE : rc = select_attribute( tag , mode ); 
		break;
	case    ROUTING :   rc = select_routing( tag , relaxed ); 
		break;
	case    SYSTEM :    rc = select_system( tag , fromTo , relaxed ); 
		break;
	case    READDRESSLIST : rc = select_readdresslist( tag , fromTo , relaxed ); 
		break;
	case    FILES :     rc = select_files( tag , mode ); 
		break;
	case    MESSAGEAGE : rc = select_msgage( tag , fromTo, mode);
		break;
	case    PROCESSAGE : rc = select_prage( tag , fromTo, mode);
		break;
	case    FILEAGE : rc = select_fage( tag , mode);
		break;
	case    PASSWORDNAME :  rc = select_passname(tag , fromTo , mode , relaxed );
		break;
	case    PASSWORDADDRESS : rc = select_passaddr(tag, fromTo, mode, relaxed );
		break;
	case    MESSAGELENGTH : rc = select_msglen(tag, fromTo, mode, relaxed );
		break;
	case    DATE_ : rc = select_date(tag , mode);
		break;
	case    DAYOFWEEK : rc = select_dow(tag , mode);
		break;
	case    KLUDGE : rc = select_kludge( tag , mode , relaxed );
		break;
	case    ECHONAME : rc = select_ename( tag , mode , relaxed );
		break;
		
	case    NODELIST4D :
	case    RENAME :
	case    READDRESSNAME :
	case    FDSEMAPHOR :
	case    BINKSEMAPHOR :
	case    ADDRESSPACKER :
	case    LENGTHADDRESS :
	case    PATH :
	case    ALIAS :
	case    FILEAREAS :
	case    ADDRESSATTR :
	case    STRINGS :
	case    TIMESLOT :
	case    DATEADDRESS :
	case    EVENT :
	case    MACRO :
	case    SUBJECTADDRESS :
	case    CONDITION :

		/* not yet implemented */
		break;
		
	default:
		
		fprintf(stderr, "%s block undefined : `%s'\n", err_diag, tag);
		abort();
	}
	
	if (negate) {
		rc = rc ? Fail : Ok;
	}
	
	sprintf(buf, "select: rc = %d\n", rc);
	writeToLog(4, buf);
	
	return rc;
}

/************************************************************/
static BOOL select_address( char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct AddrTags *AddrTag = NULL;
	ADDR addr, targ;
	BOOL found = 0;
	char buf[256], caddr1[64], caddr2[64];
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_address error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (mode == Direct) {
		
		if ( !str_to_Waddr(tag, &targ) ) {
			fprintf(stderr, "incorrect address in `SelectAddr': `%s'\n", tag);
			abort();
		}

		if (relaxed) {
			targ.point = Wild_1;
		}
		
		if ( comp_addrs(addr, targ) ) {
			return Ok;
		}
		else {
			return Fail;
		}
	}
	/*////////////////////////////////////////////////////////////*/
	for( AddrTag = AddrTagFirst; AddrTag; AddrTag = AddrTag->next )
	{
		if (!stricmp(tag, AddrTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_address: no such <Data Address> : `%s'\n", tag);
		abort();
		
	}
		
	AddrTag->rec = AddrTag->recfirst;
	for(; AddrTag->rec; AddrTag->rec = AddrTag->rec->next )
	{
		targ = AddrTag->rec->addr;
		if (relaxed) {
			targ.point = Wild_1;
		}
		addr_to_str( addr , caddr1 );
		addr_to_str( targ , caddr2 );
		sprintf( buf , "select address: compare `%s' in msg and `%s' in block\n", caddr1, caddr2 );
		writeToLog(4, buf);
		if ( comp_addrs(addr, targ) ) {
			return Ok;
		}
	}
		
	return Fail;
}
/************************************************************/
static BOOL select_readdresslist(char *tag, BOOL fromTo, BOOL relaxed)
{
	struct ReAddrTags *ReAddrTag = NULL;
	ADDR addr, targ;
	BOOL found = No;
	char buf[256], caddr1[64], caddr2[64];
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_address error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	/*////////////////////////////////////////////////////////////*/
	for( ReAddrTag = ReAddrTagFirst; ReAddrTag; ReAddrTag = ReAddrTag->next )
	{
		if (!stricmp(tag, ReAddrTag->tag)) {
			found = Yes;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_readdresslist: no such <Data ReAddress> : `%s'\n", tag);
		abort();
		
	}
		
	ReAddrTag->rec = ReAddrTag->recfirst;
	for(; ReAddrTag->rec; ReAddrTag->rec = ReAddrTag->rec->next )
	{
		targ = ReAddrTag->rec->addr1;
		if (relaxed) {
			targ.point = Wild_1;
		}
		addr_to_str( addr , caddr1 );
		addr_to_str( targ , caddr2 );
		sprintf( buf , "select_readdresslist: compare `%s' in msg and `%s' in block \n", caddr1, caddr2 );
		writeToLog(4, buf);
		if ( comp_addrs(addr, targ) ) {
			return Ok;
		}
	}
		
	return Fail;
}
/************************************************************/
static BOOL select_system( char *system , BOOL fromTo , BOOL relaxed)
{
	struct SystemTags *SystemTag = NULL;
	ADDR addr, targ;
	BOOL found = 0;
	char buf[256], caddr1[64], caddr2[64];
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_system error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
	{
		if (!stricmp(system, SystemTag->tag)) {
			found = 1;
			break;
		}
	}
	
	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_system: no such <Data System> : `%s'\n", system);
		abort();
		
	}
	
	SystemTag->rec = SystemTag->recfirst;
	for( ; SystemTag->rec ; SystemTag->rec = SystemTag->rec->next)
	{
		targ = SystemTag->rec->addr;
		if (relaxed) {
			targ.point = Wild_1;
		}
		addr_to_str( addr , caddr1 );
		addr_to_str( targ , caddr2 );
		sprintf( buf , "select_system: compare `%s' in msg and `%s' in block \n", caddr1, caddr2 );
		writeToLog(4, buf);
		if (comp_addrs(addr, targ)) {
			return Ok;
		}
	} /* sys loop */

	return Fail;

}
/************************************************************/
static BOOL select_name( char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct NameTags *NameTag = NULL;
	char name1[36], name2[36], buf[256];
	BOOL found = 0;
	
	if (fromTo == TO) {
		strcpy( name1, ftn.toName);
	}
	else {

		if (fromTo == FROM) {
			strcpy( name1, ftn.fromName);
		}
		else {
			writeToLog(3, "select_name error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (mode == Direct) {
		
		strncpy( name2, tag, 35 ); name2[35] = 0;
		
		if ( comp_strings(name1, name2, relaxed) ) {
			return Ok;
		}
		else {
			return Fail;
		}
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( NameTag = NameTagFirst; NameTag; NameTag = NameTag->next )
	{
		if (!stricmp(tag, NameTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_name: no such <Data Name> : `%s'\n", tag);
		abort();
		
	}
		
	NameTag->rec = NameTag->recfirst;
	for(; NameTag->rec; NameTag->rec = NameTag->rec->next )
	{
		if (NameTag->rec->preg) {
			if (0 == regexec (NameTag->rec->preg, name1, 0, NULL, 0))
				return Ok;
		}
		else {
			strcpy(name2 , NameTag->rec->name);
			expand_str(name2);
			sprintf( buf , "select_name: compare `%s' in msg and `%s' in block \n", name1, name2 );
			writeToLog(4, buf);
			if ( comp_strings(name1, name2, relaxed) ) {
				return Ok;
			}
		}
	}
		
	return Fail;
	
}
/************************************************************/
static BOOL select_addrname( char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct AddrNameTags *AddrNameTag = NULL;
	char name1[36], name2[36], caddr[128];
	ADDR addr1, addr2;
	BOOL found = 0;
	
	if (fromTo == TO) {
		strcpy( name1, ftn.toName);
		addr1 = ftn.to;
	}
	
	else {

		if (fromTo == FROM) {
			strcpy( name1, ftn.fromName);
			addr1 = ftn.from;
		}
		else {
			writeToLog(3, "select_addrname error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (mode == Direct) {
		
		get_next_word(tag, caddr);
		if (!str_to_addr(caddr, &addr2)) {
			fprintf(stderr, "select_addrname error: incorrect address in Data AddressName : `%s' (%s)\n",
							caddr, tag);
			abort();
		}
		if (!strlen(tag)) {
			fprintf(stderr, "select_addrname error: incorrect name in Data AddressName : `%s'\n", tag);
			abort();
		}
		
		strncpy( name2, tag, 35 ); name2[35] = 0;

		if (relaxed)
			addr2.point = Wild_1;
		
		if (comp_strings(name1, name2, relaxed) && comp_addrs(addr1, addr2)) {
			return Ok;
		}
		else {
			return Fail;
		}
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( AddrNameTag = AddrNameTagFirst; AddrNameTag; AddrNameTag = AddrNameTag->next )
	{
		if (!stricmp(tag, AddrNameTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_addrname: no such <Data AddressName> : `%s'\n", tag);
		abort();
		
	}
		
	AddrNameTag->rec = AddrNameTag->recfirst;
	for(; AddrNameTag->rec; AddrNameTag->rec = AddrNameTag->rec->next )
	{
		addr2 = AddrNameTag->rec->addr;
		if (relaxed)
			addr2.point = Wild_1;

		if (comp_addrs(addr1, addr2)) {

			if (AddrNameTag->rec->preg) {
				if (0 == regexec (AddrNameTag->rec->preg, name1, 0, NULL, 0))
					return Ok;
			}
			else {
		
				strcpy( name2 , AddrNameTag->rec->name );
				expand_str(name2);
				if (comp_strings(name1, name2, relaxed))
					return Ok;
			}
			
		} /* addrs ok */
	}
		
	return Fail;
	
}
/************************************************************/
static BOOL select_subject( char *tag , BOOL mode , BOOL relaxed)
{
	struct SubjectTags *SubjectTag = NULL;
	char subj1[72], subj2[72], buf[256];
	BOOL found = 0;
	
	strcpy( subj1, ftn.subject);
	
	if (mode == Direct) {
		
		strncpy( subj2, tag, 72 ); subj2[71] = 0;
		
		
		if ( comp_strings(subj1, subj2, relaxed) ) {
			return Ok;
		}
		else {
			return Fail;
		}
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( SubjectTag = SubjectTagFirst; SubjectTag; SubjectTag = SubjectTag->next )
	{
		if (!stricmp(tag, SubjectTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_subject: no such <Data Subject> : `%s'\n", tag);
		abort();
		
	}
		
	SubjectTag->rec = SubjectTag->recfirst;
	for(; SubjectTag->rec; SubjectTag->rec = SubjectTag->rec->next )
	{
		if (SubjectTag->rec->preg) {
			if (0 == regexec (SubjectTag->rec->preg, subj1, 0, NULL, 0))
				return Ok;
		}
		else {
			strcpy( subj2 , SubjectTag->rec->subject);
			expand_str(subj2);
			sprintf( buf , "select_subject: compare `%s' in msg and `%s' in block \n",
							subj1, subj2 );
			writeToLog(4, buf);
			if ( comp_strings(subj1, subj2, relaxed) ) {
				return Ok;
			}
		}
	}
		
	return Fail;
	
}
/************************************************************/
static BOOL select_textstr( char *tag , BOOL mode , BOOL relaxed)
{
	struct TextStrTags *TextStrTag = NULL;
	char tstr[256];
	BOOL found = 0;
	
	if (mode == Direct) {
		
		strncpy( tstr, tag, 255 ); tstr[255] = 0;
		
		if ( search_str(tstr, relaxed) ) {
			return Ok;
		}
		else {
			return Fail;
		}
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( TextStrTag = TextStrTagFirst; TextStrTag; TextStrTag = TextStrTag->next )
	{
		if (!stricmp(tag, TextStrTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_textstr error: no such <Data TextString> : `%s'\n", tag);
		abort();
		
	}
		
	TextStrTag->rec = TextStrTag->recfirst;
	for(; TextStrTag->rec; TextStrTag->rec = TextStrTag->rec->next )
	{
		if (TextStrTag->rec->preg) {
			if ( rx_search_str(TextStrTag->rec->preg) )
				return Ok;
		}
		else {
			expand_str(tstr);
			strcpy(tstr , TextStrTag->rec->str);
			if ( search_str(tstr, relaxed) )
				return Ok;
		}
	}
		
	return Fail;
	
}
/************************************************************/
static BOOL select_nodelist( char *tag , BOOL fromTo )
{
	struct ndlTags *ndlTag = NULL;
	ADDR addr;
	int found = 0;
	word j1, j2;
	
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
	{
		if (!stricmp(tag, ndlTag->tag)) {
			found = Yes;
			break;
		}
	}

	if (!found) {
		fprintf(stderr, "select_nodelist: no such <Data Nodelist> : `%s'\n", tag);
		abort();
	}
		
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_nodelist error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}

	return is_listed(ndlTag, addr, &j1, &j2);

}

/************************************************************/
static BOOL select_v9_nodelist( char *tag , BOOL fromTo )
{
	struct ndlTags *v9_ndlTag = NULL;
	ADDR addr;
	int found = 0;
	int j1, j2, j3;
	
	for( v9_ndlTag = v9_ndlTagFirst; v9_ndlTag; v9_ndlTag = v9_ndlTag->next )
	{
		if (!stricmp(tag, v9_ndlTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) {
		fprintf(stderr, "select_v9: no such <Data Version9> : `%s'\n", tag);
		abort();
	}
		
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_v9 error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}

	return is_v9_listed(v9_ndlTag, addr, &j1, &j2, &j3);

}


/************************************************************/
static BOOL select_routing( char *tag , BOOL relaxed)
{
	struct RoutingTags *RoutingTag = NULL;
	BOOL found = 0;
	ADDR addr, targ;

	for(RoutingTag = RoutingTagFirst; RoutingTag; RoutingTag = RoutingTag->next )
	{
		if ( !stricmp(RoutingTag->tag, tag) ) {
			found = 1;
			break;
		}
	}
	
	if (!found) {
		fprintf(stderr, "select_routing: no such <Data Routing> : `%s'\n", tag);
		abort();
	}
		
	addr = ftn.to;
	
	found = 0;
	
	RoutingTag->rec = RoutingTag->recfirst;
	for(; RoutingTag->rec; RoutingTag->rec = RoutingTag->rec->next )
	{
		RoutingTag->rec->targlist = RoutingTag->rec->targfirst;
		while(RoutingTag->rec->targlist)
		{
			/*///////////////////////////////////*/
			targ = RoutingTag->rec->targlist->targ;
			/*///////////////////////////////////*/
			if (relaxed) {
				targ.point = Wild_1;
			}
			
			/* do address matching */
			if ( comp_addrs(addr, targ) ) {
				return Ok;
			}
			
			RoutingTag->rec->targlist = RoutingTag->rec->targlist->next;
		}
	} /* RoutingTag->rec */

	return Fail;
}
/************************************************************/
static BOOL select_attribute( char *tag , BOOL mode )
{
	struct AttrTags *AttrTag = NULL;
	BOOL found = 0, true = 0;
	char cattr[256], *result[20];
	int i, k;
	AllAttr aa;

	aa.attr = ftn.attr;
	aa.flag = ftn.flag;
	aa.stat = ftn.stat;

	if (mode == Direct) {
		
		for( i = 0; i < 20; ++i )
		{
			result[i] = (char *)calloc(256, sizeof(char));
		}

		k = splitf( tag , result );
		k = (k>20) ? 20 : k ;

		for( i = 0; i < k; ++i )
		{
			strcpy( cattr , result[i] );
			true = check_all_attrs( cattr , aa );
			if (!true) {
				break;
			}
		} 

		for( i = 0; i < 20; ++i )
		{
			free( result[i] );
		}
		
		return true;

	}  /* end of direct data */
	
	/*///////////////////////////////////////////////////////////*/
	for( AttrTag = AttrTagFirst; AttrTag; AttrTag = AttrTag->next )
	{
		if (!stricmp(tag, AttrTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) {
		fprintf(stderr, "select_attr: no such <Data Attribute> : `%s'\n", tag);
		abort();
	}
	
	AttrTag->rec = AttrTag->recfirst;
	for(; AttrTag->rec; AttrTag->rec = AttrTag->rec->next )
	{
		AttrTag->rec->attr = AttrTag->rec->attrfirst;
		for( ; AttrTag->rec->attr; AttrTag->rec->attr = AttrTag->rec->attr->next )
		{
			strcpy( cattr , AttrTag->rec->attr->cattr );
			true = check_all_attrs( cattr , aa );
			if (!true) {
				break;
			}
		} /* list loop */
		
		if (true) {
			return Ok;
		}
		
	} /* list loop */

	return Fail;
}
/************************************************************/
static BOOL select_files( char *tag , BOOL mode )
{
	struct FilesTags *FilesTag = NULL;
	BOOL rc = Fail;
	char fullname[256], fname[256], buf[256], fdir[256];
	BOOL fDir = Yes;

	if (mode == Direct) {

		get_next_word(tag, fname);
		if (*fname == '#') {
			get_next_word(tag, fname);
			fDir = Yes;
			if (strlen(fname) > 1) {
				strcpy(fdir, &fname[1]);
				if (!setpath(fdir)) {
					sprintf(buf, "select_files error: path doesn't exist : `%s'\n", fdir);
					writeToLog(0, buf);
					return Fail;
				}
				strcpy(fdir, FlagDir);
			}
		
			memset(fname, 0, 256);
			get_next_word(tag, fname);
		}
	
		if (!strlen(fname)) {
			fprintf(stderr, "select_files error: no filename defined\n");
			abort();
		}

		do  {
			if (fDir == Yes)
				sprintf(fullname, "%s%s", fdir, fname);
			else
				strcpy(fullname, fname);

			if (Fail == fsearch(fname))
				return Fail;
		} while (get_next_word(tag, fname));

		return Ok;

	}  /* end of direct data */
	
	/*//////////////////////////////////////////////////////////////*/
	for( FilesTag = FilesTagFirst; FilesTag; FilesTag = FilesTag->next )
	{
		if (!stricmp(tag, FilesTag->tag)) {
			break;
		}
	}

	if (NULL == FilesTag) {
		fprintf(stderr, "select_files error: no such <Data Files> : `%s'\n", tag);
		abort();
	}
	
	FilesTag->rec = FilesTag->recfirst;
	for(; FilesTag->rec; FilesTag->rec = FilesTag->rec->next )
	{
		FilesTag->rec->file = FilesTag->rec->filefirst;
		for( ; FilesTag->rec->file; FilesTag->rec->file = FilesTag->rec->file->next )
		{
			strcpy(fname, FilesTag->rec->file->fname);
			expand_str(fname);
			if (FilesTag->rec->file->preg)
				rc = rx_fsearch(fname, FilesTag->rec->file->path, FilesTag->rec->file->preg);
			else
				rc = fsearch(fname);
			printf("fsearch rc=%d\n", rc);
			if (rc == Fail)
				break;
		} /* list loop */
		
		if (rc == Ok) {
			return Ok;
		}
		
	} /* list loop */

	return Fail;
}
/************************************************************/
static BOOL select_msgage( char *tag , BOOL fromTo , BOOL mode)
{
	struct MsgAgeTags *MsgAgeTag = NULL;
	ADDR addr, targ;
	BOOL found = 0;
	char cage[10];
	time_t age_secs;
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_msgage error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (mode == Direct) {

		get_next_word(tag, cage);
		
		if ((dword)-1L == (age_secs = get_age(cage))) {
			fprintf(stderr, "select_msgage error: (direct data), incorrect age : `%s'", cage);
			abort();
		}
		
		targ.zone = targ.net = targ.node = targ.point = Wild_1;
		
		if (strlen(tag)) {
			if ( !str_to_Waddr(tag, &targ) ) {
				fprintf(stderr, "select_msgage error: (direct data) incorrect address : `%s'\n",
								tag);
				abort();
			}
		}

		return (ftn.utime < (time(NULL)-age_secs)) &&  comp_addrs(addr, targ);
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( MsgAgeTag = MsgAgeTagFirst; MsgAgeTag; MsgAgeTag = MsgAgeTag->next )
	{
		if (!stricmp(tag, MsgAgeTag->tag)) {
			found = Yes;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_msgage error: no such <Data MsgAge> : `%s'\n", tag);
		abort();
		
	}
		
	MsgAgeTag->rec = MsgAgeTag->recfirst;
	for(; MsgAgeTag->rec; MsgAgeTag->rec = MsgAgeTag->rec->next )
	{
		if ( ftn.utime < time(NULL)-MsgAgeTag->rec->secs  &&
				comp_addrs(addr, MsgAgeTag->rec->addr) )
			return Ok;
	}
		
	return Fail;
}
/************************************************************/
static BOOL select_prage( char *tag , BOOL fromTo , BOOL mode)
{
	struct ProcAgeTags *ProcAgeTag = NULL;
	ADDR addr, targ;
	BOOL found = 0;
	char cage[10];
	time_t age_secs;

	if ( ! (ftn.stat & i_Lastvia) || ftn.lvia_utime==0)
		return Fail;
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_prage error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (mode == Direct) {

		get_next_word(tag, cage);
		
		if ((dword)-1L == (age_secs = get_age(cage))) {
			fprintf(stderr, "select_msgage error: (direct data), incorrect age : `%s'", cage);
			abort();
		}
		
		targ.zone = targ.net = targ.node = targ.point = Wild_1;
		
		if (strlen(tag)) {
			if ( !str_to_Waddr(tag, &targ) ) {
				fprintf(stderr, "select_msgage error: (direct data) incorrect address : `%s'\n",
								tag);
				abort();
			}
		}

		return (ftn.lvia_utime < (time(NULL)-age_secs)) &&  comp_addrs(addr, targ);
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( ProcAgeTag = ProcAgeTagFirst; ProcAgeTag; ProcAgeTag = ProcAgeTag->next )
	{
		if (!stricmp(tag, ProcAgeTag->tag)) {
			found = Yes;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_prage error: no such <Data ProcessAge> : `%s'\n", tag);
		abort();
		
	}
		
	ProcAgeTag->rec = ProcAgeTag->recfirst;
	for(; ProcAgeTag->rec; ProcAgeTag->rec = ProcAgeTag->rec->next )
	{
		if (ftn.lvia_utime < time(NULL)-ProcAgeTag->rec->secs  &&
				comp_addrs(addr, ProcAgeTag->rec->addr) )
			return Ok;
	}
		
	return Fail;
}
/************************************************************/
static BOOL select_fage( char *tag , BOOL mode)
{
	struct FileAgeTags *FileAgeTag = NULL;
	BOOL found = 0, done;
	char buf[256], fname[256], cage[256]={0};
	time_t secs;
	struct find_t ffblk;
/*	FFIND *ff; */
	
	if (mode == Direct) {
		
		get_next_word(tag, cage);
		
		if ((dword)-1L == (secs = get_age(cage))) {
			fprintf(stderr, "select_fage error: (direct data), incorrect age : `%s'", cage);
			abort();
		}
		
		while (get_next_word(tag, fname))
		{
			if (!strcmp("#", fname)) {
				if (!get_next_word(tag, buf))
					return Fail;
				sprintf(fname, "%s%s", FlagDir, buf);
			}
			
			done = _dos_findfirst(fname, 0, &ffblk);
		
			while (!done) {

				if (fdate_to_unix(&ffblk.wr_time, &ffblk.wr_date) < 
						(time(NULL)-secs) )
				{
					sprintf(buf, "select_fage: `%s' is older than %lu days\n",
									ffblk.name, secs/(dword)SECS_IN_DAY);
					writeToLog(4, buf);
					return Ok;
				}
		
				done = _dos_findnext(&ffblk);
			}
			_dos_findclose(&ffblk);
		}  /* end of loop */
		
		return Fail;
		
	} /* direct */
	/*//////////////////////////////////////////////////////////////*/
	for( FileAgeTag = FileAgeTagFirst; FileAgeTag; FileAgeTag = FileAgeTag->next )
	{
		if (!stricmp(tag, FileAgeTag->tag)) {
			found = Yes;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_fage error: no such <Data FileAge> : `%s'\n", tag);
		abort();
		
	}
		
	FileAgeTag->rec = FileAgeTag->recfirst;
	for(; FileAgeTag->rec; FileAgeTag->rec = FileAgeTag->rec->next )
	{
		secs = FileAgeTag->rec->secs;
		FileAgeTag->rec->file = FileAgeTag->rec->filefirst;
		for(;
				FileAgeTag->rec->file; 
				FileAgeTag->rec->file = FileAgeTag->rec->file->next)
		{
			strcpy(fname, FileAgeTag->rec->file->fname);
			expand_str(fname);

			done = _dos_findfirst(fname, 0, &ffblk);
		
			while (!done) {

				if (fdate_to_unix(&ffblk.wr_time, &ffblk.wr_date) < 
						(time(NULL)-secs) )
				{
					sprintf(buf, "select_fage: `%s' is older than %lu days\n",
									ffblk.name, secs/(dword)SECS_IN_DAY);
					writeToLog(4, buf);
					return Ok;
				}
		
				done = _dos_findnext(&ffblk);
			}
			_dos_findclose(&ffblk);
		} /* next set of files */
	}
		
	return Fail;
	
}

/************************************************************/
static BOOL select_passaddr( char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct PassAddrTags *PassAddrTag = NULL;
	ADDR addr, targ;
	BOOL found = 0, pass_ok = Yes;
	char caddr[64], pass[36]={0};
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_passaddr error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	
	if (mode == Direct) {
		
		get_next_word(tag, pass);
		remove_quotes(pass);
		
		get_next_word(tag, caddr);
		
		if ( !str_to_Waddr(caddr, &targ) ) {
			fprintf(stderr, "select_passaddr error: incorrect address in `SelectPasswordAddr': `%s'\n", caddr);
			abort();
		}
		
		if (relaxed == No)
			pass_ok = (0 == stricmp(pass, ftn.subject));
		
		if (comp_addrs(addr, targ) && pass_ok) {
			return Ok;
		}
		else {
			return Fail;
		}
	}
	
	/*////////////////////////////////////////////////////////////*/
	for( PassAddrTag = PassAddrTagFirst; PassAddrTag; PassAddrTag = PassAddrTag->next )
	{
		if (!stricmp(tag, PassAddrTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_passaddr error: no such <Data PasswordAddress> : `%s'\n", tag);
		abort();
		
	}
		
	PassAddrTag->rec = PassAddrTag->recfirst;
	for(; PassAddrTag->rec; PassAddrTag->rec = PassAddrTag->rec->next )
	{
		if (relaxed == No)
			pass_ok = (0 == stricmp(PassAddrTag->rec->password, ftn.subject));
		if (comp_addrs(addr, PassAddrTag->rec->addr) && pass_ok)
			return Ok;
	}
		
	return Fail;
}

/************************************************************/
static BOOL select_passname( char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct PassNameTags *PassNameTag = NULL;
	BOOL found = 0, pass_ok = Yes, name_ok = Yes;
	char name[36], pass[36]={0};
	
	if (fromTo == TO) {
		strcpy(name, ftn.toName);
	}
	else {

		if (fromTo == FROM) {
			strcpy(name, ftn.fromName);
		}
		else {
			writeToLog(3, "select_passname error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	
	if (mode == Direct) {
		
		get_next_word(tag, pass);
		remove_quotes(pass);

		if (relaxed == No)
			pass_ok = (0 == stricmp(pass, ftn.subject));

		name_ok = (0 == stricmp(tag, name));

		if (pass_ok && name_ok)
			return Ok;
		else
			return Fail;
		
	}
	
	/*//////////////////////////////////////////////////////////////*/
	for( PassNameTag = PassNameTagFirst; PassNameTag; PassNameTag = PassNameTag->next )
	{
		if (!stricmp(tag, PassNameTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_passname error: no such <Data PasswordName> : `%s'\n", tag);
		abort();
		
	}
		
	PassNameTag->rec = PassNameTag->recfirst;
	for(; PassNameTag->rec; PassNameTag->rec = PassNameTag->rec->next )
	{
		if (relaxed == No)
			pass_ok = (0 == stricmp(PassNameTag->rec->password, ftn.subject));
		name_ok = (0 == stricmp(PassNameTag->rec->name, name));
		if (name_ok && pass_ok)
			return Ok;
	}
		
	return Fail;
}

/************************************************************/
static BOOL select_msglen(char *tag , BOOL fromTo , BOOL mode , BOOL relaxed)
{
	struct MsgLenTags *MsgLenTag = NULL;
	ADDR addr, targ;
	char buf[256], clen[256];
	dword len;
	BOOL len_ok = Yes;
	
	if (fromTo == TO) {
		addr = ftn.to;
	}
	else {

		if (fromTo == FROM) {
			addr = ftn.from;
		}
		else {
			writeToLog(3, "select_msglen error: fromTo is not FROM or TO\n");
			return Fail;
		}
	}
	
	if (!ftn.msglen)
		ftn.msglen = textlen(TextFirst);
	
	if (mode == Direct) {
		
		get_next_word(tag, clen);
		
		if (!__isnumber(clen)) {
			fprintf(stderr, "select_msglen error: incorrect length: %s", clen);
			abort();
		}

		len = atol(clen);
		
		targ.zone = targ.net = targ.node = targ.point = Wild_1;
		
		if (strlen(tag)) {
		
			if ( !str_to_Waddr(tag, &targ) ) {
				fprintf(stderr, "select_msglen error: incorrect address: `%s'\n", tag);
				abort();
			}
			
		}
		
		if (relaxed == No) {
			
			if (ftn.msglen > len)
				len_ok = Yes;
			else
				len_ok = No;
		}
		
		if (comp_addrs(addr, targ) && len_ok) {
			return Ok;
		}
		else {
			return Fail;
		}
		
	}
	
	/*////////////////////////////////////////////////////////////*/
	for( MsgLenTag = MsgLenTagFirst; MsgLenTag; MsgLenTag = MsgLenTag->next )
	{
		if (!stricmp(tag, MsgLenTag->tag)) {
			break;
		}
	}

	if (MsgLenTag == NULL) { /* hardly possible, but who knows ;-) */

		fprintf(stderr, "select_msglen error: no such <Data MessageLenth> : `%s'\n", tag);
		abort();

	}

	MsgLenTag->rec = MsgLenTag->recfirst;
	for(; MsgLenTag->rec; MsgLenTag->rec = MsgLenTag->rec->next )
	{
		addr_to_str(MsgLenTag->rec->addr, buf);
		
		if (relaxed == No) {
			if (ftn.msglen > MsgLenTag->rec->len)
				len_ok = Yes;
			else
				len_ok = No;
		}
		
		if (comp_addrs(addr, MsgLenTag->rec->addr) && len_ok) {
			return Ok;
		}
	}
		
	return Fail;
}

/************************************************************/
static BOOL select_date(char *tag , BOOL mode)
{
	struct DateTags *DateTag = NULL;
	char cdate[256]={0};
	int dow[7]={0};
	DD_MM_YY date;
	
	if (mode == Direct) {
		
		/* Date */
		get_next_word(tag, cdate);
	
		if (strlen(cdate) > 8 || !str_to_date(cdate, &date)) {
		
			fprintf(stderr, "select_date error: incorrect date: `%s'", cdate);
			abort();
		
		}
		
		/* dows */
		if (strlen(tag)) {
			if (0 == get_dows(tag, dow)) {
				fprintf(stderr, "select_date error: incorrect day of week: `%s'", tag);
				abort();
			}
		}

		return is_this_date(date, dow);
		
		
	} /* direct */
	/*///////////////////////////////////////////////////////////*/
	for( DateTag = DateTagFirst; DateTag; DateTag = DateTag->next )
	{
		if (!stricmp(tag, DateTag->tag)) {
			break;
		}
	}

	if (NULL == DateTag) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_date error: no such <Data Date> : `%s'\n", tag);
		abort();
		
	}
		
	DateTag->rec = DateTag->recfirst;
	for(; DateTag->rec; DateTag->rec = DateTag->rec->next )
	{
		if (is_this_date(DateTag->rec->date, DateTag->rec->dow))
			return Ok;
	}
		
	return Fail;
	
}

/************************************************************/
static BOOL select_dow(char *tag , BOOL mode)
{
	struct DowTags *DowTag = NULL;
	int dow[7]={0};
	
	if (mode == Direct) {
		
		if (0 == get_dows(tag, dow)) {
			fprintf(stderr, "select_dow error: incorrect day of week: `%s'", tag);
			abort();
		}

		return is_this_dow(dow);
		
		
	} /* direct */
	/*////////////////////////////////////////////////////////////*/
	for( DowTag = DowTagFirst; DowTag; DowTag = DowTag->next )
	{
		if (!stricmp(tag, DowTag->tag)) {
			break;
		}
	}

	if (NULL == DowTag) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_dow error: no such <Data DayOfWeek> : `%s'\n", tag);
		abort();
		
	}
		
	DowTag->rec = DowTag->recfirst;
	for(; DowTag->rec; DowTag->rec = DowTag->rec->next )
	{
		if (is_this_dow(DowTag->rec->dow))
			return Ok;
	}
		
	return Fail;
	
}

/************************************************************/
static BOOL select_kludge( char *tag , BOOL mode , BOOL relaxed)
{
	struct KludgeTags *KludgeTag = NULL;
	char ckl[256], buf[256], *pckl;
	BOOL set, rc, success = No;

	if (mode == Direct) {

		while (get_next_word(tag, buf))
		{
			pckl = buf;
			if ( *pckl == '!' ) {
				pckl++;
				set = No;
			}
			else
				set = Yes;
			
			sprintf(ckl, "\1%s", pckl);
			
			rc = is_kl_found(ckl, NULL, relaxed);

			if (rc != set)
				return Fail;
			
		}  /* end of loop */
		
		return Ok;

	}  /* end of direct data */
	
	/*////////////////////////////////////////////////////////////*/
	for( KludgeTag = KludgeTagFirst; KludgeTag; KludgeTag = KludgeTag->next )
		if (!stricmp(tag, KludgeTag->tag))
			break;

	if (!KludgeTag) {
		fprintf(stderr, "select_kludge error: no such <Data Kludge> : `%s'\n", tag);
		abort();
	}
	
	KludgeTag->rec = KludgeTag->recfirst;
	for(; KludgeTag->rec; KludgeTag->rec = KludgeTag->rec->next )
	{
		success = Yes;
		KludgeTag->rec->kl = KludgeTag->rec->klfirst;
		for( ; KludgeTag->rec->kl; KludgeTag->rec->kl = KludgeTag->rec->kl->next )
		{
			rc = is_kl_found(KludgeTag->rec->kl->kl,
					    KludgeTag->rec->kl->preg, relaxed);

			if (rc != KludgeTag->rec->kl->set) {
				success = No;
				break;
			}
		} /* list loop */
		
		if (success) {
			return Ok;
		}

	} /* list loop */

	return Fail;
}

/************************************************************/
static BOOL select_ename( char *tag , BOOL mode , BOOL relaxed)
{
	struct EchoNameTags *EchoNameTag = NULL;
	char *msg_name;
	char select_name[20];
	BOOL found = 0;

	if (NULL == (msg_name = get_echo_name()))
		return Fail;
	
	
	if (mode == Direct) {
		
		strncpy( select_name, tag, 19 ); select_name[19] = 0;
		
		if ( comp_strings(msg_name, select_name, relaxed) )
			return Ok;
		else
			return Fail;
		
		
	}
	/*////////////////////////////////////////////////////////////*/
	for( EchoNameTag = EchoNameTagFirst;
			EchoNameTag; EchoNameTag = EchoNameTag->next )
	{
		if (!stricmp(tag, EchoNameTag->tag)) {
			found = 1;
			break;
		}
	}

	if (!found) { /* hardly possible, but who knows ;-) */
		
		fprintf(stderr, "select_name: no such <Data EchoName> : `%s'\n", tag);
		abort();
		
	}
		
	EchoNameTag->rec = EchoNameTag->recfirst;
	for(; EchoNameTag->rec; EchoNameTag->rec = EchoNameTag->rec->next )
	{
		if (EchoNameTag->rec->preg) {
			if (0 == regexec (EchoNameTag->rec->preg, msg_name, 0, NULL, 0))
				return Ok;
		}
		else {
			strcpy(select_name , EchoNameTag->rec->name);
			expand_str(select_name);
			if ( comp_strings(msg_name, select_name, relaxed) ) {
				return Ok;
			}
		}
	}
		
	return Fail;
	
}



/************************************************************************/
static BOOL check_all_attrs( char *str , AllAttr ftn_aa)
{
	char buf[256], *p = str;
	BOOL setClear;
	word at;
	dword dat;
	
	if (*p == '!') {
		setClear = attrClear;
		++p;
	}
	else {
		setClear = attrSet;
	}
	
	at = str_to_attr(p);
	if ( at != minus_one ) {
		if (setClear == attrSet) {
			if ( ftn_aa.attr & at ) {
				return Ok;
			}
		}
		else {
			if ( !(ftn_aa.attr & at) ) {
				return Ok;
			}
		}
		return Fail;
	}

	at = str_to_flag(p);
	if ( at != minus_one ) {
		if (setClear == attrSet) {
			if ( ftn_aa.flag & at ) {
				return Ok;
			}
		}
		else {
			if ( !(ftn_aa.flag & at) ) {
				return Ok;
			}
		}
		return Fail;
	}

	dat = str_to_stat(p);
	if ( dat != minus_one_long ) {
		if (setClear == attrSet) {
			if ( ftn_aa.stat & dat ) {
				return Ok;
			}
		}
		else {
			if ( !(ftn_aa.stat & dat) ) {
				return Ok;
			}
		}
		return Fail;
	}

	sprintf( buf, "check_all_attr error: incorrect attribute : `%s'\n", p);
	writeToLog(0, buf);
	abort();

	/*NOT REACHED */
	return Fail;
}
/*****************************/
static BOOL fsearch(char *fname)
{
/*	struct find_t ffblk; */
	BOOL Pos = Yes;

	if (*fname == '!') {
		++fname;
		Pos = No;
	}
	
/*	return ((0 == _dos_findfirst(fname, 0, &ffblk)) == Pos); */
	return ((!access(fname, 0) == Ok) == Pos);
}
/************************************************************/
static BOOL rx_fsearch(char *fname, char *path, regex_t *preg)
{
	struct find_t ffblk;
	char s[256];
	int i;

	sprintf(s, "%s*", path);
	
	/* file should NOT exist */
	if (*fname == '!') {
		for(i = _dos_findfirst(s, 0, &ffblk); i == 0; i = _dos_findnext(&ffblk)) {
			if (0 == regexec (preg, ffblk.name, 0, NULL, 0)) {
				_dos_findclose(&ffblk);
				return Fail;
			}
		}
		_dos_findclose(&ffblk);
		return Ok;
	}
	else {
		for(i = _dos_findfirst(s, 0, &ffblk); i == 0; i = _dos_findnext(&ffblk)) {
			if (0 == regexec (preg, ffblk.name, 0, NULL, 0)) {
				_dos_findclose(&ffblk);
				return Ok;
			}
		}
		_dos_findclose(&ffblk);
		return Fail;
	}
}


/**************************************************************************/
BOOL comp_strings( char *str, char *substr, BOOL relaxed )
{
	if (relaxed) {
		char tmp_str[256], tmp_substr[256];
		
		if (!strlen(substr)) {
			return Ok;
		}
		strcpy(tmp_str, str);
		strcpy(tmp_substr, substr);
		
		strLower(tmp_str);
		strLower(tmp_substr);
		if ( strstr(tmp_str, tmp_substr) && strlen(tmp_str) ) {
			return Ok;
		}
		else {
			return Fail;
		}
	}
	else { /* strict */
		if ( !strcmp(str, substr) )
			return Ok;
		else
			return Fail;
	}
}

/**********************************************************************/
static BOOL search_str( char *substr, BOOL relaxed)
{
	TEXT *Text = NULL;
	char str[256];
	
	for(Text = TextFirst ; Text; Text = Text->next)
	{
		strcpy( str, Text->str);
		strip_CRLF(str);
		if ( comp_strings(str, substr, relaxed) ) {
			return Ok;
		}
		
	}
	
	return Fail;
	
}
/**********************************************************************/
static BOOL rx_search_str(const regex_t *preg)
{
	TEXT *Text = NULL;
	char str[256];

	
	for(Text = TextFirst ; Text; Text = Text->next)
	{
		strcpy( str, Text->str);
		strip_CRLF(str);
		if (0 == regexec(preg, str, 0, NULL, 0))
			return Ok;
	}
	
	return Fail;
	
}
/**************************************************************************/
BOOL comp_addrs( ADDR addr, ADDR targ )
/* compares 2 addrs, * allowed
   first - address from message
   second - address from search data
*/
{
	ADDR tmp_addr;
	char saddr[128], saddr2[128], buf[256];
	
	/* do address matching */
	if ( (addr.zone == targ.zone) || isW(targ.zone) ) {
		
		/* zone independent mask */
		if (targ.net == Wild_2)
			return is_zone_indep(addr);
		
		if ( (addr.net == targ.net) || isW(targ.net) ) {
			
			/* region independent mask */
			if (targ.node == Wild_2)
				return is_reg_indep(addr);
			
			if (targ.point == Wild_2) {
				tmp_addr.zone = isW(targ.zone) ? addr.zone : targ.zone;
				tmp_addr.net = isW(targ.net) ? addr.net : targ.net;
				tmp_addr.node = isW(targ.node) ? addr.node : targ.node;
				tmp_addr.point = 0;
				addr_to_str( addr , saddr );
				addr_to_str( tmp_addr , saddr2 );
				sprintf(buf, "comp_aadrs: check if node %s is in hub segment of %s\n",
								saddr, saddr2);
				writeToLog(5, buf);
				if (from_hub(addr, tmp_addr)) {
					writeToLog(5, "comp_aadrs: addrs match ok\n");
					return Ok;
				}
				else {
					writeToLog(5, "comp_aadrs: addrs don`t match\n");
				}
			} /* from hub segment */
			else {
				if ( (addr.node == targ.node) || isW(targ.node) ) {
					if ( (addr.point == targ.point) || targ.point == Wild_1 ) {
						return Ok;
					}  /* points match */
				}    /* node  match */
			}    /* not from hub segment */
		}   /* nets   match */
	}    /* zones  match */
	
	return Fail;
}


/*************************************************/
static BOOL is_this_date(DD_MM_YY date, int dow[])
{
	struct tm *tm;
	time_t t = time(NULL);
	
	tm = localtime(&t);

	if (  (date.day   == Wild_1 || date.day   == tm->tm_mday) &&
				(date.month == Wild_1 || date.month == tm->tm_mon)  &&
				(date.year  == Wild_1 || date.year  == tm->tm_year)  )
	{

		return is_this_dow(dow);
		
	}
	
	return Fail;
}

/*************************************************/
static BOOL is_this_dow(int dow[])
{
	BOOL rc = Ok;
	int i;
	struct tm *tm;
	time_t t = time(NULL);
	
	tm = localtime(&t);

	for (i = 0; i < 7; ++i)
	{
		if (dow[i] == 0)
			break;

		if (dow[i] == tm->tm_wday+1) {
			rc = Ok;
			break;
		}
		else
			rc = Fail;
	}

	return rc;
}

/*********************************************************************/
static BOOL is_kl_found(char *substr, regex_t *preg, BOOL relaxed)
{
	TEXT *Text = NULL;
	char str[256], ckl[256];

	if (!UseRegExp) {
		preg = NULL;
	}
	
	for(Text = TextFirst ; Text; Text = Text->next)
	{
		if (is_kludge(Text->str)) {
			
			strcpy( str, Text->str);
			get_next_word(str, ckl);

			if (!preg) {
				if ( comp_strings(ckl, substr, relaxed ) )
					return Ok;
			}
			else {
				if (0 == regexec (preg, ckl, 0, NULL, 0))
					return Ok;
			}
			
		} /* kludge found */
	}
	
	return Fail;
}
