typedef struct _pkthdr PKT_HDR;
typedef struct _fidomsg FIDO_MSG;
typedef struct _ftn FTN;

#define PKT_HDR_SIZE     58
#define PKT_MSG_HDR_SIZE 34
#define OMSG_SIZE 190


typedef struct {
  int day, month, year;
} DD_MM_YY;

struct _fidomsg {
  char fromName[36], toName[36], subject[72], DateTime[20];
  word TimesRead, toNode, fromNode, cost, fromNet, toNet;
  word toZone, fromZone, toPoint, fromPoint, replyTo, attr, nextReply;
};

struct _pkthdr {
  word fromNode, toNode, year, month, day, hour, minute;
  word second, baud, pktType, fromNet, toNet;
  char prodCodeL;   /* FTSC product code (low) */
  char verMajor;    /* product version (major) */
  char password[8];
  word QfromZone;
  word QtoZone;
  word AuxNet;      /* = fromNet if from point (fromNet == 0xffff)       */
  word CapWordS;    /* 28h capability word swapped                       */
  char prodCodeH;   /* 2Ah FTSC product code (high)                      */
  char verMinor;    /* 2Bh product version (minor)                       */
  word CapWord;     /* 2Ch capability word (0x0001 - fot packet type 2+) */
  word fromZone;    /* 2Eh                                               */
  word toZone;      /* 30h                                               */
  word fromPoint;   /* 32h                                               */
  word toPoint;     /* 34h                                               */
  dword prodData;   /* 36h product specific data (whatever)              */
};

struct pkt_msg_hdr {
    word type, fromNode, toNode, fromNet, toNet, attr, cost;
    char DateTime[20];
};

struct _ftn {
  ADDR to;
  ADDR from;
  char fromName[36];
  char toName[36];
  char subject[72];
  char DateTime[20];
  word attr;
  /**/
  word    flag;
  dword   stat; /* (internal attributes) */
  word    flavor;
  ADDR    routeTo;
  char    domain[128];
  char    intl[128];
  char    reply[128];
  char    msgid[128];
  int     loop;
  int     lines;
  dword   msglen;
	dword   utime;      /* local unix time (seconds since epoch) */
	dword   lvia_utime; /* if lastvia belongs to this system     */
  ADDR    echo_from;
};

typedef struct _fidoarea {
  
  FIDO_MSG  msgHdr; 
  struct Fnames *fnames;  /*  array of names              */
  int n;                  /*  current n                   */
  int total;              /*  total msg                   */
	dword ts;         /*  current timestamp threshold */
	
} FIDO_AREA;


typedef struct _sqarea {
  
  int   sqd;
  int   sqi;
  XMSG  xmsg;   
  int   n;           /* last read msg no (0 if just open) */
  int   num_msg;     /* the number of msgs in area */
  FOFS  cur_frame;   /* cur frame offset   */
  SQIDX *idxlist;
  int   total;       /* total msg after end (!=n in case we added some) */
  
  dword ts;          /* current timestamp threshold */
  
} SQAREA;


typedef struct _jamarea {
  
  int HdrHandle, TxtHandle, IdxHandle, LrdHandle;
  dword nMsgRead, curMsgRead, totalMsg, ts;
  JAMIDXREC Idx;
  JAMHDR Hdr;
  JAMHDRINFO HdrInfo;
  JAMMSGHDR MsgHdr;   /* toName, fromName, toAddr, fromAddr, subject */
  
} JAM_AREA;

typedef struct pktarea {
  
  struct    Pktnames *fnames;  /*  array of pkt names */
  int       icur;                 /*  current array element (pkt fname) */
  int       total;             /*  total pkt files in area */
  PKT_HDR   pkt_hdr; /* pkt hdr */
  struct    pkt_msg_hdr pkt_msg_hdr; /* pkt msg hdr */
	ADDR      to, from; /* pkt hdr */
  char      fromName[36], toName[36], subject[72];
  char      fname[256];      /* filename to read */
  char      new_fname[256];  /* filename to store changed mails */
  dword     off0; /* offset of the read msg */
  dword     off; /* current offset in the old_name */
  char      *text;
  dword     len; /* total pkt length */
  char      *new_buf;  /* buffer to store unmoved mails */
  dword     new_off;   /* new_buf pointer */
	int       n;  /* the number of the processed msg */
	BOOL      rinc;  /* true if ``i'' has been incremented by readmsg() */
  
} PKT_AREA;


struct pktcache {
  struct pktcache *next;
  ADDR to;
  ADDR from;
  char fname[13];
  char areapath[256];
  char flavor;        /* for temporary *.OIP packets   */
  char mode;          /* 'P' for oiP, 'F' for oiF      */
  dword len;          /* for packet-saw (pkt &arcmail) */
};
extern struct pktcache *PktCache, *PktCacheLast, *OipCache, *OipCacheLast;
extern struct pktcache *ArcCache, *ArcCacheLast;


struct CurMsg {
  int type;           /* FIDO, SQUISH, BINK, TXT, PKT, JAM */
  char areaname[20];
  char areapath[128];
  /**/
  FIDO_AREA fido;
  SQAREA squish; 
  PKT_AREA pkt;
  PKT_AREA bink;
  JAM_AREA jam;
	int n;        /* cur msg no (0 if no active msg)               */
	int realNo;   /* real msg number for *.msg and JAM area types  */
	int lrno;     /* needed to update lastreads when deleting msgs */
  dword maxts;        /* highest timestamp so far for current area     */
};
