#include "stamp.h"

#define MSGUID 0x00020000

#define SQHDR_SIZE  28
#define SQIDX_SIZE  12
#define SQBASE_SIZE 256

#define XMSG_FROM_SIZE  36
#define XMSG_TO_SIZE    36
#define XMSG_SUBJ_SIZE  72

#define XMSG_SIZE (94 + XMSG_FROM_SIZE + XMSG_TO_SIZE + XMSG_SUBJ_SIZE)

struct _sqhdr;
struct _sqidx;
struct _xmsg;
struct _netaddr;

typedef struct _sqidx SQIDX;
typedef struct _sqhdr SQHDR;
typedef struct _xmsg XMSG;

typedef struct _netaddr
{
  word zone;
  word net;
  word node;
  word point;
} NETADDR;

typedef dword FOFS;
typedef dword UMSGID;

struct _sqhdr
{
  #define SQHDRID       0xafae4453L

  dword id;             /* sqhdr.id must always equal SQHDRID */

  FOFS next_frame;      /* Next frame in the linked list */
  FOFS prev_frame;      /* Prior frame in the linked list */

  dword frame_length;   /* Length of this frame */
  dword msg_length;     /* Length used in this frame by XMSG, ctrl and text */
  dword clen;           /* Length used in this frame by ctrl info only */

  word frame_type;      /* Type of frame -- 0-normal 1-free 2-reserv 3-locked*/
  word rsvd;            /* Reserved for future use */
};



/* An individual index entry in <area>.SQI */

struct _sqidx
{
  FOFS ofs;                   /* Offset of the frame relating to this msg */
  UMSGID umsgid;              /* Unique message identifier of this msg */
  dword hash;                 /* SquishHash of msg.to for this msg */
};


/* Header block at offset 0 of <area>.SQD */

typedef struct _sqbase
{
  word len;               /* LENGTH OF THIS STRUCTURE! */           /*   0 */
  word rsvd1;             /* reserved */                            /*   2 */

  dword num_msg;          /* Number of messages in area */          /*   4 */
  dword high_msg;         /* Highest msg in area. Same as num_msg*/ /*   8 */
  dword skip_msg;         /* Skip killing first x msgs in area */   /*  12 */
  dword high_water;       /* Msg# (not umsgid) of HWM */            /*  16 */

  dword uid;              /* Number of the next UMSGID to use */    /*  20 */

  char base[80];          /* Base name of SquishFile */             /*  24 */

  FOFS begin_frame;       /* Offset of first frame in file */       /* 104 */
  FOFS last_frame;        /* Offset to last frame in file */        /* 108 */
  FOFS free_frame;        /* Offset of first FREE frame in file */  /* 112 */
  FOFS last_free_frame;   /* Offset of last free frame in file */   /* 116 */
  FOFS end_frame;         /* Pointer to end of file */              /* 120 */

  dword max_msg;          /* Max # of msgs to keep in area */       /* 124 */
  word keep_days;         /* Max age of msgs in area (SQPack) */    /* 128 */
  word sz_sqhdr;          /* sizeof(SQHDR) */                       /* 130 */
  char rsvd2[124];        /* Reserved by Squish for future use*/    /* 132 */
                                                             /* total: 256 */
} SQBASE;

struct _xmsg
{
  dword attr;

  char from[XMSG_FROM_SIZE];
  char to[XMSG_TO_SIZE];
  char subj[XMSG_SUBJ_SIZE];

  NETADDR orig;        /* Origination and destination addresses             */
  NETADDR dest;

  struct _stamp date_written;   /* When user wrote the msg (UTC)            */
  struct _stamp date_arrived;   /* When msg arrived on-line (UTC)           */
  sword utc_ofs;                /* Offset from UTC of message writer, in    *
                                 * minutes.                                 */

#define MAX_REPLY 9            /* Max number of stored replies to one msg */

  UMSGID replyto;               /* This is a reply to message #x            */
  UMSGID replies[MAX_REPLY];            /* Replies to this message                  */
  dword umsgid;                 /* UMSGID of this message, if (attr&MSGUID) */
                                /* This field is only stored on disk -- it  *
                                 * is not read into memory.                 */

  char __ftsc_date[20];/* Obsolete date information.  If it weren't for the *
                        * fact that FTSC standards say that one cannot      *
                        * modify an in-transit message, I'd be VERY         *
                        * tempted to axe this field entirely, and recreate  *
                        * an FTSC-compatible date field using               *
                        * the information in 'date_written' upon            *
                        * export.  Nobody should use this field, except     *
                        * possibly for tossers and scanners.  All others    *
                        * should use one of the two binary datestamps,      *
                        * above.                                            */
};

enum { SQ_NOERR = 0, 
  LOW_MEMO,                   /*  1    */
  SQBASE_INVALID ,            /*  2    */
  SQD_OPEN_ERR ,              /*  3    */
  SQD_SEEK_ERR ,              /*  4    */
  SQD_READ_ERR ,              /*  5    */
  SQD_WRITE_ERR ,             /*  6    */
  SQD_INVALID_ID ,            /*  7    */
  SQD_INVALID_FRAME_TYPE,     /*  8    */
  SQD_INVALID_NUM_MSG ,       /*  9    */
  SQI_OPEN_ERR ,              /*  10   */
  SQI_SEEK_ERR ,              /*  11   */
  SQI_READ_ERR ,              /*  12   */
  SQI_WRITE_ERR ,             /*  13   */
  SQI_INVALID_OFS ,           /*  14   */
  SQI_INVALID_UMSGID ,        /*  15   */
  SQI_INVALID_NUM_MSG,        /*  16   */
  XMSG_READ_ERR,              /*  17   */
  CTRL_READ_ERR,              /*  18   */
  TEXT_READ_ERR,              /*  19   */
  XMSG_WRITE_ERR,             /*  20   */
  CTRL_WRITE_ERR,             /*  21   */
  TEXT_WRITE_ERR,             /*  22   */
  MSG_UPDATED,                /*  23   */
  BASE_LOCKED,                /*  24   */
  SQ_NOMOREMSGS               /*  25   */
};

enum { FRAME_NORMAL = 0, FRAME_FREE, RESERVED, FRAME_UPDATE };
