/*
 * Copyright (c) 1996, serge terekhov.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by serge terekhov.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY SERGE TEREKHOV AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SERGE TEREKHOV OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * Version 9 NODELIST index is designed for use in FTN mail processing
 * in order to provide fast seaching for nodes and their respective
 * zone/region/hub numbers.
 * Especially useful for message tracking and routing.
 *
 * This implementation is 32-bit by nature.
 * If you implement 16-bit version of this idea - please drop me a line.
 *
 * 960706, serge.terekhov@f13.n5000.z2.fidonet.org
 */

/*
   nodelist.v9 structure in pseudo-C:

   read_v9 () {
      V9HEADER hdr;
      V9NET nets[];
      V9NODE nodes[][];
      V9NODE hubs[][];
      int i, j, hubs;

      read (hdr);

      for (i = 0; i < hdr.nets; ++i) {
         read (nets[i]);            // sorted by V9NET.net, then V9NET.zone
      }

      for (i = 0; i < hdr.nets; ++i) {
         hubs = 0;
         for (j = 0; j < nets[i].size; ++j) {
            read (nodes[i][j]);     // sorted by V9NODE.node
            if (nodes[i][j].hub > hubs)
               hubs = nodes[i][j].hub;
         }

         for (j = 0; j < hubs; ++j)
            read (hubs[i][j]);      // unsorted, in order as in NODELIST

         // hub for node computed like:
         // hub = (nodes[i][j].hub == 0)? 0 : hubs[i][nodes[i][j].hub - 1].node
      }
   }

   we have ONE limit: maximum of 255 hubs per net

   V9NET.off points to V9NODEs as byte offset from the beginning of .v9 file

   V9NET.size is number of (sorted) V9NODEs of nodes in net
   and at V9NODE[V9NET.size] .. V9NODE[V9NET.size + n] we have
      unspecified number of hubs
 */

#define  V9_SIGNATURE   0xB9060796l
#define  V9_VERSION     1

typedef unsigned long uint32;
typedef unsigned short uint16;
/* typedef unsigned char byte; */
typedef void V9NODELIST;

typedef struct {
   uint16 ver;          /* == 1 for this moment */
   uint32 sig;          /* == V9_SIGNATURE */
   uint32 nets;         /* total number of V9NET records */
   uint32 nodes;        /* just info: total number of V9NODE records */
   uint16 hubs;         /* just info: max number of hubs per net */
   uint32 time;         /* time() of compilation */
   byte dummy[64-20];
} V9HEADER;

typedef struct {
   uint16 net;          /* net number */
   uint16 zone;         /* zone number */
   uint16 region;       /* region number */
   uint16 size;         /* number of nodes in net */
   uint32 off;          /* SEEK_SET offset of net in .v9 file */
} V9NET;

typedef struct {
   uint16 node;         /* node number */
   byte hub;            /* hub _sequential_ number (see above) */
} V9NODE;

/* v9_open() returns 0 on failure
 * if days != 0 - .v9 becomes invalid after specified number of days
 *    after compilation
 */

V9NODELIST *v9_open (char *n, int days);

/*
 * v9_search() returns:
 *    1  - found, but ambiguos (only if izone == 0)
 *    0  - found
 *    -1 - zone/net not found, or bad nodelist, or inet == 0
 *    -2 - zone/net found, but no node in it
 */

int v9_search (V9NODELIST *nl, int izone, int inet, int inode,
               int *ozone, int *oregion, int *ohub);

