
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"
#include "func.h"

struct ATTRS Attrs[] = {
	
	{ "AUDIT"     , a_Audit     },
	{ "CRASH"     , a_Crash     },
	{ "FILE"      , a_File      },
	{ "HOLD"      , a_Hold      },
	{ "INTRANSIT" , a_Intransit },
	{ "ISRR"      , a_Isrr      },
	{ "KILLSENT"  , a_Killsent  },
	{ "LOCAL"     , a_Local     },
	{ "ORPHAN"    , a_Orphan    },
	{ "PRIVATE"   , a_Private   },
	{ "RECEIVED"  , a_Received  },
	{ "REQUEST"   , a_Request   },
	{ "RES1"      , a_Res1      },
	{ "RRQ"       , a_Rrq       },
	{ "SENT"      , a_Sent      },
	{ "UPDATE"    , a_Update    },
	{ ""          , 0           }
	
};

struct FLAGS Flags[] = {
	
	{ "CFM", f_Cfm },
	{ "DIR", f_Dir },
	{ "FPU", f_Fpu },
	{ "HUB", f_Hub },
	{ "IMM", f_Imm },
	{ "IT1", f_It1 },
	{ "IT2", f_It2 },
	{ "IT3", f_It3 },
	{ "K/S", f_Ks  },
	{ "KFS", f_Kfs },
	{ "LOK", f_Lok },
	{ "TFS", f_Tfs },
	{ "XMA", f_Xma },
	{ "ZON", f_Zon },
	{ "NPD", f_Npd },
	{ ""   , 0     }
	
};

struct FLAVORS Flavors[] = {
	
	{ "CRASH"  , flv_Crash  },
	{ "DIR"    , flv_Dir    },
	{ "HOLD"   , flv_Hold   },
	{ "IMM"    , flv_Imm    },
	{ "NORMAL" , flv_Normal },
	{ ""       , 0          }
	
};

struct KLUDGES Kludges[] = {
	
	{ "AREA"     , k_Area   ,1 },
	{ "DOMAIN"   , k_Domain ,1 },
	{ "FLAGS"    , k_Flags  ,1 },
	{ "FMPT"     , k_Fmpt   ,1 },
	{ "INTL"     , k_Intl   ,1 },
	{ "MSGID"    , k_Msgid  ,1 },
	{ "PATH"     , k_Path   ,2 },
	{ "PID"      , k_Pid    ,1 },
	{ "REPLY"    , k_Reply  ,1 },
	{ "SEEN-BY"  , k_SeenBy ,2 },
	{ "TID"      , k_Tid    ,1 },
	{ "TOPT"     , k_Topt   ,1 },
	{ "Via"      , k_Via    ,2 },
	{ "Recv"     , k_Recv   ,2 },
	{ ""         , 0        ,0 }
	
};


struct STATS Stats[] = {
	{ "Addlf"       , i_Addlf        },
	{ "Changed"     , i_Changed      },
	{ "Checkedpath" , i_Checkedpath  },
	{ "Dummyattach" , i_Dummyattach  },
	{ "Echomail"    , i_Echomail     },
	{ "Flag"        , i_Flag         },
	{ "From2d"      , i_From2d       },
	{ "Ignoreattach", i_Ignoreattach },
	{ "Interzone"   , i_Interzone    },
	{ "Intl"        , i_Intl         },
	{ "Invalidvias" , i_Invalidvias  },
	{ "Invkludges"  , i_Invkludges   },
	{ "Lastvia"     , i_Lastvia      },
	{ "Loop"        , i_Loop         },
	{ "Loop2"       , i_Loop2        },
	{ "Loop3"       , i_Loop3        },
	{ "Normal"      , i_Normal       },
	{ "Null"        , i_Null         },
	{ "Repeatscan"  , i_Repeatscan   },
	{ "Routedecho"  , i_Routedecho   },
	{ "Seen"        , i_Seen         },
	{ "To2d"        , i_To2d         },
	{ "User1"       , i_User1        },
	{ "User2"       , i_User2        },
	{ "User3"       , i_User3        },
	{ "ZaLoop"      , i_ZaLoop       },
	{ ""            , 0              }
	
};

struct TAGDESC TagDesc[] = {
	{ "ADDRESS"        ,  ADDRESS        , defAddress},
	{ "ADDRESSATTR"    ,  ADDRESSATTR    , NULL},
	{ "ADDRESSNAME"    ,  ADDRESSNAME    , defAddrName},
	{ "ADDRESSPACKER"  ,  ADDRESSPACKER  , defAddrPack},
	{ "ALIAS"          ,  ALIAS          , NULL},
	{ "ATTRIBUTE"      ,  ATTRIBUTE      , defAttribute},
	{ "BINKSEMAPHOR"   ,  BINKSEMAPHOR   , NULL},
	{ "CONDITION"      ,  CONDITION      , NULL},
	{ "DATE"           ,  DATE_           , defDate},
	{ "DATEADDRESS"    ,  DATEADDRESS    , NULL},
	{ "DAYOFWEEK"      ,  DAYOFWEEK      , defDow},
	{ "EVENT"          ,  EVENT          , NULL},
	{ "FDSEMAPHOR"     ,  FDSEMAPHOR     , NULL},
	{ "FILEAGE"        ,  FILEAGE        , defFileAge},
	{ "FILEAREAS"      ,  FILEAREAS      , NULL},
	{ "FILES"          ,  FILES          , defFiles},
	{ "LENGTHADDRESS"  ,  LENGTHADDRESS  , defLengthAddr},
	{ "PKTLENGTH"      ,  PKTLENGTH      , defPlenAddr},
	{ "MACRO"          ,  MACRO          , defMacro},
	{ "MESSAGEAGE"     ,  MESSAGEAGE     , defMsgAge},
	{ "MESSAGELENGTH"  ,  MESSAGELENGTH  , defMsgLen},
	{ "NAME"           ,  NAME           , defName},
	{ "NODELIST"       ,  NODELIST       , defNodelist},
	{ "NODELIST4D"     ,  NODELIST4D     , NULL},
	{ "PASSWORDADDRESS",  PASSWORDADDRESS , defPassAddr},
	{ "PASSWORDNAME"   ,  PASSWORDNAME    , defPassName},
	{ "PATH"           ,  PATH            , defPath},
	{ "PROCESSAGE"     ,  PROCESSAGE      , defProcAge},
	{ "READDRESSLIST"  ,  READDRESSLIST   , defReAddressList},
	{ "READDRESSNAME"  ,  READDRESSNAME   , NULL},
	{ "RENAME"         ,  RENAME          , NULL},
	{ "ROUTING"        ,  ROUTING        , defRouting},
	{ "STRINGS"        ,  STRINGS        , NULL},
	{ "SUBJECT"        ,  SUBJECT        , defSubject},
	{ "SUBJECTADDRESS" ,  SUBJECTADDRESS , NULL},
	{ "SYSTEM"         ,  SYSTEM         , defSystem},
	{ "TEXTSTRING"     ,  TEXTSTRING     , defTextString},
	{ "TIMESLOT"       ,  TIMESLOT       , NULL},
	{ "VERSION9"       ,  VERSION9       , defNodelist_v9},
	{ "KLUDGE"         ,  KLUDGE         , defKludge},
	{ "ECHONAME"       ,  ECHONAME       , defEchoName},
	{ ""               ,  0              , NULL}
};

struct MACROS Macros[] = {
	{ "ATTRIBUTES"     , m_Attributes     },
	{ "FROM"           , m_from           },
	{ "FROMDOMAIN"     , m_fromDomain     },
	{ "FROMNAME"       , m_fromName       },
	{ "INTLFROM"       , m_intlFrom       },
	{ "INTLTO"         , m_intlTo         },
	{ "LINECOUNT"      , m_lineCount      },
	{ "LOCALDATE"      , m_Localdate      },
	{ "LOCALTIME"      , m_Localtime      },
	{ "MSGIDFROM"      , m_msgidFrom      },
	{ "MSGLENGTH"      , m_msgLength      },
	{ "NODELIST"       , m_Nodelist       },
	{ "PROGRAM"        , m_Program        },
	{ "REPLYTO"        , m_replyTo        },
	{ "ROUTETO"        , m_routeTo        },
	{ "SUBJECT"        , m_Subject        },
	{ "SYSOP"          , m_Sysop          },
	{ "SYSTEM"         , m_System         },
	{ "SYSTEMTIME"     , m_Systemtime     },
	{ "TIME"           , m_Time           },
	{ "TO"             , m_to             },
	{ "TODOMAIN"       , m_toDomain       },
	{ "TONAME"         , m_toName         },
	{ "TZ"             , m_TZ             },
	{ "OLDATTRIBUTES"  , m_old_Attributes },
	{ "OLDFROM"        , m_old_from       },
	{ "OLDFROMDOMAIN"  , m_old_fromDomain },
	{ "OLDFROMNAME"    , m_old_fromName   },
	{ "OLDINTLFROM"    , m_old_intlFrom   },
	{ "OLDINTLTO"      , m_old_intlTo     },
	{ "OLDLINECOUNT"   , m_old_lineCount  },
	{ "OLDMSGIDFROM"   , m_old_msgidFrom  },
	{ "OLDMSGLENGTH"   , m_old_msgLength  },
	{ "OLDREPLYTO"     , m_old_replyTo    },
	{ "OLDROUTETO"     , m_old_routeTo    },
	{ "OLDSUBJECT"     , m_old_Subject    },
	{ "OLDTIME"        , m_old_Time       },
	{ "OLDTO"          , m_old_to         },
	{ "OLDTODOMAIN"    , m_old_toDomain   },
	{ "OLDTONAME"      , m_old_toName     },
	{ "PKTFROM"        , m_pktFrom        },
	{ "PKTTO"          , m_pktTo          },
	{ ""               , 0                }
};

struct MONTHS Months[] = {
	{"Jan",0},
	{"Feb",1},
	{"Mar",2},
	{"Apr",3},
	{"May",4},
	{"Jun",5},
	{"Jul",6},
	{"Aug",7},
	{"Sep",8},
	{"Oct",9},
	{"Nov",10},
	{"Dec",11},
	{"", 12}
};

struct DOW Dow[] = {
	{"Sun",0},
	{"Mon",1},
	{"Tue",2},
	{"Wed",3},
	{"Thu",4},
	{"Fri",5},
	{"Sat",6},
	{""   ,0}
};


/* define globals */

ADDR MainAddress;
char ItraxPath[128], TemplatePath[128], FlagDir[128];
char *SeqDir = NULL;
unsigned long SeqOutRun = 0;
char *TsPath = NULL;
int  TsHandle = -1;
char CompressFile[256];
char *LogFile = NULL;
int Sharing;
int DebugLevel;
int nRead, nWritten;
FLock lockFile;
FTN ftn, old_ftn;
struct CurMsg CurMsg;

TEXT *TextFirst = NULL, *ViaTextFirst = NULL;

struct Macro *MacroFirst = NULL, *MacroLast = NULL;

struct FidoAreas *FidoAreaFirst = NULL, *FidoAreaLast = NULL;
struct PktAreas *PktAreaFirst = NULL, *PktAreaLast = NULL;
struct BinkAreas *BinkAreaFirst = NULL, *BinkAreaLast = NULL;
struct SquishAreas *SquishAreaFirst = NULL, *SquishAreaLast = NULL;
struct JamAreas *JamAreaFirst = NULL, *JamAreaLast = NULL;
struct TextAreas *TextAreaFirst = NULL, *TextAreaLast = NULL;
struct EchoAreas *EchoAreas = NULL;

struct Tags *TagFirst = NULL, *TagLast = NULL;

struct ndlTags *ndlTagFirst = NULL, *ndlTagLast = NULL;
struct ndlTags *v9_ndlTagFirst = NULL, *v9_ndlTagLast = NULL;
struct SystemTags *SystemTagFirst = NULL, *SystemTagLast = NULL;
struct AddrAttrTags *AddrAttrTagFirst = NULL, *AddrAttrTagLast = NULL;
struct AddrNameTags *AddrNameTagFirst = NULL, *AddrNameTagLast = NULL;
struct AddrPackerTags *AddrPackerTagFirst = NULL, *AddrPackerTagLast = NULL;
struct AddrTags *AddrTagFirst = NULL, *AddrTagLast = NULL;
struct AliasTags *AliasTagFirst = NULL, *AliasTagLast = NULL;
struct AttrTags *AttrTagFirst = NULL, *AttrTagLast = NULL;
struct DateAddrTags *DateAddrTagFirst = NULL, *DateAddrTagLast = NULL;
struct DateTags *DateTagFirst = NULL, *DateTagLast = NULL;
struct DowTags *DowTagFirst = NULL, *DowTagLast = NULL;
struct FileAgeTags *FileAgeTagFirst = NULL, *FileAgeTagLast = NULL;
struct FilesTags *FilesTagFirst = NULL, *FilesTagLast = NULL;
struct LengthAddrTags *LengthAddrTagFirst = NULL, *LengthAddrTagLast = NULL;
struct LengthAddrTags *PktLenAddrTagFirst = NULL, *PktLenAddrTagLast = NULL;
struct MsgAgeTags *MsgAgeTagFirst = NULL, *MsgAgeTagLast = NULL;
struct ProcAgeTags *ProcAgeTagFirst = NULL, *ProcAgeTagLast = NULL;
struct MsgLenTags *MsgLenTagFirst = NULL, *MsgLenTagLast = NULL;
struct NameTags *NameTagFirst = NULL, *NameTagLast = NULL;
struct PassAddrTags *PassAddrTagFirst = NULL, *PassAddrTagLast = NULL;
struct PassNameTags *PassNameTagFirst = NULL, *PassNameTagLast = NULL;
struct PathTags *PathTagFirst = NULL, *PathTagLast = NULL;
struct ReAddrTags *ReAddrTagFirst = NULL, *ReAddrTagLast = NULL;
struct RoutingTags *RoutingTagFirst = NULL, *RoutingTagLast = NULL;
struct SubjAddrTags *SubjAddrTagFirst = NULL, *SubjAddrTagLast = NULL;
struct SubjectTags *SubjectTagFirst = NULL, *SubjectTagLast = NULL;
struct TextStrTags *TextStrTagFirst = NULL, *TextStrTagLast = NULL;
struct KludgeTags *KludgeTagFirst = NULL, *KludgeTagLast = NULL;
struct EchoNameTags *EchoNameTagFirst = NULL, *EchoNameTagLast = NULL;

struct pktcache *PktCache = NULL, *PktCacheLast = NULL;
struct pktcache *OipCache = NULL, *OipCacheLast = NULL;
struct pktcache *ArcCache = NULL, *ArcCacheLast = NULL;

struct MaxMsgNos *MaxMsgNoFirst = NULL, *MaxMsgNoLast = NULL;

struct Compress *CompressFirst = NULL, *CompressLast = NULL;


struct AREACACHE AreaCache[10]; /* not used so far */
struct ADDRCACHE AddrCache[AddrCacheSize]; /* ditto */

TSL *TSlist = NULL;

time_t tStart;

unsigned long MsgId = 0;

int writeY, writeX = 1;
BOOL Chat_On, LogOFF, ChatOFF;

struct filter *LogIncl = NULL, *LogInclLast = NULL;
struct filter *LogExcl = NULL, *LogExclLast = NULL;

struct _gv *GlobalVarList = NULL;

int sq_errno, jam_errno;

BOOL IgnoreBSY, TouchEchomail, UseRegExp, rxIcase;

BOOL IsInited = No;

char *FBox = NULL;

struct _dbcache *DupeBase = NULL;
int DupesTotal, DupesMaxMsg, DupesMaxOpen;
char *DupeBasePath = NULL, OutFileName[256];

BOOL BadAreaDefined;

/* HEV Hev; */
