
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"

#if defined (UNIX)
#define TSEP "./"
#else
#define TSEP ".\\"
#endif

/* called from config.c */
void defFidoArea(char *str, char *err_diag);
void defBinkArea(char *str, char *err);
void defSquishArea(char *str, char *err);
void defTextArea(char *str, char *err);
void defPktArea(char *str, char *err);
void defJamArea(char *str, char *err);

void  set_ts(char *areaname, dword ts);
dword fetch_ts(char *areaname);

void create_new_sqbase(char *areapath);
void create_new_jambase(char *areapath);

static void fill_tslist(char *areaname);
static BOOL is_Password_Tag(char *tmp_str);
static BOOL is_full_path(char *p);


/************************************************************************/
void defFidoArea(char *str, char *err)
{
/*                1           2         [ 3 ]      [4]
   FIDOAREA:  ( "Path" , "AreaName" , "System" , <EchoMail> )
*/	
	struct FidoAreas *FidoArea = NULL;
	struct SystemTags *SystemTag = NULL;
	char areapath[128], areaname[20], buf[256], echo[9], system[20] = {0};
	char w[256];
	BOOL found = No, echomail = No;

	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}
	
	sprintf(areapath, "%.63s%.63s", ItraxPath, w);

	if ( !setpath(areapath) ) {
		strncpy(areapath, w, 127); areapath[127] = 0;
		if ( !setpath(areapath) ) {
			sprintf(buf, "%s path not found: `%s'\n",
							err, areapath);
			fprintf(stderr, "\7%s\n", buf);
			writeToLog(0, buf);
			abort();
		}
	}
	
	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;
	
	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	/* argc > 2 (3 or 4) */
	if (strlen(str)) {

		char *p;
		
		/* 3 */
		get_next_word(str, w);
		strncpy(system, w, 19); system[19] = 0;

		/* 4 */
		p = get_next_word(str, w);
		
		for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
			if (!stricmp(system, SystemTag->tag))
				break;
		
		/* system not foung */
		if (NULL == SystemTag) {
			
			if (p || (!p && 0 != stricmp(system, "echomail")))
				found = No;
			else
				found = Yes;
			
			if (found == No) {
				sprintf(buf, "%s data system block undefined : `%s'", err, system);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			memset(system, '\0', 20);
			p = (char *)Ok;
		} /* system not found */

		if (p) {
			strncpy(echo, w, 8); echo[8] = 0;
			if (0 != stricmp(echo, "echomail")) {
				sprintf(buf, "%s unknown argument : `%s'", err, echo);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			echomail = Yes;
		}
		
	}  /* argc > 2 (system and/or echomail defined) */
	
	/*//////////////////////////////////////////////////////////////////////////*/
	for( FidoArea = FidoAreaFirst; FidoArea; FidoArea = FidoArea ->next )
	{
		if (!stricmp(areaname, FidoArea->name)) {
			break;
		}
	}

	if (FidoArea == NULL)
		FidoArea = (struct FidoAreas *)calloc(1, sizeof(struct FidoAreas));
	FidoArea->next  = NULL;
	/*//////////////////////*/
	strcpy(FidoArea->name, areaname);
	strcpy(FidoArea->path, areapath);
	strcpy(FidoArea->system, system);
	FidoArea->EchoMail = echomail; /* yes/no */
	/*//////////////////////*/

	if (FidoAreaLast) {
		FidoAreaLast->next = FidoArea;
	}
	else {
		FidoAreaFirst = FidoArea;
	}

	FidoAreaLast = FidoArea;

	(void) fill_tslist(areaname);
	
	return;
}

/************************************************************************/
void defSquishArea(char *str, char *err)
{
/*               1             2        [ 3 ]        [ 4 ]
  SquishArea  ( "Path" , "AreaName" , "System" , "EchoMail" )
*/
	struct SquishAreas *SquishArea = NULL;
	struct SystemTags *SystemTag = NULL;
	char areapath[128], areaname[20], buf[256], sqd_name[256], sqi_name[256];
	char system[20] = {0}, echo[9], w[256];
/*	struct find_t ffblk; */
	BOOL found = No, echomail = No;

	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}
	
	if (!is_full_path(w))
		sprintf(areapath, "%.63s%.63s", ItraxPath, w);
	else
		strcpy(areapath, w);
	
	sprintf( sqd_name , "%s.sqd" , areapath );
	sprintf( sqi_name , "%s.sqi" , areapath );
		
	if ( access(sqd_name, 0) && access(sqi_name, 0) )
	{
		create_new_sqbase(areapath);
	}
	else {
		if ( access(sqd_name, 0) ) {
			sprintf(buf, "squish area file not found:`%s'\n", sqd_name);
			fprintf(stderr, "\7%s", buf);
			writeToLog(0, buf);
			abort();
		}
	
		if ( access(sqi_name, 0) ) {
			sprintf(buf, "squish area file not found:`%s'\n", sqi_name);
			fprintf(stderr, "\7%s", buf);
			writeToLog(0, buf);
			abort();
		}
	}
	
	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;
	
	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	/* argc > 2 (3 or 4) */
	if (strlen(str)) {

		char *p;
		
		/* 3 */
		get_next_word(str, w);
		strncpy(system, w, 19); system[19] = 0;

		/* 4 */
		p = get_next_word(str, w);
		
		for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
		{
			if (!stricmp(system, SystemTag->tag)) {
				break;
			}
		}

		/* system not foung */
		if (NULL == SystemTag) {
			if (p || (!p && 0 != stricmp(system, "echomail"))) {
				found = No;
			}
			else
				found = Yes;
			
			if (found == No) {
				sprintf(buf, "%s data system block undefined : `%s'", err, system);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			memset(system, '\0', 20);
			p = (char *)Ok;
		}
		
		if (p) {
			strncpy(echo, w, 8); echo[8] = 0;
			if (0 != stricmp(echo, "echomail")) {
				sprintf(buf, "%s unknown argument : `%s'", err, echo);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			echomail = Yes;
		}
	}  /* argc > 2 (system and/or echomail defined) */

	for( SquishArea = SquishAreaFirst; SquishArea; SquishArea = SquishArea ->next )
	{
		if (!stricmp(areaname, SquishArea->name)) {
			break;
		}
	}

	if (SquishArea == NULL)
		SquishArea = (struct SquishAreas *)calloc(1, sizeof(struct SquishAreas));
	SquishArea->next  = NULL;
	/*//////////////////////*/
	strcpy(SquishArea->name, areaname);
	strcpy(SquishArea->path, areapath);
	strcpy(SquishArea->system, system);
	SquishArea->EchoMail = echomail; /* yes/no */
	/*//////////////////////*/

	if (SquishAreaLast) {
		SquishAreaLast->next = SquishArea;
	}
	else {
		SquishAreaFirst = SquishArea;
	}

	SquishAreaLast = SquishArea;
	
	(void) fill_tslist(areaname);
	
	return;
}

/************************************************************************/
void defBinkArea(char *str, char *err)
{
/*               1         2             3             4             5             6
   BinkArea ( <Path> , <AreaName> , [<System>] , [<Password>] , [NoMainZone] [MultiDomain] )
*/
	struct BinkAreas *BinkArea = NULL;
	struct SystemTags *SystemTag = NULL;
	char areapath[128] = {0}, areaname[20] = {0}, buf[256], w[256];
	char system[20] = {0}, passw_tag[20] = {0};
	BOOL NoMainZone = No, MultiDomain = No;
	int type = tBINK;
	
	
	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}
	
	sprintf(areapath, "%.63s%.63s", ItraxPath, w);
	sprintf(buf, "def_bink_area: try areapath: `%s'\n", areapath);
	writeToLog(4, buf);
	
	if ( !setpath(areapath) ) {
		strncpy(areapath, w, 127); areapath[127] = 0;
		sprintf(buf, "def_bink_area: try areapath: `%s'\n", areapath);
		writeToLog(4, buf);
		if ( !setpath(areapath) ) {
			sprintf(buf, "%s path not found: `%s'\n",
							err, areapath);
			fprintf(stderr, "\7%s\n", buf);
			writeToLog(0, buf);
			abort();
		}
	}

	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;

	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	/* below this line every parameter is optional */

	while(get_next_word(str, w))
	{
		w[19] = 0;
		
		/* system */
		if (!strlen(system)) {
			for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
			{
				if (!stricmp(w, SystemTag->tag)) {
					strcpy(system, w);
					break;
				}
			}
			if (strlen(system))
				continue;
		}
		
		/* password */
		if (!strlen(passw_tag)) {
			if (Yes == is_Password_Tag(w)) {
				strcpy(passw_tag, w);
				continue;
			}
		}
		
		if (NoMainZone == No) {
			if (!stricmp(w, "NoMainZone")) {
				NoMainZone = Yes;
				continue;
			}
		}
		
		if (MultiDomain == No) {
			if (!stricmp(w, "MultiDomain")) {
				MultiDomain = Yes;
				continue;
			}
		}
		
		if (!stricmp(w, "BrakeOutbound")) {
			type = tBRAKE;
			continue;
		}
	
		if (!stricmp(w, "Amiga4D")) {
			type = tAMIGA4D;
			continue;
		}
		
		/* bad luck */
		sprintf(buf, "%s data not found or syntax error: `%s'", err, w);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	for( BinkArea = BinkAreaFirst; BinkArea; BinkArea = BinkArea ->next )
	{
		if (!stricmp(areaname, BinkArea->name)) {
			break;
		}
	}

	if (BinkArea == NULL)
		BinkArea = (struct BinkAreas *)calloc(1, sizeof(struct BinkAreas));
	BinkArea->next  = NULL;
	/*/////////////////////*/
	strcpy(BinkArea->name, areaname);
	strcpy(BinkArea->path, areapath);
	strcpy(BinkArea->system, system);
	strcpy(BinkArea->passw_tag, passw_tag);
	BinkArea->no_main_zone = NoMainZone;
	BinkArea->multi_domain = MultiDomain;
	BinkArea->type = type;
	/*//////////////////////*/

	if (BinkAreaLast) {
		BinkAreaLast->next = BinkArea;
	}
	else {
		BinkAreaFirst = BinkArea;
	}

	BinkAreaLast = BinkArea;
	
	
	return;
}


/************************************************************************/
void defTextArea(char *str, char *err)
{
/*              1          2
   TextArea( "Path" , "AreaName"  )
*/
	struct TextAreas *TextArea = NULL;
	char areapath[128], areaname[20], buf[256], w[256];
	char pathname[256], path[128] = {0}, fname[128], *p;
	
	
	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}
	strncpy(areapath, w, 127); areapath[127] = 0;
	
	if (NULL == (p = strrchr(areapath, DELIM)))
		strcpy(fname, areapath);
	else {
		strncpy(path, areapath, p-areapath+1);
		strcpy(fname, &areapath[p-areapath+1]);
	}
	
	if (!strlen(fname)) {
		sprintf(buf, "%s incorrect filename `%s'\n", err, fname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	sprintf(areapath, "%.63s%.63s", ItraxPath, path);

	if ( !setpath(areapath) ) {
		strcpy(areapath, path);
		if ( !setpath(areapath) ) {
			sprintf(buf, "%s path not found: `%s'\n",
							err, areapath);
			fprintf(stderr, "\7%s\n", buf);
			writeToLog(0, buf);
			abort();
		}
	}
	
	sprintf(pathname, "%.127s%.127s", areapath, fname);
	
	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;

	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	for( TextArea = TextAreaFirst; TextArea; TextArea = TextArea ->next )
	{
		if (!stricmp(areaname, TextArea->name)) {
			break;
		}
	}

	if (TextArea == NULL)
		TextArea = (struct TextAreas *)calloc(1, sizeof(struct TextAreas));
	TextArea->next  = NULL;
	/*//////////////////////*/
	strcpy(TextArea->name, areaname);
	strcpy(TextArea->pathname, pathname);
	/*//////////////////////*/

	if (TextAreaLast) {
		TextAreaLast->next = TextArea;
	}
	else {
		TextAreaFirst = TextArea;
	}

	TextAreaLast = TextArea;
	
	return;
}

/************************************************************************/
void defPktArea(char *str, char *err)
{
/*               1         2             3             4             5
   PktArea ( <Path> , <AreaName> , [<System>] , [<Password>] , [EchoMail] )
*/	
	struct PktAreas *PktArea = NULL;
	struct SystemTags *SystemTag = NULL;
	char areapath[128], areaname[20], buf[256], system[20] = {0};
	char passw_tag[20] = {0}, pktlentag[20] = {0}, w[256];
	BOOL echomail = No;
	int fbox = NOBOX;
	
	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}
	sprintf(areapath, "%.63s%.63s", ItraxPath, w);
	sprintf(buf, "def_pkt_area: try areapath: `%s'\n", areapath);
	writeToLog(4, buf);
	
	
	if ( !setpath(areapath) ) {
		
		strncpy(areapath, w, 127); areapath[127] = 0;
		sprintf(buf, "def_pkt_area: try areapath: `%s'\n", areapath);
		writeToLog(4, buf);
		
		if ( !setpath(areapath) ) {
			sprintf(buf, "%s path not found: `%s'\n",
							err, areapath);
			fprintf(stderr, "\7%s\n", buf);
			writeToLog(0, buf);
			abort();
		}
	}
	
	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;
	
	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	/* below this line every parameter is optional */

	while(get_next_word(str, w))
	{
		w[19] = 0;
		
		/* system */
		if (!strlen(system)) {
			for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
			{
				if (!stricmp(w, SystemTag->tag)) {
					strcpy(system, w);
					break;
				}
			}
			if (strlen(system))
				continue;
		}
		
		/* password */
		if (!strlen(passw_tag)) {
			if (Yes == is_Password_Tag(w)) {
				strcpy(passw_tag, w);
				continue;
			}
		}
		
		if (echomail == No) {
			if (!stricmp(w, "echomail")) {
				echomail = Yes;
				continue;
			}
		}
		
		if (!strlen(pktlentag)) {
			
			struct LengthAddrTags *P;

			for( P = PktLenAddrTagFirst; P; P = P->next )
			{
				if (!stricmp(w, P->tag)) {
					strcpy(pktlentag, w);
					break;
				}
			}
			if (strlen(pktlentag))
				continue;
		}

		if (!fbox) {
			
			if (!stricmp(w, "b-box")) {
				fbox = BBOX;
				continue;
			}
			else if (!stricmp(w, "t-box")) {
				fbox = TBOX;
				continue;
			}
			
		}
	
		/* bad luck */
		sprintf(buf, "%s data not found or syntax error: `%s'", err, w);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	
	/*//////////////////////////////////////////////////////////////////////////*/
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea ->next )
	{
		if (!stricmp(areaname, PktArea->name)) {
			break;
		}
	}

	if (PktArea == NULL)
		PktArea = (struct PktAreas *)calloc(1, sizeof(struct PktAreas));
	PktArea->next  = NULL;
	/*//////////////////////*/
	strcpy(PktArea->name, areaname);
	strcpy(PktArea->path, areapath);
	strcpy(PktArea->system, system);
	strcpy(PktArea->passw_tag, passw_tag);
	strcpy(PktArea->pktlentag, pktlentag);
	PktArea->EchoMail = echomail; /* yes/no */
	PktArea->fbox = fbox;
	/*//////////////////////*/
	if (PktAreaLast) {
		PktAreaLast->next = PktArea;
	}
	else {
		PktAreaFirst = PktArea;
	}

	PktAreaLast = PktArea;
	
	return;
}

/*********************************/
static BOOL is_Password_Tag(char *tmp_str)
{
	struct PassAddrTags *PassAddrTag = PassAddrTagFirst;
	char buf[256];
	
	for( ; PassAddrTag; PassAddrTag = PassAddrTag->next )
	{
		if (!stricmp(tmp_str, PassAddrTag->tag)) {
			sprintf(buf, "PasswordAddress tag found: `%s'\n", tmp_str);
			writeToLog(4, buf);
			return Ok;
		}
	}
	
/*
	if (NULL == PassNameTag) {
		PassNameTag = PassNameTagFirst;
		for( ; PassNameTag; PassNameTag = PassNameTag->next )
		{
			if (!stricmp(tmp_str, PassNameTag->tag)) {
				sprintf(buf, "PasswordName tag found: `%s'\n", tmp_str);
				writeToLog(4, buf);
				return Ok;
			}
		}
	}
	*/
	
	return Fail;
	
}

/************************************************************************/
void defJamArea(char *str, char *err)
{
/*               1             2        [ 3 ]        [ 4 ]
   JamArea  ( "Path" , "AreaName" , "System" , "EchoMail" )
*/
	struct JamAreas *JamArea = NULL;
	struct SystemTags *SystemTag = NULL;
	char areapath[128], areaname[20], buf[256];
	char jhr_name[256], jdt_name[256], jdx_name[256];
	char system[20] = {0}, echo[9], w[256];
/*	struct find_t ffblk; */
	BOOL found = No, echomail = No;

	/* PATH */
	get_next_word(str, w);
	if (!strcmp(w, ".") || !strcmp(w, TSEP))
		memset(w, '\0', 256);
	else if (w[0] == '.' && w[1] == DELIM) {
		strcpy(buf, &w[2]);
		strcpy(w, buf);
	}

	if (!is_full_path(w))
		sprintf(areapath, "%.63s%.63s", ItraxPath, w);
	else
		strcpy(areapath, w);

	sprintf( jhr_name , "%s.jhr" , areapath );
	sprintf( jdt_name , "%s.jdt" , areapath );
	sprintf( jdx_name , "%s.jdx" , areapath );
		
	if ( access(jhr_name, 0) && access(jdt_name, 0) && access(jdx_name, 0) )  {
		create_new_jambase(areapath);
	}
	else {
		
		if ( access(jhr_name, 0) ) {
			sprintf(buf, "JAM area file not found:`%s'\n", jhr_name);
			fprintf(stderr, "\7%s", buf);
			writeToLog(0, buf);
			abort();
		}
	
		if ( access(jdt_name, 0) ) {
			sprintf(buf, "JAM area file not found:`%s'\n", jdt_name);
			fprintf(stderr, "\7%s", buf);
			writeToLog(0, buf);
			abort();
		}
	
		if ( access(jdx_name, 0) ) {
			sprintf(buf, "JAM area file not found:`%s'\n", jdx_name);
			fprintf(stderr, "\7%s", buf);
			writeToLog(0, buf);
			abort();
		}

	}
	/* AREANAME */
	get_next_word(str, w);
	strncpy(areaname, (const char*)(strUpper(w)), 19); areaname[19] = 0;
	
	if ( !is_legal_tag(areaname) ) {
		sprintf(buf, "%s incorrect area tag : `%s'", err, areaname);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	}

	/* argc > 2 (3 or 4) */
	if (strlen(str)) {

		char *p;
		
		/* 3 */
		get_next_word(str, w);
		strncpy(system, w, 19); system[19] = 0;

		/* 4 */
		p = get_next_word(str, w);
		
		for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
			if (!stricmp(system, SystemTag->tag))
				break;

		/* system not foung */
		if (NULL == SystemTag) {
			if (p || (!p && 0 != stricmp(system, "echomail"))) {
				found = No;
			}
			else
				found = Yes;
			
			if (found == No) {
				sprintf(buf, "%s data system block undefined : `%s'", err, system);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			memset(system, '\0', 20);
			p = (char *)Ok;
			
		}
		
		if (p) {
			strncpy(echo, w, 8); echo[8] = 0;
			if (0 != stricmp(echo, "echomail")) {
				sprintf(buf, "%s unknown argument : `%s'", err, echo);
				fprintf(stderr, "\7%s", buf);
				writeToLog(0, buf);
				abort();
			}
			echomail = Yes;
		}
	}  /* argc > 2 (system and/or echomail defined) */

	for( JamArea = JamAreaFirst; JamArea; JamArea = JamArea ->next )
	{
		if (!stricmp(areaname, JamArea->name)) {
			break;
		}
	}

  if (JamArea == NULL)
		JamArea = (struct JamAreas *)calloc(1, sizeof(struct JamAreas));
	JamArea->next  = NULL;
	/*//////////////////////*/
	strcpy(JamArea->name, areaname);
	strcpy(JamArea->path, areapath);
	strcpy(JamArea->system, system);
	JamArea->EchoMail = echomail; /* yes/no */
	/*//////////////////////*/

	if (JamAreaLast) {
		JamAreaLast->next = JamArea;
	}
	else {
		JamAreaFirst = JamArea;
	}

	JamAreaLast = JamArea;
	
	(void) fill_tslist(areaname);
	
	return;
}

/******************************/
static void fill_tslist(char *areaname)
{
	TSL *tsl = TSlist;
	static TSL *Last = NULL;

	tsl = (TSL *)calloc(1, sizeof(TSL));
	tsl->next = NULL;
	strcpy(tsl->ts.areaname, areaname);
	tsl->ts.ts = 0L;
	if (Last)
		Last->next = tsl;
	else
		TSlist = tsl;
	Last = tsl;

	return;
}


/******************************/
void set_ts(char *areaname, dword ts)
{
	char buf[256];
	
	TSL *tsl = TSlist;

	for( ; tsl ; tsl = tsl->next)
		if (!stricmp(tsl->ts.areaname, areaname)) {
			tsl->ts.ts = ts;
      return;
		}

	sprintf (buf, "set_ts: area `%s' not found in tslist\n", areaname);
	writeToLog(4, buf);
	return;
}

/*****************************/
dword fetch_ts(char *areaname)
{
	TSL *tsl = TSlist;
	char buf[256];

	for( ; tsl ; tsl = tsl->next)
		if (!stricmp(tsl->ts.areaname, areaname)) {
			sprintf (buf, "get_ts: %s -> %d\n", areaname, (unsigned)tsl->ts.ts);
			writeToLog(4, buf);
			return tsl->ts.ts;
		}
	
	sprintf (buf, "get_ts: no entry found for area %s\n", areaname);

	return 0L;
}

static BOOL is_full_path(char *p)
{
	return (p[0]==DELIM || p[1]==':');
}
