
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"

#define MAXNDL 10

BOOL ndl_build(int nndl, char **ndlnames, char *path, char *ext);  /* ndlbld.c */


/* DATA NODELIST
             1       2       [3]        4         5       [6]     [...]
   call: ( "Tag" , "Ext" , ["Zone"] , "Path" , "Ndl1" , ["Ndl2"] , ...)
	
   char exc_[3]={"Abort", "Ignore", "Disable", ""};
	
   argc = 4 or more
*/
int getndlnames(char *path, char *arg, char *ndllist[]);

/***********************************************************************/
void defNodelist(char *tag, char *arg, char *err_diag)
{
	
	struct ndlTags *ndlTag = NULL, *mark = NULL;
	char ext[128], path[128], s[256], str[256], w[256];
	char *ndlnames[MAXNDL], ndlname[13];
	char allnames[256];
	word zone = 0;
	int i, nndl, nread;
	BOOL rebuild = No;
/*	struct find_t ffblk; */
	int h_inc;
	struct nodeinc nodeinc;
	char n_inc_buf[NODEINC_SIZE], *pbuf;
	long size;

	/* TAG */
	strUpper(tag); tag[19] = 0;
	
	if ( !is_legal_tag(tag) ) {
		sprintf(s, "%s incorrect data block name : `%s'", err_diag, tag);
		fprintf(stderr, "\7%s", s);
		writeToLog(0, s);
		abort();
	}

	/*///////////////////*/
	addtag(NODELIST, tag, err_diag);  /* add tag to the global taglist */
	/*///////////////////*/

	/* EXT */
	get_next_word(arg, ext);
	ext[127] = 0;

	get_next_word(arg, w);

	/* ZONE */
	if (__isnumber(w)) {
		zone = (word) atoi(w);
		get_next_word(arg, w);
	}

	/* PATH */
	strncpy(path, w, 127);  path[127] = 0;
	
	if ( !setpath(path) ) {
		sprintf(s, "%s path not found : `%s'\n", err_diag, path);
		fprintf(stderr, "\7%s", s);
		writeToLog(0, s);
		abort();
	}

	/* NODELIST NAMES */
	
	strcpy(allnames, arg);

	nndl = getndlnames(path, arg, ndlnames);
	if (!nndl) {
		sprintf(str, "%s nodelist file(s) template not found: `%s'\n",
						err_diag, allnames);
		fprintf(stderr, "\7%s", str);
		writeToLog(0, str);
		abort();
	}

	
	
	/* see if we need to rebuild indexes */
	rebuild = No;
	while (1)
	{
		sprintf(s, "%snodeidx.%s", path, ext);
		sprintf(str, "trying to find index file `%s'... ", s);
		
		if (access(s, 0)) {
			strcat(str, "not found.\n");
			writeToLog(3, str);
			rebuild = Yes;
			break;
		}
		else {
			strcat(str, "ok\n");
			writeToLog(3, str);
		}
		
		sprintf(s, "%snodeinc.%s", path, ext);
		sprintf(str, "trying to open index file `%s'... ", s);
		if ( (h_inc = open(s, O_RDONLY|O_BINARY)) == -1 ) {
			strcat(str, "failed.\n");
			writeToLog(3, str);
			rebuild = Yes;
			break;
		}
		else {
			strcat(str, "ok\n");
			writeToLog(3, str);
		}
		
		if (Chat_On)
			printf("checking nodelist indexes...  ");
		
		i = 0;
		
		size = filelength(h_inc);
		while( size > tell( h_inc ) )
		{
			nread = NODEINC_SIZE;
			if (nread != (read(h_inc, &n_inc_buf, nread))) {
				writeToLog(3, "ndl: nodeinc record is of incorrect size\n");
				rebuild = Yes;
				break;
			}
			
			pbuf = n_inc_buf;
			nodeinc.namelen = *pbuf;
			pbuf++;
			memmove(&(nodeinc.name), pbuf, 12);
			pbuf += 12;
			nodeinc.crc32 = get_dword(pbuf);
			
			memset(ndlname, 0, 13);
			strncpy(ndlname, nodeinc.name, 12);
			ndlname[12] = 0;
			
			++i; /* the number of records read */
			
			if (i > nndl) {
				writeToLog(3, "ndl: the number of nodelist files in config and indexfile don`t match\n");
				rebuild = Yes;
				break;
			}
			
			if (0 != stricmp(ndlnames[i-1], ndlname)) {
				sprintf(str, "ndl: nodelist names changed: old:`%s' new:`%s'\n",
								ndlname, ndlnames[i-1]);
				writeToLog(2, str);
				rebuild = Yes;
				break;
			}
			
			sprintf(s, "%s%s", path, ndlname);
			
			if (access(s, 0)) {
				sprintf(str, "ndl: incorrect filename stored in nodeinc : `%s'\n", s);
				writeToLog(3, str);
				rebuild = Yes;
				break;
			}
			if ( crc32(s) != nodeinc.crc32 ) {
				sprintf(str, "ndl: CRC don`t match (%s)\n", s);
				writeToLog(3, str);
				rebuild = Yes;
				break;
			}
		}
		
		if (Chat_On)
			printf("done\n");
		
		close(h_inc);
		
		if (rebuild == Yes)
			break;
		
		if (i != nndl) {
			writeToLog(3, "ndl: the number of nodelist files in config and indexfile don`t match\n");
			rebuild = Yes;
			break;
		}

		rebuild = No;
		break;  /* no need to rebuild */
	}
	
	if (rebuild == Yes) {
		strcpy(str, "rebuilding nodelist indexes\n");
		writeToLog(5, str);
		if (Chat_On)
			printf("%s", str);
		if ( ndl_build(nndl, ndlnames, path, ext) == Fail ) {
			strcpy(str, "ndl error: nodelist rebuild failed\n");
			fprintf(stderr, "\7%s", str);
			writeToLog(0, str);
			abort();
		}
		writeToLog(4, "ndl: done rebuilding nodelist indexes\n");
	}
	else {
		writeToLog(3, "ndl: index files are ok and up-to-date\n");
	}
	
	for( i = 0; i < nndl; i++ )
		free( ndlnames[i] );
	

/* end of argument parsing */
/*/////////////////////////////////////////////////////////////////*/
	
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
	{
		if (!stricmp(tag, ndlTag->tag)) {
			mark = ndlTag;
			break;
		}
	}

	if (mark) { /* tag already defined */
		ndlTag = mark;
	}
	else {    /* create new tag */
		ndlTag = (struct ndlTags *)calloc(1, sizeof(struct ndlTags));
		ndlTag->next = NULL;
		/*////////////////////*/
		strcpy(ndlTag->tag, tag);
		ndlTag->active = Yes;
		/*////////////////////*/
		ndlTag->rec = ndlTag->recfirst = ndlTag->reclast = NULL;
		
		if (ndlTagLast) {
			ndlTagLast->next = ndlTag;
		}
		else {
			ndlTagFirst = ndlTag;
		}
		
		ndlTagLast = ndlTag;
	}

	ndlTag->rec = (struct nodelist *)calloc(1, sizeof(struct nodelist));
	ndlTag->rec->next = NULL;
	
	strcpy(ndlTag->rec->ext,  ext);
	strcpy(ndlTag->rec->path, path);
	ndlTag->rec->zone = zone;
	ndlTag->rec->reg = 0;
	strcpy(ndlTag->rec->allnames, allnames);
	
	if (ndlTag->reclast) {
		ndlTag->reclast->next = ndlTag->rec;
	}
	else {
		ndlTag->recfirst = ndlTag->rec;
	}
	ndlTag->reclast = ndlTag->rec;
	
	
	return;
}

int getndlnames(char *path, char *arg, char *ndllist[])
{
	char *p, w[256], s[256];
	int i, done, n, n0;
	short ftime, fdate0, fdate, ftime0;
	BOOL wild = No, found = No;
	struct find_t ffblk;
/*	int rc; */
	
	/* verifying all the names defined */
	for(i = 0; get_next_word(arg, w); ++i)
	{
		wild = No;
		if (NULL == (p = strchr(w, '.')))
		{
			/* no extention */
			strcat(w, ".???");
			wild = Yes;
		}
		else {
			++p;
			if (*p == '*' || !strcmp(p, "???")) {
				wild = Yes;
			}
		}
		
		sprintf(s, "%s%s", path, w);
		
		found = No;
		
		fdate0 = fdate = 0;
		ftime0 = ftime = 0;
		n = n0 = 0;
		done = _dos_findfirst(s, 0, &ffblk);
		
		while (0 == done) {             /* if the nodelist name is wildcarded */

			p = strrchr(ffblk.name, '.');
			if (p)
				n = atoi(++p);
			
			fdate = ffblk.wr_date;   /* select the newest file */
			ftime = ffblk.wr_time;
			if ( ( (fdate > fdate0) ||
						(fdate==fdate0 && ftime>ftime0) ||
						(fdate==fdate0 && ftime==ftime0 && wild==Yes && n > n0) ) &&
					( wild==No || (p && __isnumber(p)) ) )
			{
				strcpy( w , ffblk.name );
				found = Yes;
				if (wild == No)
					break;
			}
			
			fdate0 = fdate;
			ftime0 = ftime;
			n0 = n;
			
			done = _dos_findnext(&ffblk);
		}
		
		_dos_findclose(&ffblk);
		
		/*  no pattern or nodelist file found */
		if (found == No) {

			return 0;
			
		}

		if (i < MAXNDL) {
			ndllist[i] = (char *)malloc(sizeof(char)*256);
			strcpy(ndllist[i], w);
		}
		
	}  /* checking nodelist names (as defined in config) */
	
	
	return i;
	
}
