
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"
/*
#include "init.h"
*/

void init(char *cfg, char *arg0);
void reset_globals(int old_type);
void final_cleanup(void);
void free_caches (void);
void free_pkt_cache (void);

/*
extern BOOL RegisterRexxFunc(int type);
extern BOOL DeregisterRexxFunc(void);
*/
extern void banner(void);
extern void config(char *path, char *cfg_name);


/* os2sem.c */
extern BOOL create_bsy_sem(void);
extern void close_bsy_sem (void);

/* util.c */
extern void open_log(void);


/* defined in squish.c */
extern void      sq_failure_reason(void); 
extern SQ_ERRNO  validate_sqd(int sqd);
extern SQ_ERRNO  validate_sqi(int sqd, int sqi);
extern BOOL      save_idx(int sqi, SQIDX [], int n);

/* defarea.c */
extern void set_ts(char *areaname, dword ts);

/* toss.c */
extern void flush_dupes(void);

/* pkt.c */
extern void update_old_pkt(char *fname, PKT_AREA pkt);

/* bink.c */
extern void bnk_to_bsy(char *s, char *bsy);

void init_globals (void);

static void set_TZ(char *tzbuf);
static void set_path(char *arg0);
static void store_tsfile(void);


void check_lock()
{
#if !defined(__OS2__) && !defined(UNIX)
    time_t   time_cur, locklife = 0;
    struct   stat stat_file;
#endif
    unsigned long pid;
   
    FILE *f;

    /* lock... */
    if (!access(lockFile.FileName, 0)) {
	f = fopen(lockFile.FileName, "rt");
	fscanf(f, "%lu\n", &pid);
	fclose(f);
	/* Checking process PID */
#ifdef __OS2__
	if (DosKillProcess(DKP_PROCESSTREE, pid) == ERROR_NOT_DESCENDANT) {
#elif UNIX
	if (kill(pid, 0) == 0) {
#else
	if (stat(lockFile.FileName, &stat_file) != -1) {
	    time_cur = time(NULL);
	    locklife = (time_cur - stat_file.st_mtime)/60;
	}
	if (locklife < 180) {
#endif
	    printf("lock file found! exit...\n");
	    abort();
	} else {
	    remove(lockFile.FileName);
	    createLockFile(lockFile.FileName);
	}
    }
    else createLockFile(lockFile.FileName);
   
}

/*******************************/
void init(char *cfg, char *arg0)
{
	char tzbuf[256]={0}, buf[256];
	
/*	_fmode = O_BINARY;  // global */
	
	
	(void)set_path(arg0);
	(void)set_TZ(tzbuf);

	tStart = time(NULL); /* global */

	(void)init_catcher();
	(void)reset_globals(0);
	
	init_globals ();

	sprintf(buf, "ItraxPath: `%s'\n", ItraxPath);
	writeToLog(4, buf);
	
	config(ItraxPath, cfg);
	
	
	if (LogFile == NULL) {
		LogFile = (char *) malloc (strlen(DefLogName)+strlen(ItraxPath)+1);
		sprintf (LogFile, "%s%s", ItraxPath, DefLogName);
/*		open_log (); */
	}

	check_lock();

	open_log ();
	
	writeToLog(1, "Configuration read\n");
	
	if (strlen(tzbuf))
		writeToLog(0, tzbuf);
	
	initCharsets();
	if (outtabName) getctab(outtab, (unsigned char*) outtabName);
	if (intabName)  getctab(intab, (unsigned char*)  intabName);
	
	return;
}

void  catcher(int code)
{
	
	switch (code) {
		
	/*/////////*/
	case SIGABRT:
/*	case SIGBREAK: */
	case SIGINT:
	/*/////////*/
		writeToLog(0, "program aborted \n");
		break;

	default:
		writeToLog(0, "unknown exception occurred, aborting\n");
		break;
	}
	
	final_cleanup();
	exit(1);
	
}

/*************************************************************/
void init_catcher(void)
{
	static int inited;

	if (inited)
		return;

	inited=1;
	
/*	signal( SIGBREAK, catcher ); */
	signal( SIGINT,   catcher );
	signal( SIGABRT,  catcher );

/*	signal( SIGILL,   catcher );
	signal( SIGSEGV,  catcher );
	signal( SIGFPE,   catcher );
*/
	return;
}
/*************************************************************/
void final_cleanup(void)
{

	writeToLog(2, "doing cleanup...\n");
	
	cleanup ();
/*	close_bsy_sem (); */

	return;
}

/********************************************/
void reset_globals(int base_type)
{
	char buf[256];

	writeToLog(4, "reset_gl: set_tlist: area `%s' -> %ld",
					 CurMsg.areaname, CurMsg.maxts);
	set_ts(CurMsg.areaname, CurMsg.maxts);
	if (TsHandle != -1) {
		store_tsfile();
	}

	CurMsg.maxts = 0L;
	
	if (base_type == FIDO) {

		free(CurMsg.fido.fnames);
		

	} else if (base_type == SQUISH) {

		int sqd = CurMsg.squish.sqd;
		int sqi = CurMsg.squish.sqi;

		if (sqi != -1)
			save_idx(CurMsg.squish.sqi, CurMsg.squish.idxlist, CurMsg.squish.total);

		free(CurMsg.squish.idxlist);
		
		if (-1 != sqd &&  -1 != sqi)
		{
			close(sqd);
			close(sqi);
		}
		
		
	} else if (base_type == JAM) {

		close(CurMsg.jam.HdrHandle);
		close(CurMsg.jam.TxtHandle);
		close(CurMsg.jam.IdxHandle);

	/*  PKT  */
	} else if (base_type == PKT) {

		if (CurMsg.pkt.off) {
			sprintf(buf, "reset_globals: unread block left, saving file: `%s'\n",
							CurMsg.pkt.fname);
			writeToLog(4, buf);
			update_old_pkt(CurMsg.pkt.fname, CurMsg.pkt);
		}
		
		free(CurMsg.pkt.text);
		free(CurMsg.pkt.new_buf);
		free(CurMsg.pkt.fnames);

		

	/*  BINK  */

	}  else if (base_type == BINK) {
		
		if (CurMsg.bink.off) {
			void   kill_bsy( char *bsy);
			char bsy[256];
			
			sprintf(buf, "reset_globals: unread block left, saving file: `%s'\n",
							CurMsg.bink.fname);
			writeToLog(4, buf);
			(void) update_old_pkt(CurMsg.bink.fname, CurMsg.bink);
			bnk_to_bsy(CurMsg.bink.fname, bsy);
			sprintf(buf, "reset_globals: killing bsy: `%s'\n", bsy);
			writeToLog(3, buf);
			kill_bsy(bsy);
		}
		free(CurMsg.bink.text);
		free(CurMsg.bink.new_buf);
		free(CurMsg.bink.fnames);
		
	}
	/* BINK specific */
	
	free_text(TextFirst);  
	TextFirst = NULL;    
	free_text(ViaTextFirst);
	ViaTextFirst = NULL;
	
	memset(&ftn, 0, sizeof(FTN));
	memset(&CurMsg, 0, sizeof(struct CurMsg));
	
	CurMsg.squish.sqd = CurMsg.squish.sqi = -1;
	CurMsg.jam.HdrHandle = CurMsg.jam.TxtHandle = CurMsg.jam.IdxHandle = -1;
	
	return;
	
}

int createLockFile() {
    char pidstr[256];
    char *lockfile = lockFile.FileName;

    if ( (lockFile.handle = open(lockfile, O_CREAT | O_RDWR | O_EXCL, S_IRUSR | S_IWUSR)) < 0 ) {
       fprintf(stderr,"\007createLockFile: cannot create lock file \"%s\"\n",lockfile);
       writeToLog(0, "createLockFile: cannot create lock file \"%s\"m", lockfile);
       abort();
    }

#if defined UNIX    
    sprintf(pidstr, "%u\n", (unsigned)getpid());
#elif defined __MINGW32__
    sprintf(pidstr, "%u\n", (unsigned)_getpid());
#endif
    write (lockFile.handle, pidstr, strlen(pidstr));

    lseek(lockFile.handle, 0L, SEEK_END);
    lockFile.len = tell(lockFile.handle);
    
    return 0;
}


void RemoveLockFile()
{
    if (lockFile.handle != -1) {
	close(lockFile.handle);
	lockFile.handle = -1;
	remove(lockFile.FileName);
    }
}

void cleanup(void)
{
	void *tmp;

	/* first log the previous msg, if it was actually read and changed */
	if (CurMsg.n && ( (ftn.stat & i_Changed) ||
			CurMsg.type == PKT || CurMsg.type == BINK) )
		update_msg(CurMsg.type);
	

	(void)reset_globals(CurMsg.type);

	flush_dupes();

	while (TagFirst) {
		tmp = (void *) TagFirst;
		TagFirst = TagFirst->next;
		free (tmp);
	}
	TagLast = NULL;

	while (FidoAreaFirst)
	{
		tmp = (void *) FidoAreaFirst;
		FidoAreaFirst = FidoAreaFirst->next;
		free (tmp);
	}
	FidoAreaLast = NULL;
	
	while (SquishAreaFirst)
	{
		tmp = (void *) SquishAreaFirst;
		SquishAreaFirst = SquishAreaFirst->next;
		free (tmp);
	}
	SquishAreaLast = NULL;
	
	while (JamAreaFirst)
	{
		tmp = (void *) JamAreaFirst;
		JamAreaFirst = JamAreaFirst->next;
		free (tmp);
	}
	JamAreaLast = NULL;
	
	while (PktAreaFirst)
	{
		tmp = (void *) PktAreaFirst;
		PktAreaFirst = PktAreaFirst->next;
		free (tmp);
	}
	PktAreaLast = NULL;
	
	while (BinkAreaFirst)
	{
		tmp = (void *) BinkAreaFirst;
		BinkAreaFirst = BinkAreaFirst->next;
		free (tmp);
	}
	BinkAreaLast = NULL;
	
	while (TextAreaFirst)
	{
		tmp = (void *) TextAreaFirst;
		TextAreaFirst = TextAreaFirst->next;
		free (tmp);
	}
	TextAreaLast = NULL;
	
	while (CompressFirst)
	{
		tmp = (void *) CompressFirst;
		CompressFirst = CompressFirst->next;
		free (tmp);
	}
	CompressLast = NULL;
	
	while (EchoAreas)
	{
		tmp = (void *) EchoAreas;
		EchoAreas = EchoAreas->next;
		free (tmp);
	}
	
	if (DupeBasePath)
		free (DupeBasePath);
	DupeBasePath = NULL;

	free_caches ();
	
	
	if (FBox)
		free (FBox);
	FBox = NULL;

	if (TsPath)
		free (TsPath);
	TsPath = NULL;
	
	while (MaxMsgNoFirst)
	{
		tmp = (void *) MaxMsgNoFirst;
		MaxMsgNoFirst = MaxMsgNoFirst->next;
		free (tmp);
	}
	MaxMsgNoLast = NULL;
	
	while (GlobalVarList)
	{
		tmp = (void *) GlobalVarList;
		GlobalVarList = GlobalVarList->next;
		free (((struct _gv*)tmp)->var);
		free (((struct _gv*)tmp)->val);
		free (tmp);
	}
	
	while(TSlist)
	{
		tmp = (void *) TSlist;
		TSlist = TSlist->next;
		free (tmp);
	}

	/* free DATA */
	while (ndlTagFirst)
	{
		tmp = ndlTagFirst;
		ndlTagFirst = ndlTagFirst->next;
		free (tmp);
	}
	ndlTagLast = NULL;
	
	while (v9_ndlTagFirst)
	{
		tmp = v9_ndlTagFirst;
		v9_ndlTagFirst = v9_ndlTagFirst->next;
		free (tmp);
	}
	v9_ndlTagLast = NULL;
	
	while (SystemTagFirst)
	{
		tmp = SystemTagFirst;
		SystemTagFirst = SystemTagFirst->next;
		free (tmp);
	}
	SystemTagLast = NULL;
	
	while (AddrAttrTagFirst)
	{
		tmp = AddrAttrTagFirst;
		AddrAttrTagFirst = AddrAttrTagFirst->next;
		free (tmp);
	}
	AddrAttrTagLast = NULL;
	
	while (AddrNameTagFirst)
	{
		tmp = AddrNameTagFirst;
		AddrNameTagFirst = AddrNameTagFirst->next;
		free (tmp);
	}
	AddrNameTagLast = NULL;
	
	while (AddrPackerTagFirst)
	{
		tmp = AddrPackerTagFirst;
		AddrPackerTagFirst = AddrPackerTagFirst->next;
		free (tmp);
	}
	AddrPackerTagLast = NULL;
	
	while (AddrTagFirst)
	{
		tmp = AddrTagFirst;
		AddrTagFirst = AddrTagFirst->next;
		free (tmp);
	}
	AddrTagLast = NULL;
	
	while (AliasTagFirst)
	{
		tmp = AliasTagFirst;
		AliasTagFirst = AliasTagFirst->next;
		free (tmp);
	}
  AliasTagLast = NULL;
	
	while (AttrTagFirst)
	{
		tmp = AttrTagFirst;
		AttrTagFirst = AttrTagFirst->next;
		free (tmp);
	}
  AttrTagLast = NULL;
	
	while (DateAddrTagFirst)
	{
		tmp = DateAddrTagFirst;
		DateAddrTagFirst = DateAddrTagFirst->next;
		free (tmp);
	}
  DateAddrTagLast = NULL;
	
	while (DateTagFirst)
	{
		tmp = DateTagFirst;
		DateTagFirst = DateTagFirst->next;
		free (tmp);
	}
  DateTagLast = NULL;
	
	while (DowTagFirst)
	{
		tmp = DowTagFirst;
		DowTagFirst = DowTagFirst->next;
		free (tmp);
	}
  DowTagLast = NULL;
	
	while (FileAgeTagFirst)
	{
		tmp = FileAgeTagFirst;
		FileAgeTagFirst = FileAgeTagFirst->next;
		free (tmp);
	}
  FileAgeTagLast = NULL;
	
	while (FilesTagFirst)
	{
		tmp = FilesTagFirst;
		FilesTagFirst = FilesTagFirst->next;
		free (tmp);
	}
  FilesTagLast = NULL;
	
	while (LengthAddrTagFirst)
	{
		tmp = LengthAddrTagFirst;
		LengthAddrTagFirst = LengthAddrTagFirst->next;
		free (tmp);
	}
  LengthAddrTagLast = NULL;
	
	while (PktLenAddrTagFirst)
	{
		tmp = PktLenAddrTagFirst;
		PktLenAddrTagFirst = PktLenAddrTagFirst->next;
		free (tmp);
	}
  PktLenAddrTagLast = NULL;
	
	while (MsgAgeTagFirst)
	{
		tmp = MsgAgeTagFirst;
		MsgAgeTagFirst = MsgAgeTagFirst->next;
		free (tmp);
	}
  MsgAgeTagLast = NULL;
	
	while (MsgLenTagFirst)
	{
		tmp = MsgLenTagFirst;
		MsgLenTagFirst = MsgLenTagFirst->next;
		free (tmp);
	}
  MsgLenTagLast = NULL;
	
	while (NameTagFirst)
	{
		tmp = NameTagFirst;
		NameTagFirst = NameTagFirst->next;
		free (tmp);
	}
  NameTagLast = NULL;
	
	while (PassAddrTagFirst)
	{
		tmp = PassAddrTagFirst;
		PassAddrTagFirst = PassAddrTagFirst->next;
		free (tmp);
	}
  PassAddrTagLast = NULL;
	
	while (PassNameTagFirst)
	{
		tmp = PassNameTagFirst;
		PassNameTagFirst = PassNameTagFirst->next;
		free (tmp);
	}
  PassNameTagLast = NULL;
	
	while (PathTagFirst)
	{
		tmp = PathTagFirst;
		PathTagFirst = PathTagFirst->next;
		free (tmp);
	}
  PathTagLast = NULL;
	
	while (ReAddrTagFirst)
	{
		tmp = ReAddrTagFirst;
		ReAddrTagFirst = ReAddrTagFirst->next;
		free (tmp);
	}
  ReAddrTagLast = NULL;
	
	while (RoutingTagFirst)
	{
		tmp = RoutingTagFirst;
		RoutingTagFirst = RoutingTagFirst->next;
		free (tmp);
	}
  RoutingTagLast = NULL;
	
	while (SubjAddrTagFirst)
	{
		tmp = SubjAddrTagFirst;
		SubjAddrTagFirst = SubjAddrTagFirst->next;
		free (tmp);
	}
  SubjAddrTagLast = NULL;
	
	while (SubjectTagFirst)
	{
		tmp = SubjectTagFirst;
		SubjectTagFirst = SubjectTagFirst->next;
		free (tmp);
	}
  SubjectTagLast = NULL;
	
	while (TextStrTagFirst)
	{
		tmp = TextStrTagFirst;
		TextStrTagFirst = TextStrTagFirst->next;
		free (tmp);
	}
  TextStrTagLast = NULL;
	
	while (KludgeTagFirst)
	{
		tmp = KludgeTagFirst;
		KludgeTagFirst = KludgeTagFirst->next;
		free (tmp);
	}
  KludgeTagLast = NULL;
	
	while (EchoNameTagFirst)
	{
		tmp = EchoNameTagFirst;
		EchoNameTagFirst = EchoNameTagFirst->next;
		free (tmp);
	}
	EchoNameTagLast = NULL;			 
	
	while(MacroFirst) {
	    tmp = MacroFirst;
	    MacroFirst = MacroFirst->next;
	    free(tmp);
	}
	MacroFirst = NULL;
	
	free(intabName);
	free(intab);
	free(outtabName);
	free(outtab);

	intabName = NULL;
	intab = NULL;
	outtabName = NULL;
	outtab = NULL;

	if (SeqDir)
	  free(SeqDir);
	SeqDir = NULL;

	close_log();
	RemoveLockFile();
}

void init_globals (void)
{
	memset(&MainAddress, 0, sizeof(ADDR));
	memset(TemplatePath, 0, 128);
	memset(FlagDir, 0, 128);
	if (SeqDir)
	  free(SeqDir);
	SeqDir = NULL;
	SeqOutRun = 0;
	DebugLevel = 0;
	strcpy (CompressFile, "compress.cfg");
	Chat_On = Yes;
	LogOFF = No;
	ChatOFF = No;
	IgnoreBSY = No;
	TouchEchomail = Yes;
	UseRegExp = No;
	rxIcase = No;
	DupesTotal = 0;
	DupesMaxMsg = 100;
	DupesMaxOpen = 15;
	BadAreaDefined = No;
  sq_errno = jam_errno = 0;
}

/*****************************/
static void store_tsfile(void)
{
	TSL *tsl;

	lseek(TsHandle, 0, SEEK_SET);  
	setfsize(TsHandle, 0);
	
	for(tsl = TSlist; tsl; tsl = tsl->next)
		write(TsHandle, &tsl->ts, sizeof(TS));
	
	close(TsHandle);
	TsHandle = -1;
}


/**************************/
static void set_path(char *arg0)
{
	char *env = getenv("ITRAXP");
	
	if (NULL != env) {
		
		strncpy(ItraxPath, env, 127);
		ItraxPath[127] = '\0';
		if (!setpath(ItraxPath)) {
			printf("incorrect path in ITRAX variable `%s', \nassuming the current dir\n",
						ItraxPath);
			memset(ItraxPath, 0, 256);
		}
		
	}
	
	if (!strlen(ItraxPath) && strlen(arg0)) {
		char *p = arg0, *q = strrchr(p, DELIM);
		if (q) {
			strncpy(ItraxPath, p, q-p+1);
		}
	}

	return;
}

/*****************************/
static void set_TZ(char *tzbuf)
{
#ifdef __OS2__
	char *env = getenv("TZ");
	char tz[]="TZ=MSK-3DST";
	BOOL tz_ok = Yes;

	if (NULL == env) {
		tz_ok = No;
	}
	else {
		if (strlen(env) < 4)
			tz_ok = No;
		else {
			env += 3;
			if (*env == '-' || *env == '+')
				++env;
			if (*env == '\0' || *env < '0' || *env > '9')
				tz_ok = No;
		}
	}

	if (tz_ok == No) {
		putenv(tz);
		sprintf(tzbuf, "TZ not set or of incorrect format, setting `%s'\n", tz);
	}
#endif	
	tzset();
}

void free_pkt_cache (void)
{
 	void *tmp;
	while (PktCache)
	{
		tmp = (void *) PktCache;
		PktCache = PktCache->next;
		free (tmp);
	}
	PktCacheLast = NULL;
}

void free_caches (void)
{
 	void *tmp;
	free_pkt_cache();
	while (OipCache)
	{
		tmp = (void *) OipCache;
		OipCache = OipCache->next;
		free (tmp);
	}
	OipCacheLast = NULL;
	
	while (ArcCache)
	{
		tmp = (void *) ArcCache;
		ArcCache = ArcCache->next;
		free (tmp);
	}
	ArcCacheLast = NULL;
}
