
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

/****************************************************************************/
/*       Functions for direct interface with the JAM message base           */
/****************************************************************************/

#include <assert.h>

#include "itrax.h"


dword ReadHdrInfo(sword handle, JAMHDRINFO *HdrInfo);
dword WriteHdrInfo(sword handle, JAMHDRINFO *HdrInfo);
dword ReadIdx(sword handle, JAMIDXREC *Idx);
dword WriteIdx(sword handle, JAMIDXREC *Idx);
dword ReadHdr(sword handle, JAMHDR *Hdr);
dword WriteHdr(sword handle, JAMHDR *Hdr);


void adjust_lastread (const char* lr, int type);
BOOL select_jam_area(char *areaname, JAM_AREA *);

/* called from msg.c */
BOOL open_jam_bases(char *areaname, JAM_AREA *);

BOOL read_jam_msg(JAM_AREA *, TEXT **);
int  add_jam_msg(JAM_AREA * , TEXT *);
int  update_jam_msg(JAM_AREA *jam, TEXT *newtext);
BOOL kill_jam_msg(JAM_AREA *);

BOOL jam_pack(char *areatag);

void create_new_jambase(char *areapath);

/* used for squish and jam: upper kludges->text */
extern TEXT * ctrl_to_text(char *str);

/* defined in util.c */
extern word   str_to_flags( char *p );

/* defined in readmsg.c */
extern TEXT * str_to_text(char *buf);

/* defined in time.c */
extern char * unix_to_ftscdate(time_t t, char *s);

/* util.c */
int    get_TZ(void);


static BOOL   write_jam_msg(JAM_AREA *jam, char *sbuf, char *t);
static void   text_to_jam(TEXT *txt, JAM_AREA *jam, char **sbuf, char **t);
static char * expand_text(char *text, char *str, dword *len);
static void   jam_to_flags(JAM_AREA *jam, word *flag);
static word   getLoID(char *p);
static char * add_sfield(word LoID, char *buf, char *p, dword *Sublen);
static void   flags_to_jam(char *cflags, JAM_AREA *jam);
static BOOL   del_old_header(int h, dword WhatOffset);
/* static void   create_backup(char *areaname, JAM_AREA *jam); */
static void   close_jam_base(JAM_AREA *);

/************************************************************/
BOOL select_jam_area(char *areaname, JAM_AREA *jam_area)
{
	dword len;
	
	if (!open_jam_bases(areaname, jam_area))
		return Fail;
	
	/***************************************************/
	
	if (0L != lseek(jam_area->HdrHandle, 0L ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}

	len = HDRINFO_SIZE;
	if (len != ReadHdrInfo(jam_area->HdrHandle, &jam_area->HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTRDFILE;
		return Fail;
	}

	if (jam_area->HdrInfo.BaseMsgNum == 0L)
		jam_area->HdrInfo.BaseMsgNum = 1L;

	jam_area->totalMsg = jam_area->HdrInfo.ActiveMsgs;
	jam_area->nMsgRead = jam_area->curMsgRead = 0L;
	
	
	return Ok;
	
}

/*************************************************************************/
BOOL open_jam_bases(char *areaname, JAM_AREA *jam)
{
	char fname[256], areapath[128];
	char buf[256];

	areaname_to_path(areaname, areapath);

	sprintf( fname , "%s.jhr" , areapath );
	if( -1 == (jam->HdrHandle = open(fname, O_RDWR|O_BINARY)) ) {
		perror(fname);
		sprintf(buf, "openJAM: error opening *.jhr file");
		writeToLog(0, buf);
		return Fail;
	}

	sprintf( fname , "%s.jdt" , areapath );
	if( -1 == (jam->TxtHandle = open(fname, O_RDWR|O_BINARY)) ) {
		close (jam->HdrHandle);
		perror(fname);
		sprintf(buf, "openJAM: error opening *.jdt file");
		writeToLog(0, buf);
		return Fail;
	}

	sprintf( fname , "%s.jdx" , areapath );
	if( -1 == (jam->IdxHandle = open(fname, O_RDWR|O_BINARY)) ) {
		close (jam->HdrHandle);
		close (jam->TxtHandle);
		perror(fname);
		sprintf(buf, "openJAM: error opening *.jdx file");
		writeToLog(0, buf);
		return Fail;
	}
	
	strcpy (OutFileName, areapath);
	return Ok;
}

/********************************************/
BOOL read_jam_msg(JAM_AREA *jam, TEXT **text)
{
	dword WhatMsg, WhatOffset;
	char *subf = NULL, *tbuf = NULL, *cbuf = NULL;
	dword len = 0L, sublen = 0L, textlen = 0L;
	TEXT *ctext = NULL, *tmp_text = NULL;

	dword maxts;
	
	jam_errno = 0;

	if (jam->curMsgRead == 0L)
		jam->curMsgRead = jam->HdrInfo.BaseMsgNum-1;

	while(1) {

		if (jam->nMsgRead == jam->totalMsg) {
			jam_errno = JAMAPIMSG_NOMOREMSGS;
			return Fail;
		}

		WhatMsg = ++jam->curMsgRead;

		/* fetch index record */
		WhatOffset = (WhatMsg - jam->HdrInfo.BaseMsgNum) * IDX_SIZE;

		if (WhatOffset != lseek(jam->IdxHandle, WhatOffset ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_NOMOREMSGS;
			return Fail;
		}

		len = IDX_SIZE;
		if (len != ReadIdx(jam->IdxHandle, &jam->Idx)) {
			if (WhatOffset >= filelength(jam->IdxHandle))
				jam_errno = JAMAPIMSG_NOMOREMSGS;
			else
				jam_errno = JAMAPIMSG_CANTRDFILE;
			return Fail;
		}

		WhatOffset = jam->Idx.HdrOffset;
		
		/* deleted msg */
		if (jam->Idx.UserCRC == (dword)-1) {
			continue;
		}
	
		/* fetch header */

		if (WhatOffset != lseek(jam->HdrHandle, WhatOffset ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			return Fail;
		}
	
		len = HDR_SIZE;
		if (len != ReadHdr(jam->HdrHandle, &jam->Hdr)) {
			jam_errno = JAMAPIMSG_CANTRDFILE;
			return Fail;
		}

		/* check header */
		if (strcmp(jam->Hdr.Signature, HEADERSIGNATURE) != 0) {
			jam_errno = JAMAPIMSG_BADHEADERSIG;
			return Fail;
		}

		if (jam->Hdr.Revision != CURRENTREVLEV) {
			jam_errno = JAMAPIMSG_BADHEADERREV;
			return Fail;
		}

		if (jam->Hdr.Attribute & MSG_DELETED) {
			writeToLog(4, "jam read: del attr set, but index != 0xffff");
			continue;
		}

		++jam->nMsgRead;
		
		/* process timestamp */
		if (TsHandle != -1) {
			
			maxts = jam->Hdr.DateProcessed;
			
			if (!maxts)
				maxts = jam->Hdr.DateReceived;
			if (!maxts)
				maxts = jam->Hdr.DateWritten;

			if (maxts <= jam->ts)
				continue;
			
			if (maxts > CurMsg.maxts)
				CurMsg.maxts = maxts;
		}
		

		/* fetch subfields */
		len = jam->Hdr.SubfieldLen;
		
		if (len) {
			if (NULL == (subf = (char *)malloc(len+1))) {
				/*      jam_errno = JAMAPIMSG_LOWMEMO; */
				return Fail;
			}
			
			if (len != read(jam->HdrHandle, subf, len)) {
				jam_errno = JAMAPIMSG_CANTRDFILE;
				return Fail;
			}
			subf[len] = 0;
		}
		
		sublen = len;
		
		/* fetch text */
		WhatOffset = jam->Hdr.TxtOffset;

		
		if (WhatOffset != lseek(jam->TxtHandle, WhatOffset ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			return Fail;
		}

		len = jam->Hdr.TxtLen;
		
		if (len) {
			if (NULL == (tbuf = (char *)malloc(len+1))) {
				/*      jam_errno = JAMAPIMSG_LOWMEMO; */
				return Fail;
			}
		
			if (len != read(jam->TxtHandle, tbuf, len)) {
				jam_errno = JAMAPIMSG_CANTRDFILE;
				return Fail;
			}
			tbuf[len] = 0;
		}


		textlen = len;
		
		break;
		
	} /* while(1) */

	/* parse subfields */
	{
		JAMSUBFIELD *SubFldPtr;
		char buf[256], *p = subf, str[256];
		char toNode[36]={0}, fromNode[36]={0};
		ADDR addr;
		dword clen = 0L;

		while ((p - subf) + SUBF_SIZE < sublen)
		{
			dword l;
			
			SubFldPtr = (JAMSUBFIELD *)p;

			p += BSUBF_SIZE;
			
			l = (SubFldPtr->DatLen < 255) ? SubFldPtr->DatLen : 255;
			memcpy(buf, p, l);
			buf[l] = '\0';
			
			p += SubFldPtr->DatLen;
			
			switch (SubFldPtr->LoID) {

			/* TO */
			case JAMSFLD_DADDRESS:

					
					/* 4D/5D address extracted ok */
					if (str_to_addr(buf, &addr)) {

						jam->MsgHdr.to = addr;
						
						if ((jam->Hdr.Attribute & MSG_TYPENET) && addr.point) {
							sprintf(str, "\1TOPT %d", addr.point);
							if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
								return Fail;
						}

						addr.point = 0;
						addr_to_str(addr, toNode);
					}
					
					/* 2D address, parse manually */
					else {
						
						word net = 0, node = 0;
						
						sscanf(buf, "%hu/%hu", &net, &node);
						jam->MsgHdr.to.zone = jam->MsgHdr.to.point = 0;
						jam->MsgHdr.to.net = net;
						jam->MsgHdr.to.node = node;
						
					}
					
					break;
					
			/* FROM */
			case JAMSFLD_OADDRESS:

					/* 4D/5D address extracted ok */
					if (str_to_addr(buf, &addr)) {

						jam->MsgHdr.from = addr;
						
						if ((jam->Hdr.Attribute & MSG_TYPENET) && addr.point) {
							sprintf(str, "\1FMPT %d", addr.point);
							if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
								return Fail;
						}

						addr.point = 0;
						addr_to_str(addr, fromNode);
					}
					
					/* 2D address, parse manually */
					else {
						
						word net = 0, node = 0;

						sscanf(buf, "%hu/%hu", &net, &node);
						jam->MsgHdr.from.zone = jam->MsgHdr.from.point = 0;
						jam->MsgHdr.from.net = net;
						jam->MsgHdr.from.node = node;

					}
					
					break;

					
			/* FROM NAME */
			case JAMSFLD_SENDERNAME:
					
					strncpy(jam->MsgHdr.fromName, buf, 35);
					jam->MsgHdr.fromName[35] = '\0';
					
					break;

			/* TO NAME */
			case JAMSFLD_RECVRNAME:

					strncpy(jam->MsgHdr.toName, buf, 35);
					jam->MsgHdr.toName[35] = '\0';
					break;

			/* SUBJECT */
			case JAMSFLD_SUBJECT:
			case JAMSFLD_ENCLFILE:

					strncpy(jam->MsgHdr.subj, buf, 71);
					jam->MsgHdr.subj[71] = '\0';

					break;
					
			case JAMSFLD_MSGID:

					buf[255-8] = '\0';
					sprintf(str, "\1MSGID: %s", buf);
					
					if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
						return Fail;
					
					break;
				
			case JAMSFLD_REPLYID:
					
					buf[255-8] = '\0';
					sprintf(str, "\1REPLY: %s", buf);
					if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
						return Fail;
					
					break;
				
			case JAMSFLD_FLAGS:
					
					{
						char cflag[64]={0};
						word flag = str_to_flags(buf);
						
						(void) jam_to_flags(jam, &flag);

						flag_to_str(flag, cflag);
						
						if (strlen(cflag))
							sprintf(str, "\1FLAGS %s", cflag);
						else
							;
/*							strset(str, 0); */
					}
					
					if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
						return Fail;
					
					break;
				
			case JAMSFLD_PID:
					
					buf[255-6] = '\0';
					sprintf(str, "\1PID: %s", buf);
					
					if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
						return Fail;
					
					break;
					
			case JAMSFLD_FTSKLUDGE:
					
					/* Via goes down */
					if (!strnicmp(buf, "Via ", 4)) {

						buf[255-4] = '\0';

						sprintf(str, "\1%s\r", buf);

						if (NULL == (tbuf = expand_text(tbuf, str, &textlen)))
							return Fail;
					}
					/* other kludges go up */
					else {
						buf[255-1] = '\0';
						sprintf(str, "\1%s", buf);
						
						if (NULL == (cbuf = expand_text(cbuf, str, &clen)))
							return Fail;
					}
					
					break;

			case JAMSFLD_SEENBY2D:
					
					sprintf(str, "SEEN-BY: %s\r", buf);
					
					if (NULL == (tbuf = expand_text(tbuf, str, &textlen)))
						return Fail;
					
					break;
				
			case JAMSFLD_PATH2D:
					
					sprintf(str, "\1PATH: %s\r", buf);
					
					if (NULL == (tbuf = expand_text(tbuf, str, &textlen)))
						return Fail;
					
					break;
				
			case JAMSFLD_UNKNOWN:
					break;
					
			default:
					break;
			} /* switch */
		} /* while sublen */


		if ( (jam->Hdr.Attribute & MSG_TYPENET) &&
				strlen(toNode) && strlen(fromNode) )
		{

			p = cbuf;

			sprintf(buf, "\1INTL %s %s", toNode, fromNode);

			clen += strlen(buf);

			if (NULL == (cbuf = (char *)malloc(clen+1)))
				return Fail;

			strcpy(cbuf, buf);
			
			if (p) {
				strcat(cbuf, p);
				free(p);
			}
		
		}
		
	} /* parse subfield block */

	if (tbuf)
		tmp_text = (TEXT *)str_to_text(tbuf);

	if (cbuf) {
		ctext = ctrl_to_text(cbuf);
		(void) add_text(&tmp_text, ctext, 1, Mode_Bin);
		free_text(ctext);
	}

	*text = tmp_text;
	
	
	free(subf);
	free(cbuf);
	free(tbuf);

	return Ok;
}
/********************************************************/
static char *expand_text(char *text, char *str, dword *len)
{
	*len += strlen(str);

	if (!text)
		text = (char *)calloc(*len + 1, sizeof(char));
	else
		text = (char *)realloc(text, *len + 1);

	if (text)
		strcat(text, str);

	return text;
}

/*************************************************/
static void jam_to_flags(JAM_AREA *jam, word *flag)
{
	if (jam->Hdr.Attribute & MSG_IMMEDIATE)
		*flag |= f_Imm;
	if (jam->Hdr.Attribute & MSG_DIRECT)
		*flag |= f_Dir;
	if (jam->Hdr.Attribute & MSG_GATE)
		*flag |= f_Zon;
	if (jam->Hdr.Attribute & MSG_TRUNCFILE)
		*flag |= f_Tfs;
	if (jam->Hdr.Attribute & MSG_KILLFILE)
		*flag |= f_Kfs;
	if (jam->Hdr.Attribute & MSG_FPU)
		*flag |= f_Fpu;
	if (jam->Hdr.Attribute & MSG_LOCKED)
		*flag |= f_Lok;
	if (jam->Hdr.Attribute & MSG_CONFIRMREQ)
		*flag |= f_Cfm;

	return;
}
/*************************************************/
static void flags_to_jam(char *cflags, JAM_AREA *jam)
{
	char buf[256], cf[256], new_cflags[256]={0};
	word attr;

	strcpy(buf, cflags);
	
	while(get_next_word(buf, cf))
	{
		attr = str_to_flag(cf);
	
		if (attr == f_Imm)
			jam->Hdr.Attribute |= MSG_IMMEDIATE;

		else if (attr == f_Dir)
			jam->Hdr.Attribute |= MSG_DIRECT;

		else if (attr == f_Zon)
			jam->Hdr.Attribute |= MSG_GATE;

		else if (attr == f_Tfs)
			jam->Hdr.Attribute |= MSG_TRUNCFILE;

		else if (attr == f_Kfs)
			jam->Hdr.Attribute |= MSG_KILLFILE;

		else if (attr == f_Fpu)
			jam->Hdr.Attribute |= MSG_FPU;

		else if (attr == f_Lok)
			jam->Hdr.Attribute |= MSG_LOCKED;

		else if (attr == f_Cfm)
			jam->Hdr.Attribute |= MSG_CONFIRMREQ;

		else {

			if (strlen(new_cflags))
				strcat(new_cflags, " ");
			strcat(new_cflags, cf);
			
		}
	}

	strcpy(cflags, new_cflags);
	
	return;
}

/*************************************/
void  jam_to_ftn(JAM_AREA *jam, FTN *ftn)
{
	char ftsc_date[20];
	word attr = 0;
	
	ftn->to = jam->MsgHdr.to;
	ftn->from = jam->MsgHdr.from;
	
	strcpy(ftn->fromName, jam->MsgHdr.fromName);
	strcpy(ftn->toName, jam->MsgHdr.toName);
	strcpy(ftn->subject, jam->MsgHdr.subj);

	unix_to_ftscdate(jam->Hdr.DateWritten - TZOFF, ftsc_date);
	
	strcpy(ftn->DateTime, ftsc_date);
	
	ftn->utime = jam->Hdr.DateWritten - TZOFF; /* GMT -> local */

	if (jam->Hdr.Attribute & MSG_LOCAL)
		attr |= a_Local;
	if (jam->Hdr.Attribute & MSG_INTRANSIT)
		attr |= a_Intransit;
	if (jam->Hdr.Attribute & MSG_PRIVATE)
		attr |= a_Private;
	if (jam->Hdr.Attribute & MSG_READ)
		attr |= a_Received;
	if (jam->Hdr.Attribute & MSG_SENT)
		attr |= a_Sent;
	if (jam->Hdr.Attribute & MSG_KILLSENT)
		attr |= a_Killsent;
	if (jam->Hdr.Attribute & MSG_HOLD)
		attr |= a_Hold;
	if (jam->Hdr.Attribute & MSG_CRASH)
		attr |= a_Crash;
	if (jam->Hdr.Attribute & MSG_FILEREQUEST)
		attr |= a_Request;
	if (jam->Hdr.Attribute & MSG_FILEATTACH)
		attr |= a_File;
	if (jam->Hdr.Attribute & MSG_RECEIPTREQ)
		attr |= a_Rrq;
	if (jam->Hdr.Attribute & MSG_ORPHAN)
		attr |= a_Orphan;
	
	ftn->attr = attr;
	
	return;
}

/*************************************/
void  ftn_to_jam(FTN *ftn, JAM_AREA *jam)
{
	jam->MsgHdr.to = ftn->to;
	jam->MsgHdr.from = ftn->from;

	memset (&jam->MsgHdr.fromName, 0, sizeof(jam->MsgHdr.fromName));
	strcpy(jam->MsgHdr.fromName, ftn->fromName);
	
	memset (&jam->MsgHdr.toName, 0, sizeof(jam->MsgHdr.toName));
	strcpy(jam->MsgHdr.toName, ftn->toName);
	
	memset (&jam->MsgHdr.subj, 0, sizeof(jam->MsgHdr.subj));
	strcpy(jam->MsgHdr.subj, ftn->subject);

	jam->Hdr.DateWritten = ftn->utime + TZOFF;    /* from local */
	jam->Hdr.DateProcessed = time(NULL) + TZOFF;  /* to GMT     */
	

	jam->Hdr.Attribute = 0;

	if (ftn->attr & a_Local)
		jam->Hdr.Attribute |= MSG_LOCAL;
	
	if (ftn->attr & a_Intransit)
		jam->Hdr.Attribute |= MSG_INTRANSIT;
	
	if (ftn->attr & a_Private)
		jam->Hdr.Attribute |= MSG_PRIVATE;
	
	if (ftn->attr & a_Received)
		jam->Hdr.Attribute |= MSG_READ;
	
	if (ftn->attr & a_Sent)
		jam->Hdr.Attribute |= MSG_SENT;
	
	if (ftn->attr & a_Killsent)
		jam->Hdr.Attribute |= MSG_KILLSENT;
	
	if (ftn->attr & a_Hold)
		jam->Hdr.Attribute |= MSG_HOLD;
	
	if (ftn->attr & a_Crash)
		jam->Hdr.Attribute |= MSG_CRASH;
	
	if (ftn->attr & a_Request)
		jam->Hdr.Attribute |= MSG_FILEREQUEST;
	
	if (ftn->attr & a_File)
		jam->Hdr.Attribute |= MSG_FILEATTACH;
	
	if (ftn->attr & a_Rrq)
		jam->Hdr.Attribute |= MSG_RECEIPTREQ;
	
	if (ftn->attr & a_Orphan)
		jam->Hdr.Attribute |= MSG_ORPHAN;
	
	return;
}

/********************************************/
int add_jam_msg(JAM_AREA *jam, TEXT *txt)
{
	BOOL rc;
	char *sbuf = NULL, *t = NULL;
	dword len;
	dword MsgNo;

	if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}

	len = HDRINFO_SIZE;
	if (len != ReadHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTRDFILE;
		return Fail;
	}

	MsgNo = (filelength(jam->IdxHandle)/IDX_SIZE) + jam->HdrInfo.BaseMsgNum;
	
	text_to_jam(txt, jam, &sbuf, &t);
	
	jam->Hdr.MsgNum = MsgNo;
	jam->Idx.HdrOffset = filelength(jam->HdrHandle);
	jam->Hdr.TxtOffset = filelength(jam->TxtHandle);


	unlock(jam->HdrHandle, 0L, 1L);
	

	if (0 == lock(jam->HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base locked OK\n");
	else {
		writeToLog(0, "error locking JAM base\n");
		jam_errno = JAMAPIMSG_CANTLKFILE;
		return Fail;
	}
	
	rc = write_jam_msg(jam, sbuf, t);

	if (rc) {

		/* adjust HDRINFO */
		
		++jam->HdrInfo.ActiveMsgs;
		++jam->HdrInfo.ModCounter;
		
		if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			rc = Fail;
		}

		len = HDRINFO_SIZE;
		if (len != WriteHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
			jam_errno = JAMAPIMSG_CANTWRFILE;
			rc = Fail;
		}
		
	}

	unlock(jam->HdrHandle, 0L, 1L);
	
	free(t);
	free(sbuf);
	
	return rc ? jam->Hdr.MsgNum : Fail;
	
}

/********************************************/
int update_jam_msg(JAM_AREA *jam, TEXT *txt)
{
	BOOL rc;
	char *sbuf = NULL, *t = NULL;
	dword len;
	dword sublen = jam->Hdr.SubfieldLen;
	dword tlen = jam->Hdr.TxtLen;
	
	text_to_jam(txt, jam, &sbuf, &t);

	if (jam->Hdr.SubfieldLen > sublen) {

		del_old_header(jam->HdrHandle, jam->Idx.HdrOffset);
		jam->Idx.HdrOffset = filelength(jam->HdrHandle);
		
	}
	
	if (jam->Hdr.TxtLen > tlen)
		jam->Hdr.TxtOffset = filelength(jam->TxtHandle);

	if (0 == lock(jam->HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base locked OK\n");
	else {
		writeToLog(0, "error locking JAM base\n");
		jam_errno = JAMAPIMSG_CANTLKFILE;
		return Fail;
	}
	
	rc = write_jam_msg(jam, sbuf, t);

	if (rc) {

		/* adjust HDRINFO */
		if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			return Fail;
		}

		len = HDRINFO_SIZE;
		if (len != ReadHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
			jam_errno = JAMAPIMSG_CANTRDFILE;
			return Fail;
		}
		
		++jam->HdrInfo.ModCounter;
		
		if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			rc = Fail;
		}

		len = HDRINFO_SIZE;
		if (len != WriteHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
			jam_errno = JAMAPIMSG_CANTWRFILE;
			rc = Fail;
		}
		
	}

	if (0 == unlock(jam->HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base unlocked OK\n");
	else
		writeToLog(0, "error unlocking JAM base\n");
	
	free(t);
	free(sbuf);
	
	return rc ? jam->Hdr.MsgNum : Fail;
	
}
/**************************/
BOOL kill_jam_msg(JAM_AREA *jam)
{
	dword len, WhatOffset;
	char lr[256];

	if (0 == lock(jam->HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base locked OK\n");
	else {
		writeToLog(0, "error locking JAM base\n");
		jam_errno = JAMAPIMSG_CANTLKFILE;
		return Fail;
	}
	
	
	jam->Hdr.Attribute |= MSG_DELETED;
	
	/* store header record */
	WhatOffset = jam->Idx.HdrOffset;

	if (WhatOffset != lseek(jam->HdrHandle, WhatOffset ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	len = HDR_SIZE;
	if (len != WriteHdr(jam->HdrHandle, &jam->Hdr)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	jam->Idx.UserCRC = jam->Idx.HdrOffset = (dword)-1;
	
	/* store index record */
	WhatOffset = (jam->Hdr.MsgNum - jam->HdrInfo.BaseMsgNum) * IDX_SIZE;
	
	if (WhatOffset != lseek(jam->IdxHandle, WhatOffset, SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	len = IDX_SIZE;
	if (len != WriteIdx(jam->IdxHandle, &jam->Idx)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	/* adjust HDRINFO */
	len = HDRINFO_SIZE;
	if (len != ReadHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTRDFILE;
		return Fail;
	}

	--jam->HdrInfo.ActiveMsgs;
	++jam->HdrInfo.ModCounter;
		
	if (0L != lseek(jam->HdrHandle, 0L ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}

	len = HDRINFO_SIZE;
	if (len != WriteHdrInfo(jam->HdrHandle, &jam->HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	/* adjust standard lastread file */
	sprintf (lr, "%s.JLR", CurMsg.areapath);
	adjust_lastread (lr, LR_JAM);
	
	if (0 == unlock(jam->HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base unlocked OK\n");
	else
		writeToLog(0, "error unlocking JAM base\n");

	return Ok;
}

/***************************************************************/
static BOOL write_jam_msg(JAM_AREA *jam, char *sbuf, char *t)
{
	dword WhatOffset, len;

	jam_errno = 0;
	
	/* store index record */
	WhatOffset = (jam->Hdr.MsgNum - jam->HdrInfo.BaseMsgNum) * IDX_SIZE;
	
	if (WhatOffset != lseek(jam->IdxHandle, WhatOffset, SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	len = IDX_SIZE;
	if (len != WriteIdx(jam->IdxHandle, &jam->Idx)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	/* store header record */
	WhatOffset = jam->Idx.HdrOffset;

	if (WhatOffset != lseek(jam->HdrHandle, WhatOffset ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	len = HDR_SIZE;
	if (len != WriteHdr(jam->HdrHandle, &jam->Hdr)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	/* store subfields */
	len = jam->Hdr.SubfieldLen;
	if (len != write(jam->HdrHandle, sbuf, len)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	/* store text */
	WhatOffset = jam->Hdr.TxtOffset;

	if (WhatOffset != lseek(jam->TxtHandle, WhatOffset ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}
	
	len = jam->Hdr.TxtLen;
	if (len != write(jam->TxtHandle, t, len)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	return Ok;
	
}

/***************************************************************/
static void text_to_jam(TEXT *txt, JAM_AREA *jam, char **sbuf, char **t)
{
	ULONG crc_str(register ULONG crc, char *str, register long len);
	
	char buf[256], *pbuf;
	word LoID;
	dword sublen = 0L;

/*  if (jam->Hdr.Attribute & MSG_TYPENET) { */
		
		LoID = JAMSFLD_OADDRESS;
		addr_to_str(jam->MsgHdr.from, buf);
		if (strlen(jam->MsgHdr.from.domain)) {
			strcat(buf, "@");
			strcat(buf, jam->MsgHdr.from.domain);
		}
		
		*sbuf = (char *)add_sfield(LoID, buf, *sbuf, &sublen);
	
		LoID = JAMSFLD_DADDRESS;
		addr_to_str(jam->MsgHdr.to, buf);
		if (strlen(jam->MsgHdr.to.domain)) {
			strcat(buf, "@");
			strcat(buf, jam->MsgHdr.to.domain);
		}
		*sbuf = (char *)add_sfield(LoID, buf, *sbuf, &sublen);
		
/*  } // netmail */

	LoID = JAMSFLD_SENDERNAME;
	*sbuf = (char *)add_sfield(LoID, jam->MsgHdr.fromName, *sbuf, &sublen);
	
	LoID = JAMSFLD_RECVRNAME;
	*sbuf = (char *)add_sfield(LoID, jam->MsgHdr.toName, *sbuf, &sublen);
	strcpy(buf, jam->MsgHdr.toName);
	strLower(buf);
	jam->Idx.UserCRC = crc_str(-1L, buf, strlen(buf));

/*  if (jam->Hdr.Attribute & MSG_FILEATTACH)
    LoID = JAMSFLD_ENCLFILE;
    else
*/	
	LoID = JAMSFLD_SUBJECT;

	*sbuf = (char *)add_sfield(LoID, jam->MsgHdr.subj, *sbuf, &sublen);
	

	for( ; txt; txt = txt->next)
	{
		BOOL long_kludge = No;

		strcpy(buf, txt->str);
		
		if (long_kludge) {
			char s[256];
			sprintf(s, "text_to_jam: ignoring long kludge's tail: `%s'\n", buf);
			writeToLog(4, buf);
			if (strlen(buf) < 255 || buf[255] == '\r' || buf[255] == '\r')
				long_kludge = No;
			continue;
		}
		

		if (is_kludge(buf)) {

			int n;

			if (strlen(buf) == 255 && buf[255] != '\r' && buf[255] != '\r') {
				char s[256];
				sprintf(s, "text_to_jam: long kludge found: `%s'\n", buf);
				writeToLog(4, buf);
				long_kludge = Yes;
			}
			
			strip_CRLF(buf);

			if (!strncmp(buf, "SEEN-BY", 7))
				n = 0;
			else
				n = 1;

			pbuf = buf + n;

			LoID = getLoID(pbuf);

			if (LoID == JAMSFLD_FLAGS)
				flags_to_jam(pbuf, jam);
			else if (LoID == JAMSFLD_MSGID) {
				strLower(pbuf);
				jam->Hdr.MsgIdCRC = crc_str(-1L, pbuf, strlen(pbuf));
			}
			else if (LoID == JAMSFLD_REPLYID) {
				strLower(pbuf);
				jam->Hdr.ReplyCRC = crc_str(-1L, pbuf, strlen(pbuf));
			}
					
			if (LoID != (word)-1)
				*sbuf = (char *)add_sfield(LoID, pbuf, *sbuf, &sublen);

		}
		else {
			if (NULL == *t)
				*t = (char *)calloc(strlen(buf) + 1, sizeof(char));
			else
				*t = (char *)realloc(*t, strlen(*t) + strlen(buf) + 1);
			
			if (*t)
				strcat(*t, buf);
			else
				writeToLog(3, "text_to_jam error: low memo\n");
		}

	} /* text loop */
	
	jam->Hdr.SubfieldLen = sublen;
	
	jam->Hdr.TxtLen = (*t) ? strlen(*t) : 0L;
	
	return;
}

/*****************************************************/
static char * add_sfield(word LoID, char *buf, char *p, dword *Sublen)
{
	JAMBINSUBFIELD subf;
	int sublen = *Sublen, dlen = 0;
			
	dlen = strlen(buf);

	/* restrict dlen as needed (see jam spec) */
	switch (dlen) {

	case JAMSFLD_FTSKLUDGE:

			dlen = (dlen < 255) ? dlen : 255;
			break;

	case JAMSFLD_OADDRESS:
	case JAMSFLD_DADDRESS:
	case JAMSFLD_SENDERNAME: /*  these two are already restricted, */
	case JAMSFLD_RECVRNAME:  /*  though... */
	case JAMSFLD_MSGID:
	case JAMSFLD_REPLYID:
	case JAMSFLD_SUBJECT:
	case JAMSFLD_ENCLFILE:

			dlen = (dlen < 100) ? dlen : 100;
			break;
			
	case JAMSFLD_PID:
			dlen = (dlen < 40) ? dlen : 40;
			break;
			
	default:
			break;
	}
	
	sublen += dlen + BSUBF_SIZE;

	subf.LoID = LoID;
	subf.HiID = 0;
	subf.DatLen = dlen;
			
	if (NULL == p)
		p = (char *)malloc(sublen*sizeof(char));
	else
		p = (char *)realloc(p, sublen);
			
	if (p) {
		memcpy(&p[sublen-dlen-BSUBF_SIZE], &subf,
					BSUBF_SIZE);
		memcpy(&p[sublen-dlen], buf, dlen);
	}
	else
		writeToLog(3, "add_sfield (jam) error: low memo (sublen)\n");

	
	*Sublen = sublen;
	
	return p;
}

/**************************/
static word getLoID(char *p)
{
	int j, found;
	char junk[128];
	word LoID;
	
	for(j = 0 , found = -1; strlen( Kludges[j].s ); ++j)
	{
		if (!strnicmp(p, Kludges[j].s, (int)strlen(Kludges[j].s))) {
			found = Kludges[j].n;
			break;
		}
	}

	switch (found)  {

	case k_Intl:  
	case k_Topt:  
	case k_Fmpt:  LoID = (word)-1;           break;
	
	case k_Msgid: LoID =  JAMSFLD_MSGID;     break;
	case k_Reply: LoID =  JAMSFLD_REPLYID;   break;
	case k_Flags: LoID =  JAMSFLD_FLAGS;     break;
	case k_Pid:   LoID =  JAMSFLD_PID;       break;
	case k_Path:  LoID =  JAMSFLD_PATH2D;    break;
	case k_SeenBy:LoID =  JAMSFLD_SEENBY2D;  break;
	
	default:      LoID =  JAMSFLD_FTSKLUDGE; break;

	}

	if (LoID != JAMSFLD_FTSKLUDGE)
		get_next_word(p, junk);
	
	return LoID;
}

/**************************************************/
static BOOL del_old_header(int h, dword WhatOffset)
{
	JAMHDR Hdr;
	dword len;
	
	if (WhatOffset != lseek(h, WhatOffset ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}

	len = HDR_SIZE;
	if (len != ReadHdr(h, &Hdr)) {
		jam_errno = JAMAPIMSG_CANTRDFILE;
		return Fail;
	}

	Hdr.Attribute |= MSG_DELETED;
	Hdr.TxtLen = 0L;
	
	if (WhatOffset != lseek(h, WhatOffset ,SEEK_SET)) {
		jam_errno = JAMAPIMSG_SEEKERROR;
		return Fail;
	}

	len = HDR_SIZE;
	if (len != WriteHdr(h, &Hdr)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		return Fail;
	}

	return Ok;
}

/************************************/
BOOL jam_pack(char *areatag)
{
	JAM_AREA jam;
	JAMIDXREC *idx = NULL;
	register dword len, total_msg = 0L, WhatOffset;
	char *hdr_buf = NULL, *txt_buf = NULL;
	char *hdr_buf0 = NULL, *txt_buf0 = NULL;
	dword t_orig;
	int delta = 0, i;
	register int cnt = 0;
	char buf[256];
	long size;

	jam_errno = 0;

	memset(&jam, 0, sizeof(JAM_AREA));
	
	if (!select_jam_area(areatag, &jam))
		return Fail;

	unlock(jam.HdrHandle, 0L, 1L);

	if (0 == lock(jam.HdrHandle, 0L, 1L))
		writeToLog(3, "JAM base locked OK\n");
	else {
		writeToLog(0, "error locking JAM base\n");
		jam_errno = JAMAPIMSG_CANTLKFILE;
		return Fail;
	}
	

	t_orig = filelength(jam.TxtHandle);
	if (!t_orig) {
		writeToLog(3, "jam_pack: nothing to do\n");
		goto end;
	}
	
	txt_buf = (char *)malloc(t_orig * sizeof(char));
	if (NULL == txt_buf) {
		writeToLog(0, "jam_pack error: low memo allocating txt_buf\n");
		jam_errno = -1;
		goto end;
	}
	txt_buf0 = txt_buf;
	
	len = filelength(jam.HdrHandle);
	hdr_buf = (char *)malloc(len * sizeof(char));
	if (NULL == hdr_buf) {
		writeToLog(0, "jam_pack error: low memo allocating hdr_buf\n");
		jam_errno = -1;
		goto end;
	}
	hdr_buf0 = hdr_buf;
	
	/* get HDRINFO */
	lseek(jam.HdrHandle, 0L ,SEEK_SET);
	len = HDRINFO_SIZE;
	if (len != ReadHdrInfo(jam.HdrHandle, &jam.HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTRDFILE;
		goto end;
	}

	if( Chat_On ) {
		strUpper(areatag);
		sprintf(buf, "\npacking JAM base (%s) ", areatag);
		delta = (filelength(jam.IdxHandle)/IDX_SIZE) / (70 - strlen(buf));
		printf("%s", buf);
		if (delta == 0)
			delta = 1;
	}

	

	lseek(jam.IdxHandle, 0L,SEEK_SET);
	
	/* index walk */
	size = filelength(jam.IdxHandle);
	while(size > tell(jam.IdxHandle))
	{
		if (Chat_On) {
			if (++cnt == delta) {
				printf("3");
/*				putch(''); */
				cnt = 0;
			}
		}
		
		len = IDX_SIZE;
		if (len != ReadIdx(jam.IdxHandle, &jam.Idx)) {
			jam_errno = JAMAPIMSG_CANTRDFILE;
			goto end;
		}


		WhatOffset = jam.Idx.HdrOffset;

		/* deleted msg */
		if (jam.Idx.UserCRC == (dword)-1 && WhatOffset == (dword)-1)
			continue;

		/* fetch header */
		if (WhatOffset != lseek(jam.HdrHandle, WhatOffset ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			goto end;
		}
	
		len = HDR_SIZE;
		if (len != ReadHdr(jam.HdrHandle, &jam.Hdr)) {
			jam_errno = JAMAPIMSG_CANTRDFILE;
			goto end;
		}

		/* check header */
		if (strcmp(jam.Hdr.Signature, HEADERSIGNATURE) != 0) {
			jam_errno = JAMAPIMSG_BADHEADERSIG;
			goto end;
		}

		if (jam.Hdr.Revision != CURRENTREVLEV) {
			jam_errno = JAMAPIMSG_BADHEADERREV;
			goto end;
		}

		if (jam.Hdr.Attribute & MSG_DELETED) {
			continue;
		}

		WhatOffset = jam.Hdr.TxtOffset;

		jam.Hdr.MsgNum = ++total_msg;
		jam.Hdr.TxtOffset = txt_buf - txt_buf0;
		
		if (!idx)
			idx = (JAMIDXREC *)malloc(IDX_SIZE);
		else
			idx = (JAMIDXREC *)realloc(idx, total_msg*IDX_SIZE);

		if (!idx) {
			writeToLog(0, "jam_pack error: low memo allocating idx\n");
			jam_errno = -1;
			goto end;
		}
		
		idx[total_msg-1].UserCRC = jam.Idx.UserCRC;
		idx[total_msg-1].HdrOffset = HDRINFO_SIZE + (hdr_buf - hdr_buf0);
		
		memcpy(hdr_buf, &jam.Hdr, HDR_SIZE);

		hdr_buf += HDR_SIZE;

		/* fetch subfields */
		len = jam.Hdr.SubfieldLen;
		
		if (len) {
			
			if (len != read(jam.HdrHandle, hdr_buf, len)) {
				jam_errno = JAMAPIMSG_CANTRDFILE;
				goto end;
			}

			hdr_buf += len;
			
		}
		
		/* fetch text */
		if (WhatOffset != lseek(jam.TxtHandle, WhatOffset ,SEEK_SET)) {
			jam_errno = JAMAPIMSG_SEEKERROR;
			goto end;
		}

		len = jam.Hdr.TxtLen;
		
		if (len) {
		
			if (len != read(jam.TxtHandle, txt_buf, len)) {
				jam_errno = JAMAPIMSG_CANTRDFILE;
				goto end;
			}

			txt_buf += len;
			
		}
			
	}  /* index walk */

	jam.HdrInfo.ActiveMsgs = total_msg;
	jam.HdrInfo.ModCounter = 0L;
	jam.HdrInfo.BaseMsgNum = 1L;

/*  close_jam_base(&jam);
	
    create_backup(areatag, &jam);
*/	
	/* store new HDRINFO */


	unlock(jam.HdrHandle, 0L, 1L);
	
	lseek(jam.HdrHandle, 0L ,SEEK_SET);
#ifdef __BORLANDC__
	if (0 != _ftruncate(jam.HdrHandle, 0L)) {
#else
	if (0 != setfsize(jam.HdrHandle, 0L)) {
#endif      
		writeToLog(0, "jam_pack: error truncating *.jhr file\n");
		perror(".jhr");
		jam_errno = -1;
		goto end;
	}

	lock(jam.HdrHandle, 0L, 1L);
	
	lseek(jam.HdrHandle, 0L ,SEEK_SET);
	
	len = HDRINFO_SIZE;
	if (len != WriteHdrInfo(jam.HdrHandle, &jam.HdrInfo)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		goto end;
	}
	len = hdr_buf - hdr_buf0;
	if (len != write(jam.HdrHandle, hdr_buf0, len)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		goto end;
	}
	
	lseek(jam.IdxHandle, 0L ,SEEK_SET);
	if (0 != setfsize(jam.IdxHandle, 0L)) {
		writeToLog(0, "jam_pack: error truncating *.jdx file\n");
		jam_errno = -1;
		goto end;
	}
	lseek(jam.IdxHandle, 0L ,SEEK_SET);
	for (i = 0; i < total_msg; i++) {
	    if (IDX_SIZE != WriteIdx(jam.IdxHandle, &idx[i])) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		goto end;
	    }
	}
	
	lseek(jam.TxtHandle, 0L ,SEEK_SET);
	if (0 != setfsize(jam.TxtHandle, 0L)) {
		writeToLog(0, "jam_pack: error truncating *.jdt file\n");
		jam_errno = -1;
		goto end;
	}
	
	lseek(jam.TxtHandle, 0L ,SEEK_SET);
	len = txt_buf - txt_buf0;
	if (len != write(jam.TxtHandle, txt_buf0, len)) {
		jam_errno = JAMAPIMSG_CANTWRFILE;
		goto end;
	}
	
	if (Chat_On)   {
		printf(" done.\n");
		sprintf(buf, "JAM area: %s.  Original size= %ld. Packed size= %ld.\n",
						areatag, t_orig, len);
		printf("%s\n", buf);
		writeToLog(1, buf);
	}
	
end:
	
	free(txt_buf0);
	free(idx);
	free(hdr_buf0);
	
	unlock(jam.HdrHandle, 0L, 1L);
	
	close_jam_base(&jam);
	
	return jam_errno ? Fail : Ok;
}
/*************************************************************************/
/*
static void create_backup(char *areaname, JAM_AREA *jam)
{
	char fname[256], areapath[128];
	char buf[256];

	areaname_to_path(areaname, areapath);

	sprintf( fname , "%s.j#r" , areapath );

	if( -1 == (jam->HdrHandle = creat(fname, ACCFILE)) ) {
		perror(fname);
		sprintf(buf, "JAM backup: error creating *.j#r file");
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	sprintf( fname , "%s.j#t" , areapath );

	if( -1 == (jam->TxtHandle = creat(fname, ACCFILE)) ) {
		perror(fname);
		sprintf(buf, "JAM backup: error creating *.j#t file");
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	sprintf( fname , "%s.j#x" , areapath );

	if( -1 == (jam->IdxHandle = creat(fname, ACCFILE)) ) {
		perror(fname);
		sprintf(buf, "JAM backup: error creating *.j#x file");
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}
	
	return;
}
*/
/****************************************/
static void  close_jam_base(JAM_AREA *jam)
{
	close(jam->HdrHandle);
	close(jam->TxtHandle);
	close(jam->IdxHandle);

	jam->HdrHandle = jam->IdxHandle = jam->TxtHandle = -1;
	
	return;
}

void create_new_jambase(char *areapath)
{
	char jhr_name[256], jdt_name[256], jdx_name[256], buf[256];
	int jdx, jhr, jdt;
	JAMHDRINFO hdr;

	sprintf( jhr_name , "%s.jhr" , areapath );
	sprintf( jdt_name , "%s.jdt" , areapath );
	sprintf( jdx_name , "%s.jdx" , areapath );
	
	if (-1 == (jdt = open(jdt_name, O_WRONLY|O_CREAT|O_BINARY|O_EXCL, ACCFILE))) {
		perror(jdt_name);
		sprintf(buf, "create_new_jambase: error creating %s", jdt_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}
	close(jdt);
	
	if (-1 == (jdx = open(jdx_name, O_WRONLY|O_CREAT|O_BINARY|O_EXCL, ACCFILE))) {
		perror(jdx_name);
		sprintf(buf, "create_new_jambase: error creating %s", jdx_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}
	close(jdx);
	
	if (-1 == (jhr = open(jhr_name, O_WRONLY|O_CREAT|O_BINARY|O_EXCL, ACCFILE))) {
		perror(jhr_name);
		sprintf(buf, "create_new_jambase: error creating %s", jhr_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	memset(&hdr, 0, HDRINFO_SIZE);

	strncpy(hdr.Signature,  HEADERSIGNATURE, 3);
	hdr.DateCreated =  time(NULL);
	hdr.PasswordCRC = -1;
	hdr.BaseMsgNum = 1;

	if (HDRINFO_SIZE != WriteHdrInfo(jhr, &hdr)) {
		perror(jhr_name);
		sprintf(buf, "create_new_jambase: error wriing %s", jhr_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		close(jhr);
		abort();
	}
	
	close(jhr);
	
	return;
}

dword ReadHdrInfo(sword handle, JAMHDRINFO *HdrInfo)
{
   byte buf[HDRINFO_SIZE], *pbuf = buf;
   dword len;

   len = read(handle, &buf, HDRINFO_SIZE);

   if (len != HDRINFO_SIZE) {
      return len;
   } /* endif */

   /* 04 bytes Signature */
   memmove(HdrInfo->Signature, pbuf, (size_t)4);
   pbuf += 4;

   /* 04 bytes DateCreated */
   HdrInfo->DateCreated = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes ModCounter */
   HdrInfo->ModCounter = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes ActiveMsgs */
   HdrInfo->ActiveMsgs = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes PasswordCRC */
   HdrInfo->PasswordCRC = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes BaseMsgNum */
   HdrInfo->BaseMsgNum = get_dword(pbuf);
   pbuf += 4;

   /* 1000 bytes RSRVD */
   memmove(HdrInfo->RSRVD, pbuf, (size_t)1000);
   pbuf += 1000;

   assert(pbuf - buf == HDRINFO_SIZE);

   return len;
}


dword WriteHdrInfo(sword handle, JAMHDRINFO *HdrInfo)
{
   byte buf[HDRINFO_SIZE], *pbuf = buf;
   
   /* 04 bytes Signature */
   memmove(pbuf, HdrInfo->Signature, (size_t)4);
   pbuf += 4;

   /* 04 bytes DateCreated */
   put_dword(pbuf, HdrInfo->DateCreated);
   pbuf += 4;

   /* 04 bytes ModCounter */
   put_dword(pbuf, HdrInfo->ModCounter);
   pbuf += 4;

   /* 04 bytes ActiveMsgs */
   put_dword(pbuf, HdrInfo->ActiveMsgs);
   pbuf += 4;

   /* 04 bytes PasswordCRC */
   put_dword(pbuf, HdrInfo->PasswordCRC);
   pbuf += 4;

   /* 04 bytes BaseMsgNum */
   put_dword(pbuf, HdrInfo->BaseMsgNum);
   pbuf += 4;

   /* 1000 bytes RSRVD */
   memmove(pbuf, HdrInfo->RSRVD, (size_t)1000);
   pbuf += 1000;

   assert(pbuf - buf == HDRINFO_SIZE);

   return write(handle, &buf, HDRINFO_SIZE);
}

dword ReadIdx(sword handle, JAMIDXREC *Idx)
{
   byte buf[IDX_SIZE], *pbuf = buf;
   dword len;

    len = read(handle, &buf, IDX_SIZE);
    
   if (len != IDX_SIZE) {
      return len;
   } /* endif */

   /* 04 bytes UserCRC */
   Idx->UserCRC = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes HdrOffset */
   Idx->HdrOffset = get_dword(pbuf);
   pbuf += 4;

   assert(pbuf - buf == IDX_SIZE);

   return len;
}

dword WriteIdx(sword handle, JAMIDXREC *Idx)
{
   byte buf[IDX_SIZE], *pbuf = buf;

   /* 04 bytes UserCRC */
   put_dword(pbuf, Idx->UserCRC);
   pbuf += 4;

   /* 04 bytes HdrOffset */
   put_dword(pbuf, Idx->HdrOffset);
   pbuf += 4;

   assert(pbuf - buf == IDX_SIZE);

   return write(handle, &buf, IDX_SIZE);
}

dword ReadHdr(sword handle, JAMHDR *Hdr)
{
   byte buf[HDR_SIZE], *pbuf = buf;
   dword len;


   len = read(handle, &buf, HDR_SIZE);

   if (len != HDR_SIZE) {
      return len;
   } /* endif */

   /* 04 bytes Signature */
   memmove(Hdr->Signature, pbuf, (size_t)4); 
   pbuf += 4;

   /* 02 bytes Revision */
   Hdr->Revision = get_word(pbuf);
   pbuf += 2;

   /* 02 bytes ReservedWord */
   Hdr->ReservedWord = get_word(pbuf);
   pbuf += 2;

   /* 04 bytes SubfieldLen */
   Hdr->SubfieldLen = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes TimesRead */
   Hdr->TimesRead = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes MsgIdCRC */
   Hdr->MsgIdCRC = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes ReplyCRC */
   Hdr->ReplyCRC = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes ReplyTo */
   Hdr->ReplyTo = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes Reply1st */
   Hdr->Reply1st = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes ReplyNext */
   Hdr->ReplyNext = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes DateWritten */
   Hdr->DateWritten = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes DateReceived */
   Hdr->DateReceived = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes DateProcessed */
   Hdr->DateProcessed = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes MsgNum */
   Hdr->MsgNum = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes Attribute */
   Hdr->Attribute = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes Attribute2 */
   Hdr->Attribute2 = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes TxtOffset */
   Hdr->TxtOffset = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes TxtLen */
   Hdr->TxtLen = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes PasswordCRC */
   Hdr->PasswordCRC = get_dword(pbuf);
   pbuf += 4;

   /* 04 bytes Cost */
   Hdr->Cost = get_dword(pbuf);
   pbuf += 4;

   assert(pbuf - buf == HDR_SIZE);

   return len;
}

dword WriteHdr(sword handle, JAMHDR *Hdr)
{
   byte buf[HDR_SIZE], *pbuf = buf;

   /* 04 bytes Signature */
   memmove(pbuf, Hdr->Signature, (size_t)4); 
   pbuf += 4;

   /* 02 bytes Revision */
   put_word(pbuf, Hdr->Revision);
   pbuf += 2;

   /* 02 bytes ReservedWord */
   put_word(pbuf, Hdr->ReservedWord);
   pbuf += 2;

   /* 04 bytes SubfieldLen */
   put_dword(pbuf, Hdr->SubfieldLen);
   pbuf += 4;

   /* 04 bytes TimesRead */
   put_dword(pbuf, Hdr->TimesRead);
   pbuf += 4;

   /* 04 bytes MsgIdCRC */
   put_dword(pbuf, Hdr->MsgIdCRC);
   pbuf += 4;

   /* 04 bytes ReplyCRC */
   put_dword(pbuf, Hdr->ReplyCRC);
   pbuf += 4;

   /* 04 bytes ReplyTo */
   put_dword(pbuf, Hdr->ReplyTo);
   pbuf += 4;

   /* 04 bytes Reply1st */
   put_dword(pbuf, Hdr->Reply1st);
   pbuf += 4;

   /* 04 bytes ReplyNext */
   put_dword(pbuf, Hdr->ReplyNext);
   pbuf += 4;

   /* 04 bytes DateWritten */
   put_dword(pbuf, Hdr->DateWritten);
   pbuf += 4;

   /* 04 bytes DateReceived */
   put_dword(pbuf, Hdr->DateReceived);
   pbuf += 4;

   /* 04 bytes DateProcessed */
   put_dword(pbuf, Hdr->DateProcessed);
   pbuf += 4;

   /* 04 bytes MsgNum */
   put_dword(pbuf, Hdr->MsgNum);
   pbuf += 4;

   /* 04 bytes Attribute */
   put_dword(pbuf, Hdr->Attribute);
   pbuf += 4;

   /* 04 bytes Attribute2 */
   put_dword(pbuf, Hdr->Attribute2);
   pbuf += 4;

   /* 04 bytes TxtOffset */
   put_dword(pbuf, Hdr->TxtOffset);
   pbuf += 4;

   /* 04 bytes TxtLen */
   put_dword(pbuf, Hdr->TxtLen);
   pbuf += 4;

   /* 04 bytes PasswordCRC */
   put_dword(pbuf, Hdr->PasswordCRC);
   pbuf += 4;

   /* 04 bytes Cost */
   put_dword(pbuf, Hdr->Cost);
   pbuf += 4;

   assert(pbuf - buf == HDR_SIZE);

   return write(handle, &buf, HDR_SIZE);
}
