
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"

/* builds itrack nodelist indexes */

enum { ZONE = 1, REG, HOST, HUB, NODE, PVT, HOLD, DOWN };


BOOL ndl_build(int nndl, char *ndlnames[], char *path, char *ext)
{

	FILE *h_ndl;       /* nodelist file */
	int h_idx, h_inc;  /* index files   */
	long pos = 0L, len;
	char buf[256], junk[256], temp[256], fname[256];
	register char c, *p = NULL, *q = NULL;
	char tag_name[4], cnode[6];
	int  zone, region, host, hub, node, pvt, hold, down;
	int  kndl, tag = 0;
	struct nodeidx nodeidx;
	struct nodeinc nodeinc;
	char n_idx_buf[NODEIDX_SIZE], *pbuf;
	char n_inc_buf[NODEINC_SIZE];
	
/*	_fmode = O_BINARY;  // should be global */
	
		
	/* first create both index files */
		
	sprintf(fname, "%snodeidx.%s", path, ext);
	if (-1 == (h_idx = open(fname, O_CREAT|O_WRONLY|O_TRUNC|O_BINARY, ACCFILE))) {
/*	if ( -1 == (h_idx = creat(fname, ACCFILE))) { */
		sprintf(buf, "ndl_build: error creating index file `%s'\n", fname);
		writeToLog(0, buf);
		return Fail;
	}
	
	sprintf(fname, "%snodeinc.%s", path, ext);
	if (-1 == (h_inc = open(fname, O_CREAT|O_WRONLY|O_TRUNC|O_BINARY, ACCFILE))) {
/*	if ( -1 == (h_inc = creat(fname, ACCFILE))) { */
		sprintf(buf, "ndl_build: error creating index file `%s'\n", fname);
		writeToLog(0, buf);
		return Fail ;
	}
	
	
	/* walk nodelist name list */
	
	for( kndl = 0; kndl < nndl; ++kndl )
	{
		
		/* form next nodelist name */
		
		sprintf(fname, "%s%s", path, ndlnames[kndl] );
		
		if ( NULL == (h_ndl = fopen(fname, "rb"))) {
			sprintf(buf, "ndl_build: error opening nodelist file `%s'\n", fname);
			writeToLog(0, buf);
			return Fail;
		}
		
		zone = region = host = hub = node = pvt = hold = down = 0;
		
		while (NULL != fgets(temp, 255, h_ndl))
		{
			/* skip comments */
			if (temp[0] == ';')
				continue;
			
			/* save previos pos */
			pos = ftell(h_ndl) - strlen(temp);
			
			/* strip trailing <LF> */
			p = &temp[strlen(temp)-1];
			if (*p == '\n')
				*p = '\0';
/*			p = &temp[strlen(temp)-1];
			if (*p == '\r')
				*p = '\0';
*/				

			/* too short (probably the end) */
			if (strlen(temp) < 2)
				continue;

			strcpy(junk, temp);
			
			/* get first nodelist field (tag name) */
			p = strchr(junk, ',');
			if (p == NULL) {
				sprintf(buf, "ndl_build: error parsing string `%s'\n", temp);
				writeToLog(3, buf);
				continue;
			}
			
			junk[p-junk] = 0;
			strncpy(tag_name, junk, 3);
			
			tag_name[3] = 0;           /* only the first 3 chars matter */

			/* get second nodelist field (value of tag - zone, net, etc.) */
			
			++p;                       /* skip over 1st comma               */
			
			q = strchr(p, ',');        /* search for next comma             */
			
			if (q == NULL) {
				sprintf(buf, "ndl_build: error parsing string `%s'\n", temp);
				writeToLog(3, buf);
				continue;              /* no comma found, discard the whole string */
			}
			
			p[q-p] = '\0';
			strncpy(cnode, p, 5);    /* cnode holds a char value of current number */
			cnode[5] = 0;            /* 5 chars max (enough to stuff ushort)       */
			
			/*  if cnode holds junk, discard the whole string  */
			if (No == __isnumber(cnode))
				continue;

			
			c = tag_name[0];
			
			/* resolve ambiguity */
			if (c == 'H')
				c = tag_name[2];  /* `s' for hoSt */
													/* `b' for huB  */
													/* `l' for hoLd */
			
			tag = 0;
			
			switch (c)
			{
			case 'Z':  /* Zone */
				++zone;
				tag = ZONE;
				break;
				
			case 'R':  /* Region */
				++region;
				tag = REG;
				break;

			case 's':  /* hoSt */
				++host;
				tag = HOST;
				break;
				
			case 'b':  /* huB  */
				++hub;
				break;
				
			case 'l':  /* hoLd */
				++hold;
				break;

			case 'P':  /* Pvt */
				++pvt;
				break;
				
			case 'D':  /* Down */
				++down;
				break;
				
			case '\0': /* Node */
				++node;
				break;
				
			default:
				break;
				
			}   /* switch */
			
			/* if Zone Reg or hoSt, write index file */
			if (tag) {
				nodeidx.stat = (char)(tag-1);
				nodeidx.number = (word)atoi(cnode);
				nodeidx.listno = (char)(kndl+1);
				nodeidx.pos = pos;
				
				memset(n_idx_buf, '\0', NODEIDX_SIZE);
				pbuf = n_idx_buf;
				memmove(pbuf, &(nodeidx.stat), 1);
				pbuf++;
				put_word(pbuf, nodeidx.number);
				pbuf += 2;
				memmove(pbuf, &(nodeidx.listno), 1);
				pbuf++;
				put_dword(pbuf, nodeidx.pos);
				
				len = NODEIDX_SIZE;
				if (len != write(h_idx, &n_idx_buf, len)) {
					close(h_idx);
					close(h_inc);
					fclose(h_ndl);
					writeToLog(3, "ndl_build: error writing h_idx\n");
					return Fail;
				}
			} /* write index */
			
		}   /* h_ndl walk  */

		fclose(h_ndl);
		sprintf(buf, "%s:\n", ndlnames[kndl]);
		printf("%s", buf);
		writeToLog(4, buf);
		sprintf(buf, "         found %d zones, %d regions, %d hosts, %d hubs, %d nodes\n",
					zone, region, host, hub, node);
		printf("%s", buf);
		writeToLog(4, buf);
		sprintf(buf, "         Pvt:%d Hold:%d Down:%d\n", pvt, hold, down);
		printf("%s", buf);
		writeToLog(4, buf);
		
		nodeinc.namelen = (char) strlen(ndlnames[kndl]);
		memset(nodeinc.name,0,12);
		strncpy(nodeinc.name, ndlnames[kndl], 12); 
		sprintf(fname, "%s%s", path, ndlnames[kndl] );
		nodeinc.crc32 = crc32(fname);
		
		memset(&n_inc_buf, '\0', NODEINC_SIZE);
		pbuf = n_inc_buf;
		*pbuf = nodeinc.namelen;
		pbuf++;
		memmove(pbuf, &(nodeinc.name), 12);
		pbuf += 12;
		put_dword(pbuf, nodeinc.crc32);
		
		len = NODEINC_SIZE;
		if (len != (write(h_inc, &n_inc_buf, NODEINC_SIZE))) {
			close(h_idx);
			close(h_inc);
			writeToLog(3, "ndl_build: error writing h_inc\n");
			return Fail;
		}
		
	}  /* end of walking nodelist name list */
	
	close(h_idx);
	close(h_inc);
	
	return Ok;
}
