
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

/****************************************************************************/
/*       Functions for direct interface with the SQUISH message base        */
/****************************************************************************/

#include <assert.h>

#include "itrax.h"

#define MAX_READ_RETRY 10
#define ONE_SEC 1000L
#define AT_END -1

enum { MSG_FORWARD = 0, MSG_BACKWARD, FREE_FORWARD, FREE_BACKWARD };

word sz_sqhdr = SQHDR_SIZE; /* the default */

dword ReadSqBase(int sqd, SQBASE *sqbase);
dword WriteSqBase(int sqd, SQBASE *psqbase);
dword ReadXMsg(sword handle, XMSG *pxmsg);
dword WriteXMsg(sword handle, XMSG *pxmsg);
dword ReadSqHdr(sword handle, SQHDR *psqhdr);
dword WriteSqHdr(sword handle, SQHDR *psqhdr);
dword ReadSqIdx(sword handle, SQIDX *psqidx);
dword WriteSqIdx(sword handle, SQIDX *psqidx);



void adjust_lastread (const char* lr, int type);
/*** main squish base procedures: select/read/kill/write ***/

/* called from scan.h */
BOOL select_squish_area(char *areaname, SQAREA *sqarea);

/* called from readmsg.h */
BOOL read_squish_msg(SQAREA *squish, TEXT **first);

/* called from msg.c */
BOOL  open_squish_bases(char *areaname, int *sqd, int *sqi);
BOOL  kill_squish_msg(SQAREA *squish);

int    add_squish_msg(int sqd, int sqi, XMSG xmsg, char *cbuf, char *tbuf);
BOOL   update_squish_msg(char *cbuf, char *tbuf);


/*** other functions ***/

/* called from readmsg.h */
void      xmsg_to_ftn( XMSG xmsg, FTN *ftn);

/* called from msg.c */
void      ftn_to_xmsg( FTN ftn, XMSG *xmsg);
void      text_to_sqbase(TEXT *txt, char **cbuf, char **tbuf);

/* used also in scan.c, readmsg.c, catcher.c */
void      sq_failure_reason(void);

/* used also in catcher.c */
SQ_ERRNO  validate_sqd(int sqd);
SQ_ERRNO  validate_sqi(int sqd, int sqi);


/* called from rewind.c */
void sq_rewind(void);

/* used for squish and jam: upper kludges->text */
TEXT * ctrl_to_text(char *str);

BOOL sq_pack(char *areatag);

/* used in catcher.c */
BOOL save_idx(int sqi, SQIDX [], int n);

void create_new_sqbase(char *areapath);

/* defined in readmsg.c */
extern TEXT * str_to_text(char *buf);


/* defined in time.c */
extern dword sqdate_to_unix(struct _stamp st);
extern dword ftscdate_to_unix(char *date);
extern char * unix_to_ftscdate(dword t, char *s);
extern void  unix_to_sqdate(time_t t, struct _stamp *st);


/*** local functions ***/

static BOOL   goto_frame(int h, FOFS a_off);
static BOOL   write_SQHDR(int h, FOFS off, SQHDR *sqhdr);
static BOOL   read_SQHDR(int h, FOFS off, int frame_type, SQHDR *sqhdr);
static BOOL   validate_SQHDR(SQHDR sqhdr, int type, FOFS off);
static BOOL   read_sqbase(int sqd, SQBASE *sqbase);
static BOOL   update_sqbase(int sqd, SQBASE sqbase);
/* static BOOL   read_SQIDX(int h, FOFS off, SQIDX *sqidx); */
static BOOL   validate_SQIDX(SQIDX sqidx);
static BOOL   write_SQIDX( int sqi, dword off, SQIDX sqidx);
static BOOL   update_chains(int sqd, FOFS d_off, SQBASE *sqbase);

static int    fill_idx_list(int sqi, SQIDX *[]);
static BOOL   add_to_idxlist(SQIDX *sqidx);
static BOOL   do_trunc(int sqi);

static SQ_ERRNO walk_sqd(int sqd, int mode);
/* static FOFS    sqdOff_to_sqiOff(int sqi, FOFS off); */

/* write functions */
static FOFS  write_msg(int sqd, SQBASE *sqbase, XMSG *xmsg, char *cbuf, char *tbuf);
static FOFS get_hole(int sqd, dword next_frame, dword mlen);
static dword squishHash(char *f);
static FOFS update_to_new_frame(int sqd, FOFS cur_off, char *cbuf, char *tbuf, XMSG xmsg);
static word rewrite_frame(int sqd, FOFS cur_off, XMSG xmsg, char *cbuf, char *tbuf);
/* static void create_backup(char *areaname, SQAREA *sq); */


/************************************************************/
BOOL select_squish_area(char *areaname, SQAREA *sqarea)
{
	int sqd = -1, sqi = -1;
	SQBASE sqbase;
	word len = SQBASE_SIZE;

	sq_errno = 0L;
	
	if (!open_squish_bases(areaname, &sqd, &sqi))
		return Fail;
	
	/* validate sqd */
	
/*
	if (SQ_NOERR != (sq_errno = validate_sqd(sqd))) {
		close(sqd);
		return Fail;
	}
*/
	
	/* validate sqi */
	
/*
	if ( SQ_NOERR != (sq_errno = validate_sqi(sqd, sqi)) ) {
		close(sqi);
		close(sqd);
		return Fail;
	}
*/
	
	/***************************************************/
	if ((sqd != -1) && filelength(sqd)) {
		if (Fail == read_sqbase(sqd, &sqbase)) {
			sq_errno = SQD_READ_ERR;
			return Fail;
		}

		if (0L != lseek(sqd ,0L ,SEEK_SET)) {
			sq_errno = SQD_SEEK_ERR;
			close(sqi);
			close(sqd);
			return Fail;
		}


		if (len != ReadSqBase(sqd, &sqbase)) {
			sq_errno = SQD_READ_ERR;
			close(sqi);
			close(sqd);
			return Fail;
		}

		sqarea->num_msg = sqbase.num_msg;
	}
	else
		sqarea->num_msg = 0L;

	sqarea->n = 0;
	sqarea->cur_frame = 0L;
	sqarea->sqd = sqd;
	sqarea->sqi = sqi;
	sqarea->total = sqarea->num_msg;
	sqarea->idxlist = NULL;
	
	if (sqarea->num_msg) {
		int n;
		if (-1 == (n = fill_idx_list(sqi, &sqarea->idxlist)))
			return Fail;
		if (n != sqarea->total) {
			char buf[256];
			sprintf(buf, "select_squish_area error: msg number, expected %d got %d\n",
							sqarea->total, n);
			writeToLog(0, buf);
			free(sqarea->idxlist);
			return Fail;
		}
	}

	return Ok;
	
}
/**************************************************************/
BOOL read_squish_msg(SQAREA *squish, TEXT **text)
{
	SQHDR sqhdr;
	SQIDX sqidx;
	dword len, clen, tlen, mlen;
	char *cbuf = NULL, *tbuf = NULL;
	TEXT *ctext = NULL, *tmp_text = NULL, *tmp;
	int sqd = squish->sqd;
	dword maxts;
	BOOL fmpt = No, topt = No;
	int iline = 0, i;
	
	sq_errno = 0;
	
	while(1)
	{
		
		if (squish->n == squish->num_msg) {
			sq_errno = SQ_NOMOREMSGS;
			return Fail;
		}

		sqidx = squish->idxlist[squish->n++];
		
		if (!read_SQHDR(sqd, sqidx.ofs, FRAME_NORMAL, &sqhdr))
			return Fail;

		

	    
		if (XMSG_SIZE != ReadXMsg(sqd, &(squish->xmsg))) {
			sq_errno = XMSG_READ_ERR;
			return Fail;
		}
		
		
/*
		if (sqidx.umsgid != squish->xmsg.umsgid) {
			writeToLog(0, "read_squish_msg, fatal error: current frame umsgid in sqd and sqi don`t match\n");
			printf("fatal: idx=%x xmsg=%x\n", sqidx.umsgid, squish->xmsg.umsgid);
			sq_errno = -1;
			return Fail;
		}
*/

		/* process timestamp */
		if (TsHandle != -1) {

			char buf[256];
			char ftsc[256];
			
			maxts = sqdate_to_unix(squish->xmsg.date_arrived);
			
			if (!maxts) {
				writeToLog(6, "arrived time was zero, try written\n");
				maxts = sqdate_to_unix(squish->xmsg.date_written);
			}

			sprintf (buf, "read_sq: cur msg ts=%d, area ts=%d\n", (unsigned)maxts, (unsigned)squish->ts);
			writeToLog(6, buf);
			unix_to_ftscdate (maxts, ftsc);
			sprintf (buf, "read_sq: cur msg ts=%s, area ts=", ftsc);
			unix_to_ftscdate (squish->ts, ftsc);
			strcat (buf, ftsc);
			strcat (buf, "\n");
			writeToLog(6, buf);

			if (maxts <= squish->ts) {
				writeToLog(6, "read_sq: cur msg ts <= area ts, skip message\n");
				continue;
			}
			
			writeToLog(6, "read_sq: cur msg ts > area ts, newer message, read it\n");

			if (maxts > CurMsg.maxts) {
				sprintf (buf, "set CurMsg.maxts to %d\n", (unsigned)maxts);
				writeToLog(6, buf);
				CurMsg.maxts = maxts;
			}
		}
		
		break;
		
	} /* while(1) */
	
	squish->cur_frame = sqidx.ofs;  /* only needed for possible update */

	clen = sqhdr.clen;
	mlen = sqhdr.msg_length;
	tlen = mlen - XMSG_SIZE - clen;

	if (NULL == (cbuf = (char *)calloc(clen+1, sizeof(char)))) {
		sq_errno = LOW_MEMO;
		return Fail;
	}

	len = clen;
	if (len != read(sqd, cbuf, clen)) {
		sq_errno = CTRL_READ_ERR;
		free(cbuf);
		return Fail;
	}

	if (NULL == (tbuf = (char *)calloc(tlen+1, sizeof(char)))) {
		sq_errno = LOW_MEMO;
		return Fail;
	}

	len = tlen;
	if (len != read(sqd, tbuf, tlen)) {
		sq_errno = TEXT_READ_ERR;
		free(cbuf);
		free(tbuf);
		return Fail;
	}

	tmp_text = str_to_text(tbuf);

	ctext = ctrl_to_text(cbuf);

	for (tmp = ctext, i=0; tmp; tmp = tmp->next, ++i)
	{
		if (!is_kludge (tmp->str)) {
			break;
		}
		if (!strncmp ("TOPT", &tmp->str[1], 4)) {
			topt = Yes;
		}
		else if (!strncmp ("FMPT", &tmp->str[1], 4))
			fmpt = Yes;
		else if (!strncmp ("INTL", &tmp->str[1], 4))
      iline = i+1;
	}

	
	
	if (!topt && squish->xmsg.dest.point) {
		TEXT* tmp = (TEXT *) calloc (1, sizeof (TEXT));
		if (tmp) {
			sprintf (tmp->str, "\1TOPT %d\r", squish->xmsg.dest.point);
			(void) add_text(&ctext, tmp, iline+1, Mode_Bin);
			free (tmp);
		}
	}
	
	if (!fmpt && squish->xmsg.orig.point) {
		TEXT* tmp = (TEXT *) calloc (1, sizeof (TEXT));
		if (tmp) {
			sprintf (tmp->str, "\1FMPT %d\r", squish->xmsg.orig.point);
			(void) add_text(&ctext, tmp, iline+1, Mode_Bin);
			free (tmp);
		}
	}

	(void) add_text(&tmp_text, ctext, 1, Mode_Bin);

	free_text(ctext);

	free(cbuf);
	free(tbuf);

	free_text(*text);

	*text = tmp_text;
	
	
	return Ok;
}
/****************************************************/
int add_squish_msg(int sqd, int sqi, XMSG xmsg, char *cbuf, char *tbuf)
{
	SQBASE sqbase;
	SQIDX sqidx;
	FOFS new_sqhdr_off;
	char buf[256];
	
	if (-1 == sqd || -1 == sqi) {
		strcpy(buf, "add_squish_msg error: squish base(s) not opened\n");
		fprintf(stderr, "\7%s\n", buf);
		writeToLog(0, buf);
		abort();
	}

	/*   Obtain  exclusive  access to  the  Squish  base.  */
	if (0 == lock(sqd, 0L, 1L)) {
		writeToLog(3, "sqd file locked OK\n");
	}
	else {
		writeToLog(0, "error locking sqd file\n");
		sq_errno = BASE_LOCKED;
		return Fail;
	}
	
	/*   Re-read a copy of the Squish base header. */
	if (Fail == read_sqbase(sqd, &sqbase))
		return Fail;
	
	if (Fail == (new_sqhdr_off =  write_msg(sqd, &sqbase, &xmsg, cbuf, tbuf)))
		return Fail;
	
	/*  Rewrite the Squish base header. */
	if (Fail == update_sqbase(sqd, sqbase))
		return Fail;

	/*  Relinquish exclusive access to the Squish base. */
	unlock(sqd, 0L, 1L);
	
	/* adjust *.SQI */
	sqidx.ofs = new_sqhdr_off;
	sqidx.umsgid = xmsg.umsgid;
	sqidx.hash = squishHash(xmsg.to);

	/* we are adding to the current squish base */
	if (sqi == CurMsg.squish.sqi) {
		if (!add_to_idxlist(&sqidx)) {
			writeToLog(0, "error adding to idxlist\n");
			return Fail;
		}
	}
	/* we are adding to another squish base */
	else {
		if (Fail == write_SQIDX( sqi, filelength(sqi), sqidx))
			return Fail;
	}
	
	if (sqbase.max_msg && sqbase.high_msg > sqbase.max_msg) {

		SQAREA squish;
		
		sprintf(buf, "high_msg(%d) > max_msg(%d), kill the 1st msg at %08lx\n",
						(unsigned)sqbase.high_msg, (unsigned)sqbase.max_msg, sqbase.begin_frame);
		writeToLog(4, buf);


		squish.sqd = sqd;
		squish.sqi = sqi;
		squish.cur_frame = sqbase.begin_frame;
		squish.n = 0;

		/* delete in the same base */
		if (sqi == CurMsg.squish.sqi) {

			int i;
			
			squish.idxlist = CurMsg.squish.idxlist;

			for(i = 0; i < CurMsg.squish.total; ++i)
				if (squish.idxlist[i].ofs == sqbase.begin_frame) {
					squish.n = i+1;
					break;
				}

			if (!squish.n) {
				writeToLog(0, "add_squish_msg error (adjust_max): this shit should never happen\n");
				abort();
			}
		}
		
		kill_squish_msg(&squish);
		--sqbase.high_msg;
		
	}
	
	return sqbase.high_msg;

}
/*********************************************/
BOOL update_squish_msg(char *cbuf, char *tbuf)
{
	int sqd = CurMsg.squish.sqd;
	int sqi = CurMsg.squish.sqd;
	word rc;
	XMSG xmsg = CurMsg.squish.xmsg;
	SQIDX sqidx;
	FOFS new_off, cur_off = CurMsg.squish.cur_frame;
	char buf[256];
	
	if (cur_off == 0L) {
		writeToLog(2, "squish_update error: no active msg, cur_off = 0L\n");
		return Fail;
	}
	
	if (-1 == sqd || -1 == sqi) {
		strcpy(buf, "update_squish_msg error: squish base(s) not opened\n");
		fprintf(stderr, "\7%s\n", buf);
		writeToLog(0, buf);
		abort();
	}

	if (NULL == cbuf)
		cbuf = (char *)calloc(1, sizeof(char));
	if (NULL == tbuf)
		tbuf = (char *)calloc(1, sizeof(char));
	
	
	/*   Obtain  exclusive  access to  the  Squish  base.  */
	if (0 == lock(sqd, 0L, 1L)) {
		writeToLog(3, "sqd file locked OK\n");
	}
	else {
		writeToLog(3, "error locking sqd file\n");
		sq_errno = BASE_LOCKED;
		return Fail;
	}
	
	
	writeToLog(3, "sq_update: try to update msg in the same frame\n");
	
	if (0 == (rc = rewrite_frame(sqd, cur_off, xmsg, cbuf, tbuf))) {
		writeToLog(3, "sq_update: rewrite_frame failed (error) \n");
		return Fail;
	}
	
	if (rc == 1) {
		writeToLog(3, "sq_update: msg updated in the same frame ok\n");
		new_off = cur_off;
	}
	else {
		writeToLog(3, "sq_update: msg doesn`t fit in the same frame, create new frame\n");
		
		if (Fail == (new_off = update_to_new_frame(sqd, cur_off, cbuf, tbuf, xmsg))) {
			writeToLog(3, "sq_update: update_to_new_frame failed (error)\n");
			return Fail;
		}
	}
	
	/*  Relinquish exclusive access to the Squish base. */
	unlock(sqd, 0L, 1L);
	
	sqidx = CurMsg.squish.idxlist[CurMsg.squish.n-1];
	sqidx.ofs = new_off;
	sqidx.hash = squishHash(xmsg.to);

	CurMsg.squish.idxlist[CurMsg.squish.n-1] = sqidx;
		
	return Ok;
}

/*******************************************************/
BOOL kill_squish_msg(SQAREA *squish)
{
	int sqd = squish->sqd;
	int sqi = squish->sqi;
	SQBASE sqbase;
	char buf[256], lr[256];

	if (-1 == sqd || -1 == sqi) {
		strcpy(buf, "kill_squish_msg panic: squish base(s) not opened\n");
		fprintf(stderr, "\7%s\n", buf);
		writeToLog(0, buf);
		abort();
	}


	/* read _sqbase header for future reference */
	if (Fail == read_sqbase(sqd, &sqbase))
		return Fail;
	
	/* update chains (effectively kill the message) */
	if (Fail == (BOOL)update_chains(sqd, squish->cur_frame, &sqbase))
		return Fail;

	/* adjust max msg numbers */
	if (sqbase.num_msg==0 || sqbase.high_msg==0) {
		writeToLog(3, "trying to do delete in empty sqbase\n");
		return Fail;
	}
				
	--sqbase.num_msg;
	--sqbase.high_msg;
	
	/* update _sqbase */
	if (Fail == update_sqbase(sqd, sqbase))
		return Fail;
	
	/* delete in another file */
	/* can happen only when adding, if we adjust then number of the msg */
	/* according to the max_msg field in sqbase */
	/* (always the 1st record) */
	if (!squish->n)
		return do_trunc(sqi);

	squish->idxlist[squish->n-1].ofs = -1;
	squish->idxlist[squish->n-1].umsgid = -1;

	/* adjust standard lastread file */
	sprintf (lr, "%s.SQL", CurMsg.areapath);
	adjust_lastread (lr, LR_SQUISH);

	
	return Ok;
	
}
/*****************************************************************************/
static FOFS write_msg(int sqd, SQBASE *sqbase, XMSG *xmsg, char *cbuf, char *tbuf)
{
	SQHDR sqhdr, new_sqhdr, hole;
	
	dword new_sqhdr_off, xmsg_off;
	dword old_last_frame, hole_off, prev_free, next_free;
	dword clen = 0L, tlen = 0L, mlen;
	char buf[256];

	if (cbuf)
		clen = strlen(cbuf)+1;
	else
		clen = 1;
	if (tbuf)
		tlen = strlen(tbuf)+1;
	else
		tlen = 1;
	
	mlen = clen+tlen+XMSG_SIZE;
	
/*  sz_sqhdr = sqbase->sz_sqhdr; */
	old_last_frame = sqbase->last_frame;
	
	/*   Scan  the messages in the  "free_frame" chain to  see if any
			frame  is  large  enough  to accommodate  the  new  message. */
	if ((dword)-1L == (hole_off = get_hole(sqd, sqbase->free_frame, mlen))) {
		writeToLog(3, "sqwrite: error finding hole in squish base\n");
		return Fail;
	}
	if (hole_off) {

		sprintf(buf, "got hole at  %08x\n", (unsigned)hole_off );
		writeToLog(3, buf);
		
		/* read SQHDR of found hole */
		if (!read_SQHDR(sqd, hole_off, FRAME_FREE, &hole)) {
			return Fail;
		}
		prev_free = hole.prev_frame;
		next_free = hole.next_frame;

		/* update free chain */
		if (!prev_free) { /* hole was first frame */
			sqbase->free_frame = next_free;
		}
		else { /* update frame before hole */
			if (!read_SQHDR(sqd, prev_free, FRAME_FREE, &sqhdr)) {
				return Fail;
			}

			sqhdr.next_frame = next_free;
			
			if (!write_SQHDR(sqd, prev_free, &sqhdr)) {
				return Fail;
			}
		} /* hole not first */

		if (!next_free) { /* hole was last */
			sqbase->last_free_frame = prev_free;
		}
		else { /* update frame after hole */
			if (!read_SQHDR(sqd, next_free, FRAME_FREE, &sqhdr)) {
				return Fail;
			}
			
			sqhdr.prev_frame = prev_free;
			
			if (!write_SQHDR(sqd, next_free, &sqhdr)) {
				return Fail;
			}
		} /* hole not last */
		new_sqhdr_off = hole_off;
		new_sqhdr = hole;
		/* free chain updated */
	}
	else {
	/*   If no free frame was  found, allocate a frame at the  end of
			the file, using the "end_frame" offset from the base header. */

		writeToLog(3, "hole not found, adding msg to sqbase\n");
		
		new_sqhdr_off = sqbase->end_frame;
		/* Increment  the  "end_frame" value  by  the  amount of  space  allocated. */
		sqbase->end_frame += SQHDR_SIZE + mlen;
		
	} /* no hole found */
	
	sprintf(buf, "writing new msg at frame off: %08X\n", (unsigned)new_sqhdr_off );
	writeToLog(3, buf);

	/* Link the new SQHDR frame into the end of the message  chain,
		adjusting  the begin_frame  and last_frame  pointers in  the
		base header (if necessary). */
	sqbase->last_frame = new_sqhdr_off;
	if (!sqbase->begin_frame) {
		sqbase->begin_frame = new_sqhdr_off;
	}
	
	/* form sqhdr for new message */
	new_sqhdr.id = SQHDRID;
	new_sqhdr.next_frame = 0L;
	new_sqhdr.prev_frame = old_last_frame;
	if (hole_off == 0)
		new_sqhdr.frame_length = mlen;
	new_sqhdr.msg_length = mlen;
	new_sqhdr.clen = clen;
	
	/*  Set  the  "frame_type"  field  of the  new  SQHDR  to "frame
			update" and write at the appropriate offset. */
	new_sqhdr.frame_type = FRAME_UPDATE; /* frame is being updated */
	
	/* write new msg at new_sqhdr_off */
	/* write SQHDR */
	if (!write_SQHDR(sqd, new_sqhdr_off, &new_sqhdr)) {
		return Fail;
	}

	/*  Copy  the "uid"  field from  the base  header into  the XMSG
			header of  the message to be  written. */
	xmsg->umsgid = sqbase->uid;
		
	/*  Also set  the MSGUID  flag in the "attr" field of the XMSG header.  */
	xmsg->attr |= MSGUID;
	
	
	/* Then increment the "uid" field in the base header. */
	++sqbase->uid;
	
	/* Increment  the  num_msg  and  high_msg values  in  the  base header. */
	++sqbase->num_msg;
	++sqbase->high_msg;

	/*  Take the offset  of the  new SQHDR, and  add the  "sz_sqhdr" value 
			from the base  header.  Seek to that offset  and write the XMSG header. */
	
	xmsg_off = new_sqhdr_off + (dword)sz_sqhdr;
	/* SEEK to new pos for XMSG */
	if (!goto_frame(sqd, xmsg_off)) {
		return Fail;
	}
	
	/* write XMSG */
	if (XMSG_SIZE != WriteXMsg( sqd, xmsg)) {
	    writeToLog(3, "sqwrite: error writing XMSG, frame at %08X\n", (unsigned)new_sqhdr_off);
	    return Fail;
	}
	/* write CTRL */
	if (cbuf) {
		if ( clen != write( sqd, cbuf, clen ) ){
			sprintf(buf, "sqwrite: error writing CTRL, frame at %08X\n", (unsigned)new_sqhdr_off);
			writeToLog(3, buf);
			return Fail;
		}
	}
	else {
		char p = '\0';
		write(sqd, &p, 1);
	}
		
	/* write Text */
	if (tbuf) {
		if ( tlen != write( sqd, tbuf, tlen ) ){
			printf("sqwrite: error writing text, frame at %08X\n", (unsigned)new_sqhdr_off);
			return Fail;
		}
	}
	else {
		char p = '\0';
		write(sqd, &p, 1);
	}
	
	/*  Set the frame type of the new message to "normal". */
	new_sqhdr.frame_type = FRAME_NORMAL; /* normal frame */
	/* update new SQHDR */
	if (!write_SQHDR(sqd, new_sqhdr_off, &new_sqhdr)) {
		return Fail;
	}

	if (old_last_frame) {
		/* read SQHDR of old last frame */
		if (!read_SQHDR(sqd, old_last_frame, FRAME_NORMAL, &sqhdr)) {
			return Fail;
		}
	
		sqhdr.next_frame = new_sqhdr_off;

		/* update SQHDR of old last frame */
		if (!write_SQHDR(sqd, old_last_frame, &sqhdr)) {
			return Fail;
		}
	}
	
	return new_sqhdr_off;
}

/**************************/
static BOOL goto_frame(int h, FOFS a_off)
{
	FOFS off;
	char buf[256];
	
	off = lseek(h, a_off, SEEK_SET);
	if (off != a_off) {
		sprintf(buf, "goto_frame: lseek error in *.sqd,  asked:%08lX, got:%08lX\n",
						a_off, off);
		writeToLog(2, buf);
		sq_errno = SQD_SEEK_ERR;
		return Fail;
	}
	sq_errno = SQ_NOERR;
	return Ok;
}
	
/*****************************************************/
static BOOL write_SQHDR(int h, FOFS off, SQHDR *sqhdr)
{
	char buf[256];
	dword nwrite;
	
	if (!goto_frame(h, off))
		return Fail;
	
	if (sz_sqhdr != (nwrite = WriteSqHdr(h, sqhdr))) {
		sprintf(buf, "write_SQHDR: error writing sqhdr, asked:%d, written:%ld, offset:%08lX\n",
					sz_sqhdr, nwrite, off);
		writeToLog(2, buf);
		sq_errno = SQD_WRITE_ERR;
		return Fail;
	}
	sq_errno = SQ_NOERR;
	return Ok;
}
	
/****************************************************/
static BOOL read_SQHDR(int h, FOFS off, int frame_type, SQHDR *sqhdr)
{
	int cnt = MAX_READ_RETRY;
	char buf[256];
	dword nread;
	
	if (!goto_frame(h, off))
		return Fail;
	
	if (sz_sqhdr != (nread = ReadSqHdr(h, sqhdr))) {
		sprintf(buf, "read_SQHDR: error reading sqhdr, asked:%d, read:%ld, offset:%08lX\n",
					sz_sqhdr, nread, off);
		writeToLog(2, buf);
		sq_errno = SQD_READ_ERR;
		return Fail;
	}
	
	/* in case the current message is being updated by another process */
	
	while (cnt--)
	{
		if (!validate_SQHDR(*sqhdr, frame_type, off)) {
			if (sq_errno == MSG_UPDATED) {
			/*  DosSleep(ONE_SEC); */
			}
			else
				return Fail;  /* really error occured */
		}
		else
			break;
	}

	/* attempt counter expired */
	if (sq_errno == MSG_UPDATED)
		return Fail;  
	
	sq_errno = SQ_NOERR;
	return Ok;
}

/****************************************************************/
static BOOL validate_SQHDR(SQHDR sqhdr, int frame_type, FOFS off)
{
	char buf[256];
	
	if (sqhdr.id != SQHDRID) {
		sprintf(buf, "validate_SQHDR error: invalid ID:%08lX in SQHDR at offset %08lX\n",
						sqhdr.id, off);
		writeToLog(2, buf);
		sq_errno = SQD_INVALID_ID;
		return Fail;
	}
	
	if (sqhdr.frame_type != frame_type) {
		if (sqhdr.frame_type == FRAME_UPDATE) {
			sprintf(buf, "msg is locked at offset %08lX\n", off);
			sq_errno = MSG_UPDATED;
		}
		else {
			sprintf(buf, "validate_SQHDR error: invalid frame type in SQHDR, expected: %d, got %d at offset %08lX\n",
						sqhdr.frame_type, frame_type, off);
			sq_errno = SQD_INVALID_FRAME_TYPE;
		}
		writeToLog(2, buf);
		return Fail;
	}
	
	sq_errno = SQ_NOERR;
	return Ok;
}
/**************************************/

static BOOL read_sqbase(int sqd, SQBASE *sqbase)
{
    dword cnt;
    word sqbase_len;
	
    if (Fail == goto_frame(sqd, 0L)) return Fail;
	
    /* read _sqbase header */
    read( sqd, &sqbase_len, 2L );
	
    if (sqbase_len != SQBASE_SIZE) {
	writeToLog(3, "read_sqbase error: _sqbase corrupted (%d read)\n", sqbase_len);
	return Fail;
    }
	
    if (!goto_frame(sqd, 0L)) return Fail;
    
    cnt = ReadSqBase(sqd, sqbase);
    
    if (SQBASE_SIZE != cnt) {
	writeToLog(3, "read_sqbase: error reading _sqbase header, asked:%d, read:%ld\n",
				SQBASE_SIZE, cnt);
	sq_errno = SQD_READ_ERR;
	return Fail;
    }
    return Ok;
}
/*****************************************/

static BOOL update_sqbase(int sqd, SQBASE sqbase)
{
	if (!goto_frame(sqd, 0L))
		return Fail;
	
	if (SQBASE_SIZE != WriteSqBase(sqd, &sqbase)) {
		writeToLog(2, "update_sqbase: error writing _sqbase\n");
		sq_errno = SQD_WRITE_ERR;
		return Fail;
	}

	return Ok;
}

/*************************************************/
/*
static BOOL read_SQIDX(int sqi, FOFS i_off, SQIDX *sqidx)
{
	register FOFS off, len = SQIDX_SIZE;
	
	off = lseek(sqi, i_off, SEEK_SET);
	if ( off != i_off ) {
		sq_errno = SQI_SEEK_ERR;
		return Fail;
	}

	if ( len != ReadSqIdx( sqi, sqidx)) {
		sq_errno = SQI_READ_ERR;
		return Fail;
	}
	
	if (!validate_SQIDX(*sqidx))
		return Fail;

	return Ok;
}
*/
/**************************************/
static BOOL validate_SQIDX(SQIDX sqidx)
{
	if (sqidx.ofs == 0L) {
		sq_errno = SQI_INVALID_OFS;
		return Fail;
	} else if (sqidx.umsgid == 0xffffffff) {
		sq_errno = SQI_INVALID_UMSGID;
		return Fail;
	}
	
	sq_errno = SQ_NOERR;
	return Ok;
}
/************************************************/
static BOOL write_SQIDX( int sqi, dword off, SQIDX sqidx)
{
	register dword cnt, len = SQIDX_SIZE;
	char buf[256];
	
	if (off != (cnt = lseek(sqi, off, SEEK_SET))) {
		sprintf(buf, "write_SQIDX: lseek error in *.sqi, asked:%08lX, got:%08lX\n", off, cnt);
		sq_errno = SQI_SEEK_ERR;
		writeToLog(3, buf);
		return Fail;
	}
	if ( len != (cnt = WriteSqIdx( sqi, &sqidx)) ){
		sprintf(buf, "write_SQIDX: error writing *.sqi, asked:%ld, written:%ld, offset:%08lX\n",
					len, cnt, off);
		sq_errno = SQI_WRITE_ERR;
		writeToLog(3, buf);
		return Fail;
	}
	return Ok;
}

/*************************************/
void xmsg_to_ftn( XMSG xmsg, FTN *ftn)
{
	ftn->to.zone    = xmsg.dest.zone;
	ftn->to.net     = xmsg.dest.net;
	ftn->to.node    = xmsg.dest.node;
	ftn->to.point   = xmsg.dest.point;
	
	ftn->from.zone  = xmsg.orig.zone;
	ftn->from.net   = xmsg.orig.net;
	ftn->from.node  = xmsg.orig.node;
	ftn->from.point = xmsg.orig.point;
	
	strcpy(ftn->fromName, xmsg.from);
	strcpy(ftn->toName, xmsg.to);
	strcpy(ftn->subject, xmsg.subj);
	strcpy(ftn->DateTime, xmsg.__ftsc_date);
	ftn->attr = (word)(xmsg.attr & 0x0000FFFF);
	ftn->utime = sqdate_to_unix(xmsg.date_written);
	
	return;
}
/*********************************************/
void ftn_to_xmsg(FTN ftn, XMSG *xmsg)
{
	xmsg->dest.zone  = ftn.to.zone;
	xmsg->dest.net   = ftn.to.net;
	xmsg->dest.node  = ftn.to.node;
	xmsg->dest.point = ftn.to.point;

	xmsg->orig.zone  = ftn.from.zone;
	xmsg->orig.net   = ftn.from.net;
	xmsg->orig.node  = ftn.from.node;
	xmsg->orig.point = ftn.from.point;
	
	memset(xmsg->from, 0, 36);
	strcpy(xmsg->from, ftn.fromName);
	memset(xmsg->to, 0, 36);
	strcpy(xmsg->to, ftn.toName);
	memset(xmsg->subj, 0, 72);
	strcpy(xmsg->subj, ftn.subject);
	strcpy(xmsg->__ftsc_date, ftn.DateTime);
	
	unix_to_sqdate(ftn.utime, &xmsg->date_written);
	unix_to_sqdate(time(NULL), &xmsg->date_arrived);
	
	xmsg->attr &= 0xffff0000;
	xmsg->attr |= (dword)(ftn.attr);
	
	return;
}
/*********************************************/
TEXT *ctrl_to_text(char *buf)
{
	register int i;
	register char cc;
	char str[256]={0};
	TEXT *Text = NULL, *First = NULL, *Last = NULL;

	i = 0;
	while (*buf)
	{
		cc = *buf++;
		str[i++] = cc;
		if ( *buf == 1 || *buf == '\0' || i == 255 ) {
			if (i < 255)
				strcat(str, "\r");
			/*/////////////////////////////////////////////*/
			Text = (TEXT *)calloc(1, sizeof(TEXT));
			if (Text) {
				Text->next = NULL;
				/*/////////////////*/
				strcpy( Text->str, str);
				/*/////////////////*/
				if (Last) {
					Last->next = Text;
				}
				else {
					First = Text;
				}
				Last = Text;
			}
			/*//////////////////////////////////////////////*/
			i = 0;
			memset(str, 0, 256);
		} /* cr found or line too long */
	} /* while (buf) */


	return First;
	
}
/*****************************************************/
void text_to_sqbase(TEXT *txt, char **cbuf, char **tbuf)
{
	char *c = NULL, *t = NULL, buf[256];
	BOOL ctrl_part = Yes;
	
	for( ; txt; txt = txt->next)
	{
		strcpy(buf, txt->str);
		if (!is_kludge(buf) || is_Via(buf)) {
			ctrl_part = No;
		}
		if (ctrl_part == Yes) {
			if (!strncmp (buf, "\1FMPT", 5) || !strncmp (buf, "\1TOPT", 5))
				continue;
			if (NULL == c)
				c = (char *)calloc(strlen(buf) + 1, sizeof(char));
			else
				c = (char *)realloc(c, strlen(c) + strlen(buf) + 1);
			if (c) {
				strip_CRLF(buf);
				strcat(c, buf);
			}
			else
				writeToLog(3, "text_to_sqbase error: low memo\n");
		}
		/* text */
		else {
			if (NULL == t)
				t = (char *)calloc(strlen(buf) + 1, sizeof(char));
			else
				t = (char *)realloc(t, strlen(t) + strlen(buf) + 1);
			if (t) {
				strcat(t, buf);
			}
			else
				writeToLog(3, "text_to_sqbase error: low memo\n");
		}
	}
	
	*cbuf = c;
	*tbuf = t;
	
	return;
}
/**********************************************/
static int fill_idx_list(int sqi, SQIDX *idxlist[])
{
	SQIDX sqidx;
	dword len = SQIDX_SIZE;
	int n = 0;
	long size;
	
	size = filelength(sqi);

	lseek(sqi, 0, SEEK_SET);
	
	while(size > tell(sqi))
	{
		if (len != ReadSqIdx(sqi, &sqidx)) {
			writeToLog(2, "fill_idx_list: error reading sqi file\n");
			sq_errno = SQI_READ_ERR;
			n = -1;
			break;
		}
		
		if (!validate_SQIDX(sqidx))
			break;

		if (!*idxlist)
			*idxlist = (SQIDX *)malloc(SQIDX_SIZE);
		else
			*idxlist = (SQIDX *)realloc(*idxlist, (n + 1) * SQIDX_SIZE);

		if (!*idxlist) {
			n = -1;
			break;
		}
			
		memcpy(&(*idxlist)[n++], &sqidx, SQIDX_SIZE);
		
	}

	lseek(sqi, 0, SEEK_SET);
	
	if (lock(sqi, 0L, filelength(sqi))) {
		writeToLog(0, "error locking sqi file\n");
		sq_errno = BASE_LOCKED;
		n = -1;
	}

	if (n == -1) {
		free(*idxlist);
		*idxlist = NULL;
	}
	
	return n;
	
}

/******************************************************/
static BOOL add_to_idxlist(SQIDX *sqidx)
{
	SQIDX *i = CurMsg.squish.idxlist;
	
	if (!i)
		i = (SQIDX *)malloc(1 * SQIDX_SIZE);
	else
		i = (SQIDX *)realloc(i, (CurMsg.squish.total+1) * SQIDX_SIZE);

	if (!i)
		return Fail;

	memcpy(&i[CurMsg.squish.total++], sqidx, SQIDX_SIZE);
	
	CurMsg.squish.idxlist = i;

	return Ok;
}

/****************************/
static BOOL do_trunc(int sqi)
{
	char *p;
	dword n = filelength(sqi) - SQIDX_SIZE;
	
	lseek(sqi, SQIDX_SIZE, SEEK_SET);

	p = (char *)malloc(n);

	if (n != read(sqi, p, n)) {
		sq_errno = SQI_READ_ERR;
		free(p);
		return Fail;
	}

	lseek(sqi, 0, SEEK_SET);
		
	if (n != write(sqi, p, n)) {
		sq_errno = SQI_WRITE_ERR;
		free(p);
		return Fail;
	}

	setfsize(sqi, n);
	free(p);
	return Ok;
	
}

/*******************************************/
BOOL save_idx(int sqi, SQIDX sqidx[], int n)
{
	int i;
	
	unlock(sqi, 0L, filelength(sqi));

	lseek(sqi, 0L, SEEK_SET);
	
	if (0 != setfsize(sqi, 0L)) {
		writeToLog(0, "save_idx: error truncating *.sqi file\n");
		return Fail;
	}

	for(i = 0; i < n; ++i)
	{
		if (sqidx[i].ofs != (dword)-1 && sqidx[i].umsgid != (dword)-1)
			if (SQIDX_SIZE != WriteSqIdx(sqi, &sqidx[i])) {
				sq_errno = SQI_WRITE_ERR;
				return Fail;
			}
	}
	return Ok;
}



/*****************************************************************************/
/*  from kill */
static BOOL update_chains(int sqd, FOFS d_off, SQBASE *sqbase)
{
	SQHDR sqhdr, d_sqhdr;
	register FOFS prev_off, next_off, d_prev_frame;
	
	/* read SQHDR of msg to kill */
	if (!read_SQHDR(sqd, d_off, FRAME_NORMAL, &d_sqhdr))
		return Fail;
	
	prev_off = d_sqhdr.prev_frame;
	next_off = d_sqhdr.next_frame;
	
	/* that's too strict :-)

	if (next_off != CurMsg.squish.next_frame) {
		writeToLog(2, "wrong message to kill\n");
		return Fail;
	}
	
*/
	
	/* update message chain */
	
	/* message to kill was first frame */
	if (0L == prev_off)
		sqbase->begin_frame = next_off;
	
	/* read frame before the frame to kill */
	else { 
		
		if (!read_SQHDR(sqd, prev_off, FRAME_NORMAL, &sqhdr))
			return Fail;

		sqhdr.next_frame = next_off;
		
		if (!write_SQHDR(sqd, prev_off, &sqhdr))
			return Fail;
	} 
	
	/* message to kill was last frame */
	if (0L == next_off)
		sqbase->last_frame = prev_off;
	
	/* go to frame after the frame to kill */
	else { 
		
		if (!read_SQHDR(sqd, next_off, FRAME_NORMAL, &sqhdr))
			return Fail;
		
		sqhdr.prev_frame = prev_off;
		
		if (!write_SQHDR(sqd, next_off, &sqhdr))
			return Fail;
		
	} 
	
	/* update free chain */
	
	/* no free chain */
	if (0L == sqbase->free_frame) { 
		sqbase->free_frame = sqbase->last_free_frame = d_off;
		d_prev_frame = 0L;
	}
	else {
		
		if (!read_SQHDR(sqd, sqbase->last_free_frame, FRAME_FREE, &sqhdr))
			return Fail;
		
		sqhdr.next_frame = d_off;
		
		if (!write_SQHDR(sqd, sqbase->last_free_frame, &sqhdr))
			return Fail;
		
		d_prev_frame = sqbase->last_free_frame;
		
		sqbase->last_free_frame = d_off;
	}

	d_sqhdr.prev_frame = d_prev_frame;
	d_sqhdr.next_frame = 0L;
	d_sqhdr.frame_type = FRAME_FREE;
	
	/* update deleted frame */
	if (!write_SQHDR(sqd, d_off, &d_sqhdr))
		return Fail;
	
	return Ok;
}

/*********************************/
void sq_failure_reason(void)
{
	switch (sq_errno) /* sq_errno - global */
	{
	case SQ_NOERR:
		printf("no errors\n");
		break;
	case SQD_OPEN_ERR:
		printf("sqd: open error\n");
		break;
	case SQBASE_INVALID:
		printf("sqd: invalid _sqbase header\n");
		break;
	case SQD_SEEK_ERR:
		printf("sqd: seek error\n");
		break;
	case SQD_READ_ERR:
		printf("sqd: read error\n");
		break;
	case SQD_INVALID_ID:
		printf("sqd: invalid SQHDR id read\n");
		break;
	case SQD_INVALID_FRAME_TYPE:
		printf("sqd: invalid frame type read\n");
		break;
	case SQD_INVALID_NUM_MSG:
		printf("sqd: wrong number of messages read\n");
		break;
	case SQI_OPEN_ERR:
		printf("sqi: open error\n");
		break;
	case SQI_SEEK_ERR:
		printf("sqi: seek error\n");
		break;
	case SQI_READ_ERR:
		printf("sqi: read error\n");
		break;
	case SQI_WRITE_ERR:
		printf("sqi: write error\n");
		break;
	case SQI_INVALID_OFS:
		printf("sqi: invalid offset value read\n");
		break;
	case SQI_INVALID_UMSGID:
		printf("sqi: invalid umsgid value read\n");
		break;
	case SQI_INVALID_NUM_MSG:
		printf("sqi: wrong number of messages read\n");
		break;
	case LOW_MEMO:
		printf("general memory allocation failure\n");
		break;
	case BASE_LOCKED:
		printf("squish base locked by another process\n");
		break;
	default:
		printf("unknown error code: %d\n", sq_errno);
		break;
	}
	return;
}


/*******************************/
/*
static BOOL copy(char *from, char *to)
{
	// if the file `to' exists, it will be overwritten!
	
	int h_from, h_to;
	dword len;
	char *tmp = NULL;
	
	if( -1 == (h_from = open(from, O_RDONLY)) ) {
		perror(from);
		return Fail;
	}
	
	len = filelength(h_from);
	
	if (NULL == (tmp = (char *)calloc(len+1, sizeof(char)))) {
		writeToLog(4, "copy error: low memo\n");
		return Fail;
	}
	
	if (len != read(h_from, tmp, len)) {
		free(tmp);
		close(h_from);
		return Fail;
	}
	
	close(h_from);

	if( -1 == (h_to = creat(to, ACCFILE))) {
		perror(to);
		free(tmp);
		return Fail;
	}
	
	if (len != write(h_to, tmp, len)) {
		free(tmp);
		close(h_to);
		return Fail;
	}

	free(tmp);
	close(h_to);
	
	return Ok;
}
*/

/**********************************/
/*
static FOFS sqdOff_to_sqiOff(int sqi, FOFS off)
{
	SQIDX sqidx;
	long size;
	
	size = filelength(sqi);
	
	if (0L != lseek(sqi, 0L, SEEK_SET)) {
		sq_errno = SQI_SEEK_ERR;
		return -1;
	}

	while(size > tell(sqi))
	{
		if (SQIDX_SIZE != ReadSqIdx(sqi, &sqidx)) {
			writeToLog(2, "sqd->sqi: error reading sqi file\n");
			sq_errno = SQI_READ_ERR;
			return -1;
		}
		
		// success
		if (sqidx.ofs == off)
			return tell(sqi) - SQIDX_SIZE;
	}
	
	return -1;

}
*/
/***************************************************/
static FOFS get_hole(int sqd, dword next_frame, dword mlen)
{
	SQHDR sqhdr;
	
	while(next_frame)
	{
		if (!read_SQHDR(sqd, next_frame, FRAME_FREE, &sqhdr)) {
			return (dword)-1L;
		}
		if (sqhdr.frame_length >= mlen) {
			return next_frame;
		}
		next_frame = sqhdr.next_frame;
	}
	return 0L;
}
/**********************************************/
static dword squishHash(char *f)
{
	dword hash=0L, g;
	char *p;

	for (p=f; *p; p++)
	{
		hash=(hash << 4) + (unsigned long)tolower(*p);

		if ((g=(hash & 0xf0000000L)) != 0L)
		{
			hash |= g >> 24;
			hash |= g;
		}
	}
	
	/* Strip off high bit */
	return (hash & 0x7fffffffLu);
}

/*******************************************************************/
static word rewrite_frame(sqd, cur_off, xmsg, cbuf, tbuf)
int sqd;
FOFS cur_off;
XMSG xmsg;
char *cbuf, *tbuf;
{
	SQHDR sqhdr;
	dword len, clen, tlen, mlen;
	char buf[256];
	FOFS xmsg_off;
	
	
	clen = strlen(cbuf)+1;
	tlen = strlen(tbuf)+1;
	mlen = clen + tlen + XMSG_SIZE;
	

	
	/* read SQHDR */
	if (!read_SQHDR(sqd, cur_off, FRAME_NORMAL, &sqhdr)) {
		return Fail;
	}

	/* too big to fit */
	if (mlen > sqhdr.frame_length) {
		return (word)-1;
	}

	sqhdr.frame_type = FRAME_UPDATE;
	/* write SQHDR */
	if (!write_SQHDR(sqd, cur_off, &sqhdr)) {
		return Fail;
	}

	/*  Take the offset  of the  new SQHDR, and  add the  "sz_sqhdr" value 
			from the base  header.  Seek to that offset  and write the XMSG header. */
	
	xmsg_off = cur_off + (dword)sz_sqhdr;
	/* SEEK to new pos for XMSG */
	if (!goto_frame(sqd, xmsg_off)) {
		return Fail;
	}
	len = XMSG_SIZE;
	/* write XMSG */
	if ( len != WriteXMsg( sqd, &xmsg) ){
		sprintf(buf, "rewrite_frame: error writing XMSG, frame at %08lX\n", cur_off);
		writeToLog(3, buf);
		return Fail;
	}
	/* write CTRL */
	if ( clen != write( sqd, cbuf, clen ) ){
		sprintf(buf, "rewrite_frame: error writing CTRL, frame at %08lX\n", cur_off);
		writeToLog(3, buf);
		return Fail;
	}
	/* write Text */
	if ( tlen != write( sqd, tbuf, tlen ) ){
		sprintf(buf, "rewrite_frame: error writing text, frame at %08lX\n", cur_off);
		writeToLog(3, buf);
		return Fail;
	}
	
	sqhdr.frame_type = FRAME_NORMAL;
	sqhdr.msg_length = mlen;
	sqhdr.clen = clen;
	/* update new SQHDR */
	if (!write_SQHDR(sqd, cur_off, &sqhdr)) {
		return Fail;
	}
	
	return Ok;
}
/********************************************************/
static FOFS update_to_new_frame(sqd, cur_off, cbuf, tbuf, xmsg)
int sqd;
FOFS cur_off;
char *cbuf, *tbuf;
XMSG xmsg;
{
	dword len, clen, tlen, mlen;
	FOFS hole_off, new_off, xmsg_off;
	FOFS prev_cur_off, next_cur_off, prev_free, next_free;
	char buf[256];
	SQHDR sqhdr, new_sqhdr, cur_sqhdr, hole_sqhdr;
	SQBASE sqbase;
	
	
	/*   Re-read a copy of the Squish base header. */
	if (Fail == read_sqbase(sqd, &sqbase))
		return Fail;

	clen = strlen(cbuf)+1;
	tlen = strlen(tbuf)+1;
	mlen = clen + tlen + XMSG_SIZE;
	
	/* read currently active msg */
	if (!read_SQHDR(sqd, cur_off, FRAME_NORMAL, &cur_sqhdr)) {
		return Fail;
	}
	
	new_sqhdr = cur_sqhdr;
	
	new_sqhdr.frame_length = mlen;
	new_sqhdr.msg_length = mlen;
	new_sqhdr.clen = clen;
		
/*  sz_sqhdr = sqbase.sz_sqhdr; */
	
	/*   Scan  the messages in the  "free_frame" chain to  see if any
	frame  is  large  enough  to accommodate  the  new  message. */
	if ((dword)-1L == (hole_off = get_hole(sqd, sqbase.free_frame, mlen))) {
		writeToLog(3, "update_to_new_frame: error finding hole in squish base\n");
		return Fail;
	}
	
	if (hole_off) {

		sprintf(buf, "got hole at  %08lX\n", hole_off );
		writeToLog(3, buf);
		
		/* read SQHDR of found hole */
		if (!read_SQHDR(sqd, hole_off, FRAME_FREE, &hole_sqhdr)) {
			return Fail;
		}
		prev_free = hole_sqhdr.prev_frame;
		next_free = hole_sqhdr.next_frame;

		/* update free chain */
		if (!prev_free) { /* hole was first frame */
			sqbase.free_frame = next_free;
		}
		else { /* update frame before hole */
			if (!read_SQHDR(sqd, prev_free, FRAME_FREE, &sqhdr)) {
				return Fail;
			}

			sqhdr.next_frame = next_free;
			
			if (!write_SQHDR(sqd, prev_free, &sqhdr)) {
				return Fail;
			}
		} /* hole not first */

		if (!next_free) { /* hole was last */
			sqbase.last_free_frame = prev_free;
		}
		else { /* update frame after hole */
			if (!read_SQHDR(sqd, next_free, FRAME_FREE, &sqhdr)) {
				return Fail;
			}
			
			sqhdr.prev_frame = prev_free;
			
			if (!write_SQHDR(sqd, next_free, &sqhdr)) {
				return Fail;
			}
		} /* hole not last */
		new_off = hole_off;
		new_sqhdr.frame_length = hole_sqhdr.frame_length;
		/* free chain updated */
	}
	else {
	/*   If no free frame was  found, allocate a frame at the  end of
			the file, using the "end_frame" offset from the base header. */

		writeToLog(3, "hole not found, adding msg to sqbase\n");
		
		new_off = sqbase.end_frame;
		/* Increment  the  "end_frame" value  by  the  amount of  space  allocated. */
		sqbase.end_frame += SQHDR_SIZE + mlen;
		
	} /* no hole found */
	
	sprintf(buf, "writing new msg at frame off: %08lX\n", new_off );
	writeToLog(3, buf);

	/* Link the new SQHDR frame into the end of the message  chain,
		adjusting  the begin_frame  and last_frame  pointers in  the
		base header (if necessary). */
	
	if (sqbase.begin_frame == cur_off)
		sqbase.begin_frame = new_off;
	
	if (sqbase.last_frame == cur_off)
		sqbase.last_frame = new_off;
	
	/*  Set  the  "frame_type"  field  of the  new  SQHDR  to "frame
			update" and write at the appropriate offset. */
	new_sqhdr.frame_type = FRAME_UPDATE; /* frame is being updated */
	
	/* write SQHDR */
	if (!write_SQHDR(sqd, new_off, &new_sqhdr)) {
		return Fail;
	}

	/*  Take the offset  of the  new SQHDR, and  add the  "sz_sqhdr" value 
			from the base  header.  Seek to that offset  and write the XMSG header. */
	
	xmsg_off = new_off + (dword)sz_sqhdr;
	/* SEEK to new pos for XMSG */
	if (!goto_frame(sqd, xmsg_off)) {
		return Fail;
	}
	len = XMSG_SIZE;
	/* write XMSG */
	if ( len != WriteXMsg( sqd, &xmsg) ){
		sprintf(buf, "update_to_new_frame: error writing XMSG, frame at %08lX\n", new_off);
		writeToLog(3, buf);
		return Fail;
	}
	/* write CTRL */
	if ( clen != write( sqd, cbuf, clen ) ){
		sprintf(buf, "update_to_new_frame: error writing CTRL, frame at %08lX\n", new_off);
		writeToLog(3, buf);
		return Fail;
	}
	/* write Text */
	if ( tlen != write( sqd, tbuf, tlen ) ){
		sprintf(buf, "update_to_new_frame: error writing text, frame at %08lX\n", new_off);
		writeToLog(3, buf);
		return Fail;
	}
	
	/*  Set the frame type of the new message to "normal". */
	new_sqhdr.frame_type = FRAME_NORMAL; /* normal frame */
	/* update new SQHDR */
	if (!write_SQHDR(sqd, new_off, &new_sqhdr)) {
		return Fail;
	}

	/* store new link offsets in the frames prior and after the updated */
	if (0 != (prev_cur_off = new_sqhdr.prev_frame)) {
		if (!read_SQHDR(sqd, prev_cur_off, FRAME_NORMAL, &sqhdr)) {
			return Fail;
		}
		sqhdr.next_frame = new_off;

		if (!write_SQHDR(sqd, prev_cur_off, &sqhdr)) {
			return Fail;
		}
	}
	
	if (0 != (next_cur_off = new_sqhdr.next_frame)) {
		if (!read_SQHDR(sqd, next_cur_off, FRAME_NORMAL, &sqhdr)) {
			return Fail;
		}
		sqhdr.prev_frame = new_off;

		if (!write_SQHDR(sqd, next_cur_off, &sqhdr)) {
			return Fail;
		}
	}
	
	/* update free chain */
	
	/* no free chain */
	if (0L == sqbase.free_frame) {
		sqbase.free_frame = sqbase.last_free_frame = cur_off;
		prev_free = 0L;
	}
	else {
		
		if (!read_SQHDR(sqd, sqbase.last_free_frame, FRAME_FREE, &sqhdr))
			return Fail;
		
		sqhdr.next_frame = cur_off;
		
		if (!write_SQHDR(sqd, sqbase.last_free_frame, &sqhdr))
			return Fail;
		
		prev_free = sqbase.last_free_frame;
		
		sqbase.last_free_frame = cur_off;
	}

	cur_sqhdr.prev_frame = prev_free;
	cur_sqhdr.next_frame = 0L;
	cur_sqhdr.frame_type = FRAME_FREE;
	
	/* update deleted frame */
	if (!write_SQHDR(sqd, cur_off, &cur_sqhdr))
		return Fail;
	
	/*  Rewrite the Squish base header. */
	if (!update_sqbase(sqd, sqbase))
		return Fail;
	

	return new_off;
}
/********************************/
SQ_ERRNO validate_sqd(int sqd)
{
	SQ_ERRNO rc;
	
	if (SQ_NOERR != (rc = walk_sqd(sqd, MSG_FORWARD)))
		return rc;
	if (SQ_NOERR != (rc = walk_sqd(sqd, MSG_BACKWARD)))
		return rc;
	if (SQ_NOERR != (rc = walk_sqd(sqd, FREE_FORWARD)))
		return rc;
	if (SQ_NOERR != (rc = walk_sqd(sqd, FREE_BACKWARD)))
		return rc;
	
	return SQ_NOERR; 
}
/***************************************/
static SQ_ERRNO walk_sqd(int sqd, int mode)
{
	register FOFS next_frame, off;
	word sqbase_len;
	SQBASE sqbase;
	SQHDR sqhdr;
	register int nframes = 0;
	char buf[256];
	long size;
	
	if (0L != lseek(sqd, 0L, SEEK_SET))
		return SQD_SEEK_ERR;
	
	read(sqd, &sqbase_len, 2L);
	if (sqbase_len != SQBASE_SIZE)
		return SQBASE_INVALID;
	
	if (0L != lseek(sqd, 0L, SEEK_SET))
		return SQD_SEEK_ERR;
	if (sqbase_len != ReadSqBase(sqd, &sqbase))
		return SQD_READ_ERR;

/*  sz_sqhdr = sqbase.sz_sqhdr; */

	/* get starting frame */
	switch(mode)
	{
	/* get offset of the 1st msg frame */
	case MSG_FORWARD: 
		next_frame = sqbase.begin_frame;
		break;
	/* get offset of the last msg frame */  
	case MSG_BACKWARD: 
		next_frame = sqbase.last_frame;
		break;
	/* get offset of the 1st free frame */  
	case FREE_FORWARD: 
		next_frame = sqbase.free_frame;
		break;
	/* get offset of the last free frame */
	case FREE_BACKWARD: 
		next_frame = sqbase.last_free_frame;
		break;
	default:
		next_frame = 0L;
		break;
	}
	
	size = filelength(sqd);
	
	while (size > tell(sqd))
	{
		/* the final frame reached */
		if (!next_frame)
			break;
		
		off = lseek(sqd, next_frame, SEEK_SET);
		if (off != next_frame)
			return SQD_SEEK_ERR;
		
		/* read SQHDR */
		if (sz_sqhdr != ReadSqHdr(sqd, &sqhdr))
			return SQD_READ_ERR;
		
		if (sqhdr.id != SQHDRID)
			return SQD_INVALID_ID;
		
		if (mode == MSG_FORWARD || mode == MSG_BACKWARD) {
			if ( sqhdr.frame_type != FRAME_NORMAL  && 
					sqhdr.frame_type != FRAME_UPDATE ) {
				return SQD_INVALID_FRAME_TYPE;
			}
		}
		/* free chain */
		else { 
			if (sqhdr.frame_type != FRAME_FREE  &&
					sqhdr.frame_type != FRAME_UPDATE) {
				return SQD_INVALID_FRAME_TYPE;
			}
		}
		
		/* the total number of frames processed */
		++nframes;

		switch(mode)
		{
		case MSG_FORWARD:
		case FREE_FORWARD:
		/* get offset of the next frame */
			next_frame = sqhdr.next_frame;
			break;
		case MSG_BACKWARD:
		case FREE_BACKWARD:
		/* get offset of the prev frame */
			next_frame = sqhdr.prev_frame;
			break;
		default:
			next_frame = 0;
			break;
		}
	} 
	
	
	if (mode == MSG_FORWARD || mode == MSG_BACKWARD)
		if (nframes != sqbase.num_msg) {
			sprintf(buf, "walk_sqd error: wrong number of msgs, should be:%ld, read:%d\n",
							sqbase.num_msg, nframes);
			writeToLog(2, buf);
			return SQD_INVALID_NUM_MSG;
		}
	
	return SQ_NOERR;
	
}

/***********************************/
SQ_ERRNO validate_sqi(int sqd, int sqi)
{
	register dword rec_len = SQIDX_SIZE, off;
	word sqbase_len;
	SQBASE sqbase;
	SQHDR sqhdr;
	SQIDX sqidx;
	register int n = 0;
	char buf[256];
	long size;
	
	if (0L != lseek(sqd ,0L ,SEEK_SET))
		return SQD_SEEK_ERR;
	
	read(sqd, &sqbase_len, 2L);
	if (sqbase_len != SQBASE_SIZE)
		return SQBASE_INVALID;
	
	if (0L != lseek(sqd ,0L ,SEEK_SET))
		return SQD_SEEK_ERR;
	
	if (sqbase_len != ReadSqBase(sqd, &sqbase))
		return SQD_READ_ERR;
	
/*  sz_sqhdr = sqbase.sz_sqhdr; */
	
	if (0L != lseek(sqi ,0L ,SEEK_SET))
		return SQI_SEEK_ERR;

	size = filelength(sqi);	
	
	while (size > tell(sqi))
	{
		
		if (rec_len != ReadSqIdx(sqi, &sqidx))
			return SQI_READ_ERR;
		
		/* invalid OFFSET  -- may be true only close to the end */
		if (sqidx.ofs == 0L)
			continue;

		/* invalid UMSGID -- may be true only close to the end  */
		if (sqidx.umsgid == 0xffffffff)
			continue;
		
		/* position sqd to the start of next frame */
		off = lseek(sqd, sqidx.ofs, SEEK_SET);
		if ( off != sqidx.ofs )
			return SQD_SEEK_ERR;
		
		/* read SQHDR */
		if (sz_sqhdr != ReadSqHdr( sqd, &sqhdr))
			return SQD_READ_ERR;
	
		/* validate SQHDR */
		if (sqhdr.id != SQHDRID)
			return SQD_INVALID_ID;
		
		if (sqhdr.frame_type != FRAME_NORMAL &&
				sqhdr.frame_type != FRAME_UPDATE)
			return SQD_INVALID_FRAME_TYPE;
		
		/* the number of the records processed */
		++n;  
		
	} 
	
	
	if (n != sqbase.num_msg) {
		sprintf(buf, "walk_sqi error: msg no:%ld, recs read:%d\n",
						sqbase.num_msg, n);
		writeToLog(2, buf);
		return SQI_INVALID_NUM_MSG;
	}

	return SQ_NOERR;
	
}
/*************************************************************************/
BOOL open_squish_bases(char *areaname, int *sqd, int *sqi)
{
	char sqd_name[256], sqi_name[256], areapath[128];
	char buf[256];

	
	areaname_to_path(areaname, areapath);
	
	sprintf( sqd_name , "%s.sqd" , areapath );
	sprintf( sqi_name , "%s.sqi" , areapath );
	
	if( -1 == (*sqd = sopen(sqd_name, O_RDWR|O_BINARY, SH_DENYNO, ACCFILE)) ) {
		perror(sqd_name);
		if (errno == EACCES) {
			sprintf (buf, "squish base `%s' locked by another process\n", areaname);
			writeToLog(0, buf);
			return Ok;
		}
		sprintf(buf, "open_sqbases: error opening *.sqd file");
		writeToLog(0, buf);
		return Fail;
	}
	
	if( -1 == (*sqi = sopen(sqi_name, O_RDWR|O_BINARY, SH_DENYNO, ACCFILE)) ) {
		close(*sqd);
		perror(sqi_name);
		if (errno == EACCES) {
			sprintf (buf, "squish base `%s' locked by another process\n", areaname);
			writeToLog(0, buf);
			*sqd = -1;
			return Ok;
		}
		sprintf(buf, "open_sqbases: error opening *.sqi file");
		writeToLog(0, buf);
		return Fail;
	}
	
	strcpy (OutFileName, areapath);
	
	return Ok;
}

/*******************/
void 
sq_rewind(void)
{
	int sqd = CurMsg.squish.sqd;
	int sqi = CurMsg.squish.sqi;
	SQBASE sqbase;
	word len = SQBASE_SIZE;
	
	if (Fail == read_sqbase(sqd, &sqbase))
		return;
	
	if (0L != lseek(sqd, 0L ,SEEK_SET))
		return;

	if (len != ReadSqBase(sqd, &sqbase))
		return;

	CurMsg.squish.n = 0;
	CurMsg.squish.num_msg = sqbase.num_msg;
	CurMsg.squish.cur_frame = 0L;
	
	lseek(sqi, 0L, SEEK_SET);
	
	return;
	
}

/**************************/
BOOL sq_pack(char *areatag)
{

	SQAREA sq;
	SQBASE sqbase, nsqbase;
	SQHDR sqhdr;
	SQIDX *sqidx = NULL;
	register FOFS next_frame, off;
	char *mbuf = NULL, *main_buf = NULL;
	dword n_idx, len, ilen, curpos = 0L, reclen, mlen, orig_len;
	int i, delta = 0;
	register int cnt = 0L;
	char buf[256];
	
	sq_errno = 0;

	memset(&sq, 0, sizeof(SQAREA));

	select_squish_area(areatag, &sq);

	if (0 == lock(sq.sqd, 0L, 1L)) {
		writeToLog(3, "sqd file locked OK\n");
	}
	else {
		writeToLog(0, "error locking sqd file\n");
		sq_errno = BASE_LOCKED;
		return Fail;
	}
	

	orig_len = filelength(sq.sqd);
	main_buf = (char *)malloc(orig_len*sizeof(char));
	if (NULL == main_buf) {
		sq_errno = LOW_MEMO;
		goto end;
	}
	

	/* read SQBASE */
	len = SQBASE_SIZE;
	lseek(sq.sqd, 0L, SEEK_SET);
	if (len != ReadSqBase(sq.sqd, &sqbase)) {
		sq_errno = SQD_READ_ERR;
		goto end;
	}
	
	if( Chat_On ) {
		strUpper(areatag);
		sprintf(buf, "\npacking squish base (%s) ", areatag);
		printf("%s", buf);
		delta = sqbase.num_msg / (70 - strlen(buf));
		if (delta == 0)
			delta = 1;
	}

	nsqbase = sqbase;
	nsqbase.num_msg = nsqbase.high_msg = 0L;
	nsqbase.begin_frame = nsqbase.last_frame = 0L;
	nsqbase.free_frame = nsqbase.last_free_frame = 0L;
	nsqbase.end_frame = SQBASE_SIZE;

	n_idx = filelength(sq.sqi) / SQIDX_SIZE;

	ilen = n_idx * SQIDX_SIZE;
	
	sqidx = (SQIDX *) malloc(ilen);

	if (!sqidx) {
		sq_errno = LOW_MEMO;
		goto end;
	}

	lseek(sq.sqi, 0L, SEEK_SET);
	for (i = 0; i < n_idx; i++) {
	    if (ReadSqIdx(sq.sqi, &sqidx[i]) != SQIDX_SIZE) {
		sq_errno = SQI_READ_ERR;
		goto end;
	    }
	}


	for( i = 0; i < n_idx; ++i)
	{
		if (Chat_On) {
			if (++cnt == delta) {
				printf("3");
/*				putch(''); */
				cnt = 0;
			}
		}
		
		next_frame = sqidx[i].ofs;
		
		off = lseek(sq.sqd, next_frame, SEEK_SET);
		if (off != next_frame) {
			sq_errno = SQD_SEEK_ERR;
			goto end;
		}
		
		/* read SQHDR */
		if (sz_sqhdr != ReadSqHdr(sq.sqd, &sqhdr)) {
			sq_errno = SQD_READ_ERR;
			goto end;
		}
		
		if (sqhdr.id != SQHDRID) {
			sq_errno =  SQD_INVALID_ID;
			goto end;
		}
		
		if (sqhdr.frame_type != FRAME_NORMAL) {
			sq_errno =  SQD_INVALID_FRAME_TYPE;
			goto end;
		}

		/* read XMSG + CTRL + TEXT*/
		mlen = sqhdr.msg_length;

		if (!mbuf)
			mbuf = (char *)malloc((mlen+1) * sizeof(char));
		else
			mbuf = (char *)realloc(mbuf, (mlen+1) * sizeof(char));
		
		if (NULL == mbuf) {
			sq_errno = LOW_MEMO;
			goto end;
		}
	
		if (mlen != read(sq.sqd, mbuf, mlen)) {
			sq_errno = CTRL_READ_ERR;
			goto end;
		}

		reclen = mlen + SQHDR_SIZE;

		len = curpos + reclen + 1;


		if (i == n_idx-1)
			sqhdr.next_frame = 0L;
		else
			sqhdr.next_frame = curpos + SQBASE_SIZE + reclen;
			
		sqhdr.prev_frame = nsqbase.last_frame;
		sqhdr.frame_type = FRAME_NORMAL;
		sqhdr.frame_length = mlen;
		
		memcpy(&main_buf[curpos], &sqhdr, SQHDR_SIZE);
		memcpy(&main_buf[curpos+SQHDR_SIZE], mbuf, mlen);

		nsqbase.last_frame = curpos + SQBASE_SIZE;
		
		if (!nsqbase.begin_frame)
			nsqbase.begin_frame = nsqbase.last_frame;

		sqidx[i].ofs = nsqbase.last_frame;

		curpos += reclen;
		
	} /* idx walk */
	
	
	lseek(sq.sqi, 0L, SEEK_SET);
	
	len = ilen/SQIDX_SIZE;
	for (i = 0; i < len; i++) {
            if (WriteSqIdx(sq.sqi, &sqidx[i]) != SQIDX_SIZE) {
		sq_errno = SQI_WRITE_ERR;
		goto end;
	    }
	}
	
	lseek(sq.sqd, 0L, SEEK_SET);
	
	if (0 != setfsize(sq.sqd, 0L)) {
		sprintf(buf, "sqpack: error truncating *.sqd file\n");
		writeToLog(0, buf);
		goto end;
	}
	
	lseek(sq.sqd, 0L, SEEK_SET);

	nsqbase.num_msg = nsqbase.high_msg = n_idx;
	nsqbase.end_frame = curpos + SQBASE_SIZE;
	
	/* write new SQBASE */
	len = SQBASE_SIZE;
	if (len != WriteSqBase(sq.sqd, &nsqbase)) {
		sq_errno = SQD_READ_ERR;
		goto end;
	}

	if (curpos != write(sq.sqd, main_buf, curpos)) {
		sq_errno = SQI_WRITE_ERR;
		goto end;
	}
		
	if (Chat_On)   {
		printf(" done.\n");
		sprintf(buf, "Squish area: %s.  Original size= %ld. Packed size= %ld.\n",
						areatag, orig_len, curpos+SQBASE_SIZE);
		fprintf(stderr, "%s\n", buf);
		writeToLog(1, buf);
	}
	

end:
	free(sqidx);
	free(main_buf);
	free(mbuf);
	
	unlock(sq.sqd, 0L, 1L);
	close(sq.sqd);
	close(sq.sqi);
	
	return sq_errno ? Fail : Ok;
}
/*************************************************************************/
/*
static void create_backup(char *areaname, SQAREA *sq)
{
	char fname[256], areapath[128];
	char buf[256];

	areaname_to_path(areaname, areapath);

	sprintf(fname , "%s.s#d" , areapath );

	if( -1 == (sq->sqd = creat(fname, ACCFILE)) ) {
		perror(fname);
		sprintf(buf, "create_sqbackup: error creating *.s#d file");
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	sprintf(fname , "%s.s#i" , areapath );

	if( -1 == (sq->sqi = creat(fname, ACCFILE)) ) {
		perror(fname);
		sprintf(buf, "create_sqbackup: error creating *.s#d file");
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	
	return;
}
*/

void create_new_sqbase(char *areapath)
{
	char sqd_name[256], sqi_name[256], buf[256];
	int sqd, sqi;
	SQBASE sqbase;
	
	sprintf( sqd_name , "%s.sqd" , areapath );
	sprintf( sqi_name , "%s.sqi" , areapath );

	if (-1 == (sqi = open(sqi_name, O_WRONLY|O_CREAT|O_BINARY|O_EXCL, ACCFILE))) {
		perror(sqi_name);
		sprintf(buf, "create_new_sqbase: error creating %s", sqi_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}
	close(sqi);
	
	if (-1 == (sqd = open(sqd_name, O_WRONLY|O_CREAT|O_BINARY|O_EXCL, ACCFILE))) {
		perror(sqd_name);
		sprintf(buf, "create_new_sqbase: error creating %s", sqd_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		abort();
	}

	memset(&sqbase, 0, SQBASE_SIZE);
	sqbase.len = SQBASE_SIZE;
	sqbase.uid = 1;
	sqbase.end_frame = SQBASE_SIZE;
	sqbase.sz_sqhdr = SQHDR_SIZE;
	if (SQBASE_SIZE != WriteSqBase(sqd, &sqbase)) {
		perror(sqd_name);
		sprintf(buf, "create_new_sqbase: error wriing %s", sqd_name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s\n", buf);
		close(sqd);
		abort();
	}
	
	close(sqd);

	return;
}

dword ReadSqBase(int sqd, SQBASE *sqbase)
{
    word cnt;
    byte buf[SQBASE_SIZE], *pbuf = buf;
    
    cnt = read( sqd, &buf, SQBASE_SIZE);
	
    if (SQBASE_SIZE != cnt) {
	return cnt;
    }
	
    sqbase->len = get_word(pbuf);
    pbuf += 2;

    sqbase->rsvd1 = get_word(pbuf);
    pbuf += 2;

    sqbase->num_msg = get_dword(pbuf);
    pbuf += 4;

    sqbase->high_msg = get_dword(pbuf);
    pbuf += 4;

    sqbase->skip_msg = get_dword(pbuf);
    pbuf += 4;

    sqbase->high_water = get_dword(pbuf);
    pbuf += 4;

    sqbase->uid = get_dword(pbuf);
    pbuf += 4;

    memmove(sqbase->base, pbuf, 80);
    pbuf += 80;

    sqbase->begin_frame = get_dword(pbuf);
    pbuf += 4;

    sqbase->last_frame = get_dword(pbuf);
    pbuf += 4;

    sqbase->free_frame = get_dword(pbuf);
    pbuf += 4;

    sqbase->last_free_frame = get_dword(pbuf);
    pbuf += 4;

    sqbase->end_frame = get_dword(pbuf);
    pbuf += 4;

    sqbase->max_msg = get_dword(pbuf);
    pbuf += 4;

    sqbase->keep_days = get_word(pbuf);
    pbuf += 2;

    sqbase->sz_sqhdr = get_word(pbuf);
    pbuf += 2;

    memmove(sqbase->rsvd2, pbuf, 124);
    pbuf += 124;

    return cnt;
}

dword WriteSqBase(int sqd, SQBASE *psqbase)
{
    char buf[SQBASE_SIZE], *pbuf = buf;

    put_word(pbuf, psqbase->len);
    pbuf += 2;

    put_word(pbuf, psqbase->rsvd1);
    pbuf += 2;

    put_dword(pbuf, psqbase->num_msg);
    pbuf += 4;

    put_dword(pbuf, psqbase->high_msg);
    pbuf += 4;

    put_dword(pbuf, psqbase->skip_msg);
    pbuf += 4;

    put_dword(pbuf, psqbase->high_water);
    pbuf += 4;

    put_dword(pbuf, psqbase->uid);
    pbuf += 4;

    memmove(pbuf, psqbase->base, 80);
    pbuf += 80;

    put_dword(pbuf, psqbase->begin_frame);
    pbuf += 4;

    put_dword(pbuf, psqbase->last_frame);
    pbuf += 4;

    put_dword(pbuf, psqbase->free_frame);
    pbuf += 4;

    put_dword(pbuf, psqbase->last_free_frame);
    pbuf += 4;

    put_dword(pbuf, psqbase->end_frame);
    pbuf += 4;

    put_dword(pbuf, psqbase->max_msg);
    pbuf += 4;

    put_word(pbuf, psqbase->keep_days);
    pbuf += 2;

    put_word(pbuf, psqbase->sz_sqhdr);
    pbuf += 2;

    memmove(pbuf, psqbase->rsvd2, 124);
    pbuf += 124;

    assert(pbuf - buf == SQBASE_SIZE);

    return write(sqd, &buf, SQBASE_SIZE);
}

dword ReadXMsg(sword handle, XMSG *pxmsg)
{
    byte buf[XMSG_SIZE], *pbuf = buf;
    word rawdate, rawtime;
    dword len;
    int i;

    len = read(handle, &buf, XMSG_SIZE);

    if (len != XMSG_SIZE)
    {
        return len;
    }

                                /* 04 bytes "attr" */
    pxmsg->attr = get_dword(pbuf);
    pbuf += 4;

                                /* 36 bytes "from" */
    memmove(pxmsg->from, pbuf, XMSG_FROM_SIZE);
    pbuf += XMSG_FROM_SIZE;

                                /* 36 bytes "to"   */
    memmove(pxmsg->to, pbuf, XMSG_TO_SIZE);
    pbuf += XMSG_TO_SIZE;

                                /* 72 bytes "subj" */
    memmove(pxmsg->subj, pbuf, XMSG_SUBJ_SIZE);
    pbuf += XMSG_SUBJ_SIZE;

                                /* 8 bytes "orig"  */
    pxmsg->orig.zone = get_word(pbuf); pbuf += 2;
    pxmsg->orig.net  = get_word(pbuf); pbuf += 2;
    pxmsg->orig.node = get_word(pbuf); pbuf += 2;
    pxmsg->orig.point= get_word(pbuf); pbuf += 2;

                                /* 8 bytes "dest"  */
    pxmsg->dest.zone = get_word(pbuf); pbuf += 2;
    pxmsg->dest.net  = get_word(pbuf); pbuf += 2;
    pxmsg->dest.node = get_word(pbuf); pbuf += 2;
    pxmsg->dest.point= get_word(pbuf); pbuf += 2;

                                /* 4 bytes "date_written" */
    rawdate = get_word(pbuf); pbuf += 2;
    rawtime = get_word(pbuf); pbuf += 2;
    pxmsg->date_written.date.da = rawdate & 31;
    pxmsg->date_written.date.mo = (rawdate >> 5) & 15;
    pxmsg->date_written.date.yr = (rawdate >> 9) & 127;
    pxmsg->date_written.time.ss = rawtime & 31;
    pxmsg->date_written.time.mm = (rawtime >> 5) & 63;
    pxmsg->date_written.time.hh = (rawtime >> 11) & 31;

                                /* 4 bytes "date_arrived" */
    rawdate = get_word(pbuf); pbuf += 2;
    rawtime = get_word(pbuf); pbuf += 2;
    pxmsg->date_arrived.date.da = rawdate & 31;
    pxmsg->date_arrived.date.mo = (rawdate >> 5) & 15;
    pxmsg->date_arrived.date.yr = (rawdate >> 9) & 127;
    pxmsg->date_arrived.time.ss = rawtime & 31;
    pxmsg->date_arrived.time.mm = (rawtime >> 5) & 63;
    pxmsg->date_arrived.time.hh = (rawtime >> 11) & 31;

                                /* 2 byte "utc_ofs" */
    pxmsg->utc_ofs = get_word(pbuf);
    pbuf += 2;

                                /* 4 bytes "replyto" */
    pxmsg->replyto = get_dword(pbuf);
    pbuf += 4;

                                /* 10 times 4 bytes "replies" */
    for (i = 0; i < MAX_REPLY; i++)
    {
        pxmsg->replies[i] = get_dword(pbuf);
        pbuf += 4;
    }

                                /* 4 bytes "umsgid" */
    pxmsg->umsgid = get_dword(pbuf);
    pbuf += 4;
    
                                /* 20 times FTSC date stamp */
    memmove(pxmsg->__ftsc_date, pbuf, 20);
    pbuf += 20;

    assert(pbuf - buf == XMSG_SIZE);
    
    return len;
}

dword WriteXMsg(sword handle, XMSG *pxmsg)
{
    byte buf[XMSG_SIZE], *pbuf = buf;
    word rawdate, rawtime;
    int i;

                                /* 04 bytes "attr" */
    put_dword(pbuf, pxmsg->attr);
    pbuf += 4;

                                /* 36 bytes "from" */
    memmove(pbuf, pxmsg->from, XMSG_FROM_SIZE);
    pbuf += XMSG_FROM_SIZE;

                                /* 36 bytes "to"   */
    memmove(pbuf, pxmsg->to, XMSG_TO_SIZE);
    pbuf += XMSG_TO_SIZE;

                                /* 72 bytes "subj" */
    memmove(pbuf, pxmsg->subj, XMSG_SUBJ_SIZE);
    pbuf += XMSG_SUBJ_SIZE;

                                /* 8 bytes "orig"  */
    put_word(pbuf, pxmsg->orig.zone);  pbuf += 2;
    put_word(pbuf, pxmsg->orig.net);   pbuf += 2;
    put_word(pbuf, pxmsg->orig.node);  pbuf += 2;
    put_word(pbuf, pxmsg->orig.point); pbuf += 2;

                                    /* 8 bytes "dest"  */
    put_word(pbuf, pxmsg->dest.zone);  pbuf += 2;
    put_word(pbuf, pxmsg->dest.net);   pbuf += 2;
    put_word(pbuf, pxmsg->dest.node);  pbuf += 2;
    put_word(pbuf, pxmsg->dest.point); pbuf += 2;


                                /* 4 bytes "date_written" */
    rawdate = rawtime = 0;

    rawdate |= (((word)pxmsg->date_written.date.da) & 31);
    rawdate |= (((word)pxmsg->date_written.date.mo) & 15) << 5;
    rawdate |= (((word)pxmsg->date_written.date.yr) & 127) << 9;

    rawtime |= (((word)pxmsg->date_written.time.ss) & 31);
    rawtime |= (((word)pxmsg->date_written.time.mm) & 63) << 5;
    rawtime |= (((word)pxmsg->date_written.time.hh) & 31) << 11;

    put_word(pbuf, rawdate); pbuf += 2;
    put_word(pbuf, rawtime); pbuf += 2;


                                /* 4 bytes "date_arrvied" */
    rawdate = rawtime = 0;

    rawdate |= (((word)pxmsg->date_arrived.date.da) & 31);
    rawdate |= (((word)pxmsg->date_arrived.date.mo) & 15) << 5;
    rawdate |= (((word)pxmsg->date_arrived.date.yr) & 127) << 9;

    rawtime |= (((word)pxmsg->date_arrived.time.ss) & 31);
    rawtime |= (((word)pxmsg->date_arrived.time.mm) & 63) << 5;
    rawtime |= (((word)pxmsg->date_arrived.time.hh) & 31) << 11;

    put_word(pbuf, rawdate); pbuf += 2;
    put_word(pbuf, rawtime); pbuf += 2;


                                /* 2 byte "utc_ofs" */
    put_word(pbuf, pxmsg->utc_ofs);
    pbuf += 2;

                                /* 4 bytes "replyto" */
    put_dword(pbuf, pxmsg->replyto);
    pbuf += 4;

                                /* 10 times 4 bytes "replies" */
    for (i = 0; i < MAX_REPLY; i++)
    {
        put_dword(pbuf, pxmsg->replies[i]);
        pbuf += 4;
    }
                                /* 4 bytes "umsgid" */
    put_dword(pbuf, pxmsg->umsgid);
    pbuf += 4;
    

                                /* 20 times FTSC date stamp */
    memmove(pbuf, pxmsg->__ftsc_date, 20);
    pbuf += 20;

    assert(pbuf - buf == XMSG_SIZE);
    return write(handle, &buf, XMSG_SIZE);
}

dword ReadSqHdr(sword handle, SQHDR *psqhdr)
{
    byte buf[SQHDR_SIZE], *pbuf = buf;
    dword len;

    len = read(handle, &buf, SQHDR_SIZE);

    if (len != SQHDR_SIZE)
    {
        return len;
    }
                                /* 4 bytes "id" */
    psqhdr->id = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "next_frame" */
    psqhdr->next_frame = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "prev_frame" */
    psqhdr->prev_frame = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "frame_length" */
    psqhdr->frame_length = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "msg_length" */
    psqhdr->msg_length = get_dword(pbuf);
    pbuf += 4;
                                /* 4 bytes "clen" */
    psqhdr->clen = get_dword(pbuf);
    pbuf += 4;

                                /* 2 bytes "frame_type" */
    psqhdr->frame_type = get_word(pbuf);
    pbuf += 2;
                                /* 4 bytes "rsvd" */
    psqhdr->rsvd = get_word(pbuf);
    pbuf += 2;

    assert(pbuf - buf == SQHDR_SIZE);

    return len;
}

dword WriteSqHdr(sword handle, SQHDR *psqhdr)
{
    byte buf[SQHDR_SIZE], *pbuf = buf;

                                /* 4 bytes "id" */
    put_dword(pbuf, psqhdr->id);
    pbuf += 4;

                                /* 4 bytes "next_frame" */
    put_dword(pbuf, psqhdr->next_frame);
    pbuf += 4;

                                /* 4 bytes "prev_frame" */
    put_dword(pbuf, psqhdr->prev_frame);
    pbuf += 4;

                                /* 4 bytes "frame_length" */
    put_dword(pbuf, psqhdr->frame_length);
    pbuf += 4;

                                /* 4 bytes "msg_length" */
    put_dword(pbuf, psqhdr->msg_length);
    pbuf += 4;
                                /* 4 bytes "clen" */
    put_dword(pbuf, psqhdr->clen);
    pbuf += 4;

                                /* 2 bytes "frame_type" */
    put_word(pbuf, psqhdr->frame_type);
    pbuf += 2;
                                /* 4 bytes "rsvd" */
    put_word(pbuf, psqhdr->rsvd);
    pbuf += 2;

    assert(pbuf - buf == SQHDR_SIZE);

    return write(handle, &buf, SQHDR_SIZE);
}


dword ReadSqIdx(sword handle, SQIDX *psqidx)
{
    byte buf[SQIDX_SIZE], *pbuf = NULL;
    dword len;

    len = read(handle, buf, SQIDX_SIZE);
    if (len != SQIDX_SIZE) {
        return len;
    }
    pbuf = buf;
        /* Begin reading in a single structure */

                                /* 4 bytes "ofs" */
    psqidx->ofs = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "umsgid" */
    psqidx->umsgid = get_dword(pbuf);
    pbuf += 4;

                                /* 4 bytes "hash" */
    psqidx->hash = get_dword(pbuf);
    pbuf += 4;

    return len;
}


dword WriteSqIdx(sword handle, SQIDX *psqidx)
{
    byte buf[SQIDX_SIZE], *pbuf = NULL;
    dword len;
    
    pbuf = buf;
                                /* 4 bytes "ofs" */
    put_dword(pbuf, psqidx->ofs);
    pbuf += 4;

                                /* 4 bytes "umsgid" */
    put_dword(pbuf, psqidx->umsgid);
    pbuf += 4;

                                /* 4 bytes "hash" */
    put_dword(pbuf, psqidx->hash);
    pbuf += 4;

    len = write(handle, &buf, SQIDX_SIZE);

    if (len != SQIDX_SIZE) {
        return len;
    }
    return len;
}

