
/************************************************************************
 *	itraxp - netmail tracker
 *      by Fedor Lizunkov 2:5020/960@Fidonet
 *
 *	itraxp - clone from itraX 
 *	by Igor Shvyrkov 2:5020/410@Fidonet
 ************************************************************************/

#include "itrax.h"

#include <stdarg.h>

#if defined UNIX || defined EMX
#include <sys/time.h>
#include <unistd.h>
#endif

#ifdef MSGIDSEQ
#include "genmsgid.h"
#endif

void writeToLog(int level, char *str, ...);
BOOL is_listed(struct ndlTags *ndlTag, ADDR addr, word* uplink, word* region);

void   fetch_systime(char *buf);
dword  textlen(TEXT *t);
word   str_to_flags( char *p );
int    get_TZ(void);

BOOL is_filter_found(char *str, struct filter *f);

extern BOOL get_global_var_old(char *in, char *out);
extern BOOL is_v9_listed();

extern void *xmalloc(size_t size);


/**************************************************/
int parse(char *str, char *delim[], char *result[])
{
	char *ptr, *buf;
	int  kwo = 0;
	
	if ( !strlen(str) ) {
		return 0;
	}
	
	buf = str;
		
	while (strlen(delim[kwo]))
	{
		ptr = strstr(buf,delim[kwo]);

		if (!ptr) 
			break;
		else {
/*                     strset(result[kwo],0); */
			strncpy(result[kwo],buf,ptr-buf);
			*(result[kwo]+(ptr-buf)) = '\0';
			ptr += strlen(delim[kwo]);
			buf = ptr;
/*                     strset(delim[kwo],0); */

		}
		if (++kwo == 20) 
			break;
	}
	strcpy(result[kwo++],buf);
		
	if (kwo == 1) {
		kwo = 0;
	}
	
	return kwo; /* return the number of result fields filled   */
		    /* (if 1 then zero is returned (nothing found) */
}  /* end of parse */

/************************************************************************/
int __isnumber(char *str)
{

	int i;
	
	if (str == NULL)
		return Fail;
		
	if (!strlen(str)) 
		return Fail;
		
	for(i=0;i<strlen(str);i++)
		if ( ((str[i] < '0') || (str[i] > '9')) && (str[i] != '-') )
			return Fail;

	return Ok;
}

/**************************************************/
/* splits a string into fields separated by space */
int splitf(char *str, char **result)
{
	int kwo, j, body = 0;
	char buf[256], *c;
	int MAXCOUNT = 20;

	if (!strlen(str)) 
		return 0;
	
	memset(buf,0,256);
	c = str;

	j = 0;   kwo = 0;
	while (*str)
	{
		if ( (*str == 32) || (*str == 9) )
		{
			if (body) {
				if (strlen(buf))  {

					strcpy(result[kwo++],buf);
					memset(buf,0,256);
					if (kwo == MAXCOUNT)
						break;
					j = 0;
				}
			}
		}
		else
		{
			body = 1;

			buf[j++] = *str;
			if (j == 256) 
				break;
		}
		str++;
	} 

	if (strlen(buf))
		strcpy(result[kwo++],buf);
	
	str = c;

	return kwo;
}  /* end of splitf */

/*************************************************************************/
char *trim(char *str)
{
	int i, j, body;
	char buf[256];

	if (!strlen(str)) return str;

	for(i=0;i<strlen(str);i++)
		if (str[i] == 9) str[i] = 32;

	for(i=0;i<256;i++)
		buf[i] = 0;

	body = 0;
	for(i=0,j=0;i<strlen(str);i++)
	{
		if (str[i] == 32) {
			if (body == 0) continue;
		}
			else body = 1;

		buf[j++]=str[i];
		if (j == 256) break;

	}

	strcpy(str,buf);
	strrev(str);

	for(i=0;i<256;i++)
				buf[i] = 0;

	body = 0;
	for(i=0,j=0;i<strlen(str);i++)
	{
		if (str[i] == 32) {
			if (body == 0) continue;
		}
			else body = 1;

		buf[j++] = str[i];
		if (j == 256) break;

	}
	strrev(buf);
	strcpy(str,buf);
	return str;
}
/*********************************/
char *get_next_word(char *str, char *w)
{
	char *p = str;
	
	if (!strlen(str))
		return NULL;
	
	while( *p && *p != ' ' && *p != '\t' )
	{
		*w++ = *p++;
	}
	*w = '\0';
	strip(p);
	if (*p) {
	    memmove(str, p, strlen(p));
	    str[strlen(p)] = '\0';
	} else {
	    *str = '\0';
	}
	return w;
}

/**********************************************************************/
BOOL addr_cmp( ADDR a1, ADDR a2 )
{
	return (a1.zone  == a2.zone &&
					a1.net   == a2.net  &&
					a1.node  == a2.node &&
					a1.point == a2.point);
}

/*** Extracts 5D address from a string ***/
int str_to_addr(char *address, ADDR *s)
{
	char buf[256];
	char zone_[8], net_[8], node_[24], point_[8];
	int i;
	char *del[4], *res[4];

	for(i=0;i<4;i++)
	{
		del[i] = (char *)calloc(256,sizeof(char));
		res[i] = (char *)calloc(256,sizeof(char));
	}

	strcpy(del[0],":");
	strcpy(del[1],"/");
	strcpy(del[2],"@");
	strcpy(buf,address);
		
	parse(buf,del,res);
		
	strncpy(zone_,res[0],7);
	strncpy(net_,res[1],7);
	strncpy(node_,res[2],11);
	strncpy(s->domain,res[3],63);

	for(i=0;i<4;i++) *(res[i]) = '\0';
/*	strset(res[i],0); */

	strcpy(del[0],".");
	parse(node_,del,res); /* extract point # (optional) */
	strncpy(node_,res[0],7);
	strncpy(point_,res[1],7);

	for(i=0;i<4;i++)
	{
		free(del[i]);
		free(res[i]);
	}
		
	if (!strlen(point_)) {
		strcpy(point_,"0");
	}

		
	if (!__isnumber(zone_) || !__isnumber(net_) ||
			!__isnumber(node_) || !__isnumber(point_) || !strlen(node_) ) {
		return Fail;
	}

	s->zone  = (word) atoi(zone_);
	s->net   = (word) atoi(net_);
	s->node  = (word) atoi(node_);
	s->point = (word) atoi(point_);

	return Ok;
	
} /* end of getaddr */

/*** Extracts 5D address from a string (wildcards allowed) ***/
int str_to_Waddr(char *address, ADDR *s)
{
	char buf[256];
	char c_zone[8], c_net[8], c_node[24], c_point[8];
	int i;
	char *del[4], *res[4];

	for(i=0;i<4;i++)
	{
		del[i] = (char *)calloc(256,sizeof(char));
		res[i] = (char *)calloc(256,sizeof(char));
	}

	strcpy(del[0],":");
	strcpy(del[1],"/");
	strcpy(del[2],"@");
	strcpy(buf,address);
		
	parse(buf,del,res);
		
	strncpy(c_zone,res[0],7);
	strncpy(c_net,res[1],7);
	strncpy(c_node,res[2],11);
	strncpy(s->domain,res[3],63);

	for(i=0;i<4;i++) *(res[i]) = '\0';
/*     strset(res[i],0); */

	strcpy(del[0],".");
	parse(c_node,del,res); /* extract point # (optional) */
	strncpy(c_node,res[0],7);
	strncpy(c_point,res[1],7);

	for(i=0;i<4;i++)
	{
		free(del[i]);
		free(res[i]);
	}
		
	if (!strlen(c_point)) 
		strcpy(c_point,"0");

		
	if (!__isnumber(c_zone)) {
		if (!strcmp(c_zone,"*"))
			s->zone = -1;
		else {
			if (!strcmp(c_zone,"#")) {
				s->zone = -2;
			}
			else {
				return Fail;
			}
		}
	}
	else {
		s->zone  = (word) atoi(c_zone);
	}
	if (!__isnumber(c_net)) {
		if (!strcmp(c_net,"*"))
			s->net = -1;
		else {
			if (!strcmp(c_net,"#")) {
				s->net = -2;
			}
			else {
				return Fail;
			}
		}
	}
	else {
		s->net  = (word) atoi(c_net);
	}
	if (!__isnumber(c_node)) {
		if (!strcmp(c_node,"*"))
			s->node = -1;
		else {
			if (!strcmp(c_node,"#")) {
				s->node = -2;
			}
			else {
				return Fail;
			}
		}
	}
	else {
		s->node  = (word) atoi(c_node);
	}
	if (!__isnumber(c_point)) {
		if (!strcmp(c_point,"*"))
			s->point = -1;
		else {
			if (!strcmp(c_point,"#")) {
				s->point = -2;
			}
			else {
				return Fail;
			}
		}
	}
	else {
		s->point  = (word) atoi(c_point);
	}

	return Ok;
	
} /* end of getaddr */
/********************************************************************/
char *addr_to_str(ADDR addr, char *string)
{
	word wild = -1;
	char buf[256];
	
	if ( addr.zone != wild )
		sprintf(buf, "%u", addr.zone);
/*		itoa( addr.zone, buf, 10); */
	else
		strcpy(buf, "*");
	
	strcpy(string, buf);
	strcat(string, ":");
	
	if ( addr.net != wild )
		sprintf(buf, "%u", addr.net);
/*		itoa( addr.net, buf, 10); */
	else
		strcpy(buf, "*");
	
	strcat(string, buf);
	strcat(string, "/");
	
	if ( addr.node != wild )
		sprintf(buf, "%u", addr.node);
/*		itoa( addr.node, buf, 10); */
	else
		strcpy(buf, "*");
	
	strcat(string, buf);
	
	if ( addr.point ) {
		
		strcat(string, ".");
		
		if ( addr.point != wild )
			sprintf(buf, "%u", addr.point);
/*			itoa( addr.point, buf, 10); */
		else
			strcpy(buf, "*");

		strcat(string, buf);
	}
	
	return string;
}
/***************************************************************************/
char *attr_to_str( const word attr, char *res )
{
	BOOL found = 0;
	int i;

	*res = '\0';	
/*	strset( res , 0 ); */
	
	/* search for real attributes */
	for( i = 0; strlen( Attrs[i].s ); ++i)
	{
		if ( Attrs[i].n & attr ) {
			if (found) {
				strcat( res , " " );
			}
			strcat( res , (const char*)(strUpper(Attrs[i].s)) );
			found = 1;
		}
	}
	
	return res;
}
/***************************************************************************/
char *flag_to_str( const word flag, char *res )
{
	BOOL found = 0;
	int i;

	*res = '\0';	
/*	strset( res , 0 ); */
	
	/* search for FLAG kludge attributes */
	for( i = 0; strlen( Flags[i].s ); ++i)
	{
		if ( Flags[i].n & flag ) {
			if (found) {
				strcat( res , " " );
			}
			strcat( res , (const char*)(strUpper(Flags[i].s)) );
			found = 1;
		}
	}
	
	return res;
}
/***************************************************************************/
char *stat_to_str( const dword stat, char *res )
{
	BOOL found = 0;
	int i;
	
	*res = '\0';
/*	strset( res , 0 ); */
	
	/* search for internal attributes */
	for( i = 0; strlen( Stats[i].s ); ++i)
	{
		if ( Stats[i].n & stat ) {
			if (found) {
				strcat( res , " " );
			}
			strcat( res , (const char*)(strUpper(Stats[i].s)) );
			found = 1;
		}
	}
	
	return res;
}
/***********************************************/
word str_to_attr( char *str)
{
	int i;
	
	/* search for real attributes */
	for( i = 0; strlen( Attrs[i].s ); ++i)
	{
		if ( !stricmp(Attrs[i].s ,str) ) {
			return Attrs[i].n;
		}
	}

	return minus_one;

}
/***********************************************/
word str_to_flag( char *str)
{
	int i;
	
	/* search for FLAG kludge attributes */
	for( i = 0; strlen( Flags[i].s ); ++i)
	{
		if ( !stricmp(Flags[i].s ,str) ) {
			return Flags[i].n;
		}
	}
			
	return minus_one;

}
/***********************************************/
dword str_to_stat( char *str)
{
	int i;
	
	/* search for internal attributes */
	for( i = 0; strlen( Stats[i].s ); ++i)
	{
		if ( !stricmp(Stats[i].s ,str) ) {
			return Stats[i].n;
		}
	}
	return minus_one_long;
}
/***********************************/
word str_to_flags( char *p )
{
	register int i;
	word flag = 0;
	char cflag[20];
	
	while(get_next_word(p, cflag))
	{
		for(i = 0; strlen(Flags[i].s); ++i)
		{
			if (!stricmp(Flags[i].s, cflag)) {
				flag |= Flags[i].n;
				break;
			}
		}
	}
	
	return flag;
	
}

/********************************************************************/
BOOL is_legal_tag(char *tag)
{
	if (!strlen(tag)) 
		return Fail;
	else
		return Ok;
/*
	for( i = 0; i < strlen(tag); ++i )
	{
		if ( tag[i] != '_' && !isalnum(tag[i]) )
				return Fail;
	}
	
	return Ok;
*/
}
/********************************************************************/
int setpath(char *path)
{
	char *c, s[128];
	
	if (!strlen(path))
		return Ok;

	c = &path[strlen(path)-1];
	if ( (*c == DELIM) && (strlen(path) > 1) ) {
		*c = '\0';
	}

#if !defined (UNIX)
	c = &path[strlen(path)-1];
	if (*c == ':') {
		strcat(path,"\\");
	}
#endif

	getcwd(s, 100);
	if ( chdir(path) ) {
		return Fail;  /* path not found */
	}

	chdir(s);

	c = &path[strlen(path)-1];
	if (*c != DELIM) {
		strcat(path,SEP);
	}

	sprintf(s, ".%c", DELIM);
	if (!strncmp(path, s, 2)) {
		c = path+2;
		strncpy(path, c, 127);
		path[127] = 0;
	}
	
	return Ok; /* ok */
}
/********************************************************************/
void writeToLog(int level, char *str, ...)
{
    va_list par;
   
    FILE *h;
    time_t cur_time;
    struct tm *loc_time;
    char logfile[256];
    int rc;
    
    if (lockFile.handle < 0) return;
	
    if (LogFile == NULL)
	return;

	if (level > DebugLevel) {

		if (!LogIncl || !is_filter_found(str, LogIncl)) {
			return;
		}
	}

	/* look for exclusions */
		
	if (LogExcl && is_filter_found(str, LogExcl)) {
		return;
	}
	
	sprintf(logfile, "%s%s", ItraxPath, LogFile);
	
	if (NULL == (h = fopen(logfile, "a+t")))
		if (NULL == (h = fopen(LogFile, "a+t"))) {
			return;
		}
		
	cur_time = time(NULL);
	loc_time = localtime(&cur_time);
	
	rc = fprintf(h, "%02u:%02u:%02u  ", loc_time->tm_hour,
						   loc_time->tm_min,
						   loc_time->tm_sec);
	if (rc == -1) {
	    fprintf(stderr, "\7fatal: can't write logfile\n");
	    abort();
	}
	va_start (par, str);
	vfprintf(h, str, par);
	va_end(par);
	if (str[strlen(str)-1] != '\n') {
	    fputc('\n', h);
	}
	
	if (level > 5)  {
		fflush(h);
		setvbuf(stderr, NULL, _IONBF, 0);
		
		fprintf(stderr, "%02u:%02u:%02u  ",
		loc_time->tm_hour, loc_time->tm_min, loc_time->tm_sec);
		va_start (par, str);
		vfprintf(stderr, str, par);
		va_end(par);
		if (str[strlen(str)-1] != '\n') {
		    fputc('\n', stderr);
		}
		fflush(stderr);
	}

	fclose (h);

	return;
}
/*******************************************/

BOOL is_filter_found(char *str, struct filter *f)
{
	for( ; f ; f = f->next)
	{
		if (f->preg) {
			if (0 == regexec(f->preg, str, 0, NULL, 0))
				return Ok;;
		}
		else {
			if (strstr(str, f->str))
				return Ok;
		}
	}

	return Fail;
}

/************************/
void open_log(void)
{
    struct tm *tm;
    time_t t;
    FILE *h;
    
    if (lockFile.handle < 0) return;
    
    if (NULL == (h = fopen(LogFile, "a+t"))) return;

	t = time(NULL);
	tm = localtime(&t);
	fprintf(h, "--------  %s %02u %s %04u, %s %s %s\n", 
	        Dow[tm->tm_wday].dow, tm->tm_mday, Months[tm->tm_mon].name, tm->tm_year+1900,
		ProdName, VER, BUILD_STR);

	fclose (h);
	
	return;
	
}

/************************************************************************/
void close_log(void)
{
    time_t cur_time;
    struct tm *loc_time;
    
    if (lockFile.handle < 0) return;
	if (LogFile) {
		FILE *h;

		if (NULL == (h = fopen(LogFile, "a+t")))
			return;

		cur_time = time(NULL);
		loc_time = localtime(&cur_time);
		fprintf(h, "%02u:%02u:%02u  Runtime: %ld Seconds - Messages  Read : %d  Written : %d\n\n",
		    loc_time->tm_hour, loc_time->tm_min, loc_time->tm_sec,
		    (unsigned long)time(NULL)-tStart, nRead, nWritten);
		fflush(h);
		fclose(h);
		free (LogFile);
		LogFile = NULL;
	}
	return;
}
/************************/
int fidoname_to_number( char *name )
{
	char *p, base[128] = {0};
	
	p = base;
	
	while( *name && (*name != '.') ) {
		*p++ = *name++;
	}
	
	return __isnumber(base) ? atoi(base) : 0 ;
}
/************************************************************************/
void log_msg_hdr(void)
{
	char buf[256], to[128], from[128];
	
	if (ftn.stat & MsgLogged) {  /* already logged */
		return;
	}
	
	addr_to_str(ftn.to, to);
	
	addr_to_str(ftn.from, from);
	
	sprintf(buf, "%d %s (%s) -> %s (%s)\n", CurMsg.n, ftn.fromName, from,
					ftn.toName, to );
	writeToLog(1, buf);
	
	ftn.stat |= MsgLogged;
	
	return;
}
/*****************************/
unsigned long get_MsgId(void)
{
#ifdef MSGIDSEQ
	return (unsigned long)GenMsgId(SeqDir, SeqOutRun);
#else
	return get_MsgId_original();
#endif
}

unsigned long get_MsgId_original(void)
{
	unsigned long msgid = 0;
	time_t s;
	int s100;

	while( msgid <= MsgId)
	{
#if defined DOS
		struct dostime_t dt;
		_dos_gettime( &dt );
		s100 = dt.hsecond;
#elif defined UNIX || defined EMX
		struct timeval tv;
		struct timezone tz;
		gettimeofday(&tv, &tz);
		s100 = tv.tv_usec/10;
#else    
		SYSTEMTIME lpt;
		GetSystemTime(&lpt);
		s100 = lpt.wMilliseconds/10;
#endif    
		s = time(NULL);
		msgid = ( s<<7 ) | s100;
	}
	MsgId = msgid; /* global */
	
	return msgid;
}

/************************************************************************/

void fix_path( char *outdir) 
{
	struct find_t ffblk;
	BOOL found = No;
	char buf[256];

	if ( !_dos_findfirst(outdir, _A_SUBDIR, &ffblk) ) {
		/* something found */
		if ( (ffblk.attrib & _A_SUBDIR) != _A_SUBDIR ) {
			/* not a directory, let's try to delete this crap */
#ifndef UNIX
			chmod(outdir, S_IWRITE);
#endif
			remove(outdir);
			found = No;
		}
		else {
			/* directory */
			found = Yes;
		}
		_dos_findclose(&ffblk);
	}

	if (!found) {
	    if ( mymkdir(outdir ) == -1) {
		perror(outdir);
		sprintf(buf, "fix_path: error making dir: `%s'\n", outdir);
		fprintf(stderr, "\7%s", buf);
		writeToLog(0, buf);
		abort();
	    }
	}
	
	return;
}

static int get_metric (const ADDR* a1, const ADDR* a2)
{
	if (a1->net != a2->net)
		return 1; /* zones match  */
	if (a1->node != a2->node)
		return 2; /* nets match   */
	if (a1->point != a2->point)
		return 3; /* nodes match  */
	return 4;         /* points match */
	
}
/*******************************************************************/
BOOL get_def_addr( char *system, const ADDR addrIn, ADDR *addrOut )
{
	/* returns true if data system found, false otherwise */
	
	struct SystemTags *SystemTag = NULL;
	struct System *sys = NULL;
	char buf[256], caddr[128];
	int i, l = 0;
	BOOL found = No, rc = Fail;
	word z;
	int metric0 = 0;
	
	if (!strlen(system)) {
		*addrOut = MainAddress;
		writeToLog(4, "Data System not defined for this area, using MainAddress\n");
		return Fail;
	}
	
	for( SystemTag = SystemTagFirst; SystemTag; SystemTag = SystemTag->next )
		if (!stricmp(system, SystemTag->tag))
			break;
	
	if (SystemTag) {

		rc = Ok;
		
		sprintf(buf, "match_aka: data system `%s' found ok\n", system);
		writeToLog(5, buf);
		
		
		for( sys = SystemTag->recfirst; sys; sys = sys->next )
		{
			++l;
			addr_to_str(sys->addr, caddr);
			sprintf(buf, "match_aka: seeking match for line %d (address: `%s', domain:`%s')\n",
							l, caddr, sys->domain);
			writeToLog(5, buf);
			
			for( i = 0; i < 20; ++i )
			{
				z = sys->zones[i];
				if (z == 0)
					break;
				sprintf(buf, "match_aka: next system zone: %d, compare to address zone: %d\n",
								z, addrIn.zone);
				writeToLog(5, buf);
				if (z == addrIn.zone) {
					int metric = get_metric (&addrIn, &sys->addr);
					if (metric > metric0) {
						metric0 = metric;
						*addrOut = sys->addr;
						strcpy(addrOut->domain, sys->domain);
						found = Yes;
						sprintf(buf, "match_aka: zones match ok (zone:%d, domain:`%s')\n",
										z, addrOut->domain);
						writeToLog(5, buf);
					}
					break;
				}
			}
/*			if (found) {
				break;
			}
*/		} /* sys loop */
		if (!found) {
			sprintf(buf, "match_aka: no matching found for zone %d, using MainAddress\n",
							addrIn.zone);
			writeToLog(5, buf);
		}
	} /* system tag found */
	else {
		sprintf(buf, "match_aka: data system `%s' not found, using MainAddress\n",
						system);
		writeToLog(5, buf);
	}
	
	if (!found) {
		*addrOut = MainAddress;
	}
	
	return rc;
}
/********************************************************************/
void add_text(TEXT **text, TEXT *subtxt, int lineno, BOOL mode)
{
	/* subtxt is NOT freed here!!! */
	
	/* if lineno = -1 then add to the end of the message */
	
	TEXT *tmp_text = *text, *newtext = NULL, *Last= NULL, *newFirst = NULL;
	BOOL forceAdd = No, txtAdded = No, nextline = No, complete_line = Yes;
	int l;

	if (lineno == 0) {
		lineno = 1;
		mode = Mode_Bin;
	}

	for(l = 0; tmp_text; tmp_text = tmp_text->next)
	{
		nextline = No;

		if (No == txtAdded) {

			if ( mode == Mode_Bin )
				nextline = Yes;
			else {
				if (is_kludge(tmp_text->str) ||
						(tmp_text == *text && !strncmp("AREA:", tmp_text->str, 5)) )
				{

					if (is_reserved_kludge(tmp_text->str) == 2)
						forceAdd = Yes;      /* lower kludge found, insert subtxt before */

				}
				/* not a kludge */
				else {
					
					/* tearline or origin found, insert subtxt before */
					if ( (is_TearLine(tmp_text->str) || is_Origin(tmp_text->str))
							&& lineno != -1 )
						forceAdd = Yes;
					else {
						if (complete_line == Yes) {
							nextline = Yes;
						}
						if (tmp_text->str[strlen(tmp_text->str)-1] == '\r')
							complete_line = Yes;
						else
							complete_line = No;
					}

				}     /* not a kludge */

			}       /* mode = Mode_Text */

			if (nextline == Yes)
				++l;

			if (l == lineno) {
				forceAdd = Yes;
			}

			/* insert new subtext */
			if (forceAdd == Yes) {
				for( ; subtxt; subtxt = subtxt->next )
				{
					newtext = (TEXT *) calloc(1, sizeof(TEXT));
					newtext->next = NULL;
					strcpy( newtext->str, subtxt->str);
					if (Last)
						Last->next = newtext;
					else
						newFirst = newtext;
					Last = newtext;
				}
				/* make sure the subtxt will not me added more than once */
				txtAdded = Yes;
			}                /* insert subtext */
		}                  /* txtAdded == No */

		/* copy next line of old text */
		newtext = (TEXT *) calloc(1, sizeof(TEXT) );
		newtext->next = NULL;
		strcpy( newtext->str, tmp_text->str);
		if (Last)
			Last->next = newtext;
		else
			newFirst = newtext;
		Last = newtext;

	}  /* main text loop */

	if (No == txtAdded) {
		/* add subtext at the end */
		for(; subtxt; subtxt = subtxt->next)
		{
			newtext = (TEXT *) calloc(1, sizeof(TEXT));
			if (newtext != NULL) {
				newtext->next = NULL;
				strcpy( newtext->str, subtxt->str);
				if (Last)
					Last->next = newtext;
				else
					newFirst = newtext;
				Last = newtext;
			}
		}
	}

	free_text(*text);
	*text = newFirst;

	return;

}
/********************************************************************/
BOOL kill_text( TEXT **text , int lineno, BOOL mode)
{
	TEXT *t_prev = NULL, *t_next = NULL;
	TEXT *t = *text, *tmp;
	int l = 0;

	while(t)
	{
		if ((mode == Mode_Bin) ||
				(!is_kludge(t->str) && !is_TearLine(t->str) && !is_Origin(t->str)))
				
		{
			++l;
		}
		
		if (l == lineno) {

			while (t && t->next) {
				if (t->str[strlen(t->str)-1] == '\r') {
					break;
				}
				tmp = t;
				t = t->next;
				free (tmp);
			}
			
			t_next = t->next;

				
			free(t);
				
			if (t_prev) {
				t = t_prev;
				t->next = t_next;
			}
			else {
				t = t_next;
				*text = t;
			}

			return Ok;

		} /* line found */
		else {

			while (t && t->next) {
				if (t->str[strlen(t->str)-1] == '\r') {
					break;
				}
				t = t->next;
			}
			
		}
			
		t_prev = t;
		t = t->next;
		
	}
	
	return Fail;
	
}
/*
BOOL kill_text( TEXT **text , int lineno, BOOL mode)
{
	TEXT *newtext = NULL, *Last= NULL, *newFirst = NULL;
	TEXT *tmp_txt = *text;
	BOOL skip = No, killed = No;
	int l;

	for(l = 0; tmp_txt; tmp_txt = tmp_txt->next)
	{
		skip = No;

		if (No == killed) {
			if (mode == Mode_Bin)
				++l;
			else {
				if (!is_kludge(tmp_txt->str)   &&
						!is_TearLine(tmp_txt->str) && 
						!is_Origin(tmp_txt->str))
				{
//        if ( strcmp(tmp_txt->str, "\r") ) { // only non-empty lines count
					++l;
//        }
				}
			}
			if (l == lineno) {
				skip = Yes;
				killed = Yes;
			}
		}
			
		if (No == skip) {
			newtext = (TEXT *) calloc(1, sizeof(TEXT) );
			newtext->next = NULL;
			strcpy(newtext->str, tmp_txt->str);
			if (Last)
				Last->next = newtext;
			else
				newFirst = newtext;
			Last = newtext;
		}
	}

	free_text(*text);
	
	*text = newFirst;
	
	return killed;
	
}
*/
/***************************************************************************/
void free_text(TEXT *Text)
{
	TEXT *curptr;
	
	while(Text) {
		curptr = Text;
		Text = Text->next;
		free(curptr);
	}
	return;
}
/***************************************************************************/
char *expand_str( char *str)
{
	int i, rec_level = 0, pos;
	char *delim[4], *res[4], *env = NULL, macro[256]={0}, buf[256];
	char save[256], strbuf[256]={0}, str_to_parse[256];
	BOOL all_done = 0, cat_ok = 0, macro_ok = 0;

	if (strlen(str) > 255)
		return str;
	
	strcpy(save, str);
	
	for( i = 0; i < 4; ++i )
	{
		res[i] = (char *)calloc(256, sizeof(char));
		delim[i] = (char *)calloc(256, sizeof(char));
	}

	memset(&strbuf, 0, 256);
	strcpy(str_to_parse, str);
	
	while(strlen(str_to_parse)) {
		
		strcpy(delim[0], "%%");  strcpy(delim[1], "%%");
		i = parse(str_to_parse, delim, res);
		
		if (i != 3) {
			if (!strncat_max(strbuf, str_to_parse, 255)) {
				sprintf(buf, "expanded string for `%s' too long, truncated\n", save);
				writeToLog(3, buf);
				all_done = 1;
			}
			break;
		}
		
		if (!get_macro( res[1] , macro ) && !get_global_var_old( res[1], macro) ) {
			sprintf(buf, "no such macro: `%s'\n", res[1]);
			writeToLog(3, buf);
			strncat( strbuf, str_to_parse, strlen(res[0]) + strlen(res[1]) + 4 );
			strcpy(buf, res[2]);
		}
		else {
			if (!strncat_max(strbuf, res[0], 255)) {
				sprintf(buf, "expanded string for `%s' too long, truncated\n", save);
				writeToLog(3, buf);
				all_done = 1;
				break;
			}
			if (!strncat_max(strbuf, macro, 255)) {
				sprintf(buf, "expanded string for `%s' too long, truncated\n", save);
				writeToLog(3, buf);
				all_done = 1;
				break;
			}
		}
		strcpy(str_to_parse, res[2]);
	} /* while (expand %%) */

	if (!all_done) {
		
		strcpy(str_to_parse, strbuf);
		memset(&strbuf, 0, 256);
		memset(&macro, 0, 256);

		pos = -1;
		while(1) {
		
			strcpy(delim[0], "%");  strcpy(delim[1], "%");
			i = parse(str_to_parse, delim, res);
		
			if (i != 3) {
				if (!strncat_max(strbuf, str_to_parse, 255)) {
					sprintf(buf, "expanded string for `%s' too long, truncated\n", save);
					writeToLog(3, buf);
					all_done = 1;
				}
				break;
			}
			
			if (strlen(res[0]) == pos) {
				++rec_level;
			}
			else {
				rec_level = 0;
			}

			if (rec_level > 256) {
				rec_level = 0;
				sprintf(buf, "recursion level exceeded for a macro in `%s'\n",
								save );
				writeToLog(2, buf);
				macro_ok = 0;
			}
			else {
				
				pos = strlen(res[0]);

				if (strlen(res[1])) {
					
					strUpper(res[1]);
					env = getenv(res[1]);

					memset(macro, 0, 256);
					if (env) {
						strncpy(macro, env, 255);
						macro[255] = 0;
						macro_ok = 1;
					}
					else {
						sprintf(buf, "no such env: `%s'\n", res[1]);
						writeToLog(3, buf);
						if (!get_macro( res[1] , macro )  &&
								!get_global_var_old( res[1], macro) ) {
							sprintf(buf, "no such macro: `%s'\n", res[1]);
							writeToLog(3, buf);
							macro_ok = 0;
						}
						else {
							macro_ok = 1;
						}
					} /* env not found */
					
				} /* res[1] is not zerolength */
				else
						macro_ok = 0;
						
			} /* recursion not too deep */

			if (!macro_ok) {
				strncat( strbuf, str_to_parse, strlen(res[0]) + strlen(res[1]) + 2 );
				strcpy(buf, res[2]);
				strcpy(str_to_parse, res[2]);
				pos = -1;
			}
			else {
				memset(&str_to_parse, 0, 256);
				cat_ok = strncat_max(str_to_parse, res[0], 255);
				cat_ok = strncat_max(str_to_parse, macro, 255);
				cat_ok = strncat_max(str_to_parse, res[2], 255);
				if (!cat_ok) {
					sprintf(buf, "expanded string for `%s' too long, truncated\n", save);
					writeToLog(3, buf);
					strncat_max(strbuf, str_to_parse, 255);
					break;
				}
			}
		} /* while (expand %) */
	} /* all done */
	
	strcpy(str, strbuf);

	for( i = 0; i < 4; ++i )
	{
		free(res[i]);
		free(delim[i]);
	}
	
	return str;
}
/*************************************************************************/
BOOL get_macro( char *str, char *macro)
{
    int i;
    char buf[256], system[20], w1[20];
    struct Macro *tmpMacro;
    ADDR addr, aka;
    BOOL rc = Fail;
	
    sprintf( buf , "macro to expand: `%s'\n", str );
    writeToLog(4, buf);
	
    for( i = 0; strlen( Macros[i].s ); ++i)
 {
	if ( !stricmp(Macros[i].s , str) ) {
			
	    if (CurMsg.n == 0) {
		if (Macros[i].n != m_Program    &&
		    Macros[i].n != m_Systemtime &&
		    Macros[i].n != m_Localtime  &&
		    Macros[i].n != m_Localdate  &&
		    Macros[i].n != m_TZ  &&
		    Macros[i].n != m_System  )
 {
		    
		    writeToLog(4, "expanding macro: no active message in memory\n");
        	    memset(macro, '\0', 256);
		    return Fail;
		}
	    }
			
	rc = Ok;
			
	switch( Macros[i].n ) {
				
	    case m_Program:
		sprintf( macro, "%s %s", ProdName, VER );
		break;
				
/*      case m_Nodelist: // ??? who the fuck ever need it? */

	    case m_Systemtime:
				
		fetch_systime(macro);
		break;
				
	    case m_Localtime: {
		struct tm *stm;
		time_t tt;
				    
		tt = time(NULL);
		stm = localtime(&tt);
				    
		sprintf(macro, "%02d:%02d:%02d", stm->tm_hour, 
						 stm->tm_min,
						 stm->tm_sec);
		}
		break;
				
	    case m_Localdate: {
		struct tm *stm;
		time_t tt;
				    
		tt = time(NULL);
		stm = localtime(&tt);
			    
		sprintf(macro, "%02d/%02d/%d", stm->tm_mon+1, 
					       stm->tm_mday,
					       stm->tm_year+1900);
		}
		break;

	    case m_TZ:
 {
		sprintf(macro, "%+d", get_TZ());
		break;
		}

			/*///////////////////////////////*/
				
	    case m_System:
		aka = MainAddress;
		if (CurMsg.n && areaname_to_system(CurMsg.areaname, system)) {
		    addr = ftn.to;
		    get_def_addr(system, addr, &aka);
		}
		addr_to_str( aka, macro );
		break;
				
	    case m_fromName:
		strcpy( macro, ftn.fromName );
		break;
				
	    case m_from:
		addr_to_str( ftn.from, macro );
		break;
				
	    case   m_toName:
		strcpy( macro, ftn.toName );
		break;
				
	    case   m_to:
		addr_to_str( ftn.to, macro );
		break;
				
	    case   m_Time:
		strcpy( macro, ftn.DateTime );
		break;
				
	    case m_Attributes:
		get_attr(macro);
		break;
				
	    case m_Subject:
		strcpy( macro, ftn.subject );
		break;
				
	    case m_lineCount:
		sprintf(macro, "%u", ftn.lines);
/*				itoa(ftn.lines, macro, 10); */
		break;
				
	    case m_msgLength:
		if (!ftn.msglen)
		    ftn.msglen = textlen(TextFirst);

		sprintf(macro, "%u", (unsigned)ftn.msglen);
/*				itoa(ftn.msglen, macro, 10); */
		break;
				
	    case m_intlFrom:
		strcpy(buf, ftn.intl);
		get_next_word(buf, w1);
		strcpy(macro, buf);
		break;
				
	    case m_intlTo:
		strcpy(buf, ftn.intl);
		get_next_word(buf, macro);
		break;
				
	    case m_msgidFrom:
		strcpy(macro, ftn.msgid);
		break;
				
	    case m_replyTo:
		strcpy(macro, ftn.reply);
		break;
				
	    case m_routeTo:
		addr_to_str(ftn.routeTo, macro);
		break;
				
	    case m_toDomain:
		if (areaname_to_system(CurMsg.areaname, system)) {
		    addr = ftn.to;
		    get_def_addr(system, addr, &aka);
		}
 else {
		    aka = MainAddress;
		}
		strcpy(macro, aka.domain);
				
		break;
				
	    case m_fromDomain:
		if (areaname_to_system(CurMsg.areaname, system)) {
		    addr = ftn.from;
		    get_def_addr(system, addr, &aka);
		}
 else {
		    aka = MainAddress;
		}
		strcpy(macro, aka.domain);
				
		break;
				
	    case m_old_fromName:
		strcpy( macro, old_ftn.fromName );
		break;
				
	    case m_old_from:
		addr_to_str( old_ftn.from, macro );
		break;
				
	    case   m_old_toName:
		strcpy( macro, old_ftn.toName );
		break;
				
	    case   m_old_to:
		addr_to_str( old_ftn.to, macro );
		break;
				
	    case   m_old_Time:
		strcpy( macro, old_ftn.DateTime );
		break;
				
	    case m_old_Attributes:
		get_attr(macro);
		break;
				
	    case m_old_Subject:
		strcpy( macro, old_ftn.subject );
		break;
				
	    case m_old_lineCount:
		sprintf(macro, "%u", old_ftn.lines);
/*				itoa(old_ftn.lines, macro, 10); */
		break;
				
	    case m_old_msgLength:
		sprintf(macro, "%u", (unsigned)old_ftn.msglen);
/*				itoa(old_ftn.msglen, macro, 10); */
		break;
				
	    case m_old_intlFrom:
		strcpy(buf, old_ftn.intl);
		get_next_word(buf, w1);
		strcpy(macro, buf);
		break;
				
	    case m_old_intlTo:
		strcpy(buf, old_ftn.intl);
		get_next_word(buf, macro);
		break;
				
	    case m_old_msgidFrom:
		strcpy(macro, old_ftn.msgid);
		break;
				
	    case m_old_replyTo:
		strcpy(macro, old_ftn.reply);
		break;
				
	    case m_old_routeTo:
		addr_to_str(old_ftn.routeTo, macro);
		break;
				
	    case m_old_toDomain:
		if (areaname_to_system(CurMsg.areaname, system)) {
		    addr = old_ftn.to;
		    get_def_addr(system, addr, &aka);
		}
 else {
		    aka = MainAddress;
		}
		strcpy(macro, aka.domain);
		break;
				
	    case m_old_fromDomain:
		if (areaname_to_system(CurMsg.areaname, system)) {
		    addr = old_ftn.from;
		    get_def_addr(system, addr, &aka);
		}
 else {
		    aka = MainAddress;
		}
		strcpy(macro, aka.domain);
		break;

	    case m_pktFrom:
		if (CurMsg.type == PKT)
		    addr_to_str(CurMsg.pkt.from, macro);
		else if (CurMsg.type == BINK)
		    addr_to_str(CurMsg.bink.from, macro);
		break;
					
	    case m_pktTo:
		if (CurMsg.type == PKT)
		    addr_to_str(CurMsg.pkt.to, macro);
		else if (CurMsg.type == BINK)
		    addr_to_str(CurMsg.bink.to, macro);
		break;
				
	    default:
		rc = Fail;
		break;
	    }
	    break;
	}
    }
    
    if (rc == Fail) {

	for (tmpMacro = MacroFirst; tmpMacro; tmpMacro = tmpMacro->next) {
	    if (!stricmp(tmpMacro->macrotag, str)) {
	        strcpy(macro, tmpMacro->macroname);
	        rc = Ok;
	        break;
	    }
	}
	if (rc == Fail) {
	    writeToLog(0, "macro not yet implemented: `%s'", str );
	}
    }
    
    if (rc) {
	sprintf( buf , "expanded macro: `%s'\n", macro );
	writeToLog(4, buf);
    }
    
    return rc;
}
/**********************************************************************/
void set_flag_kludge(TEXT **txt, char *cflag)
{
	TEXT *tmp_txt = *txt;
	register int flagLine = 0, l, i;

	
	for( l = 1, i = 1; tmp_txt; tmp_txt = tmp_txt->next, ++i )
	{
		/* not a kludge */
		if (l==1 && !strncmp("AREA", tmp_txt->str, 4)) {
			++l;
		}
		else 	if (!is_kludge(tmp_txt->str)) {
			break;
		} else if ( !strnicmp("\1FLAGS ", tmp_txt->str, 7) ) {
			flagLine = i;
			break;
		} else if (!strncmp("\1INTL", tmp_txt->str, 5) ||
							 !strncmp("\1TOPT", tmp_txt->str, 5) ||
							 !strncmp("\1FMPT", tmp_txt->str, 5))	 {
			
			++l;
		}
	}
	
	tmp_txt = NULL;

	if (strlen(cflag)) {
		tmp_txt = (TEXT *) calloc(1, sizeof(TEXT) );
		tmp_txt->next = NULL;
		sprintf(tmp_txt->str, "\1FLAGS %s\r", cflag );
	}
	
	/* FLAGS already exist */
	if (flagLine) {  
		
		/* kill old FLAGS line */
		kill_text(txt, flagLine, Mode_Bin);
		
		/*  there's FLAGS in the new message */
		if (tmp_txt)
			(void) add_text(txt, tmp_txt, flagLine, Mode_Bin);
	}
	else {
		/*  there's FLAGS in the new message */
		if (tmp_txt)  
			(void) add_text(txt, tmp_txt, l, Mode_Bin);
	}

	free_text(tmp_txt);
				
	return;
}
/**********************************************************************/
char *remove_quotes(char *str)
{
	char *p = str;
	
	if ( p[0] == '"' && p[strlen(p)-1] == '"' ) {
		p[strlen(p)-1] = '\0';
		++p;
	}
	else {
		if ( p[0] == '\'' && p[strlen(p)-1] == '\'' ) {
			p[strlen(p)-1] = '\0';
			++p;
		}
	}
	strcpy( str , p );
	
	return str;
}

/******************************************************/
BOOL is_TearLine( char *str)
{
	if ( !strncmp(str, "--- ", 4) || !strcmp(str, "---\r") ) {
		return Ok;
	}
	else {
		return Fail;
	}
}

/******************************************************/
BOOL is_Origin( char *str)
{
	if ( !strncmp(str, " * Origin:", 10) ) {
		return Ok;
	}
	else {
		return Fail;
	}
}
/******************************************************/
BOOL is_Via( char *str)
{
	if ( !strnicmp(str, "\001Via", 4) ) {
		return Ok;
	}
	else {
		return Fail;
	}
}
/******************************************************/
BOOL is_kludge( char *str)
{
	if ( str[0] == '\001' || !strncmp(str, "SEEN-BY", 7)) {
		return Ok;
	}
	else {
		return Fail;
	}
}
/**********************/
int is_reserved_kludge( char *kl)
{
	int i;

	if (*kl == '\001')
		++kl;
	
	for(  i = 0 ; strlen( Kludges[i].s ); ++i )
		if ( !strncmp( kl , Kludges[i].s , (int) strlen(Kludges[i].s) ) )
			return Kludges[i].type;

	return Fail;
}
/******************************************************/
char *strip(char *str) /* skip leading&trailing spaces&tabs */
{
    char *p = str;
	
    /* skip leading spaces &tabs */
    while( *p && (*p == ' ' || *p == '\t') ) ++p;
	
	/* skip trailing spaces & tabs */
    while((p[strlen(p)-1] == ' ') || (p[strlen(p)-1] == '\t'))
			    p[strlen(p)-1] = '\0';
	
    if (*p) {
	memmove(str, p, strlen(p));
	str[strlen(p)] = '\0';
    } else {
	*str = '\0';
    }
	
    return str;
}
/*****************************/
BOOL isW(word c)
{
	if ( c == Wild_1 || c == Wild_2 ) {
		return Ok;
	}
	else {
		return Fail;
	}
}
/*************************************/
BOOL from_hub(ADDR addr, ADDR hub)
{
	struct ndlTags *ndlTag = NULL;
	int z, r, h;

	if (addr.zone != hub.zone || addr.net != hub.net)
		return Fail;
	
	for( ndlTag = v9_ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
		if (is_v9_listed(ndlTag, addr, &z, &r, &h))
			if ((short)h == hub.node)
				return Ok;
	
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
	{
		word j, uplink;
		if (is_listed(ndlTag, addr, &uplink, &j) && uplink== hub.node)
			return Ok;
	}
	
	return Fail;
}
/*************************************/
BOOL is_reg_indep(ADDR addr)
{
	struct ndlTags *ndlTag = NULL;
	int z, r, h;
	
	for( ndlTag = v9_ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
		if (is_v9_listed(ndlTag, addr, &z, &r, &h))
			if ((short)h == 0 && addr.net == r)
				return Ok;

	return Fail;
}
/*************************************/
BOOL is_zone_indep(ADDR addr)
{
	struct ndlTags *ndlTag = NULL;
	int z, r, h;
	
	for( ndlTag = v9_ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
		if (is_v9_listed(ndlTag, addr, &z, &r, &h))
			if ((short)h == 0 && addr.net == r && addr.net == z)
				return Ok;

	return Fail;
}
/*************************************/
BOOL get_ndl_info(ADDR addr, word* hub, word* region)
{
	struct ndlTags *ndlTag = NULL;
	int z;
	
	for( ndlTag = v9_ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
		if (is_v9_listed(ndlTag, addr, &z, region, hub))
			return Ok;
	
	for( ndlTag = ndlTagFirst; ndlTag; ndlTag = ndlTag->next )
		if (is_listed(ndlTag , addr, hub, region))
			return Ok;
	
	return Fail;
}

/*********************************/
BOOL strncat_max(char *to_str, char *from_str, int max)
{
	int free = max - strlen(to_str), len;
	BOOL rc = Ok;

	len = strlen(from_str);
	if (len >= free) {
		len = free;
		rc = Fail;
	}
	
	if (len) {
		strncat(to_str, from_str, len);
	}

	return rc;
}
/**************************************************/
int areaname_to_type(char *areatag)
{
	struct FidoAreas *FidoArea;
	struct BinkAreas *BinkArea;
	struct SquishAreas *SquishArea;
	struct JamAreas *JamArea;
	struct PktAreas  *PktArea;
	struct TextAreas *TextArea;
	
	for( FidoArea = FidoAreaFirst; FidoArea; FidoArea = FidoArea->next )
	{
		if ( !stricmp( FidoArea->name, areatag ) ) {
			return FIDO;
		}
	}
	
	for( BinkArea = BinkAreaFirst; BinkArea; BinkArea = BinkArea->next )
	{
		if ( !stricmp( BinkArea->name, areatag ) ) {
			return BINK;
		}
	}
	
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea->next )
	{
		if ( !stricmp( PktArea->name, areatag ) ) {
			return PKT;
		}
	}
	
	for( SquishArea = SquishAreaFirst; SquishArea; SquishArea = SquishArea->next )
	{
		if ( !stricmp( SquishArea->name, areatag ) ) {
			return SQUISH;
		}
	}
	
	for( JamArea = JamAreaFirst; JamArea; JamArea = JamArea->next )
	{
		if ( !stricmp( JamArea->name, areatag ) ) {
			return JAM;
		}
	}
	
	for( TextArea = TextAreaFirst; TextArea; TextArea = TextArea->next )
	{
		if ( !stricmp( TextArea->name, areatag ) ) {
			return TXT;
		}
	}
	
	return -1;
}
/**************************************************/
BOOL areaname_to_system(char *areatag, char *system)
{
	struct FidoAreas *FidoArea;
	struct BinkAreas *BinkArea;
	struct SquishAreas *SquishArea;
	struct JamAreas *JamArea;
	struct PktAreas  *PktArea;
	
	for( FidoArea = FidoAreaFirst; FidoArea; FidoArea = FidoArea->next )
	{
		if ( !stricmp( FidoArea->name, areatag ) ) {
			strcpy( system , FidoArea->system );
			return Ok;
		}
	}
	
	for( BinkArea = BinkAreaFirst; BinkArea; BinkArea = BinkArea->next )
	{
		if ( !stricmp( BinkArea->name, areatag ) ) {
			strcpy( system , BinkArea->system );
			return Ok;
		}
	}
	
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea->next )
	{
		if ( !stricmp( PktArea->name, areatag ) ) {
			strcpy( system , PktArea->system );
			return Ok;
		}
	}
	
	for( SquishArea = SquishAreaFirst; SquishArea; SquishArea = SquishArea->next )
	{
		if ( !stricmp( SquishArea->name, areatag ) ) {
			strcpy( system , SquishArea->system );
			return Ok;
		}
	}
	
	for( JamArea = JamAreaFirst; JamArea; JamArea = JamArea->next )
	{
		if ( !stricmp( JamArea->name, areatag ) ) {
			strcpy( system , JamArea->system );
			return Ok;
		}
	}
	
	return Fail;
	
}
/**************************************************/
BOOL areaname_to_path(char *areatag, char *path)
{
	struct FidoAreas *FidoArea;
	struct BinkAreas *BinkArea;
	struct SquishAreas *SquishArea;
	struct JamAreas    *JamArea;
	struct PktAreas  *PktArea;
	struct TextAreas *TextArea;
	
	for( FidoArea = FidoAreaFirst; FidoArea; FidoArea = FidoArea->next )
	{
		if ( !stricmp( FidoArea->name, areatag ) ) {
			strcpy( path , FidoArea->path );
			return Ok;
		}
	}
	
	for( BinkArea = BinkAreaFirst; BinkArea; BinkArea = BinkArea->next )
	{
		if ( !stricmp( BinkArea->name, areatag ) ) {
			strcpy( path , BinkArea->path );
			return Ok;
		}
	}
	
	for( PktArea = PktAreaFirst; PktArea; PktArea = PktArea->next )
	{
		if ( !stricmp( PktArea->name, areatag ) ) {
			strcpy( path , PktArea->path );
			return Ok;
		}
	}
	
	for( SquishArea = SquishAreaFirst; SquishArea; SquishArea = SquishArea->next )
	{
		if ( !stricmp( SquishArea->name, areatag ) ) {
			strcpy( path , SquishArea->path );
			return Ok;
		}
	}
	
	for( JamArea = JamAreaFirst; JamArea; JamArea = JamArea->next )
	{
		if ( !stricmp( JamArea->name, areatag ) ) {
			strcpy( path , JamArea->path );
			return Ok;
		}
	}
	
	for( TextArea = TextAreaFirst; TextArea; TextArea = TextArea->next )
	{
		if ( !stricmp( TextArea->name, areatag ) ) {
			strcpy( path , TextArea->pathname );
			return Ok;
		}
	}
	
	return Fail;
	
}
/************************************************************/
void fetch_systime(char *buf)
{
	char ctz[5]={0};
	struct tm *tm;
	time_t t;
/*  long tz = - ((_timezone/3600L) % 12L ); */
	int tz;
	int year;

	
	t = time(NULL);
	tm = localtime(&t);
	year = 1900 + tm->tm_year;
	if (tm->tm_year < 70)
		year = 2000 + tm->tm_year;
	
	
	sprintf(buf, "@%d%02d%02d.%02d%02d%02d.UTC",
					year, tm->tm_mon+1, tm->tm_mday,
					tm->tm_hour, tm->tm_min, tm->tm_sec);

	if (0 != (tz = get_TZ())) {
		sprintf(ctz, "%+d", tz);
		strcat(buf, ctz);
	}
	
	return;
}
/**********************/
int get_TZ(void)
/*   get time difference between local time and UTC     */
/*   with DST taken into account                        */
{
	int local_hour = 0, gmt_hour = 0;
	struct tm *tm;
	time_t t;
	int tz;
	static int TZ = -1L;

	if (TZ != -1L)
		return TZ;

	t = time(NULL);
	tm = gmtime(&t);
	gmt_hour = tm->tm_hour;
	tm = localtime(&t);
	local_hour = tm->tm_hour;
	
	tz = local_hour - gmt_hour;
	
	if ((tz+12) < 0) {
		tz = local_hour - gmt_hour + 24;
	}
	return TZ = tz;
}


/***********************/
void strip_CRLF(char *buf)
{
	while(*buf && (buf[strlen(buf)-1] == '\n' || buf[strlen(buf)-1] == '\r'))
		buf[strlen(buf)-1] = '\0';

	return;
}

/***************************/
dword textlen(TEXT *t)
{
	dword len = 0L;
	for( ; t ; t = t->next)
		len += strlen(t->str);
	return len;
}

BOOL is_full_path(char *p)
{
	return (p[0]==DELIM || p[1]==':');
}


void adjust_lastread (const char* lr, int type)
{
	struct stat st;
	
	if (0 == stat (lr, &st) && st.st_size) {
		
		int fd = sopen ((char*)lr, O_RDWR|O_BINARY, SH_DENYNO, ACCFILE);

		
		if (fd != -1) {
			int ptr;
			
			char* lrbuf = (char*)xmalloc (st.st_size);
			
			if (-1 != read (fd, lrbuf, st.st_size)) {

				for (ptr = 0; ptr < st.st_size; )
				{
					if (type == LR_JAM) {
						JAMLREAD jlr;
						memcpy (&jlr, &lrbuf[ptr], sizeof jlr);
						if (jlr.LastReadMsg == CurMsg.realNo)
							jlr.LastReadMsg = CurMsg.lrno;
						if (jlr.HighReadMsg == CurMsg.realNo)
							jlr.HighReadMsg = CurMsg.lrno;
						memcpy (&lrbuf[ptr], &jlr, sizeof jlr);
						ptr += sizeof (JAMLREAD);
					}
					else if (type == LR_FIDO) {
						word lrno;
						memcpy (&lrno, &lrbuf[ptr], sizeof (word));
						if (lrno == CurMsg.realNo) {
							lrno = (word) CurMsg.lrno;
							memcpy (&lrbuf[ptr], &lrno, sizeof (word));
						}
						ptr += sizeof (word);
					}
					else if (type == LR_SQUISH) {
						dword lrno;
						memcpy (&lrno, &lrbuf[ptr], sizeof (dword));
						if (lrno == CurMsg.realNo) {
							lrno = (dword) CurMsg.lrno;
							memcpy (&lrbuf[ptr], &lrno, sizeof (dword));
						}
						ptr += sizeof (dword);
					}
					else
						break; /* error: invalid type, get out */
				}
				
				lseek (fd, 0, SEEK_SET);
				write (fd, lrbuf, st.st_size);
				
			}
			free (lrbuf);
			close (fd);
		}
	}
}
