/*
* Copyright (c) 1996, serge terekhov.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. All advertising materials mentioning features or use of this software
*    must display the following acknowledgement:
*      This product includes software developed by serge terekhov.
* 4. Neither the name of the author nor the names of any co-contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY SERGE TEREKHOV AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL SERGE TEREKHOV OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*
*/

/*       NODELIST.V9 compiler       */
/*       Revision 1.0               */

#ifdef __WATCOMC__
#include <direct.h>
#else
#include <sys/types.h>
#include <dirent.h>
#endif

#include "itrax.h"


uint16 hub[256];

V9NET *net;
V9NODE *fido;
static int h_num, max_hub, n_num, n_size = 0;
static int f_num, f_size, f_hub;

long sizeof_V9NET = sizeof (V9NET);
long sizeof_V9NODE = sizeof (V9NODE);
long sizeof_V9HEADER = sizeof (V9HEADER);


int fido_cmp (V9NODE *a, V9NODE *b)
{
	return (a->node < b->node)? -1 : (a->node > b->node);
}



int net_cmp (V9NET *a, V9NET *b)
{
	if (a->net != b->net)
			return (a->net < b->net)? -1 : 1;
	return (a->zone < b->zone)? -1 : (a->zone > b->zone);
}

static void build_stub(char *path, char *ext, char *ndlnames[], int nndl);


BOOL compile (char *name, word DefZone, word DefReg)
{
	FILE *fp;
	char junk[512], save_junk[512];
	uint16 z = 0, r = 0, n, f;
	int i;
	BOOL last = Yes, nozone = No;
	int fake = 0;

	
	if ((fp = fopen (name, "r")) == 0) {
		perror(name);
		sprintf(junk, "v9 compile, can't open nodelist file `%s'\n", name);
		writeToLog(0, junk);
		return Fail;
	}

	setvbuf (fp, 0, _IOFBF, 0x4000);

	while (fgets (junk, 511, fp) && junk[0] == ';')
		continue;

	if (strncmp (junk, "Zone,", 5)) {
		if (DefZone == Wild_1 || DefReg == Wild_1) {
			sprintf (junk, "no default zone/reg defined for nodelist `%s'\n", name);
			writeToLog(0, junk);
			fclose (fp);
			return Fail;
		}
		nozone = Yes;
		strcpy(save_junk, junk);
	}

	do {
	start:
		if (nozone) {
			if (fake == 0) {
				sprintf(junk, "Zone,%u", DefZone);
			}
			else if (fake == 1) {
				sprintf(junk, "Region,%u", DefReg);
			}
			else {
				nozone = 0;
				strcpy(junk, save_junk);
			}
			++fake;
		}
			/* Zone Region Host Hub 0 Pvt Hold Down */
			switch (junk[0]) {
				case 'Z':   /* Zone */
						if (!isdigit (junk[5]))
							continue;
						z = r = n = (uint16) atoi (&junk[5]);
						goto nw0;

				case 'R':   /* Region */
						if (!isdigit (junk[7]))
							continue;
						r = n = (uint16) atoi (&junk[7]);
				nw0:
						printf ("Zone %-5u Region %-5u\r", z, r);
						goto nw;

				case 'H':   /* Host Hub Hold */
						if (junk[1] == 'u')
							goto hub;
						if (junk[2] == 'l')
							goto hold;
						if (!isdigit (junk[5]))
							continue;
						n =  (uint16) atoi (&junk[5]);
				nw:
						if (n_num) {
							i = net[n_num - 1].off;
							qsort (&fido[i], f_num - i, sizeof_V9NODE, (int (*)(const void*, const void*)) fido_cmp);
							net[n_num - 1].size = (uint16) (f_num - i);
							for (i = 0; i < h_num; ++i) {
									if (f_num >= f_size) {
										/* assert: f_size != 0 && fido != 0 */
										fido = realloc (fido, (f_size <<= 1) * sizeof_V9NODE);
										if (!fido) {
											sprintf (junk, "v9 compile error: low memo allocating v9 fido\n");
											writeToLog(2, junk);
											abort();
										}
									}
									fido[f_num].node = hub[i];
									fido[f_num].hub = (char) (i + 1);
									++f_num;
							}
						}
						if (n_num >= n_size) {
							if (net)
									net = realloc (net, (n_size <<= 1) * sizeof_V9NET);
							else
									net = malloc ((n_size = 64) * sizeof_V9NET);
							if (!net) {
								sprintf (junk, "v9 compile error: low memo allocating v9 net\n");
								writeToLog(2, junk);
								abort();
							}
						}
						net[n_num].zone = z;
						net[n_num].region = r;
						net[n_num].net = n;
						net[n_num].off = f_num;
						++n_num;

						f = 0;
						f_hub = 0;
						h_num = 0;

						goto nd;

				hub:
						if (!isdigit (junk[4]))
							continue;
						f =  (uint16) atoi (&junk[4]);
						if (h_num < 255) {
							hub[h_num++] = f;
							f_hub = h_num;
							if (h_num > max_hub)
									max_hub = h_num;
						} else
							f_hub = 0;
						goto nd;

				case 'P':   /* Pvt */
						if (!isdigit (junk[4]))
							continue;
						f =  (uint16) atoi (&junk[4]);
						goto nd;

				case 'D':   /* Down */
				hold:
						if (!isdigit (junk[5]))
							continue;
						f =  (uint16) atoi (&junk[5]);
						goto nd;

				case ',':
						if (!isdigit (junk[1]))
							continue;
						f =  (uint16) atoi (&junk[1]);
				nd:
						if (f_num >= f_size) {
							if (fido)
									fido = realloc (fido, (f_size <<= 1) * sizeof_V9NODE);
							else
									fido = malloc ((f_size = 256) * sizeof_V9NODE);
							if (!fido) {
								sprintf (junk, "v9 compile error: low memo allocating v9 fido\n");
								writeToLog(2, junk);
								abort();
							}
						}
						fido[f_num].node = f;
						fido[f_num].hub = (char) f_hub;
						++f_num;
						break;
			}
		if (nozone) {
			goto start;
		}
	} while (fgets (junk, 511, fp));
	
	if (last) {
		last = 0;
		sprintf(junk, "Host,%hu", -1);
		goto start;
	}

	fclose (fp);
	return Ok;
}



void find_last (char *nm)
{
	DIR *dir;
	struct dirent *ent;
	int last = 0, day, len;
	char name[256], *p, *q;

	q = strrchr (strcpy (name, nm), '\\');
	if ((p = strrchr ((q)? q : name, '/')) != 0)
			q = p;
	if (q)
			*q++ = 0;
	else {
			q = nm;
			name[0] = '.';
			name[1] = 0;
	}
	len = strlen (q);

	if ((dir = opendir (name)) == 0)
			return;

	while ((ent = readdir (dir)) != 0) {
			p = ent->d_name;
#ifdef __OS2__
			strlwr (p);
#endif
			if (!strncmp (p, q, len) && *(p += len) == '.' &&
					isdigit (p[1]) && isdigit (p[2]) && isdigit (p[3]) && !p[4] &&
					(day = atoi (&p[1])) > 0 && day < 367) {
				if (!last ||
									(day > last && day - last < 300) ||
									(day < last && last - day > 300))
						last = day;
				}
			}
	closedir (dir);

	if (last)
			sprintf (nm + strlen (nm), ".%03d", last);
}


/***********************************************************************/
BOOL v9_ndl_build(int nndl, char *ndlnames[],
									char *path, char *ext, word DefZone, word DefReg)
{
	V9HEADER hd;
	int i, j;
	long off;
	char v9name[256];
	char nlname[256];
	char buf[256];

	h_num = max_hub = n_num = n_size = f_num = f_size = 0;
	net = NULL;
	fido = NULL;

	sprintf(v9name, "%sver9idx.%s", path, ext);

	for( i = 0; i < nndl; ++i )
	{
		sprintf(nlname, "%s%s", path, ndlnames[i] );
		printf("compiling:`%s'\n", nlname);
		if (Fail == compile(nlname, DefZone, DefReg))
			return Fail;
	}


	i = net[n_num - 1].off;
	
	qsort (&fido[i], f_num - i, sizeof_V9NODE, (int (*)(const void*, const void*)) fido_cmp);
	net[n_num - 1].size = (uint16) (f_num - i);

	qsort (net, n_num, sizeof_V9NET, (int (*)(const void*, const void*)) net_cmp);
	
	printf ("%d nets, %d nodes       \n", n_num, f_num);
	
	i = open (v9name, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, ACCFILE);
	if (i == -1) {
		sprintf (buf, "v9_build: error opening version 9 index file `%s'\n", v9name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s", buf);
		abort();
	}
	
	printf ("writing %s\n", v9name);
	memset (&hd, 0, sizeof (hd));
	hd.sig = V9_SIGNATURE;
	hd.ver = V9_VERSION;
	hd.nets = n_num;
	hd.nodes = f_num;
	hd.hubs = (uint16) max_hub;
	hd.time = time (0);
	
	if (write (i, &hd, sizeof (hd)) != sizeof (hd)) {
		sprintf (buf, "v9_build: error writing version 9 index file `%s'\n", v9name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s", buf);
		abort();
	}
	
	off = sizeof (hd) + n_num * sizeof_V9NET;
	
	for (j = 0; j < n_num; ++j)
		net[j].off = off + net[j].off * sizeof_V9NODE;
	
	if (write (i, net, n_num * sizeof_V9NET) != n_num * sizeof_V9NET ||
			write (i, fido, f_num * sizeof_V9NODE) != f_num * sizeof_V9NODE)
	{
		sprintf (buf, "v9_build: error writing version 9 index file `%s'\n", v9name);
		writeToLog(0, buf);
		fprintf(stderr, "\7%s", buf);
		close (i);
		abort();
	}
	
	close (i);
	
	printf ("size: header %d nets %ld nodes %ld\n",
					sizeof (hd), n_num * sizeof_V9NET, f_num * sizeof_V9NODE);

	free(fido);
	free(net);
	writeToLog(4, "done making v9 nodelist\n");

	build_stub(path, ext, ndlnames, nndl);
	

	return Ok;
}

/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/*       NODELIST.V9 searcher       */
/*       Revision 1.0               */


V9NODELIST *v9_open (char *n, int days)
{
	struct stat st;
	int h;
	V9HEADER *hdr;
	char *data;

	if (stat (n, &st) == -1 || st.st_size < sizeof_V9HEADER)
			return 0;
	if ((h = open (n, O_RDONLY|O_BINARY)) == -1)
			return 0;
	data = malloc (st.st_size);
	if (!data) {
			close (h);
			return 0;
	}
	if (read (h, data, st.st_size) != st.st_size) {
			close (h);
			free (data);
			return 0;
	}
	close (h);
	hdr = (V9HEADER *)data;
	if (hdr->sig != V9_SIGNATURE ||
			hdr->ver != V9_VERSION ||
			sizeof_V9HEADER + hdr->nets * sizeof_V9NET
				+ hdr->nodes * sizeof_V9NODE != st.st_size ||
			(days && (time (0) - hdr->time) / 86400l > days)) {
			free (data);
			return 0;
	}
	return (V9NODELIST *)data;
}



int v9_search (V9NODELIST *nl, int izone, int inet, int inode,
							int *ozone, int *oregion, int *ohub)
{
	V9HEADER *hdr = (V9HEADER *)nl;
	V9NET *pn, *net = (V9NET *)&hdr[1];
	V9NODE *pf, *fido;
	int i, j, off, num;
	int ret = 0;

	if (!hdr || hdr->sig != V9_SIGNATURE || hdr->ver != V9_VERSION || !inet)
			return -1;

	for (off = 0, num = hdr->nets; num > 0; ) {
			i = num >> 1;
			pn = &net[i + off];

			if ((uint16)inet == pn->net) {
				if (izone) {
						if ((uint16)izone == pn->zone)
							goto net_found;
						j = -((uint16)izone < pn->zone);
				} else {
						if ((i + off > 0 && pn[-1].net == (uint16)inet) ||
								(i + off < hdr->nets - 1 && pn[1].net == (uint16)inet)) {
							/* ambiguous nets: resolve at node level */
							ret = 1;
							while (pn != net && pn[-1].net == (uint16)inet)
									--pn;
						}
						goto net_found;
				}
			} else
				j = -((uint16)inet < pn->net);

			if (j < 0)
				num = i;
			else {
				off += i + 1;
				num -= i + 1;
			}
	}

	/* zone/net not found */
	return -1;

net_found:
	fido = (V9NODE *)((char *)nl + pn->off);
	for (off = 0, num = pn->size; num > 0; ) {
			i = num >> 1;
			pf = &fido[i + off];

			if ((uint16)inode == pf->node) {
				if (ozone)
						*ozone = pn->zone;
				if (oregion)
						*oregion = pn->region;
				if (ohub)
						*ohub = (!pf->hub)? 0 : fido[pn->size + pf->hub - 1].node;
				return ret; /* 0 if OK, 1 if OK but ambiguous */
			}

			if ((uint16)inode < pf->node)
				num = i;
			else {
				off += i + 1;
				num -= i + 1;
			}
	}

	/* resolve ambiguous nets (when no zone specified) at node level */
	if (ret && (++pn)->net == (uint16)inet)
			goto net_found;

	/* node not found, but net exists */
	return -2;
}


/********************************************************/
BOOL is_v9_listed(struct ndlTags *v9_ndlTag, ADDR addr,
									int *z, int *r, int *h)
{
	int zo = addr.zone, ne = addr.net, no = addr.node;
	char *nl;
	int rc;
	char v9name[256], buf[256];
	char caddr[128];

	addr_to_str(addr, caddr);

	/*  ndl disabled  */
	if (!v9_ndlTag->active) {
		sprintf(buf, "v9 nodelist `%s' is inactive, ignored\n", v9_ndlTag->tag);
		writeToLog(4, buf);
		return Fail;
	}
	
	v9_ndlTag->rec = v9_ndlTag->recfirst;

	/* scanning the list of records in the selected tag */
	for (; v9_ndlTag->rec ; v9_ndlTag->rec = v9_ndlTag->rec->next)
	{
		sprintf(v9name, "%sver9idx.%s",
						v9_ndlTag->rec->path, v9_ndlTag->rec->ext);
		
		if ((nl = v9_open (v9name, 0)) == NULL) {
			sprintf(buf, "is_v9_listed: error opening v9 nodelist:`%s'\n", v9name);
			writeToLog(0, buf);
			return Fail;
		}

		sprintf(buf, "looking for %s in %s\n", caddr, v9name);
		writeToLog(4, buf);
		
		rc = v9_search (nl, zo, ne, no, z, r, h);
		free (nl);
		
		if (rc >= 0) {
			writeToLog(4, "v9 search ok\n");
			return Ok;
		}
		else
			writeToLog(4, "v9 search failed\n");
		
	} /* outmost loop - scanning records in tag */

	
	return Fail;
	
}

static void build_stub(char *path, char *ext, char *ndlnames[], int nndl)
{
	int kndl, h_inc;
	dword len;
	char fname[256], str[256];
	struct nodeinc nodeinc;
	char n_inc_buf[NODEINC_SIZE], *pbuf;


	sprintf(fname, "%sver9inc.%s", path, ext);

	if (Chat_On)
		printf("building v9 index `%s'...", fname);


	if ( -1 == (h_inc = creat(fname, ACCFILE))) {
		sprintf(str, "v9 error: can`t create v9 index file `%s'\n", fname);
		fprintf(stderr, "\7%s", str);
		writeToLog(0, str);
		abort();
	}

	for( kndl = 0; kndl < nndl; ++kndl )
	{

		nodeinc.namelen = (char) strlen(ndlnames[kndl]);
		memset(nodeinc.name,0,12);
		strncpy(nodeinc.name, ndlnames[kndl], 12);
		sprintf(fname, "%s%s", path, ndlnames[kndl] );
		nodeinc.crc32 = crc32(fname);
		writeToLog(4, "v9: crc done\n");
		
		memset(&n_inc_buf, '\0', NODEINC_SIZE);
		pbuf = n_inc_buf;
		*pbuf = nodeinc.namelen;
		pbuf++;
		memmove(pbuf, &(nodeinc.name), 12);
		pbuf += 12;
		put_dword(pbuf, nodeinc.crc32);

		len = NODEINC_SIZE;
		if (len != (write(h_inc, &n_inc_buf, NODEINC_SIZE))) {
			sprintf(str, "v9: error writing v9 index file `%s'\n", fname);
			close(h_inc);
			fprintf(stderr, "\7%s", str);
			writeToLog(0, str);
			abort();
		}
		
		
	}

	close(h_inc);

	if (Chat_On)
		printf("  done\n");

	return;
	
}
