/*
* Copyright (c) 1996, serge terekhov.  All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 3. All advertising materials mentioning features or use of this software
*    must display the following acknowledgement:
*      This product includes software developed by serge terekhov.
* 4. Neither the name of the author nor the names of any co-contributors
*    may be used to endorse or promote products derived from this software
*    without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY SERGE TEREKHOV AND CONTRIBUTORS ``AS IS'' AND
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED.  IN NO EVENT SHALL SERGE TEREKHOV OR CONTRIBUTORS BE LIABLE
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
* OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
* LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
* OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
* SUCH DAMAGE.
*
*/

/*       NODELIST.V9 compiler       */
/*       Revision 1.0               */


#include <stdio.h>
#include <ctype.h>
#include <mem.h>
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <time.h>

#include "ver9.h"

#define DefZone 2
#define DefReg 50

uint16 hub[256];
int h_num = 0;
int max_hub = 0;

V9NET *net = 0;
int n_num = 0;
int n_size = 0;

V9NODE *fido = 0;
int f_num = 0;
int f_size = 0;
int f_hub;



int fido_cmp (V9NODE *a, V9NODE *b)
{
	return (a->node < b->node)? -1 : (a->node > b->node);
}



int net_cmp (V9NET *a, V9NET *b)
{
	if (a->net != b->net)
			return (a->net < b->net)? -1 : 1;
	return (a->zone < b->zone)? -1 : (a->zone > b->zone);
}



void compile (char *name)
{
	FILE *fp;
	char junk[512], save_junk[512];
	uint16 z,  r, n, f;
	int i;
	int last = 1, nozone = 0, fake = 0;

	printf ("compiling %s\n", name);

	if ((fp = fopen (name, "r")) == 0) {
			printf ("cant open\n");
			exit (1);
	}

	setvbuf (fp, 0, _IOFBF, 0x4000);

	while (fgets (junk, 511, fp) && junk[0] == ';')
			continue;
	if (strncmp (junk, "Zone,", 5)) {
//      printf ("first line is not Zone\n");
//      fclose (fp);
//      return;
		nozone = 1;
		strcpy(save_junk, junk);
	}

	do {
	start:
		if (nozone) {
			printf("nozone, fake=%d\n", fake);
			if (fake == 0)
				sprintf(junk, "Zone,%u", DefZone);
			else if (fake == 1)
				sprintf(junk, "Region,%u", DefReg);
			else {
				nozone = 0;
				strcpy(junk, save_junk);
			}
			++fake;
		}
			/* Zone Region Host Hub 0 Pvt Hold Down */
			switch (junk[0]) {
				case 'Z':   /* Zone */
						if (!isdigit (junk[5]))
							continue;
						z = r = n = atoi (&junk[5]);
						printf("got zone %d\n", r);
						goto nw0;

				case 'R':   /* Region */
						if (!isdigit (junk[7]))
							continue;
						r = n = atoi (&junk[7]);
						printf("got region %d\n", r);
				nw0:
						printf ("Zone %-5u Region %-5u\n", z, r);
						goto nw;

				case 'H':   /* Host Hub Hold */
						if (junk[1] == 'u')
							goto hub;
						if (junk[2] == 'l')
							goto hold;
						if (!isdigit (junk[5]))
							continue;
						n = atoi (&junk[5]);
						printf("got host %d\n", n);
					nw:
						printf("nw: n_num=%d h_num=%d f_num=%d\n",
									n_num, h_num, f_num);
						if (n_num) {
							i = net[n_num - 1].off;
							qsort (&fido[i], f_num - i, sizeof (*fido), (int (*)(const void*, const void*)) fido_cmp);
							net[n_num - 1].size = f_num - i;
							for (i = 0; i < h_num; ++i) {
									if (f_num >= f_size) {
										/* assert: f_size != 0 && fido != 0 */
										fido = realloc (fido, (f_size <<= 1) * sizeof (*fido));
										if (!fido) {
												printf ("low memo\n");
												exit (1);
										}
									}
									fido[f_num].node = hub[i];
									fido[f_num].hub = i + 1;
									printf("i=%d hub[i]=%d\n", i, hub[i]);
									++f_num;
							}
						}
						if (n_num >= n_size) {
							if (net)
									net = realloc (net, (n_size <<= 1) * sizeof (*net));
							else
									net = malloc ((n_size = 64) * sizeof (*net));
							if (!net) {
									printf ("low memo\n");
									exit (1);
							}
						}
						net[n_num].zone = z;
						net[n_num].region = r;
						net[n_num].net = n;
						net[n_num].off = f_num;
						++n_num;

						f = f_hub = 0;
						h_num = 0;
						goto nd;

				hub:
						if (!isdigit (junk[4]))
							continue;
						f = atoi (&junk[4]);
						printf("got hub %d\n", f);
						if (h_num < 255) {
							hub[h_num++] = f;
							f_hub = h_num;
							if (h_num > max_hub)
									max_hub = h_num;
						} else
							f_hub = 0;
						goto nd;

				case 'P':   /* Pvt */
						if (!isdigit (junk[4]))
							continue;
						f = atoi (&junk[4]);
						goto nd;

				case 'D':   /* Down */
				hold:
						if (!isdigit (junk[5]))
							continue;
						f = atoi (&junk[5]);
						goto nd;

				case ',':
						if (!isdigit (junk[1]))
							continue;
						f = atoi (&junk[1]);
				nd:
						if (f_num >= f_size) {
							if (fido)
									fido = realloc (fido, (f_size <<= 1) * sizeof (*fido));
							else
									fido = malloc ((f_size = 256) * sizeof (*fido));
							if (!fido) {
									printf ("low memo\n");
									exit (1);
							}
						}
						fido[f_num].node = f;
						fido[f_num].hub = f_hub;
						++f_num;
						break;
			}
		if (nozone) {
			goto start;
		}
		
	} while (fgets (junk, 511, fp));

	if (last) {
		last = 0;
		sprintf(junk, "Host,%hu", -1);
		goto start;
	}

	fclose (fp);
}



void find_last (char *nm)
{
	DIR *dir;
	struct dirent *ent;
	int last = 0, j, day, len;
	char *p, *q;
	char name[256];

	q = strrchr (strcpy (name, nm), '\\');
	if ((p = strrchr ((q)? q : name, '/')) != 0)
			q = p;
	if (q)
			*q++ = 0;
	else {
			q = nm;
			name[0] = '.';
			name[1] = 0;
	}
	len = strlen (q);

	if ((dir = opendir (name)) == 0)
			return;

	while ((ent = readdir (dir)) != 0) {
			p = ent->d_name;
#ifdef __OS2__
			strlwr (p);
#endif
			if (!strncmp (p, q, len) && *(p += len) == '.' &&
					isdigit (p[1]) && isdigit (p[2]) && isdigit (p[3]) && !p[4] &&
					(day = atoi (&p[1])) > 0 && day < 367) {
				if (!last ||
									(day > last && day - last < 300) ||
									(day < last && last - day > 300))
						last = day;
				}
			}
	closedir (dir);

	if (last)
			sprintf (nm + strlen (nm), ".%03d", last);
}



void main (int argc, char *argv[])
{
	V9HEADER hd;
	int i, j, err = 1;
	long off;
	char *q, *p;
	char v9name[256];
	char nlname[256];

	v9name[0] = 0;
	nlname[0] = 0;

	for (i = 1; i < argc; ++i) {
			if (argv[i][0] == '-' && strchr ("hH?", argv[i][1])) {
				printf ("usage: ver9 [ outname.v9 ] { nodelistname }\n");
				exit (1);
			}
			j = strlen (argv[i]);
			if (j > 3 && !strcmp (argv[i] + j - 3, ".v9"))
				strcpy (v9name, argv[i]);
			else {
				q = strcpy (nlname, argv[i]);
				if ((p = strrchr (q, '\\')) != 0)
						q = p + 1;
				if ((p = strrchr (q, '/')) != 0)
						q = p + 1;
				if (!*q)
						strcpy (q, "nodelist");
				if (q != nlname && !v9name[0]) {
						p = strcpy (v9name, nlname) + (int)(q - nlname);
						if ((p = strchr (p, '.')) == 0)
							p = v9name + strlen (v9name);
						strcpy (p, ".v9");
				}
				if (!strchr (q, '.'))
						find_last (nlname);
				compile (nlname);
			}
	}

	if (!nlname[0]) {
			if (!v9name[0])
				strcpy (v9name, "nodelist.v9");
			find_last (strcpy (nlname, "nodelist"));
			compile (nlname);
	}

	if (!n_num)
			printf ("no nodelist to compile\n");
	else {
			i = net[n_num - 1].off;
			qsort (&fido[i], f_num - i, sizeof (*fido), (int (*)(const void*, const void*)) fido_cmp);
			net[n_num - 1].size = f_num - i;

			qsort (net, n_num, sizeof (*net), (int (*)(const void*, const void*)) net_cmp);

			printf ("%d nets, %d nodes       \n", n_num, f_num);

#if 0    /* print ambiguous nets */
			for (i = 1; i < n_num; ++i)
				if (net[i - 1].net == net[i].net) {
						j = net[i].net;
						printf ("net %-5u - zones %u", j, net[i - 1].zone);
						for (; i < n_num && net[i].net == j; ++i)
							printf (", %u", net[i].zone);
						printf ("\n");
						--i;
				}
#endif

			unlink (v9name);
			i = open (v9name, O_WRONLY|O_BINARY|O_CREAT|O_TRUNC, ACCFILE);
			if (i == -1)
				printf ("cant write %s\n", v9name);
			else {
				err = 0;
				printf ("writing %s\n", v9name);
				memset (&hd, 0, sizeof (hd));
				hd.sig = V9_SIGNATURE;
				hd.ver = V9_VERSION;
				hd.nets = n_num;
				hd.nodes = f_num;
				hd.hubs = max_hub;
				hd.time = time (0);
				if (write (i, &hd, sizeof (hd)) != sizeof (hd))
						err = 1;
				off = sizeof (hd) + n_num * sizeof (*net);
				for (j = 0; j < n_num; ++j)
						net[j].off = off + net[j].off * sizeof (*fido);
				if (write (i, net, n_num * sizeof (*net)) != n_num * sizeof (*net) ||
						write (i, fido, f_num * sizeof (*fido)) != f_num * sizeof (*fido))
						err = 1;
				if (close (i) == -1)
						err = 1;
				printf ("size: header %ld nets %ld nodes %ld\n",
						sizeof (hd), n_num * sizeof (*net), f_num * sizeof (*fido));
				if (err) {
						unlink (v9name);
						printf ("error writing %s, deleted\n", v9name);
				}
			}
	}
	exit (err);
}

