//#include "squish.h"

#include "itrax.h"
#define OK 0

int walk_sqd(char *sqd_name, int mode);
void print_err(int rc);

enum { MSG_FORWARD = 0, MSG_BACKWARD, FREE_FORWARD, FREE_BACKWARD };


/**********************/
int main(int argc, char **argv)
{
	char sqd_name[256];
	int rc;
	
	if (!argv[1]) {
		return 0;
	}
	
	sprintf( sqd_name , "%s.sqd" , argv[1] );
	
	if (rc = walk_sqd(sqd_name, MSG_FORWARD)) {
		print_err(rc);
	}
	if (rc = walk_sqd(sqd_name, MSG_BACKWARD)) {
		print_err(rc);
	}
	if (rc = walk_sqd(sqd_name, FREE_FORWARD)) {
		print_err(rc);
	}
	if (rc = walk_sqd(sqd_name, FREE_BACKWARD)) {
		print_err(rc);
	}
	
	return 1;
}


int walk_sqd(char *sqd_name, int mode)
{
	int sqd; // sqd handle
	dword next_frame, nread, len, off, mlen = 0L;
	word sqbase_len, sz_sqhdr;
	SQBASE sqbase;
	SQHDR sqhdr;
	XMSG xmsg;
	SQIDX sqidx;
	int rc = OK, nframes = 0;
	
	if( -1 == (sqd = open(sqd_name, O_RDONLY|O_BINARY)) ) {
		printf("error opening `%s'\n", sqd_name );
		perror(sqd_name);
		return SQD_OPEN_ERR;
	}
	
	lseek( sqd , 0L , SEEK_SET );
	// read msg hdr
	read( sqd, &sqbase_len, 2L );
	lseek( sqd , 0L , SEEK_SET );
	if ( sqbase_len != (len = (read( sqd, &sqbase, sqbase_len )) ) ){
		close(sqd);
		return SQD_READ_ERR;
	}
	
	if (len != sizeof(SQBASE)) {
		printf("error reading _sqbase hdr in `%s'\n", sqd_name);
	}
	
	// len = sizeof(SQHDR);
	sz_sqhdr = sqbase.sz_sqhdr;

	// get starting frame
	switch(mode)
	{
	case MSG_FORWARD:
		// get offset of the 1st msg frame
		next_frame = sqbase.begin_frame;
		break;
	case MSG_BACKWARD:
		// get offset of the last msg frame
		next_frame = sqbase.last_frame;
		break;
	case FREE_FORWARD:
		// get offset of the 1st free frame
		next_frame = sqbase.free_frame;
		break;
	case FREE_BACKWARD:
		// get offset of the last msg frame
		next_frame = sqbase.last_free_frame;
		break;
	default:
		printf("unknown mode\n");
		next_frame = 0;
		break;
	}
	
	while (!eof(sqd))
	{
		
		if (!next_frame) { // final frame reached
			break;
		}

//    printf("%08X\n", next_frame);
		// position *.sqd to the start of next frame
		off = lseek( sqd , next_frame , SEEK_SET );
		if ( off != next_frame ) {
			printf("lseek error in `%s', requested:%08X, returned:%08X\n",
						sqd_name, next_frame, off);
			rc = SQD_SEEK_ERR;
			break;
		}
		
		// read SQHDR
		if ( sz_sqhdr != (nread = read( sqd, &sqhdr, sz_sqhdr )) ){
			printf("error reading sqhdr in `%s'\n", sqd_name);
			printf("chars requested:%d, read:%d, offset:%08X\n",
						sz_sqhdr, nread, off);
			rc = SQD_READ_ERR;
			break;
		}
		
		if ( sqhdr.id != SQHDRID ) {
			rc = SQD_INVALID_ID;
			break;
		}
		
		if (mode == MSG_FORWARD || mode == MSG_BACKWARD) {
			if ( sqhdr.frame_type != 0  && sqhdr.frame_type != 3 ) {
				printf("invalid frame type in `%s' at offset %08X\n",
							sqd_name, next_frame);
				rc = SQD_INVALID_FRAME_TYPE;
				break;
			}
			mlen += sqhdr.msg_length;
		}
		else { // FREE CHAIN
			if ( sqhdr.frame_type != 1  && sqhdr.frame_type != 3 ) {
				printf("invalid frame type in `%s' at offset %08X\n",
							sqd_name, next_frame);
				rc = SQD_INVALID_FRAME_TYPE;
				break;
			}
		}
				
		++nframes;

		//get next frame offset
		switch(mode)
		{
		case MSG_FORWARD:
		case FREE_FORWARD:
		// get offset of the next frame
			next_frame = sqhdr.next_frame;
			break;
		case MSG_BACKWARD:
		case FREE_BACKWARD:
		// get offset of the prev frame
			next_frame = sqhdr.prev_frame;
			break;
		default:
			printf("unknown mode\n");
			next_frame = 0;
			break;
		}
	} // sqd walk
	
	close(sqd);
	
	if (!rc) {// ok
		switch(mode)
		{
		case MSG_FORWARD:
			printf("msg: walk forward, frames read: %d, num_msg: %d high_msg: %d\n",
						nframes, sqbase.num_msg, sqbase.high_msg );
			break;
		case MSG_BACKWARD:
			printf("msg: walk backward, frames read: %d, num_msg: %d high_msg: %d\n",
						nframes, sqbase.num_msg, sqbase.high_msg );
			break;
		case FREE_FORWARD:
			printf("free: walk forward, frames read: %d\n", nframes);
			break;
		case FREE_BACKWARD:
			printf("free: walk backward, frames read: %d\n", nframes);
			break;
		}
	}

	printf("mlen=%d\n", mlen);
	return rc;
	
}
/***************************/
void print_err(int rc)
{
	switch(rc) {
	case SQD_OPEN_ERR:
		printf("sqd: open error\n");
		break;
	case SQD_SEEK_ERR:
		printf("sqd: seek error\n");
		break;
	case SQD_READ_ERR:
		printf("sqd: read error\n");
		break;
	case SQD_INVALID_ID:
		printf("sqd: invalid SQHDR id read\n");
		break;
	case SQD_INVALID_FRAME_TYPE:
		printf("sqd: invalid frame type read\n");
		break;
	default:
		printf("unknown error, rc = %d\n" , rc);
		break;
	}
	return;
}
	
