#include "itrax.h"
#define OK 0

int walk_sqi(char *sqi_name, char *sqd_name);

int main(int argc, char **argv)
{
	char sqd_name[256], sqi_name[256];
	int rc;
	
	if (!argv[1]) {
		return 0;
	}
	
	sprintf( sqd_name , "%s.sqd" , argv[1] );
	sprintf( sqi_name , "%s.sqi" , argv[1] );

	if (rc = walk_sqi(sqi_name, sqd_name)) {
		
		switch(rc) {
		case SQD_OPEN_ERR:
			printf("sqd: open error, `%s'\n", sqd_name);
			break;
		case SQI_OPEN_ERR:
			printf("sqi: open error, `%s'\n", sqi_name);
			break;
		case SQI_READ_ERR:
			printf("sqi: read error\n");
			break;
		case SQI_INVALID_OFS:
			printf("sqi: invalid offset value read\n");
			break;
		case SQI_INVALID_UMSGID:
			printf("sqi: invalid umsgid value read\n");
			break;
		case SQD_SEEK_ERR:
			printf("sqd: seek error\n");
			break;
		case SQD_READ_ERR:
			printf("sqd: read error\n");
			break;
		case SQD_INVALID_ID:
			printf("sqd: invalid SQHDR id read\n");
			break;
		case SQD_INVALID_FRAME_TYPE:
			printf("sqd: invalid frame type read\n");
			break;
		default:
			printf("rc = %d\n" , rc);
			break;
		}
	}
	else {
		printf("ok\n");
	}
	return 1;
	
}

/**********************/
int walk_sqi(char *sqi_name, char *sqd_name)
{
	int sqd, sqi; /* sqd, sqi handles */
	dword rec_len = sizeof(SQIDX), off, nread, len;
	word sqbase_len, sz_sqhdr;
	SQBASE sqbase;
	SQHDR sqhdr;
	XMSG xmsg;
	SQIDX sqidx;
	int rc = OK;
	dword n = 0, estim = 0;
	
	if( -1 == (sqi = open(sqi_name, O_RDONLY|O_BINARY)) ) {
		printf("error opening `%s'\n", sqi_name );
		perror(sqi_name);
		return SQI_OPEN_ERR;
	}
	if( -1 == (sqd = open(sqd_name, O_RDONLY|O_BINARY)) ) {
		printf("error opening `%s'\n", sqd_name );
		perror(sqi_name);
		return SQD_OPEN_ERR;
	}

	lseek( sqd , 0L , SEEK_SET );
	/* read msg hdr */
	read( sqd, &sqbase_len, 2L );
	lseek( sqd , 0L , SEEK_SET );
	if ( sqbase_len != (len = (read( sqd, &sqbase, sqbase_len )) ) ){
		close(sqi);
		close(sqd);
		return SQD_READ_ERR;
	}
	
	if (len != sizeof(SQBASE)) {
		printf("error reading _sqbase hdr in `%s'\n", sqd_name);
	}
	
/*	len = sizeof(SQHDR); */
	sz_sqhdr = sqbase.sz_sqhdr;

	estim = filelength(sqi) / rec_len;
	while (!eof(sqi))
	{
		
		if ( rec_len != (nread = read( sqi, &sqidx, rec_len )) ){
			printf("error reading `%s', read %d chars, offset:%08X\n",
						sqi_name, nread, n/rec_len);
			rc = SQI_READ_ERR;
			break;
		}
		if (!sqidx.ofs) {
			rc = SQI_INVALID_OFS;
			printf("invalid offset\n");
			continue;
		}
		if (sqidx.umsgid == 0xffffffff) {
			rc = SQI_INVALID_UMSGID;
			printf("invalid umsgid\n");
			continue;
		}
		/* position sqd to the start of next frame */
		off = lseek( sqd , sqidx.ofs , SEEK_SET );
		if ( off != sqidx.ofs ) {
			printf("lseek error in `%s', requested:%08X, returned:%08X\n", 
						sqd_name, sqidx.ofs, off);
			rc = SQD_SEEK_ERR;
			break;
		}
		
		/* read SQHDR */
		if ( sz_sqhdr != (nread = read( sqd, &sqhdr, sz_sqhdr )) ){
			printf("error reading sqhdr in `%s'\n", sqd_name);
			printf("chars requested:%d, read:%d, offset:%08X\n",
						sz_sqhdr, nread, off);
			rc = SQD_READ_ERR;
			break;
		}
		if ( sqhdr.id != SQHDRID ) {
			rc = SQD_INVALID_ID;
			break;
		}
		if ( sqhdr.frame_type != 0  && sqhdr.frame_type != 3 ) {
			rc = SQD_INVALID_FRAME_TYPE;
			break;
		}
	
		++n;
	} /* sqi walk */
	
	if (!rc) { /* ok */
		printf("records, estimated: %d, read: %d, num_msg: %d high_msg: %d\n",
					estim, n, sqbase.num_msg, sqbase.high_msg );
	}
	
	close(sqd);
	close(sqi);
	
	return rc;
	
}
